#include "standard.h"

#include "files.h"
#include "tip_seqs.h"

LoggerPtr Tip_Seqs::s_logger(Logger::getLogger("IO.tip_seqs"));



// parse file to Tip_Seqs
Tip_Seqs::Tip_Seqs(string filename)
{
	OUT_FNCTN_HEADER(logger());

	ifstream ifs(filename.c_str());
	if (!ifs.good()) {
		stringstream ss;
		ss << "Could not open " << filename;
		throw ERROR(logger(), ss.str());
	}

	LOG4CXX_DEBUG(logger(), "Parsing: " << filename);
	BOOL reading_seq_name = true;
	string str;
	Seq seq;
	VECTOR(int) seq_len;
	do
	{
		getline(ifs, str);
		LOG4CXX_DEBUG(logger(), str);
		if (reading_seq_name) {
			if (str[0] == '>')
				seq.name = str.substr(1);
			else
				if (str.length() > 0)
					throw ERROR(logger(), "File format error ");
		}
		else {
			seq.genome = str;
			m_seq_of_tips.push_back(seq);
			seq_len.push_back(str.length());
		}

		reading_seq_name = !reading_seq_name;
	}
	while (!ifs.eof() && (str.length() > 0));
	LOG4CXX_DEBUG(logger(), "Parsing finished ");

	stringstream ss;
	ss << "seq_len: ";
	for (VECTOR(int)::iterator it = seq_len.begin(); it != seq_len.end(); ++it)
		ss << *it << " ";
	LOG4CXX_DEBUG(logger(), ss.str());

	m_num_tips = m_seq_of_tips.size();
	if (m_num_tips < 2)
		throw ERROR(logger(), "Trying to construct tip_seqs with less the two sequences");

	BOOL is_first = true;
	for (VECTOR(int)::iterator it = seq_len.begin(); it != seq_len.end(); ++it)
	{
		if (is_first)
		{
			m_num_sites = *it;
			is_first = false;
		}
		else
			if (*it != m_num_sites)
				throw ERROR(logger(), "Trying to construct tip_seqs with sequences of different length");
	}

	OUT_FNCTN_BOTTOM(logger());
}



// convert VECTOR(string) to Tip_Seqs
Tip_Seqs::Tip_Seqs(VECTOR(string) seqs)
{
	OUT_FNCTN_HEADER(logger());

	m_num_sites = seqs[0].length();
	m_num_tips = seqs.size();

	int n = 0;
	for (VECTOR(string)::iterator it = seqs.begin(); it != seqs.end(); ++it)
	{
		if (it->length() != m_num_sites)
			throw ERROR(logger(), "Trying to construct tip_seqs with sequences of different length");

		stringstream s;
		s << "Seq" << n;
		Seq ss = { s.str(), *it };
		m_seq_of_tips.push_back(ss);
		n++;
	}

	OUT_FNCTN_BOTTOM(logger());
}



// construct Tip_Seqs by choosing from a list of examples
Tip_Seqs::Tip_Seqs(const int no, const int num_sites) : m_num_sites(num_sites)
{
	OUT_FNCTN_HEADER(logger());

	switch (no)
	{
		case 0:
		{
			m_num_tips = 3;

			m_seq_of_tips.resize(m_num_tips);
			m_seq_of_tips[0].genome = string(m_num_sites, 'A');
			m_seq_of_tips[1].genome = string(m_num_sites, 'C');
			m_seq_of_tips[2].genome = string(m_num_sites / 2, 'A') + string(m_num_sites / 2, 'C');
			break;
		}
		case 1:
		{
			m_num_tips = 4;

			m_seq_of_tips.resize(m_num_tips);
			m_seq_of_tips[0].genome = string(m_num_sites, 'A');
			m_seq_of_tips[1].genome = string(m_num_sites, 'C');
			m_seq_of_tips[2].genome = string(m_num_sites / 2, 'A') + string(m_num_sites / 2, 'C');
			m_seq_of_tips[3].genome = string(m_num_sites / 2, 'A') + string(m_num_sites / 2, 'C');
			break;
		}
		case 2:
		{
			m_num_tips = 4;

			m_seq_of_tips.resize(m_num_tips);
			m_seq_of_tips[0].genome = string(m_num_sites, 'A');
			m_seq_of_tips[1].genome = string(m_num_sites, 'C');
			m_seq_of_tips[2].genome = string(m_num_sites / 2, 'A') + string(m_num_sites / 2, 'C');
			m_seq_of_tips[3].genome = string(m_num_sites / 2, 'A') + string(m_num_sites / 2, 'C');
			break;
		}
		case 3:
		{
			m_num_sites = 6;
			m_num_tips = 5;

			m_seq_of_tips.resize(m_num_tips);
			m_seq_of_tips[0].genome = string(m_num_sites, 'A');
			m_seq_of_tips[1].genome = string(m_num_sites, 'C');
			m_seq_of_tips[2].genome = string(m_num_sites, 'T');
			m_seq_of_tips[3].genome = string(m_num_sites / 3, 'A') + string(m_num_sites / 3, 'C') + string(m_num_sites / 3, 'T');
			m_seq_of_tips[4].genome = string(m_num_sites / 3, 'A') + string(m_num_sites / 3, 'C') + string(m_num_sites / 3, 'T');
			break;
		}
		case 4:
		{
			m_num_tips = 5;

			m_seq_of_tips.resize(m_num_tips);
			m_seq_of_tips[0].genome = string(m_num_sites, 'A');
			m_seq_of_tips[1].genome = string(m_num_sites, 'C');
			m_seq_of_tips[2].genome = string(m_num_sites, 'T');
			m_seq_of_tips[3].genome = string(m_num_sites / 2, 'A') + string(m_num_sites / 2, 'C');
			m_seq_of_tips[4].genome = string(m_num_sites / 2, 'C') + string(m_num_sites / 2, 'T');
			break;
		}
		case 5:
		{
			m_num_tips = 8;

			m_seq_of_tips.resize(m_num_tips);
			m_seq_of_tips[0].genome = string(m_num_sites, 'A');
			m_seq_of_tips[1].genome = string(m_num_sites, 'A');
			m_seq_of_tips[2].genome = string(m_num_sites, 'A');
			m_seq_of_tips[3].genome = string(m_num_sites, 'A');
			m_seq_of_tips[4].genome = string(m_num_sites, 'C');
			m_seq_of_tips[5].genome = string(m_num_sites, 'C');
			m_seq_of_tips[6].genome = string(m_num_sites, 'C');
			m_seq_of_tips[7].genome = string(m_num_sites, 'C');
			break;
		}
		case 6:
		{
			m_num_tips = 6;

			m_seq_of_tips.resize(m_num_tips);
			m_seq_of_tips[0].genome = string(m_num_sites, 'A');
			m_seq_of_tips[1].genome = string(m_num_sites, 'C');
			m_seq_of_tips[2].genome = string(m_num_sites / 2, 'A') + string(m_num_sites / 2, 'C');
			m_seq_of_tips[3].genome = string(m_num_sites / 2, 'A') + string(m_num_sites / 2, 'C');
			m_seq_of_tips[4].genome = string(m_num_sites / 2, 'A') + string(m_num_sites / 2, 'C');
			m_seq_of_tips[5].genome = string(m_num_sites / 2, 'A') + string(m_num_sites / 2, 'C');
			break;
		}
		case 7:
		{
			m_num_tips = 14;

			m_seq_of_tips.resize(m_num_tips);
			m_seq_of_tips[0].genome = string(m_num_sites, 'A');
			m_seq_of_tips[1].genome = string(m_num_sites, 'C');
			m_seq_of_tips[2].genome = string(m_num_sites, 'C');
			m_seq_of_tips[3].genome = string(m_num_sites, 'T');
			m_seq_of_tips[4].genome = string(m_num_sites, 'T');
			m_seq_of_tips[5].genome = string(m_num_sites, 'T');
			m_seq_of_tips[6].genome = string(m_num_sites, 'T');
			m_seq_of_tips[7].genome = string(m_num_sites / 2, 'A') + string(m_num_sites / 2, 'C');
			m_seq_of_tips[8].genome = string(m_num_sites / 2, 'A') + string(m_num_sites / 2, 'C');
			m_seq_of_tips[9].genome = string(m_num_sites / 2, 'A') + string(m_num_sites / 2, 'C');
			m_seq_of_tips[10].genome = string(m_num_sites / 2, 'A') + string(m_num_sites / 2, 'C');
			m_seq_of_tips[11].genome = string(m_num_sites / 3, 'A') + string(m_num_sites / 3, 'C') + string(m_num_sites / 3, 'T');
			m_seq_of_tips[12].genome = string(m_num_sites / 3, 'A') + string(m_num_sites / 3, 'C') + string(m_num_sites / 3, 'T');
			m_seq_of_tips[13].genome = string(m_num_sites / 2, 'T') + string(m_num_sites / 2, 'A');
			break;
		}
		case 8:
		{
			m_num_tips = 2;

			m_seq_of_tips.resize(m_num_tips);
			m_seq_of_tips[0].genome = string(m_num_sites, 'A');
			m_seq_of_tips[1].genome = string(m_num_sites, 'C');
			break;
		}
		case 9:
		{
			m_num_tips = 3;

			m_seq_of_tips.resize(m_num_tips);
			m_seq_of_tips[0].genome = string(m_num_sites / 2, 'A') + string(m_num_sites / 2, 'C');
			m_seq_of_tips[1].genome = string(m_num_sites / 2, 'A') + string(m_num_sites / 2, 'G');
			m_seq_of_tips[2].genome = string(m_num_sites / 2, 'A') + string(m_num_sites / 2, 'T');
			break;
		}
		case 10:
		{
			m_num_tips = 5;

			m_seq_of_tips.resize(m_num_tips);
			m_seq_of_tips[0].genome = string(m_num_sites, 'A');
			m_seq_of_tips[1].genome = string(m_num_sites, 'A');
			m_seq_of_tips[2].genome = string(m_num_sites, 'A');
			m_seq_of_tips[3].genome = string(m_num_sites, 'C');
			m_seq_of_tips[4].genome = string(m_num_sites / 2, 'A') + string(m_num_sites / 2, 'C');
			break;
		}
		default:
			break;
	}

	OUT_FNCTN_BOTTOM(logger());
}



// save to file
void Tip_Seqs::save(string filename) const
{
	ofstream file_strm(filename.c_str());
	for (int i = 0; i < num_tips(); ++i) {
		file_strm << ">" << name(i) << endl;
		file_strm << genome(i) << endl;
	}
}



// output of Tip_Seqs
ostream& operator<<(ostream& s, const Tip_Seqs& ts)
{
	s << "*** TIP_SEQUENCES ***" << endl;
	s << "num_tips: " << ts.num_tips() << endl;
	s << "num_sites: " << ts.num_sites() << endl << endl;


	for (VECTOR(Seq)::const_iterator it = ts.m_seq_of_tips.begin(); it != ts.m_seq_of_tips.end(); ++it)
	{
		s << it->name << endl;
		s << it->genome << endl;
	}

	s << endl;

	return s;
}
