#include "standard.h"

#include "constants.h"
#include "dot.h"
#include "files.h"
#include "MCMC.h"
#include "nc.h"
#include "nd.h"
#include "random.h"
#include "tr_rearr_brch.h"
#include "tr_rearr_coal.h"
#include "tr_rearr_rec.h"
#include "tr_rearr_scl.h"
#include "tr_rearr_sngl.h"
#include "tree.h"

const int NUM_OPS = 5;   // number of different alteration operations
const double FRACTION_OPS[NUM_OPS] = { 1, 1, 0.1, 2, 1 };   // proportion of each operation
														    // (COAL_OP, RECOMB_OP, SCALE_OP, BRANCH_OP, SINGLE_OP)


ostream& MCMC::operator<<(ostream& s, const MCMC::Results& res)
{
	s << ">tree_log_probs   # prior log-prob, acid log-prob, tree log-prob" << endl;
	s << setprecision(20) << res.log_p_prior() << endl;
	s << setprecision(20) << res.log_p_acids() << endl;
	s << setprecision(20) << res.log_p() << endl;

	return s;
}



// apply MCMC (not for users)
MCMC::Results MCMC_impl::call(Rev_Tree* tr, const BOOL num_steps_fixed, const int num_steps, const int break_length, const double eps)
{
	OUT_FNCTN_HEADER(mcmc_logger);
	enum enum_operations { COAL_OP, RECOMB_OP, SCALE_OP, BRANCH_OP, SINGLE_OP };

	const BOOL OUTPUT_MCMC = 0;
	const BOOL OUTPUT_ITER = Constants::instance()->messages_on();
	const BOOL OUTPUT_DOT = 0;
	const BOOL OUTPUT_DOT_NEW_MAX = 0;

	if (OUTPUT_DOT)
		Dot::to_dot(tr, "MCMC begin");

	ofstream* mcmc_strm = In_Out_Streams::instance()->ostrm(MCMC_FILE);
	VECTOR(double) prob;

	if (tr->num_act_nodes() > 1)
		throw ERROR(mcmc_logger, "MCMC: Trying to apply MCMC to tree with more than one active node");

	if (tr->num_nodes() - tr->num_act_nodes() - tr->num_tips() < 1)
		throw ERROR(mcmc_logger, "MCMC: Trying to apply MCMC to tree without interior nodes");

	if (!tr->acid_prob_calced())
		throw ERROR(mcmc_logger, "MCMC: Trying to apply MCMC to tree without having calculated acid probabilities");

	if (num_steps_fixed && (num_steps <= 0))
		throw ERROR(mcmc_logger, "MCMC: Illegal number of iteration steps");

	if (SAFE_MODE)
		tr->check();

	double log_prob_acids_old = tr->log_prob_acids();
	double log_prior_prob_old = tr->log_prior_prob();
	double log_prob_before = tr->log_prob();
	double log_prob_old_tree = log_prob_before;
	double curr_max_log_prob = -numeric_limits<double>::max();
	double log_prior_prob_max = 0;
	double log_prob_acids_max = 0;
	VECTOR(double) log_prob_acids_pos_max(tr->num_sites(), 0);
	VECTOR(double) max_log_probs;
	int arg_max_log_prob;

	int num_ops[NUM_OPS], num_successful_ops[NUM_OPS];
	for (int i = 0; i < NUM_OPS; i++)
		num_ops[i] = num_successful_ops[i] = 0;

	if (OUTPUT_MCMC)
	{
		*mcmc_strm << "log_prob_acids: " << tr->log_prob_acids() << endl;
		*mcmc_strm << "log_prior_prob: " << tr->log_prior_prob() << endl;
		*mcmc_strm << "log_prob: " << tr->log_prob() << endl << endl;
	}

	int no_step = 0;
	double rel_diff_prob = 0;
	clock_t t1 = clock();
	while (num_steps_fixed && (no_step < num_steps) ||
		   !num_steps_fixed && ((no_step <= break_length) || (rel_diff_prob > eps)))
	{
		if (OUTPUT_ITER)
			cout << "MCMC step: " << no_step << flush;

		LOG4CXX_DEBUG(mcmc_logger, "MCMC step: " << no_step);

		// choose rearranging operation randomly
		gsl_ran_discrete_t* grd = gsl_ran_discrete_preproc (NUM_OPS, FRACTION_OPS);
		enum_operations chosen_operation = (enum_operations)gsl_ran_discrete (Random::rng(), grd);
		gsl_ran_discrete_free(grd);

		num_ops[(int)chosen_operation]++;
		double scale = 0;

		switch (chosen_operation)
		{
			// rearranging operation "coalescence"
			case COAL_OP:
			{
				LOG4CXX_DEBUG(mcmc_logger, "COAL_OP");
				if (OUTPUT_MCMC)
					*mcmc_strm << "COAL_OP" << endl;

				// find possible targets
				VECTOR(Rev_Coal_Nd*) elig_targets;
				Rev_Nd* n = tr->first_node();
				while (n)
				{
					if (Rev_Coal_Nd* m = dynamic_cast<Rev_Coal_Nd*>(n))
						if (m->has_coal_anc())
							elig_targets.push_back(m);
					n = n->next();
				}

				// output possible targets
				stringstream ss;
				ss << "elig_targets (coal): " << elig_targets;
				LOG4CXX_DEBUG(mcmc_logger, ss.str());

				// choose target randomly
				if (elig_targets.size() > 0)
				{
					int r = gsl_rng_uniform_int(Random::rng(), elig_targets.size());
					Rev_Coal_Nd* target = elig_targets[r];
					if (target->is_pure_node() == target->parent_node()->is_pure_node())
						tree_rearranger_coal::coal(tr, target, true);
					else
						tree_rearranger_coal::coal(tr, target, false);
				}
				break;
			}

			// rearranging operation "recombination"
			case RECOMB_OP:
			{
				LOG4CXX_DEBUG(mcmc_logger, "RECOMB_OP");
				if (OUTPUT_MCMC)
					*mcmc_strm << "RECOMB_OP" << endl;

				// find possible targets
				VECTOR(Rev_Nd*) elig_targets;
				Rev_Nd* n = tr->first_node();
				while (n)
				{
					if (n->has_rec_anc() && n->is_crf_node())
						tree_rearranger_rec::get_recomb_nodes_rearr(tr, elig_targets, n);
					n = n->next();
				}

				// output possible targets
				stringstream ss;
				ss << "elig_targets (recomb): ";
				for (VECTOR(Rev_Nd*)::iterator it = elig_targets.begin(); it != elig_targets.end(); it++)
					ss << (*it)->id() << " ";
				ss << endl;
				LOG4CXX_DEBUG(mcmc_logger, ss.str());

				// choose target randomly
				if (elig_targets.size() > 0)
				{
					int r = gsl_rng_uniform_int(Random::rng(), elig_targets.size());
					Rev_Nd* target = elig_targets[r];
					if (false)
						tree_rearranger_rec::recomb(tr, target, true);
					else
						tree_rearranger_rec::recomb(tr, target);
				}
				break;
			}

			// rearranging operation "scale"
			case SCALE_OP:
			{
				LOG4CXX_DEBUG(mcmc_logger, "SCALE_OP");
				if (OUTPUT_MCMC)
					*mcmc_strm << "SCALE_OP" << endl;
				tree_rearranger_scl::scale(tr);
				break;
			}

			// rearranging operation "branch"
			case BRANCH_OP:
			{
				LOG4CXX_DEBUG(mcmc_logger, "BRANCH_OP");
				if (OUTPUT_MCMC)
					*mcmc_strm << "BRANCH_OP" << endl;

				// find possible targets
				VECTOR(Rev_Coal_Nd*) elig_targets;
				Rev_Nd* n = tr->first_node();
				while (!n->is_last())
				{
					if (Rev_Coal_Nd* m = dynamic_cast<Rev_Coal_Nd*>(n))
						if ((m->is_crf_node() && m->has_coal_anc() ||
							 m->is_pure_node() && m->parent_node()->is_pure_node()))
							elig_targets.push_back(m);
					n = n->next();
				}

				// output possible targets
				stringstream ss;
				ss << "elig_targets (branch): ";
				for (VECTOR(Rev_Coal_Nd*)::iterator it = elig_targets.begin(); it != elig_targets.end(); it++)
					ss << (*it)->id() << " ";
				ss << endl;
				LOG4CXX_DEBUG(mcmc_logger, ss.str());

				// choose target randomly
				if (elig_targets.size() > 0)
				{
					int r = gsl_rng_uniform_int(Random::rng(), elig_targets.size());
					Rev_Coal_Nd* target = elig_targets[r];
					tree_rearranger_brch::branch(tr, target);
				}
				break;
			}

			// rearranging operation "single"
			case SINGLE_OP:
			{
				LOG4CXX_DEBUG(mcmc_logger, "SINGLE_OP");
				if (OUTPUT_MCMC)
					*mcmc_strm << "SINGLE_OP" << endl;
				tree_rearranger_sngl::single(tr);
				break;
			}
		}

		double log_prob_acids_new = tr->log_prob_acids();
		double log_prior_prob_new = tr->log_prior_prob();
		double log_prob_new_tree = log_prob_acids_new + log_prior_prob_new;

		if (OUTPUT_MCMC)
		{
			*mcmc_strm << "MCMC step: " << no_step << endl;
			*mcmc_strm << "log_prob_acids: " << log_prob_acids_new << endl;
			*mcmc_strm << "log_prior_prob: " << log_prior_prob_new << endl;
			*mcmc_strm << "log_prob: " << log_prob_new_tree << endl;
		}
		// output Dot file of Rev_Tree
		if (OUTPUT_DOT || (log_prob_new_tree > curr_max_log_prob) && OUTPUT_DOT_NEW_MAX)
		{
			stringstream ss;
			ss << "MCMC " << no_step << " ";
			switch (chosen_operation)
			{
				case COAL_OP:
					ss << "COAL";
					break;
				case RECOMB_OP:
					ss << "REC";
					break;
				case SCALE_OP:
					ss << "SCL";
					break;
				case BRANCH_OP:
					ss << "BRN";
					break;
				case SINGLE_OP:
					ss << "SGN";
					break;
			}
			ss << "\\n";
			ss.precision(4);
			ss << "old: " << log_prob_old_tree << "\\n";
			ss << "ac/pr: " << log_prob_acids_old << ", " << log_prior_prob_old << "\\n";
			ss << "new: " << log_prob_new_tree << "\\n";
			ss << "ac/pr: " << log_prob_acids_new << ", " << log_prior_prob_new;
			Dot::to_dot(tr, ss.str());
		}

		// output Dot file of Rev_Tree
		stringstream ss;
		if (OUTPUT_DOT || (log_prob_new_tree > curr_max_log_prob) && OUTPUT_DOT_NEW_MAX)
		{
			ss.str("");
			ss << "MCMC " << no_step << " ";
		}

		// decide whether changes are accepted
		double x = gsl_rng_uniform(Random::rng());
		double y;
		if ((chosen_operation == SCALE_OP) || (chosen_operation == BRANCH_OP))
			y = exp(log_prob_new_tree - log_prob_old_tree);
		else
			y = exp(log_prob_acids_new - log_prob_acids_old);			
		if (!tr->check_pure_final())
			y = -1;
		if (y > x)
		{
			num_successful_ops[(int)chosen_operation]++;

			// accept changes
			if (OUTPUT_DOT || (log_prob_new_tree > curr_max_log_prob) && OUTPUT_DOT_NEW_MAX)
				ss << "ACC";
			log_prob_old_tree = log_prob_new_tree;
			log_prob_acids_old = log_prob_acids_new;
			log_prior_prob_old = log_prior_prob_new;

			tr->accept_changes();

		}
		else
		{
			// reject changes
			if (OUTPUT_DOT || (log_prob_new_tree > curr_max_log_prob) && OUTPUT_DOT_NEW_MAX)
				ss << "REJ";
			tr->reverse_changes();
		}

		tr->recalc_act_sites_grps();

		if (SAFE_MODE)
			tr->check();

		// output Dot file of Rev_Tree
		if (OUTPUT_DOT || (log_prob_new_tree > curr_max_log_prob) && OUTPUT_DOT_NEW_MAX) {
			ss << "\\n";
			ss << log_prob_old_tree << "\\n";
			ss << "ac/pr: " << log_prob_acids_old << ", " << log_prior_prob_old << "\\n";
			Dot::to_dot(tr, ss.str());
			/*if (Constants::instance()->compile_dot() && Constants::instance()->compile_dot_immediately())  {
				stringstream ss;
				ss << "dot -Tps <output/arg" << In_Out_Streams::instance()->no_arg_Dot() - 1 << ".txt >output/arg"
				   << In_Out_Streams::instance()->no_arg_Dot() - 1 << ".ps";
				std::system(ss.str().c_str());
			}*/
		}

		prob.push_back(log_prob_old_tree);
		if (log_prob_new_tree > curr_max_log_prob)
		{
			curr_max_log_prob = log_prob_new_tree;
			arg_max_log_prob = no_step;
			log_prior_prob_max = log_prior_prob_new;
			log_prob_acids_max = log_prob_acids_new;
			log_prob_acids_pos_max = tr->log_prob_acids_pos();
			tr->save();
		}
		if (OUTPUT_MCMC)
			*mcmc_strm << "max_log_prob: " << curr_max_log_prob << endl << endl;

		max_log_probs.push_back(curr_max_log_prob);

		if (OUTPUT_ITER) {
			cout << " (";
			if (no_step >= break_length) {
				rel_diff_prob = (max_log_probs[no_step] - max_log_probs[no_step - break_length]) /
								abs(max_log_probs[no_step - break_length]);
				cout << setprecision(4) << rel_diff_prob << ", ";
			}
			clock_t t2 = clock();
			cout << double(t2 - t1) / CLOCKS_PER_SEC;
			cout << ")" << endl;
		}

		no_step++;
	}

	// output maximum tree probs
	if (OUTPUT_MCMC)
	{
		*mcmc_strm << "maximum log_prob:" << endl;
		for (VECTOR(double)::iterator it = max_log_probs.begin(); it != max_log_probs.end(); ++it)
			*mcmc_strm << "step " << setw(4) << it - max_log_probs.begin() << ": "
					   << fixed << setprecision(4) << *it << endl;
		*mcmc_strm << endl;
	}

	// output Rev_Tree probability
	if (OUTPUT_MCMC)
	{
		*mcmc_strm << "MCMC prob" << endl;
		*mcmc_strm << "before   " << setprecision(4) << log_prob_before << endl;
	}

	if (OUTPUT_MCMC)
	{
		// mcmcput maximum tree probability
		*mcmc_strm << "max log_prob: "<< curr_max_log_prob << endl;
		*mcmc_strm << "argmax: "<< arg_max_log_prob << endl << endl;

		// mcmcput number of (successful) operations
		*mcmc_strm << "coal ops:   " << num_successful_ops[COAL_OP] << "/" << num_ops[COAL_OP] << endl;
		*mcmc_strm << "recomb ops: " << num_successful_ops[RECOMB_OP] << "/" << num_ops[RECOMB_OP] << endl;
		*mcmc_strm << "scale ops:  " << num_successful_ops[SCALE_OP] << "/" << num_ops[SCALE_OP] << endl;
		*mcmc_strm << "branch ops:  " << num_successful_ops[BRANCH_OP] << "/" << num_ops[BRANCH_OP] << endl;
		*mcmc_strm << "single ops:  " << num_successful_ops[SINGLE_OP] << "/" << num_ops[SINGLE_OP] << endl;
		*mcmc_strm << endl;
	}

	MCMC::Results res (log_prior_prob_max, log_prob_acids_max, log_prob_acids_pos_max);

	OUT_FNCTN_BOTTOM(mcmc_logger);

	return res;
}



// apply MCMC (fixed number of steps)
MCMC::Results MCMC::call(Rev_Tree* tr, const int num_steps) { return MCMC_impl::call(tr, true, num_steps, 0, 0); }

// apply MCMC (variable number of steps)
MCMC::Results MCMC::call(Rev_Tree* tr, const int break_length, const double eps) { return MCMC_impl::call(tr, false, 0, break_length, eps); }

