#include "standard.h"

#include "classftn.h"
#include "cls_changer.h"
#include "crf.h"
#include "dot.h"
#include "files.h"
#include "jpHMM.h"
#include "MCMC.h"
#include "nc.h"
#include "nd.h"
#include "random.h"
#include "sim_tree.h"
#include "tip_seqs.h"
#include "tree.h"

LoggerPtr Classftn::m_logger(Logger::getLogger("classification.classftn"));

using namespace Classftn_impl;



int Classftn::sub_id(int i) const
{
	if (m_tips[i].is_pure)
		return m_tips[i].id;
	else
		throw ERROR(logger(), "Trying to get subtype of CRF tip");
}



int Classftn::crf_id(int i) const
{
	if (!m_tips[i].is_pure)
		return m_tips[i].id;
	else
		throw ERROR(logger(), "Trying to get CRF of pure tip");
}



void Classftn::get_occur_subs(set<int>* occur_subs) const
{
	for (map<int, set<int> >::const_iterator it = m_sub_containing_tips.begin();
	     it != m_sub_containing_tips.end(); ++it)
		occur_subs->insert(it->first);
}



void Classftn::get_occur_crfs(set<int>* occur_crfs) const
{
	for (map<int, set<int> >::const_iterator it = m_crf_containing_tips.begin();
	     it != m_crf_containing_tips.end(); ++it)
		occur_crfs->insert(it->first);
}



// set sub of a pure tip
void Classftn::set_sub_id(int i, int val)
{
	if (is_pure(i))
	{
		m_tips.at(i).id = val;
		recalc_internal();
	}
	else
	{
		ofstream* out_strm = In_Out_Streams::instance()->ostrm(OUTPUT_FILE);
		*out_strm << "Trying to set sub of a CRF tip" << endl;
		abort();
	}
}



// convert a crf tip and set its sub
void Classftn::set_to_pure(int idx, const int val)
{
	if (!is_pure(idx))
	{
		m_tips.at(idx).is_pure = true;
		m_tips.at(idx).id = val;
	}
	else
	{
		ofstream* out_strm = In_Out_Streams::instance()->ostrm(OUTPUT_FILE);
		*out_strm << "Trying to perform conversion to a pure tip of a tip which already is pure" << endl;
		abort();
	}
}



// get an id for a new sub
int Classftn::new_sub_id() const
{
	set<int> s;
	get_occur_subs(&s);
	int i = 0;
	do
		if (s.find(i) == s.end())
			return i;
		else
			++i;
	while(true);
}



// set whether a sample sequence is pure subtype or crf and its sub/crf id
void Classftn::add_tip(BOOL is_pure, int id)
{
	Tip s = { is_pure, id };
	int index = num_tips();
	m_tips.push_back(s);

	if (is_pure)
	{
		if (m_sub_containing_tips.find(id) == m_sub_containing_tips.end())
			m_sub_containing_tips[id] = set<int>();
		m_sub_containing_tips[id].insert(index);
	}
	else
	{
		if (m_crf_containing_tips.find(id) == m_crf_containing_tips.end())
			m_crf_containing_tips[id] = set<int>();
		m_crf_containing_tips[id].insert(index);
	}
}



// resize after sequences have been added
void Classftn::resize_aft_seq()
{
	if (contains_crfs())
	{
		m_bps.resize(num_crfs());
		m_segm_subs.resize(num_crfs());
	}
}



// resize after breakpoints have been added
void Classftn::resize_aft_bp()
{
	if (contains_crfs())
		m_segm_subs.resize(num_crfs());
}



// recalculate internal variables after deletion of CRFs
void Classftn::recalc_internal()
{
	OUT_FNCTN_HEADER(logger());

	// renumber crfs such that their ids are gapless (0, 1, 2, ...)
	set<int> occur_crfs;
	for (VECTOR(Tip)::iterator it = m_tips.begin(); it != m_tips.end(); ++it)
		if (!it->is_pure)
			occur_crfs.insert(it->id);
	map<int, int> mp;
	int i = 0;
	for (set<int>::iterator it = occur_crfs.begin(); it != occur_crfs.end(); ++it, ++i)
		mp[*it] = i;
	for (VECTOR(Tip)::iterator it = m_tips.begin(); it != m_tips.end(); ++it)
		if (!it->is_pure)
			it->id = mp[it->id];

	// recalculate m_crf_containing_tips and m_sub_containing_tips
	m_crf_containing_tips.clear();
	m_sub_containing_tips.clear();
	for (int i = 0; i < num_tips(); ++i)
	{
		if (is_pure(i))
		{
			if (m_sub_containing_tips.find(sub_id(i)) == m_sub_containing_tips.end())
				m_sub_containing_tips[sub_id(i)] = set<int>();
			m_sub_containing_tips[sub_id(i)].insert(i);
		}
		else
		{
			if (m_crf_containing_tips.find(crf_id(i)) == m_crf_containing_tips.end())
				m_crf_containing_tips[crf_id(i)] = set<int>();
			m_crf_containing_tips[crf_id(i)].insert(i);
		}
	}

	OUT_FNCTN_BOTTOM(logger());
}



// add a breakpoint
void Classftn::add_bp(int index, double val)
{
	int i = (int)floor(val * num_sites() - 0.5);
	m_bps[index].push_back(i);
}



// add a subtype of segment
void Classftn::add_segment(int index, int val)
{
	m_segm_subs[index].push_back(val);
}



Classftn::Classftn(const string& filename, int num_sites_para) : m_num_sites(num_sites_para)
{
	OUT_FNCTN_HEADER(logger());
	
	ifstream* cl_strm = new ifstream(filename.c_str());

	LOG4CXX_DEBUG(logger(),	"Parsing: " << filename);
	if (cl_strm && !cl_strm->eof())
	{
		string str;
		getline(*cl_strm, str);
		LOG4CXX_DEBUG(logger(),	str);
		if (str.compare(">seqs") != 0)
			throw ERROR(logger(), "File format error in constructor of Classftn (>seqs)");

		// read sequences
		do
		{
			getline(*cl_strm, str);
			LOG4CXX_DEBUG(logger(),	str);
			if (str[0] != '>')
			{
				stringstream ss(str);
				string str_type;
				int id;
				ss >> str_type;
				ss >> id;
				if (!str_type.compare("P"))
					add_tip(true, id);
				else
					if (!str_type.compare("C"))
						add_tip(false, id);
					else
						throw ERROR(logger(), "File format error in constructor of Classftn (P/C ?)");

			}
		}
		while (!cl_strm->eof() && (str[0] != '>'));
		if (cl_strm->eof())
			throw ERROR(logger(), "File format error in constructor of Classftn (EOF)");

		if (str.compare(">bps") != 0)
			throw ERROR(logger(), "File format error in constructor of Classftn (>bps)");

		// read breakpoints
		resize_aft_seq();
		if (num_crfs() > 0)
		{
			int index = 0;
			do
			{
				getline(*cl_strm, str);
				LOG4CXX_DEBUG(logger(),	str);
				if (str[0] != '>')
				{
					stringstream ss(str);
					while (!ss.eof())
					{
						double x;
						ss >> x;
						add_bp(index, x);
					}
				}
				++index;
			}
			while (!cl_strm->eof() && (str[0] != '>'));
			if (cl_strm->eof())
				throw ERROR(logger(), "File format error in constructor of Classftn (EOF)");
		}

		for (int i = 0; i < num_crfs(); ++i) {
			for (int j = 0; j < num_bps(i) - 1; ++j)
				if (m_bps[i][j] == m_bps[i][j + 1])
					throw ERROR(logger(), "File format error in constructor of Classftn: A segment has length 0");
			if (m_bps[i].back() == num_sites() - 1)
				throw ERROR(logger(), "File format error in constructor of Classftn: Last segment has length 0");
		}

		// read subtypes of segments
		resize_aft_bp();
		if (num_crfs() > 0)
		{
			int index = 0;
			do
			{
				getline(*cl_strm, str);
				LOG4CXX_DEBUG(logger(),	str);
				if (str[0] != '>')
				{
					stringstream ss(str);
					while (!ss.eof())
					{
						int i;
						ss >> i;
						add_segment(index, i);
					}
				}
				++index;
			}
			while (!cl_strm->eof() && (str[0] != '>'));
		}

		for (int i = 0; i < num_crfs(); i++)
			if (m_bps[i].size() + 1 != m_segm_subs[i].size())
				throw ERROR(logger(), "File format error in constructor of Classftn (number of breakpoints -- number of segments)");
	}
	else
		throw ERROR(logger(), "File not found by constructor of Classftn");

	LOG4CXX_DEBUG(logger(),	"Parsing finished");
	OUT_FNCTN_BOTTOM(logger());
}


// generates a Random tree according to the Classftn (tip sequences given)
void Classftn::generate_tree(Rev_Tree* tr, const variant<Tip_Seqs, Empty>& tip_seqs) const
{
	const BOOL OUTPUT_DOT = 0;
	OUT_FNCTN_HEADER(logger());
	if (const Tip_Seqs* ts = get<Tip_Seqs>(&tip_seqs)) {
		LOG4CXX_DEBUG(logger(), *ts);
	}
	else
		LOG4CXX_DEBUG(logger(), "tip_seqs is empty");

	VECTOR(CRF*) crfs(num_crfs());
	for (int i = 0; i < num_crfs(); i++) {
		crfs[i] = new CRF(tr, m_bps[i].size(), m_bps[i], m_segm_subs[i]);
	}
	tr->calc_part_alias();

	for (int i = 0; i < num_tips(); i++)
	{
		LOG4CXX_DEBUG(logger(), "Generating tip " << i);
		int id = is_pure(i) ? sub_id(i) : crf_id(i);

		if (const Tip_Seqs* ts = get<Tip_Seqs>(&tip_seqs))
			new Rev_Tip_NC(tr, ts->genome(i), id, is_pure(i));
		else
			new Rev_Tip_NC(tr, tr->num_sites(), id, is_pure(i));
	}

	tr->set_to_all_tips_added();
	tr->calc_tip_laps();
	if (OUTPUT_DOT)
		Dot::to_dot(tr);

	tr->generate_final();
	tr->log_prob_acids();
	tr->accept_changes();

	if (OUTPUT_DOT)
		Dot::to_dot(tr);

	OUT_FNCTN_BOTTOM(logger());
}



// generates tree according to Classftn, then simulates mutation in order to get tip sequences
tuple<VECTOR( VECTOR(string) ),
	  VECTOR( VECTOR(double) ),
	  VECTOR(double)> Classftn::sim_seqs(int num_sims, BOOL is_optim_para)
{
	const BOOL OUTPUT_DOT = 0;
	const BOOL IS_MUT_RATE_CONST = 0;
	OUT_FNCTN_HEADER(logger());

	Rev_Tree tr(num_sites(), Constants::instance()->num_mutation_rates());
	string str(num_sites(), 'A');
	VECTOR(string) v(num_tips(), str);
	generate_tree(&tr, v);
	tr.log_prior_prob();
	if (Constants::instance()->compile_sim_tree()) {
		Dot::to_dot(&tr, "", Dot::SIM);
		stringstream ss;
		ss << "dot -Tps <" << In_Out_Streams::instance()->filename(SIM_DOT_FILE)
		   << " >" << In_Out_Streams::instance()->filename(SIM_PS_FILE);
		std::system(ss.str().c_str());
	}
	if (OUTPUT_DOT)
		Dot::to_dot(&tr, "SIM");

	Sim_Tree s_tr(tr);
	VECTOR( VECTOR(string) ) seqs;
	for (int i = 0; i < num_sims; ++i) {
		VECTOR(string) v = s_tr.sim_seq(IS_MUT_RATE_CONST);
		seqs.push_back(v);
	}

	// calculate distance between final nodes
	VECTOR(VECTOR(double) ) sep_dist(tr.num_subtypes());
	for (int i = 0; i < tr.num_subtypes(); ++i)
		for (int j = 0; j < tr.num_subtypes(); ++j)
			if (i == j)
				sep_dist[i].push_back(0);
			else {
				Rev_Nd* n = tr.final_node(i);
				while ((!n->occur_subs().count(i) || !n->occur_subs().count(j)) && n)
					n = n->parent_node();
				if (n)
					sep_dist[i].push_back(2 * n->birth_time() - tr.final_node(i)->birth_time()
								          - tr.final_node(j)->birth_time());
				else
					throw ERROR(logger(), "Error during calculation of distance matrix of final nodes");
			}

	// calculate distance between final nodes
	VECTOR(double) noise_dist;
	for (int i = 0; i < tr.num_subtypes(); ++i) {
		Rev_Nd* n = tr.final_node(i);
		noise_dist.push_back(n->birth_time());
	}

	if (is_optim_para)
		optim_para(&s_tr);

	OUT_FNCTN_BOTTOM(logger());

	return make_tuple(seqs, sep_dist, noise_dist);
}



// get the log-prob of the Classftn (not for users)
MCMC::Results Classftn::log_prob_impl(const Tip_Seqs& tip_seqs, BOOL num_steps_fixed, int num_steps,
								     int break_length, double eps) const
{
	OUT_FNCTN_HEADER(logger());
	const BOOL OUTPUT_DOT = 0;

	if (num_sites() != tip_seqs.num_sites())
	{
		stringstream ss;
		ss << "Classification and tip_seqs have different number of sites: "
		   << num_sites() << " " << tip_seqs.num_sites() << endl;
		throw ERROR(logger(), ss.str());
	}
	if (num_tips() != tip_seqs.num_tips())
	{
		stringstream ss;
		ss << "Classification and tip_seqs have different number of tips "
		   << num_tips() << " " << tip_seqs.num_tips() << endl;
		throw ERROR(logger(), ss.str());
	}

	Rev_Tree* tr = new Rev_Tree(m_num_sites, Constants::instance()->num_mutation_rates());
	generate_tree(tr, tip_seqs);

	clock_t t1 = clock();
	MCMC::Results res;
	if (num_steps_fixed)
		res = MCMC::call(tr, num_steps);
	else
		res = MCMC::call(tr, break_length, eps);
	clock_t t2 = clock();
	if (Constants::instance()->messages_on())
		cout << "Time: " << double(t2 - t1) / CLOCKS_PER_SEC << endl;

	ofstream* lap_strm = In_Out_Streams::instance()->ostrm(LOG_ACID_PROB_FILE);
	*lap_strm << res.log_p_acids_pos() << endl;

	if (Constants::instance()->compile_best_tree()) {
		Rev_Tree* max_tr = load_tree();
		Dot::to_dot(max_tr, "", Dot::BEST);
		delete max_tr;

		string str(In_Out_Streams::instance()->filename(BEST_DOT_FILE));
		stringstream ss;
		ss << "dot -Tps <" << In_Out_Streams::instance()->filename(BEST_DOT_FILE)
		   << " >" << In_Out_Streams::instance()->filename(BEST_PS_FILE);
		std::system(ss.str().c_str());
	}
	if (OUTPUT_DOT) {
		Rev_Tree* max_tr = load_tree();
		Dot::to_dot(max_tr, "MAX");
		delete max_tr;
	}

	delete tr;

	OUT_FNCTN_BOTTOM(logger());

	return res;
}



// get the log-prob of the Classftn (fixed number of steps)
MCMC::Results Classftn::log_prob_fix(const Tip_Seqs& tip_seqs, int num_steps) const
{ return log_prob_impl(tip_seqs, true, num_steps, 0, 0); }

// get the log-prob of the Classftn (variable number of steps)
MCMC::Results Classftn::log_prob_var(const Tip_Seqs& tip_seqs, int break_length, double eps) const
{ return log_prob_impl(tip_seqs, false, 0, break_length, eps); }

// get the log-prob of the Classftn (variable number of steps)
MCMC::Results Classftn::log_prob_var(const Tip_Seqs& tip_seqs, double eps) const
{ return log_prob_impl(tip_seqs, false, 0, MCMC_BREAK_LENGTH, eps); }



// get crf containing segments of unknown subtype and the boundaries of these segments
void Classftn::get_unknown(int* crf_containing_unknown, VECTOR(pair<int COMMA int> )* boundaries_unknown)
{
	OUT_FNCTN_HEADER(logger());

	ifstream* in_strm = In_Out_Streams::instance()->istrm(UNKNOWN_SUB_FILE);
	BOOL is_first = true;
	BOOL is_start = true;
	int start = 0;
	boundaries_unknown->clear();
	while (!in_strm->eof()) {
		int i;
		*in_strm >> i;
		if (is_first) {
			*crf_containing_unknown = i;
			is_first = false;
		}
		else
			if (is_start) {
				start = i;
				is_start = false;
			}
			else {
				int end = i;
				boundaries_unknown->push_back(pair<int, int>(start, end));
				is_start = true;
			}
	}

	OUT_FNCTN_BOTTOM(logger());
}



// calculates the breakpoints of the CRFs using jpHMM and remove crfs which have only one subtype (according to jpHMM)
void Classftn::calc_bps(const Tip_Seqs& tip_seqs, BOOL use_unknown_segms)
{
	OUT_FNCTN_HEADER(logger());

	if (contains_crfs()) {
		jphmm::gen_algnmt_file(&tip_seqs, this);

		// read data for unknown subtype
		int crf_unknown;   // CRF containing the unknown subtype
		VECTOR(pair<int COMMA int> ) boundaries_unknown;
		if (use_unknown_segms)
			get_unknown(&crf_unknown, &boundaries_unknown);

		// get breakpoints and segment subtypes from jpHMM
		set<int> occur_crfs;
		get_occur_crfs(&occur_crfs);
		VECTOR(Crf_To_Delete) crfs_to_delete;
		for (set<int>::iterator it = occur_crfs.begin(); it != occur_crfs.end() ; ++it) {
			int idx = *m_crf_containing_tips[*it].begin();
			jphmm jhmm(tip_seqs.genome(idx));
			if (jhmm.is_crf()) {
				m_segm_subs[*it].clear();
				for (int i = 0; i < jhmm.size(); ++i)
					m_segm_subs[*it].push_back(jhmm.sub(i));
				m_bps[*it].clear();
				for (int i = 0; i < jhmm.size() - 1; ++i)
					m_bps[*it].push_back(jhmm.end(i) - 1);
			}
			else
				if (!use_unknown_segms || (*it != crf_unknown)) {
					Crf_To_Delete s = { *it, jhmm.sub() };
					crfs_to_delete.push_back(s);
				}
			LOG4CXX_DEBUG(logger(), "crf " << *it << ":");
			LOG4CXX_DEBUG(logger(), jhmm);
		}

		stringstream ss;
		ss << "crfs to be deleted (target sub): ";
		for (VECTOR(Crf_To_Delete)::iterator it = crfs_to_delete.begin();
			 it != crfs_to_delete.end(); ++it)
			ss << it->crf << "(" << it->sub << ") ";
		LOG4CXX_DEBUG(logger(), ss.str());

		// insert segments belonging to unknown subtype
		if (use_unknown_segms) {
			int no_unknown_sub = m_sub_containing_tips.rbegin()->first + 1;
			m_sub_containing_tips[no_unknown_sub] = set<int>();
			for (VECTOR(pair<int COMMA int> )::iterator it = boundaries_unknown.begin();
				 it != boundaries_unknown.end(); ++it) {
				VECTOR(int)::iterator jt =
					find_if(m_bps[crf_unknown].begin(), m_bps[crf_unknown].end(),
							bind2nd(greater_equal<int>(), it->first));   // first bp after unknown segment
				VECTOR(int)::iterator kt = m_segm_subs[crf_unknown].begin() + (jt - m_bps[crf_unknown].begin()) + 1;   // first segment after unknown segment

				// insert unknown segment
				jt = m_bps[crf_unknown].insert(jt, it->second);
				jt = m_bps[crf_unknown].insert(jt, it->first);
				jt += 2;
				kt = m_segm_subs[crf_unknown].insert(kt, *(kt - 1));
				kt = m_segm_subs[crf_unknown].insert(kt, no_unknown_sub);
				++kt;

				// remove segments erased by unknown segment
				while ((jt != m_bps[crf_unknown].end()) && (*jt <= it->second)) {
					m_bps[crf_unknown].erase(jt);
					m_segm_subs[crf_unknown].erase(kt);
				}
			}
			LOG4CXX_DEBUG(logger(), "crf containing unknown: " << crf_unknown);
			LOG4CXX_DEBUG(logger(), "bps: ");
			FOREACH (int i, m_bps[crf_unknown])
				LOG4CXX_DEBUG(logger(), i);
			LOG4CXX_DEBUG(logger(), "segm_subs: ");
			FOREACH (int i, m_segm_subs[crf_unknown])
				LOG4CXX_DEBUG(logger(), i);
		}

		// delete unnecessary CRFs
		for (VECTOR(Crf_To_Delete)::iterator it = crfs_to_delete.begin(); it != crfs_to_delete.end(); ++it) {
			m_bps.erase(m_bps.begin() + it->crf);
			m_segm_subs.erase(m_segm_subs.begin() + it->crf);
			set<int> s = m_crf_containing_tips[it->crf];
			for (set<int>::iterator jt = s.begin(); jt != s.end(); ++jt)
				set_to_pure(*jt, it->sub);
		}

		if (crfs_to_delete.size() > 0)
			recalc_internal();
	}

	OUT_FNCTN_BOTTOM(logger());
}



// make a change at Classftn
void Classftn::change(const Tip_Seqs& tip_seqs)
{
	OUT_FNCTN_HEADER(logger());

	LOG4CXX_DEBUG(logger(), *this);

	const int NUM_OPS = 2;
	double FRACTION_OPS[NUM_OPS] = { 1, 1 };
	gsl_ran_discrete_t* grd = gsl_ran_discrete_preproc (NUM_OPS, FRACTION_OPS);
	enum_cls_chng_ops chosen_op = (enum_cls_chng_ops)gsl_ran_discrete (Random::rng(), grd);
	gsl_ran_discrete_free(grd);
	//classftn_changer::change(chosen_op, *this);

	classftn_changer::change((enum_cls_chng_ops)0, this);

	LOG4CXX_DEBUG(logger(), *this);

	classftn_changer::change((enum_cls_chng_ops)1, this);

	LOG4CXX_DEBUG(logger(), *this);

	VECTOR(Crf_To_Delete) crfs_to_delete;
	calc_bps(tip_seqs, false);

	LOG4CXX_DEBUG(logger(), *this);

	OUT_FNCTN_BOTTOM(logger());
}



ostream& Classftn_impl::operator<<(ostream& s, const Classftn_impl::Tip& cs)
{
	if (cs.is_pure)
		s << "P ";
	else
		s << "C ";
	s << cs.id << endl;
	return s;
}



ostream& operator<<(ostream& s, const Classftn& cls)
{
	s << "*** CLASSIFICATION ***" << endl;
	s << "num_tips: " << cls.num_tips() << endl;
	s << "num_crfs: " << cls.num_crfs() << endl;
	s << "contains_subs: " << boolalpha << cls.contains_subs() << endl;
	s << "contains_crfs: " << boolalpha << cls.contains_crfs() << endl;

	// output occuring subs
	s << "occuring subs: ";
	set<int> occur_subs;
	cls.get_occur_subs(&occur_subs);
	for (set<int>::iterator it = occur_subs.begin(); it != occur_subs.end(); ++it)
		s << *it << " ";
	s << endl;

	// output occuring crfs
	s << "occuring crfs: ";
	set<int> occur_crfs;
	cls.get_occur_crfs(&occur_crfs);
	for (set<int>::iterator it = occur_crfs.begin(); it != occur_crfs.end(); ++it)
		s << *it << " ";
	s << endl << endl;

	// output which tips belong to which sub
	for (map<int, set<int> >::const_iterator it = cls.m_sub_containing_tips.begin();
	     it != cls.m_sub_containing_tips.end(); ++it)
	{
		s << "sub " << it->first << " contains the tips: ";
		for (set<int>::const_iterator jt = it->second.begin(); jt != it->second.end(); ++jt)
			s << *jt << " ";
		s << endl;
	}
	s << endl;

	// output which tips belong to which CRF
	for (map<int, set<int> >::const_iterator it = cls.m_crf_containing_tips.begin();
	     it != cls.m_crf_containing_tips.end(); ++it)
	{
		s << "crf " << it->first << " contains the tips: ";
		for (set<int>::const_iterator jt = it->second.begin(); jt != it->second.end(); ++jt)
			s << *jt << " ";
		s << endl;
	}
	s << endl;

	// output sequences
	s << "seqs:" << endl;
	int i = 0;
	for (VECTOR(Tip)::const_iterator it = cls.m_tips.begin(); it != cls.m_tips.end(); ++it, ++i)
		s << "tip " << i << ": " << *it;
	s << endl;

	// output breakpoints
	s << "bps:" << endl;
	for (VECTOR( VECTOR(int) )::const_iterator it = cls.m_bps.begin(); it != cls.m_bps.end(); ++it)
	{
		for (VECTOR(int)::const_iterator jt = it->begin(); jt != it->end(); ++jt)
			s << setw(4) << left << *jt << " ";
		s << endl;
	}
	s << endl;

	// output subtypes of segments
	s << "segms:" << endl;
	for (VECTOR( VECTOR(int) )::const_iterator it = cls.m_segm_subs.begin(); it != cls.m_segm_subs.end(); ++it)
	{
		for (VECTOR(int)::const_iterator jt = it->begin(); jt != it->end(); ++jt)
			s << setw(2) << left << *jt << " ";
		s << endl;
	}
	s << endl;

	return s;
}


		/*case 6:
		{
			num_tips = 6;
			num_crfs = 1;

			VECTOR(string) seq_of_tips(num_tips);
			seq_of_tips[0] = string(num_sites, 'A');
			seq_of_tips[1] = string(num_sites, 'C');
			seq_of_tips[2] = string(num_sites / 2, 'A') + string(num_sites / 2, 'C');
			seq_of_tips[3] = string(num_sites / 2, 'A') + string(num_sites / 2, 'C');
			seq_of_tips[4] = string(num_sites / 2, 'A') + string(num_sites / 2, 'C');
			seq_of_tips[5] = string(num_sites / 2, 'A') + string(num_sites / 2, 'C');

			is_pure_tip.resize(num_tips);
			is_pure_tip[0] = true;
			is_pure_tip[1] = true;
			is_pure_tip[2] = false;
			is_pure_tip[3] = false;
			is_pure_tip[4] = false;
			is_pure_tip[5] = false;

			sub_crf.resize(num_tips);
			subtype[0] = 0;
			subtype[1] = 1;
			crf[2] = 0;
			crf[3] = 0;
			crf[4] = 0;
			crf[5] = 0;

			VECTOR( VECTOR(int) ) bp(num_crfs);
			bp[0].resize(1);
			bp[0][0] = num_sites / 2 - 1;

			VECTOR( VECTOR(int) ) segm_subs(num_crfs);
			for (int i = 0; i < num_crfs; i++)
				segm_subs[i].resize(bp[i].size() + 1);
			segm_subs[0][0] = 0;
			segm_subs[0][1] = 1;
			break;
		}*/

