#include "standard.h"

#include "constants.h"
#include "classftn.h"
#include "files.h"
#include "jpHMM.h"
#include "tip_seqs.h"

using namespace jpHMM_impl;

BOOL jphmm::s_alignmt_file_generated = false;
LoggerPtr jphmm::s_logger(Logger::getLogger("classification.jpHMM"));



jphmm::jphmm(const string query_seq_para) : m_query_seq(query_seq_para)
{
	calc_bps();
}



int jphmm::sub() const
{
	if (!is_crf())
		return sub(0);
	else
	{
		ofstream* out_strm = In_Out_Streams::instance()->ostrm(OUTPUT_FILE);
		*out_strm << "Trying to get the subtype of an actual crf" << endl;
		abort();
	}
}



// generates alignment file
void jphmm::gen_algnmt_file(const Tip_Seqs* tip_seqs, const Classftn* cls)
{
	OUT_FNCTN_HEADER(logger());

	// generate alignment file
	ofstream* algnmt_strm = In_Out_Streams::instance()->ostrm(JPHMM_ALIGNMENT_FILE);
	map<int, set<int> > mp = cls->sub_containing_tips();
	for (map<int, set<int> >::const_iterator it = mp.begin(); it != mp.end(); ++it)
	{
		*algnmt_strm << ">>sub" << it->first << endl;
		for (set<int>::const_iterator jt = it->second.begin(); jt != it->second.end(); ++jt)
		{
			*algnmt_strm << ">seq" << *jt << endl;
			*algnmt_strm << tip_seqs->genome(*jt) << endl;
		}
	}

	s_alignmt_file_generated = true;
	OUT_FNCTN_BOTTOM(logger());
}



// calculates the breakpoints of the CRFs using jpHMM
void jphmm::calc_bps()
{
	OUT_FNCTN_HEADER(logger());

	if (!alignmt_file_generated())
		throw ERROR(logger(), "Trying to calculate breakpoints by jpHMM without having generated an alignment file");

	// generate query file
	ofstream* query_strm = In_Out_Streams::instance()->ostrm(JPHMM_QUERY_FILE, true);
	*query_strm << ">seq" << endl;
	*query_strm << m_query_seq << endl;
	LOG4CXX_DEBUG(logger(),	"Query file generated");

	// call jpHMM
	stringstream ss;
	ss << In_Out_Streams::instance()->jphmm_exe_dir_name() << "/jpHMM" << " -i " 
	   << In_Out_Streams::instance()->filename(JPHMM_ALIGNMENT_FILE) << " -s "
	   << In_Out_Streams::instance()->filename(JPHMM_QUERY_FILE) << " -v HIV " 
	   << " -j " << Constants::instance()->jump_prob() << " -X " << Constants::instance()->appl_id();
	if (Constants::instance()->messages_on())
		cout << "Calling " << ss.str() << endl;
	std::system(ss.str().c_str());
	LOG4CXX_DEBUG(logger(),	"jpHMM called\n");

	// parse results of jpHMM
	LOG4CXX_DEBUG(logger(),	"Parsing results of jpHMM");
	In_Out_Streams::instance()->set_jphmm_results_filename(Constants::instance()->appl_id());
	ifstream* results_strm = In_Out_Streams::instance()->istrm(JPHMM_RESULTS_FILE);
	string str;
	do {
		getline(*results_strm, str);
		LOG4CXX_DEBUG(logger(),	str);
	}
	while (!results_strm->eof() && (str[0] == '#'));
	if (results_strm->eof())
		throw ERROR(logger(), "File format error in results of jpHMM");

	getline(*results_strm, str);
	LOG4CXX_DEBUG(logger(),	str);
	if (results_strm->eof() || (str.compare("seq") != 0))
		throw ERROR(logger(), "File format error in results of jpHMM");

	while (!results_strm->eof() && (str.length() > 0))
	{
		getline(*results_strm, str);
		LOG4CXX_DEBUG(logger(),	str);
		stringstream ss(str);
		int begin, end;
		string str_sub;

		ss >> begin;
		ss >> end;
		ss >> str_sub;
		if (str_sub.compare(0, 3, "sub") == 0)
		{
			stringstream ts(str_sub.substr(3));
			int sub;
			ts >> sub;
			Segm segm = {begin, end, sub, true};
			m_segments.push_back(segm);
		}
		else
		{
			if ((str_sub.compare("3'-Insertion") == 0) || (str_sub.compare("5'-Insertion") == 0))
			{
				Segm segm = {begin, end, 0, false};
				m_segments.push_back(segm);
			}
		}
	}

	LOG4CXX_DEBUG(logger(),	"Results of jpHMM parsed\n");

	// take care of undefined m_segments
	for (int i = m_segments.size() - 1; i >= 0; --i)
		if (!m_segments[i].is_defined)
			if ((i < m_segments.size() - 1) && m_segments[i + 1].is_defined) {
				m_segments[i + 1].begin = m_segments[i].begin;
				m_segments.erase(m_segments.begin() + i);
			}
			else
				if ((i > 0) && m_segments[i - 1].is_defined) {
					m_segments[i].end = m_segments[i - 1].end;
					m_segments.erase(m_segments.begin() + i);
				}

	OUT_FNCTN_BOTTOM(logger());
}



ostream& operator<<(ostream& s, const jphmm& jhmm)
{
	s << "*** JPHMM ***" << endl;
	s << "m_segments: " << endl;
	for (VECTOR(Segm)::const_iterator it = jhmm.m_segments.begin(); it != jhmm.m_segments.end(); ++it)
		s << it->begin << " " << it->end << " " << it->sub << " " << boolalpha << it->is_defined << endl;
	s << "is_crf: " << boolalpha << jhmm.is_crf() << endl;

	return s;
}

