#include "standard.h"

#include "files.h"
#include "part_node.h"
#include "part_tree.h"

LoggerPtr Partition_Node::s_logger(Logger::getLogger("crf.part_node"));



ostream& operator<<(ostream& s, const Suc_Rec& suc)
{
	s << "target id: " << suc.target->id() << endl;
	s << "source subs: " << suc.source_subs << endl;
	s << "first target subs: " << suc.first_target_subs << endl;
	s << "second target subs: " << suc.second_target_subs << endl;
	return s;
}



Partition_Node::Partition_Node(const set<int>& coal_subs, const set<int>& sep_subs, const set< set<int> >& grouped_subs,
							   int level, BOOL is_first)
: m_subs(Subs(coal_subs, sep_subs, grouped_subs)), m_level(level)
{
	if (is_first)
		m_id = 0;
}



void Partition_Node::generate_suc(Partition_Tree* tr)
{
	OUT_FNCTN_HEADER(logger());

	// generation of node by coalescent event
	for (set<int>::iterator it = m_subs.sep_subs.begin(); it != m_subs.sep_subs.end(); it++)
		if (!tr->sub_only_in_crfs(*it)) {
			Subs subs_new = m_subs;
			subs_new.sep_subs.erase(*it);
			subs_new.coal_subs.insert(*it);

			Partition_Node* n = tr->find_node(subs_new);
			if (n == NULL)
			{
				n = new Partition_Node(subs_new, level() + 1);
				tr->m_part_nodes.push_back(n);
				n->m_id = tr->m_part_nodes.size() - 1;
			}
			Suc_Coal s = {n, *it};
			m_suc_coal.insert(s);
			n->generate_suc(tr);
		}

	// generation of node by recombination event
	for (set< set<int> >::iterator it = m_subs.grouped_subs.begin(); it != m_subs.grouped_subs.end(); it++)
	{
		VECTOR(int) mask(it->size(), 0);
		mask[0] = 1;
		do
		{
			set<int> first_set;
			set<int> second_set;

			int j = 0;
			for (set<int>::iterator jt = it->begin(); jt != it->end(); jt++, j++)
			{
				if (mask[j])
					first_set.insert(*jt);
				else
					second_set.insert(*jt);
			}

			Subs subs_new = m_subs;
			subs_new.grouped_subs.erase(*it);

			if (first_set.size() == 1)
				subs_new.sep_subs.insert(*first_set.begin());
			else
				subs_new.grouped_subs.insert(first_set);
			if (second_set.size() == 1)
				subs_new.sep_subs.insert(*second_set.begin());
			else
				subs_new.grouped_subs.insert(second_set);

			Partition_Node* n = tr->find_node(subs_new);
			if (n == NULL)
			{
				n = new Partition_Node(subs_new, level() + 1);
				tr->m_part_nodes.push_back(n);
				n->m_id = tr->m_part_nodes.size() - 1;
			}
			Suc_Rec s = {n, *it, first_set, second_set};
			m_suc_recomb.insert(s);
			n->generate_suc(tr);

			int i;
			for (i = 0; (i < mask.size()) && mask[i]; i++)
				mask[i] = 0;
			mask[i] = 1;
		}
		while (mask[mask.size() - 1] == 0);
	}

	OUT_FNCTN_BOTTOM(logger());
}


// get all direct and indirect successors
void Partition_Node::get_all_sucs(set<Partition_Node*>& all_sucs) const
{
	for (set<Suc_Coal>::const_iterator it = suc_coal().begin(); it != suc_coal().end(); it++)
	{
		all_sucs.insert(it->target);
		it->target->get_all_sucs(all_sucs);
	}
	for (set<Suc_Rec>::const_iterator it = suc_recomb().begin(); it != suc_recomb().end(); it++)
	{
		all_sucs.insert(it->target);
		it->target->get_all_sucs(all_sucs);
	}
}



// convert to string
string Partition_Node::to_string() const
{
	stringstream ss;

	ss << "no. " << m_id << "\\n";

	BOOL is_first = true;
	for (set<int>::iterator it = m_subs.coal_subs.begin(); it != m_subs.coal_subs.end(); it++)
	{
		if (!is_first)
			ss << " ";
		ss << *it;
		is_first = false;
	}
	if (m_subs.coal_subs.size() == 0)
		ss << "-";
	ss << "\\n";

	is_first = true;
	for (set<int>::iterator it = m_subs.sep_subs.begin(); it != m_subs.sep_subs.end(); it++)
	{
		if (!is_first)
			ss << " ";
		ss << *it;
		is_first = false;
	}
	if (m_subs.sep_subs.size() == 0)
		ss << "-";
	ss << "\\n";

	BOOL is_first_in_list = true;
	for (set< set<int> >::const_iterator it = m_subs.grouped_subs.begin(); it != m_subs.grouped_subs.end(); it++)
	{
		if (!is_first_in_list)
			ss << "  ";
		BOOL is_first_in_set = true;
		for (set<int>::iterator jt = it->begin(); jt != it->end(); jt++)
		{
			if (!is_first_in_set)
				ss << "";
			ss << *jt;
			is_first_in_set = false;
		}
		is_first_in_list = false;
	}
	if (m_subs.grouped_subs.size() == 0)
		ss << "-";
	ss << "\\n";

	return ss.str();
}



// get all connections from this node to all direct and indirect successors
void Partition_Node::conns(VECTOR(Part_Node_Conn)* conns, Part_Node_Conn* curr_conn, int req_length) const
{
	OUT_FNCTN_HEADER(logger());

	LOG4CXX_DEBUG(logger(), *this);
	LOG4CXX_DEBUG(logger(), "curr_conn:");
	LOG4CXX_DEBUG(logger(), *curr_conn);

	for (set<Suc_Coal>::const_iterator it = suc_coal().begin(); it != suc_coal().end(); it++)
	{
		curr_conn->push_back(it->target->m_id, it->source_sub);
		if (curr_conn->size() == req_length)
			conns->push_back(*curr_conn);
		it->target->conns(conns, curr_conn, req_length);
		curr_conn->pop_back();
	}

	for (set<Suc_Rec>::const_iterator it = suc_recomb().begin(); it != suc_recomb().end(); it++)
	{
		curr_conn->push_back(it->target->m_id, false, it->source_subs, it->first_target_subs, it->second_target_subs);
		if (curr_conn->size() == req_length)
			conns->push_back(*curr_conn);
		it->target->conns(conns, curr_conn, req_length);
		curr_conn->pop_back();
	}

	OUT_FNCTN_BOTTOM(logger());
}



BOOL Partition_Node::operator==(Subs subs_para)
{
	return (m_subs.coal_subs == subs_para.coal_subs) && (m_subs.sep_subs == subs_para.sep_subs) &&
	       (m_subs.grouped_subs == subs_para.grouped_subs);
}



// output partition node
ostream& operator<<(ostream& s, const Partition_Node& n)
{
	s << "*** PARTITION_NODE ***" << endl;
	s << "id: " << n.id() << endl;

	s << "coal_subs: ";
	for (set<int>::iterator it = n.subs().coal_subs.begin(); it != n.subs().coal_subs.end(); it++)
		s << *it << " ";
	s << endl;

	s << "sep_subs: ";
	for (set<int>::iterator it = n.subs().sep_subs.begin(); it != n.subs().sep_subs.end(); it++)
		s << *it << " ";
	s << endl;

	s << "group_subs: ";
	for (set< set<int> >::iterator it = n.subs().grouped_subs.begin(); it != n.subs().grouped_subs.end(); it++)
	{
		for (set<int>::iterator jt = it->begin(); jt != it->end(); jt++)
			s << *jt << " ";
		s << "  ";
	}
	s << endl;

	s << "suc_coal: ";
	for (set<Suc_Coal>::const_iterator it = n.suc_coal().begin();
		 it != n.suc_coal().end(); it++)
		s << it->target->id() << " ";
	s << endl;

	s << "suc_recomb: ";
	for (set<Suc_Rec>::const_iterator it = n.suc_recomb().begin();
		 it != n.suc_recomb().end(); it++)
		s << it->target->id() << " ";
	s << endl;

	return s;
}



BOOL operator<(const Suc_Coal& s, const Suc_Coal& t)
{
	return s.target->id() < t.target->id();
}



BOOL operator<(const Suc_Rec& s, const Suc_Rec& t)
{
	return s.target->id() < t.target->id();
}



