#include "standard.h"

#include "nd.h"
#include "part_node.h"
#include "part_tree.h"
#include "prt_nd_conn.h"

LoggerPtr Partition_Tree::s_logger(Logger::getLogger("crf.part_tree"));



Partition_Tree::Partition_Tree(const set<int>& occur_subs, const set<int>& subs_only_in_crfs)
{
	OUT_FNCTN_HEADER(logger());

	set< set<int> > grouped_occur_subs;
	grouped_occur_subs.insert(occur_subs);
	Partition_Node* n = new Partition_Node(set<int>(), set<int>(), grouped_occur_subs, 0, true);
	m_part_nodes.push_back(n);
	m_subs_only_in_crf = set<int>(subs_only_in_crfs.begin(), subs_only_in_crfs.end());
	root()->generate_suc(this);

	// find last node
	int max = 0;
	int argmax = 0;
	for (VECTOR(Partition_Node*)::iterator it = m_part_nodes.begin(); it != m_part_nodes.end(); it++)
		if ((*it)->level() >= max)
		{
			max = (*it)->level();
			argmax = (*it)->id();
		}
	m_last = m_part_nodes[argmax];

	// calculate all direct and indirect successor of the nodes
	m_all_sucs.resize(num_part_nodes());
	for (int i = 0; i < num_part_nodes(); i++)
	{
		set<Partition_Node*> s;
		m_part_nodes[i]->get_all_sucs(s);
		for (set<Partition_Node*>::iterator it = s.begin(); it != s.end(); it++)
			m_all_sucs[i].push_back((*it)->id());
	}

	m_max_level = 2 * occur_subs.size() - 1 - m_subs_only_in_crf.size();

	// calculate all possible connections between first and last node
	Part_Node_Conn c;
	m_part_nodes[0]->conns(&m_conns, &c, max_level());

	OUT_FNCTN_BOTTOM(logger());
}



Partition_Tree::~Partition_Tree()
{
	OUT_FNCTN_HEADER(logger());
	for (VECTOR(Partition_Node*)::iterator it = m_part_nodes.begin(); it != m_part_nodes.end(); it++)
		delete *it;
	OUT_FNCTN_BOTTOM(logger());
}



Partition_Node* Partition_Tree::find_node(Subs subs)
{
	Partition_Node* n = 0;
	for (VECTOR(Partition_Node*)::iterator it = m_part_nodes.begin(); it != m_part_nodes.end(); it++)
	{
		if (**it == subs)
			n = *it;
	}
	return n;
}


// setting node of a set of subtypes taking into account that a recomb node has to be assigned to a 1-element set
void Partition_Tree::set_node(const set<int>& s, Rev_Nd* n)
{
	if (s.size() > 1)
		m_nodes_grouped[s] = n;
	else
		if (Rev_Rec_Nd* m = dynamic_cast<Rev_Rec_Nd*>(n))
			ref_node(*s.begin()) = m;
		else
			throw Error("Trying to assign a node not of type Rev_Rec_Nd to m_nodes_sep");
}



void Partition_Tree::set_root(Rev_Nd* n)
{
	set_node(*m_part_nodes[0]->subs().grouped_subs.begin(), n);
}



Rev_Nd* Partition_Tree::node(const set<int>& s) const
{
	if (s.size() > 1)
		return m_nodes_grouped.find(s)->second;
	else
		return m_nodes_sep.find(*s.begin())->second;
}



ostream& operator<<(ostream& s, const Partition_Tree& tr)
{
	s << "*** PARTITION_TREE ***" << endl;
	s << "number of nodes: " << tr.num_part_nodes() << endl;

	// print nodes
	s << "nodes:" << endl;
	for (VECTOR(Partition_Node*)::const_iterator it = tr.part_nodes().begin(); it != tr.part_nodes().end(); it++)
	{
		s << **it;
	}
	s << endl;

	// check integrity of Rev_Nd ids
	for (int i = 0; i < tr.num_part_nodes(); i++)
	{
		if (i != tr.part_nodes()[i]->id())
			s << "data error: Rev_Nd has position " << tr.part_nodes()[i]->id()
			   << "in m_part_nodes of Partition_Tree, but id " << i << endl;
	}

	// print all successors of each Rev_Nd
	s << "succesors:" << endl;
	for (int i = 0; i < tr.num_part_nodes(); i++)
	{
		s << "source: " << setw(3) << i << "  sucs: ";
		for (VECTOR(int)::const_iterator jt = tr.all_sucs(i).begin(); jt != tr.all_sucs(i).end(); jt++)
		{
			s << *jt << " ";
		}
		s << endl;
	}
	s << endl;

	// print all connections from one Rev_Nd to another
	s << "connections:" << endl;
	for (VECTOR(Part_Node_Conn)::const_iterator kt = tr.conns().begin();
	     kt != tr.conns().end(); kt++)
	{
		s << *kt;
	}
	s << endl;

	return s;
}

