#include "standard.h"

#include "alphabet.h"
#include "error.h"

LoggerPtr Alphabet::s_logger(Logger::getLogger("IO.tip_seqs"));



// initialize with characters
void Alphabet::init(VECTOR(char) characters)
{
	m_characters = characters;
	for (VECTOR(char)::iterator it = m_characters.begin(); it != m_characters.end(); ++it) {
		m_char_to_idx[*it] = it - m_characters.begin();
	}
	m_is_init = true;
}



// convert index (int) to nucleotide (char)
char Alphabet::character(int i) const
{
	CHECK_INTGRTY;
	try {
		return m_characters[i];
	}
	catch (out_of_range) {
		throw ERROR(logger(), "Trying to get character for illegal index");
	}
}



// convert vector of indices (int) to sequence string
string Alphabet::characters(VECTOR(int) v) const
{
	CHECK_INTGRTY;
	string s;
	for (VECTOR(int)::iterator it = v.begin(); it != v.end(); ++it)
		s = s + character(*it);
	return s;
}



// convert nucleotide (char) to vector of indices (int)
int Alphabet::idx(char c) const
{
	CHECK_INTGRTY;
	map<char, int>::const_iterator it = m_char_to_idx.find(c);
	if (it == m_char_to_idx.end()) {
		stringstream ss;
		ss << "Trying to get index of character not in the alphabet: " << c;
		throw ERROR(logger(), ss.str());
	}
	return it->second;
}



// convert sequence string to vector of indices (int)
VECTOR(int) Alphabet::indices(string s) const
{
	CHECK_INTGRTY;
	VECTOR(int) v;
	for (string::iterator it = s.begin(); it != s.end(); ++it)
		v.push_back(idx(*it));
	return v;
}



// check whether vector of indices only contains legal values
BOOL Alphabet::is_legal_indices(VECTOR(int) v) const
{
	for (VECTOR(int)::iterator it = v.begin(); it != v.end(); ++it)
		if (!is_legal_idx(*it))
			return false;
	return true;
}



// check integrity
void Alphabet::check() const
{
	if (!m_is_init)
		throw ERROR(logger(), "Alphabet in illegal state");
}



ostream& operator<<(ostream& s, const Alphabet& al)
{
	s << "characters: " << al.m_characters << endl;
	s << "char_to_idx: " << al.m_char_to_idx << endl;
	s << "is_init: " << boolalpha << al.m_is_init << endl;
	return s;
}


