#ifndef BASE_NODE_CONTR_H
#define BASE_NODE_CONTR_H

#include "standard.h"

using namespace std;


struct Dot_Para;
class Rev_Tree;
class Changes;
class S_and_P;
template<class N, class C, class PC, class T> class Base_NC;
template<class T> class Bound_Det_Manip;
template<class N, class C, class PC, class T>
ostream& operator<<(ostream& s, const Bound_Det_Manip<Base_NC<N, C, PC, T> >& bdm);



template<class N, class C, class PC, class T> class Base_NC : private noncopyable
{
	friend ostream& operator<<<N, C, PC, T>(ostream& s, const Bound_Det_Manip<Base_NC<N, C, PC, T> >& bdm);
private:
	T* m_tree;
	double m_birth_time;

	// order
	C* m_next;
	C* m_prev;

	// prob
	double m_coal_intensity, m_recomb_intensity;
	double m_P_prior, m_P_gen_event;
	int m_num_elig_nodes;
	int m_num_cum_elig_sites;

	static LoggerPtr m_logger;
protected:
	Base_NC() { }
	Base_NC(T* tree);
	Base_NC(T* tree, double birth_time, N* first_old_node, N* second_old_node);
	Base_NC(T* tree, double birth_time, N* old_node);
	virtual ~Base_NC() { }
	static LoggerPtr& logger() { return m_logger; }

	double& ref_birth_time() { return m_birth_time; }

	// prob
	double& ref_coal_intensity() { return m_coal_intensity; }
	double& ref_recomb_intensity() { return m_recomb_intensity; }
	double& ref_P_prior() { return m_P_prior; }
	double& ref_P_gen_event() { return m_P_gen_event; }
	int& ref_num_elig_nodes() { return m_num_elig_nodes; }
	int& ref_num_cum_elig_sites() { return m_num_cum_elig_sites; }
public:
	T* tree() const { return m_tree; }

	double birth_time() const { return m_birth_time; }
	virtual VECTOR(N*) unspec_nodes() const = 0;
	virtual BOOL is_single() const = 0;
	virtual BOOL is_double() const = 0;

	// anc
	virtual PC* youngest_parent_contnr() const = 0;

	// suc
	virtual VECTOR(N*) unspec_children() const = 0;
	virtual BOOL has_coal_suc() const = 0;
	virtual BOOL has_rec_suc() const = 0;
	virtual BOOL is_tip() const = 0;

	// order
	C* next() const { return m_next; }
	C* prev() const { return m_prev; }
	C*& next() { return m_next; }
	C*& prev() { return m_prev; }
	BOOL is_first() const { return prev() == 0; }
	BOOL is_last() const { return next() == 0; }
	virtual void connect_prev_and_next() = 0;

	// prob
	int num_elig_nodes() const { return m_num_elig_nodes; }
	int num_cum_elig_sites() const { return m_num_cum_elig_sites; }
	virtual int num_elig_sites() const = 0;
	double coal_intensity() const { return m_coal_intensity; }
	double recomb_intensity() const { return m_recomb_intensity; }
	double P_prior() const { return m_P_prior; }
	double P_gen_event() const { return m_P_gen_event; }
	virtual void calc_log_prior_prob() = 0;
	virtual void calc_num_elig_nd_st(int* curr_num_elig_sites, int* curr_num_elig_nodes) = 0;

	// output
	virtual string str_id() const = 0;
	virtual string to_dot_node_name() const = 0;
	virtual string to_dot_string(Dot_Para dot_para) const = 0;

	virtual void check() const;

	BOOL operator==(const Base_NC& c) const { return str_id() == c.str_id(); }
	BOOL operator!=(const Base_NC& c) const { return !operator==(c); }
};



template<class N, class ON, class C, class PC, class T> class NC_with_Nds
: public virtual Base_NC<N, C, PC, T>
{
private:
public:
	virtual VECTOR(ON*) nodes() const = 0;
	void check() const { Base_NC<N, C, PC, T>::check(); }
};



template<class N, class ON, class C, class PC, class T> class NC_with_1_Nd
: public virtual NC_with_Nds<N, ON, C, PC, T>
{
private:
	ON* m_node;
	BOOL m_is_init;
protected:
	void init_node(ON* node)  { m_node = node; m_is_init = true; }
public:
	~NC_with_1_Nd() { checked_delete(m_node); }

	using Base_NC<N, C, PC, T>::logger;

	ON* node() const { return m_node; }
	VECTOR(N*) unspec_nodes() const;
	VECTOR(ON*) nodes() const;
	PC* parent_contnr() const { return node()->parent_contnr(); }
	PC* youngest_parent_contnr() const;
	BOOL has_rec_suc() const { return false; }
	BOOL is_single() const { return true; }
	BOOL is_double() const { return false; }
	int num_elig_sites() const { return node()->num_elig_sites(); }
	string str_id() const;
	string to_dot_node_name() const;

	void check() const;
};



template<class N, class ON, class C, class PC, class T> class NC_with_2_Nds
: public virtual NC_with_Nds<N, ON, C, PC, T>
{
private:
	pair<ON*, ON*> m_nodes;
	BOOL m_is_init;
protected:
	void init_nodes(ON* first_node, ON* second_node)  { m_nodes = make_pair(first_node, second_node); m_is_init = true; }
public:
	~NC_with_2_Nds() { checked_delete(m_nodes.first); checked_delete(m_nodes.second); }

	using Base_NC<N, C, PC, T>::logger;

	ON* first_node() const { return m_nodes.first; }
	ON* second_node() const { return m_nodes.second; }
	VECTOR(N*) unspec_nodes() const;
	VECTOR(ON*) nodes() const;
	PC* youngest_parent_contnr() const;
	BOOL has_coal_suc() const { return false; }
	BOOL has_rec_suc() const { return true; }
	BOOL is_tip() const { return false; }
	BOOL is_single() const { return false; }
	BOOL is_double() const { return true; }
	int num_elig_sites() const { return first_node()->num_elig_sites() + second_node()->num_elig_sites(); }
	string str_id() const;
	string to_dot_node_name() const;

	void check() const;
};



template<class N, class ON, class C, class PC, class T> class Tip_NC : public virtual NC_with_1_Nd<N, ON, C, PC, T>
{
private:
protected:
	Tip_NC();
public:
	VECTOR(N*) unspec_children() const { return VECTOR(N*)(); };
	BOOL has_coal_suc() const { return false; }
	BOOL is_tip() const { return true; }
	void calc_log_prior_prob() { Base_NC<N, C, PC, T>::ref_P_prior() = 1; Base_NC<N, C, PC, T>::ref_P_gen_event() = 1; }
	void calc_num_elig_nd_st(int* curr_num_elig_sites, int* curr_num_elig_nodes);
	void check() const { NC_with_1_Nd<N, ON, C, PC, T>::check(); }
};



template<class C> class Intern_NC
{
private:
public:
	virtual C* oldest_child_contnr() const = 0;
};



template<class N, class ON, class C, class CC, class PC, class T> class Coal_NC
: public virtual Intern_NC<C>, public virtual NC_with_1_Nd<N, ON, C, PC, T>
{
private:
	pair<N*, N*> m_children;
protected:
	pair<N*, N*> children() const { return m_children; }
	void connect_to_suc(const double birth_time, CC* node_contnr, N* first_old_node, N* second_old_node);
public:
	Coal_NC(double birth_time, CC* node_contnr, N* first_old_node, N* second_old_node);

	using Base_NC<N, C, PC, T>::logger;
	using Base_NC<N, C, PC, T>::str_id;

	N* first_child() const { return m_children.first; }
	N*& first_child() { return m_children.first; }
	N* second_child() const { return m_children.second; }
	N*& second_child() { return m_children.second; }
	C* oldest_child_contnr() const;
	VECTOR(N*) unspec_children() const;
	BOOL has_coal_suc() const { return true; }
	BOOL is_tip() const { return false; }
	void calc_log_prior_prob();
	void calc_num_elig_nd_st(int* curr_num_elig_sites, int* curr_num_elig_nodes);
	void check() const;
};



template<class N, class ON, class C, class RC, class PC, class T> class Rec_NC
: public virtual Intern_NC<C>, public virtual NC_with_2_Nds<N, ON, C, PC, T>
{
private:
	N* m_child;
protected:
	void connect_to_suc(double birth_time, RC* node_contnr, N* old_node);
public:
	Rec_NC(double birth_time, RC* node_contnr, N* old_node);

	using Base_NC<N, C, PC, T>::logger;
	using Base_NC<N, C, PC, T>::str_id;

	N* child() const { return m_child; }
	N*& child() { return m_child; }
	C* oldest_child_contnr() const { return child()->unspec_node_contnr(); }
	VECTOR(N*) unspec_children() const;
	void calc_log_prior_prob();
	void calc_num_elig_nd_st(int* curr_num_elig_sites, int* curr_num_elig_nodes);
	void check() const;
};

#endif
