#ifndef BASE_NC_DEF_H_
#define BASE_NC_DEF_H_

#include "standard.h"

#include "base_nc.h"
#include "constants.h"
#include "dot.h"
#include "error.h"
#include "files.h"

template<class N, class C, class PC, class T> LoggerPtr Base_NC<N, C, PC, T>::m_logger(Logger::getLogger("node.base_nc"));



// *****************************************************************************
// * Base_NC
// *****************************************************************************

template<class N, class C, class PC, class T>
Base_NC<N, C, PC, T>::Base_NC(T* tr)
: m_tree(tr), m_birth_time(0), m_num_elig_nodes(0), m_num_cum_elig_sites(0)
{
	OUT_FNCTN_HEADER(logger());
	OUT_FNCTN_BOTTOM(logger());
}


template<class N, class C, class PC, class T>
Base_NC<N, C, PC, T>::Base_NC(T* tr, double birth_time, N* first_old_node, N* second_old_node)
: m_tree(tr), m_birth_time(birth_time), m_num_elig_nodes(0), m_num_cum_elig_sites(0)
{
	OUT_FNCTN_HEADER(logger());
	LOG4CXX_DEBUG(logger(),	"birth_time: " << birth_time);
	LOG4CXX_DEBUG(logger(),	"first_old_node: " << first_old_node->id());
	LOG4CXX_DEBUG(logger(),	"second_old_node: " << second_old_node->id() << endl);

	if (first_old_node->is_active())
		tree()->set_num_act_nodes(tree()->num_act_nodes() - 1);
	if (second_old_node->is_active())
		tree()->set_num_act_nodes(tree()->num_act_nodes() - 1);

	OUT_FNCTN_BOTTOM(logger());
}



template<class N, class C, class PC, class T>
Base_NC<N, C, PC, T>::Base_NC(T* tr, double birth_time, N* old_node)
: m_tree(tr), m_birth_time(birth_time), m_num_elig_nodes(0), m_num_cum_elig_sites(0)
{
	OUT_FNCTN_HEADER(logger());
	LOG4CXX_DEBUG(logger(), "birth_time: " << birth_time);
	LOG4CXX_DEBUG(logger(), "old_node: " << old_node->id() << endl);

	if (old_node->is_active())
		tree()->set_num_act_nodes(tree()->num_act_nodes() - 1);

	OUT_FNCTN_BOTTOM(logger());
}



// check integrity
template<class N, class C, class PC, class T> void Base_NC<N, C, PC, T>::check() const
{
	BOOL is_ok = true;
	ostringstream oss;

	if (!tree() || birth_time() < 0) {
		is_ok = false;
		oss << "Base_NC in illegal state";
	}

	// next and prev ok?
	if ((tree()->first_node_contnr() != this) && !prev() ||
		(tree()->last_node_contnr() != this) && !next()) {
		is_ok = false;
		oss << "next or prev is NULL";
	}

	// check integrity of position and birth time
	if (next() && (birth_time() > next()->birth_time()))
	{
		is_ok = false;
		oss << "time of generation inconsistent (pos): " << str_id() << " (" << birth_time() << ")"
		   << " -> " << next()->str_id() << " (" << next()->birth_time() << ")" << endl;
	}
	if (prev() && (birth_time() < prev()->birth_time()))
	{
		is_ok = false;
		oss << "time of generation inconsistent (pos): " << str_id() << " (" << birth_time() << ")"
		   << " -> " << prev()->str_id() << " (" << prev()->birth_time() << ")" << endl;
	}

	if (!is_ok)
		throw ERROR(logger(), oss.str());
}



// *****************************************************************************
// * NC_with_1_Nd
// *****************************************************************************

template<class N, class ON, class C, class PC, class T> VECTOR(N*) NC_with_1_Nd<N, ON, C, PC, T>::unspec_nodes() const
{
	VECTOR(N*) v;
	v.push_back(node());
	return v;
}



template<class N, class ON, class C, class PC, class T> VECTOR(ON*) NC_with_1_Nd<N, ON, C, PC, T>::nodes() const
{
	VECTOR(ON*) v;
	v.push_back(node());
	return v;
}



template<class N, class ON, class C, class PC, class T> PC* NC_with_1_Nd<N, ON, C, PC, T>::youngest_parent_contnr() const
{
	if (!node()->is_active())
		return node()->parent_contnr();
	else
		throw Error("Trying to get the youngest parent container of a node container not having one");
}



// check integrity
template<class N, class ON, class C, class PC, class T> void NC_with_1_Nd<N, ON, C, PC, T>::check() const
{
	NC_with_Nds<N, ON, C, PC, T>::check();

	BOOL is_ok = true;
	stringstream oss;
	oss << "Node container " << str_id() << ": ";

	if (!m_is_init) {
		is_ok = false;
		oss << "NC_with_1_Nd in illegal state" << endl;
	}

	// is node ok?
	if (!node()->node_contnr() || (node()->node_contnr()->str_id() != str_id())) {
		is_ok = false;
		oss << "node is NULL or wrong" << endl;
	}

	if (!is_ok)
		throw ERROR(logger(), oss.str());
}



template<class N, class ON, class C, class PC, class T> string NC_with_1_Nd<N, ON, C, PC, T>::str_id() const
{
	stringstream ss;
	ss << node()->id();
	return ss.str();
}



template<class N, class ON, class C, class PC, class T> string NC_with_1_Nd<N, ON, C, PC, T>::to_dot_node_name() const
{
	stringstream ss;
	ss << "C_" << node()->id();
	return ss.str();
}



// *****************************************************************************
// * NC_with_2_Nds
// *****************************************************************************

template<class N, class ON, class C, class PC, class T> VECTOR(N*) NC_with_2_Nds<N, ON, C, PC, T>::unspec_nodes() const
{
	VECTOR(N*) v;
	v.push_back(first_node());
	v.push_back(second_node());
	return v;
}



template<class N, class ON, class C, class PC, class T> VECTOR(ON*) NC_with_2_Nds<N, ON, C, PC, T>::nodes() const
{
	VECTOR(ON*) v;
	v.push_back(first_node());
	v.push_back(second_node());
	return v;
}



template<class N, class ON, class C, class PC, class T> PC* NC_with_2_Nds<N, ON, C, PC, T>::youngest_parent_contnr() const
{
	PC* c = NULL;
	if (!first_node()->is_active())
		c = first_node()->parent_contnr();
	if (!second_node()->is_active() &&
		((c == NULL) || (second_node()->time_of_parent() < first_node()->time_of_parent())))
		c = second_node()->parent_contnr();

	if (c == NULL)
		throw Error("Trying to get the youngest parent container of a node container not having one");
	else
		return c;
}



// check integrity
template<class N, class ON, class C, class PC, class T> void NC_with_2_Nds<N, ON, C, PC, T>::check() const
{
	NC_with_Nds<N, ON, C, PC, T>::check();
	BOOL is_ok = true;
	stringstream oss;

	oss << "Node container " << str_id() << ": ";

	if (!m_is_init) {
		is_ok = false;
		oss << "NC_with_2_Nds in illegal state" << endl;
	}

	// are nodes ok?
	if (!first_node()->node_contnr() || (first_node()->node_contnr()->str_id() != str_id()) ||
		!second_node()->node_contnr() || (second_node()->node_contnr()->str_id() != str_id()) ||
		!first_node()->is_first_in_NC() || second_node()->is_first_in_NC()) {
		is_ok = false;
		oss << "nodes are NULL or wrong" << endl;
	}

	if (!is_ok)
		throw ERROR(logger(), oss.str());
}



template<class N, class ON, class C, class PC, class T> string NC_with_2_Nds<N, ON, C, PC, T>::str_id() const
{
	stringstream ss;
	ss << first_node()->id() << "_" << second_node()->id();
	return ss.str();
}



template<class N, class ON, class C, class PC, class T> string NC_with_2_Nds<N, ON, C, PC, T>::to_dot_node_name() const
{
	stringstream ss;
	ss << "C_" << first_node()->id() << "_" << second_node()->id();
	return ss.str();
}



// *****************************************************************************
// * Tip_NC
// *****************************************************************************

template<class N, class ON, class C, class PC, class T> Tip_NC<N, ON, C, PC, T>::Tip_NC()
{
	Base_NC<N, C, PC, T>::tree()->m_num_tips++;
}



// calculates number of eligible nodes and number of cumulative eligible sites
// input provides data about previous node
template<class N, class ON, class C, class PC, class T>
void Tip_NC<N, ON, C, PC, T>::calc_num_elig_nd_st(int* curr_num_elig_sites, int* curr_num_elig_nodes)
{
	(*curr_num_elig_nodes)++;
	*curr_num_elig_sites += NC_with_1_Nd<N, ON, C, PC, T>::num_elig_sites();
	Base_NC<N, C, PC, T>::ref_num_elig_nodes() = *curr_num_elig_nodes;
	Base_NC<N, C, PC, T>::ref_num_cum_elig_sites() = *curr_num_elig_sites;
}



// *****************************************************************************
// * Coal_NC
// *****************************************************************************

template<class N, class ON, class C, class CC, class PC, class T>
Coal_NC<N, ON, C, CC, PC, T>::Coal_NC(double birth_time, CC* node_contnr, N* first_old_node, N* second_old_node)
{
	connect_to_suc(birth_time, node_contnr, first_old_node, second_old_node);
}



template<class N, class ON, class C, class CC, class PC, class T>
void Coal_NC<N, ON, C, CC, PC, T>::connect_to_suc(const double birth_time, CC* node_contnr, N* first_old_node, N* second_old_node)
{
	OUT_FNCTN_HEADER(logger());

	LOG4CXX_DEBUG(logger(), "first_old_node: " << first_old_node->id());
	LOG4CXX_DEBUG(logger(), "second_old_node: " << second_old_node->id());


	first_old_node->set_to_first_coal_child();
	first_old_node->ref_parent_contnr() = node_contnr;
	second_old_node->set_to_second_coal_child();
	second_old_node->ref_parent_contnr() = node_contnr;

	first_child() = first_old_node;
	second_child() = second_old_node;

	OUT_FNCTN_BOTTOM(logger());
}



template<class N, class ON, class C, class CC, class PC, class T> C* Coal_NC<N, ON, C, CC, PC, T>::oldest_child_contnr() const
{
	if (first_child()->birth_time() > second_child()->birth_time())
		return first_child()->unspec_node_contnr();
	else
		return second_child()->unspec_node_contnr();
}



template<class N, class ON, class C, class CC, class PC, class T>
VECTOR(N*) Coal_NC<N, ON, C, CC, PC, T>::unspec_children() const
{
	VECTOR(N*) v;
	v.push_back(first_child());
	v.push_back(second_child());
	return v;
}



template<class N, class ON, class C, class CC, class PC, class T> void Coal_NC<N, ON, C, CC, PC, T>::calc_log_prior_prob()
{
	OUT_FNCTN_HEADER(logger());

	LOG4CXX_DEBUG(logger(), "id: " << str_id());
	double coal_intensity = Base_NC<N, C, PC, T>::num_elig_nodes() * (Base_NC<N, C, PC, T>::num_elig_nodes() - 1);
	double recomb_intensity = Constants::instance()->recomb_const() * Base_NC<N, C, PC, T>::num_cum_elig_sites();
	double intensity = coal_intensity + recomb_intensity;
	double time = Base_NC<N, C, PC, T>::prev()->birth_time();
	double e = exp(-(Base_NC<N, C, PC, T>::birth_time() - time) * intensity);

	double q = 2 / Constants::instance()->theta();
	Base_NC<N, C, PC, T>::ref_P_gen_event() = q * e;

	double p = Base_NC<N, C, PC, T>::prev()->P_prior();
	p *= Base_NC<N, C, PC, T>::P_gen_event();
	Base_NC<N, C, PC, T>::ref_P_prior() = p;
	Base_NC<N, C, PC, T>::ref_coal_intensity() = coal_intensity;
	Base_NC<N, C, PC, T>::ref_recomb_intensity() = recomb_intensity;

	OUT_FNCTN_BOTTOM(logger());
}



template<class N, class ON, class C, class CC, class PC, class T>
void Coal_NC<N, ON, C, CC, PC, T>::calc_num_elig_nd_st(int* curr_num_elig_sites, int* curr_num_elig_nodes)
{
	(*curr_num_elig_nodes)--;
	*curr_num_elig_sites += NC_with_1_Nd<N, ON, C, PC, T>::num_elig_sites()
							- first_child()->num_elig_sites() - second_child()->num_elig_sites();
	Base_NC<N, C, PC, T>::ref_num_elig_nodes() = *curr_num_elig_nodes;
	Base_NC<N, C, PC, T>::ref_num_cum_elig_sites() = *curr_num_elig_sites;
}



// check integrity
template<class N, class ON, class C, class CC, class PC, class T>
void Coal_NC<N, ON, C, CC, PC, T>::check() const
{
	NC_with_1_Nd<N, ON, C, PC, T>::check();

	BOOL is_ok = true;
	stringstream oss;
	oss << "Node " << str_id() << ": ";

	// is child ok?
	if (!first_child() || !second_child()) {
		is_ok = false;
		oss << "at least one child is NULL" << endl;
	}
	else
		if (!first_child()->parent_contnr() || (*first_child()->parent_contnr() != *this) ||
			!second_child()->parent_contnr() || (*second_child()->parent_contnr() != *this)) {
			is_ok = false;
			oss << "parent container of children is NULL or wrong" << endl;
		}

	if (!is_ok)
		throw ERROR(logger(), oss.str());
}



// *****************************************************************************
// * Rec_NC
// *****************************************************************************

template<class N, class ON, class C, class RC, class PC, class T>
Rec_NC<N, ON, C, RC, PC, T>::Rec_NC(double birth_time, RC* node_contnr, N* old_node)
{
	connect_to_suc(birth_time, node_contnr, old_node);
}



template<class N, class ON, class C, class RC, class PC, class T>
void Rec_NC<N, ON, C, RC, PC, T>::connect_to_suc(double birth_time, RC* node_contnr, N* old_node)
{
	old_node->set_to_rec_child();
	old_node->ref_parent_contnr() = node_contnr;
	child() = old_node;
}



template<class N, class ON, class C, class RC, class PC, class T>
VECTOR(N*) Rec_NC<N, ON, C, RC, PC, T>::unspec_children() const
{
	VECTOR(N*) v;
	v.push_back(child());
	return v;
}



template<class N, class ON, class C, class RC, class PC, class T> void Rec_NC<N, ON, C, RC, PC, T>::calc_log_prior_prob()
{
	OUT_FNCTN_HEADER(logger());

	LOG4CXX_DEBUG(logger(), "id: " << str_id());
	double coal_intensity = Base_NC<N, C, PC, T>::num_elig_nodes() * (Base_NC<N, C, PC, T>::num_elig_nodes() - 1);
	double recomb_intensity = Constants::instance()->recomb_const() * Base_NC<N, C, PC, T>::num_cum_elig_sites();
	double intensity = coal_intensity + recomb_intensity;
	double time = Base_NC<N, C, PC, T>::prev()->birth_time();
	double e = exp(-(Base_NC<N, C, PC, T>::birth_time() - time) * intensity);

	double q = Constants::instance()->recomb_const();
	Base_NC<N, C, PC, T>::ref_P_gen_event() = q * e;

	double p = Base_NC<N, C, PC, T>::prev()->P_prior();
	p *= Base_NC<N, C, PC, T>::P_gen_event();
	Base_NC<N, C, PC, T>::ref_P_prior() = p;
	Base_NC<N, C, PC, T>::ref_coal_intensity() = coal_intensity;
	Base_NC<N, C, PC, T>::ref_recomb_intensity() = recomb_intensity;

	OUT_FNCTN_BOTTOM(logger());
}



template<class N, class ON, class C, class RC, class PC, class T>
void Rec_NC<N, ON, C, RC, PC, T>::calc_num_elig_nd_st(int* curr_num_elig_sites, int* curr_num_elig_nodes)
{
	OUT_FNCTN_HEADER(logger());
	*curr_num_elig_sites -= child()->num_elig_sites();
	if (NC_with_2_Nds<N, ON, C, PC, T>::first_node()->is_active())
		(*curr_num_elig_nodes)--;
	else
		*curr_num_elig_sites  += NC_with_2_Nds<N, ON, C, PC, T>::first_node()->num_elig_sites();
	if (NC_with_2_Nds<N, ON, C, PC, T>::second_node()->is_active())
		(*curr_num_elig_nodes)--;
	else
		*curr_num_elig_sites += NC_with_2_Nds<N, ON, C, PC, T>::second_node()->num_elig_sites();
	(*curr_num_elig_nodes)++;
	Base_NC<N, C, PC, T>::ref_num_elig_nodes() = *curr_num_elig_nodes;
	Base_NC<N, C, PC, T>::ref_num_cum_elig_sites() = *curr_num_elig_sites;
	OUT_FNCTN_BOTTOM(logger());
}



// check integrity
template<class N, class ON, class C, class RC, class PC, class T>
void Rec_NC<N, ON, C, RC, PC, T>::check() const
{
	NC_with_2_Nds<N, ON, C, PC, T>::check();

	BOOL is_ok = true;
	stringstream oss;

	// is child ok?
	if (!child()) {
		is_ok = false;
		oss << "child is NULL" << endl;
	}
	else
		if (!child()->parent_contnr() || (*child()->parent_contnr() != *this)) {
			is_ok = false;
			oss << "parent container of child is NULL or wrong" << endl;
		}

	if (!is_ok)
		throw ERROR(logger(), oss.str());
}


#endif
