#ifndef BASE_NODE_H
#define BASE_NODE_H

#include "standard.h"

#include "constants.h"
#include "empty.h"
#include "error.h"
#include "parent_visitors.h"

class Dot;
struct Dot_Para;
class Log_Acid_Prob;
template<class N, class C, class T> class Base_Nd;
template<class N, class C, class T> BOOL operator==(const Base_Nd<N, C, T>& lhs, const Base_Nd<N, C, T>& rhs);

enum enum_node_type { PURE_NODE, CRF_NODE, RECOMB_NODE, FINAL_NODE };
enum enum_anc_type { COAL_ANC_FIRST, COAL_ANC_SECOND, RECOMB_ANC, ACTIVE };



template<class N, class C, class T> class Base_Nd : private noncopyable, private equality_comparable<Base_Nd<N, C, T> >
{
	friend BOOL operator==<N, C, T>(const Base_Nd<N, C, T>& lhs, const Base_Nd<N, C, T>& rhs);
private:
	static long m_count_id;
	T* m_tree;
	long m_id;
	int m_num_sites;
	int m_num_elig_sites, m_left_most_act_site, m_right_most_act_site;
	VECTOR(BOOL) m_is_act_site;

	BOOL m_is_act_sites_init;

	void init(variant<int, Empty> id);
	int calc_num_sites_new_coal_nd(N* first_old_node, N* second_old_node);
	static LoggerPtr s_logger;
protected:
	Base_Nd() { }
	Base_Nd(T* tree, int num_sites, variant<int, Empty> id);
	Base_Nd(T* tree, N* first_old_node, N* second_old_node, variant<int, Empty> id);
	virtual ~Base_Nd() { }
	static LoggerPtr& logger() { return s_logger; }

	void init_act_sites(const VECTOR(BOOL)& is_act_site, int left_most_act_site, int right_most_act_site);
	virtual void correct_after_move_hook_node(Base_Nd* dest, BOOL is_move_down);
	void recalc_act_sites(N* first_child, N* second_child);

	virtual N* var_next(C* c) const = 0;
	virtual N* var_prev(C* c) const = 0;
public:
	T* tree() const { return m_tree; }
	long id() const { return m_id; }
	int num_sites() const { return m_num_sites; }

	BOOL is_act_site(const int i) const { return m_is_act_site[i]; }
	int left_most_act_site() const { return m_left_most_act_site; }
	int right_most_act_site() const { return m_right_most_act_site; }
	int num_elig_sites() const { return m_num_elig_sites; }

	BOOL is_first() const { return (this == tree()->first_node()); }
	BOOL is_last() const { return (this == tree()->last_node()); }
	virtual BOOL has_coal_suc() const = 0;
	virtual BOOL has_rec_suc() const = 0;
	virtual BOOL is_tip() const = 0;
	virtual BOOL is_single() const = 0;
	virtual BOOL is_double() const = 0;
	double birth_time() const;
	virtual C* unspec_node_contnr() const = 0;
	VECTOR(N*) unspec_children() const { return unspec_node_contnr()->unspec_children(); }

	virtual N* next() const = 0;
	virtual N* prev() const = 0;

	virtual string acid_seq() const;
	virtual string to_dot_string(Dot_Para dot_para) const = 0 ;
	virtual string to_dot_node_name() const = 0;

	virtual void check() const;

	//BOOL operator==(const Base_Nd<N, C, T>& n) const { return id() == n.id(); }
	//BOOL operator!=(const Base_Nd<N, C, T>& n) const { return id() != n.id(); }
};

template<class N, class C, class T> BOOL operator==(const Base_Nd<N, C, T>& lhs, const Base_Nd<N, C, T>& rhs)
{ return lhs.id() == rhs.id(); }



template<class N, class C, class T> class Prob_Nd : public virtual Base_Nd<N, C, T>
{
private:
	VECTOR( VECTOR(Log_Acid_Prob*) ) m_log_acid_prob_ptr;   // pointers to log-prob of nucleic acids (group, index)
	VECTOR(Log_Acid_Prob*) m_site_to_prob;   // pointers to log-prob of nucleic acids
	set<int> m_act_groups;
	set<int> m_changed_groups;
	VECTOR(gsl_matrix*) m_log_subst_prob;

	BOOL m_log_acid_probs_calced;
	BOOL m_is_act_groups_init;
protected:
	Prob_Nd();
	~Prob_Nd();

	const VECTOR( VECTOR(Log_Acid_Prob*) )& log_acid_prob_ptr() const { return m_log_acid_prob_ptr; }
	VECTOR( VECTOR(Log_Acid_Prob*) )& ref_log_acid_prob_ptr()  { return m_log_acid_prob_ptr; }
	const VECTOR(Log_Acid_Prob*)& site_to_prob() const { return m_site_to_prob; }
	VECTOR(Log_Acid_Prob*)& ref_site_to_prob()  { return m_site_to_prob; }
	const set<int>& changed_groups() const { return m_changed_groups; }
	set<int>& ref_changed_groups() { return m_changed_groups; }

	void init_act_groups(const set<int> act_groups);
	void correct_after_move_hook_node(Prob_Nd<N, C, T>* dest, BOOL is_move_down);
	void recalc_act_grps(N* first_child, N* second_child);
	void restore_lap(VECTOR( VECTOR(Log_Acid_Prob*) )* log_acid_prob_ptr_old, VECTOR(Log_Acid_Prob*)* site_to_prob_old);
public:
	using Base_Nd<N, C, T>::logger;

	const VECTOR(gsl_matrix*)& log_subst_prob() const { return m_log_subst_prob; }
	const set<int>& act_groups() const { return m_act_groups; }

	virtual void calc_log_subst_probs() = 0;
	virtual void calc_log_acid_probs() = 0;
	VECTOR(double) log_acid_probs() const;
	double sum_acid_probs() const;

	double get_log_subst_prob(const int index, const int first_acid, const int second_acid) const;
	BOOL log_acid_probs_calced() const { return m_log_acid_probs_calced; }
	void set_log_acid_probs_calced() { m_log_acid_probs_calced = true; }

	double lap(const int mut_rate, const int site, const int acid) const;
	double lap_anc(const int mut_rate, const int site, const int acid) const;

	void reset_changed_sites();

	BOOL is_changed_group(const int i) const { return m_changed_groups.find(i) != m_changed_groups.end(); }
	BOOL is_act_group(const int i) const { return m_act_groups.find(i) != m_act_groups.end(); }
	int num_act_groups() const { return m_act_groups.size(); }

	void check() const;
};



template<class N, class C, class OC, class T> class Nd_with_NC : public virtual Base_Nd<N, C, T>
{
private:
	OC* m_node_contnr;
protected:
	Nd_with_NC() { }
	Nd_with_NC(OC* node_contnr) : m_node_contnr(node_contnr) { }
public:
	OC* node_contnr() const { return m_node_contnr; }
	using Base_Nd<N, C, T>::logger;

	N* next() const;
	N* prev() const;

	void check() const;
};




template<class N, class C, class OC, class T> class Nd_alone_in_NC
: public virtual Nd_with_NC<N, C, OC, T>
{
private:
protected:
	N* var_next(C* c) const { return c ? c->unspec_nodes().front() : 0; }
	N* var_prev(C* c) const { return c ? c->unspec_nodes().back() : 0; }
public:
	BOOL has_rec_suc() const { return false; }
	BOOL is_single() const { return true; }
	BOOL is_double() const { return false; }
	string to_dot_node_name() const;
	void check() const { Nd_with_NC<N, C, OC, T>::check(); }
};



template<class N, class C, class OC, class T> class Tip_Nd : public virtual Nd_alone_in_NC<N, C, OC, T>
{
private:
public:
	BOOL has_coal_suc() const { return false; }
	BOOL is_tip() const { return true; }
	void check() const{ Nd_alone_in_NC<N, C, OC, T>::check(); }
};



template<class N, class C, class OC, class T> class Coal_Nd : public virtual Nd_alone_in_NC<N, C, OC, T>
{
private:
public:
	using Base_Nd<N, C, T>::logger;

	BOOL has_coal_suc() const { return true; }
	BOOL is_tip() const { return false; }

	N*& first_child_node() const;
	N*& second_child_node() const;
	void calc_log_subst_probs();
	void check() const { Nd_alone_in_NC<N, C, OC, T>::check(); }
};



template<class N, class C, class OC, class T> class Rec_Nd
: public virtual Nd_with_NC<N, C, OC, T>
{
private:
	BOOL m_is_first_in_NC;
protected:
	N* var_next(C* c) const;
	N* var_prev(C* c) const;
public:
	Rec_Nd(BOOL b) : m_is_first_in_NC(b) { }
	using Base_Nd<N, C, T>::logger;

	BOOL has_coal_suc() const { return true; }
	BOOL has_rec_suc() const { return false; }
	BOOL is_tip() const { return false; }
	BOOL is_single() const { return false; }
	BOOL is_double() const { return true; }

	N*& child_node() const;
	BOOL is_first_in_NC() const { return m_is_first_in_NC; }
	void calc_log_subst_probs();
	string to_dot_node_name() const;
	void check() const { Nd_with_NC<N, C, OC, T>::check(); }
};



template<class N, class CN, class RN, class C, class CC, class RC, class PC, class T> class Nd_with_Anc
: public virtual Base_Nd<N, C, T>
{
	friend class Changes;
	friend class tree_rearranger_brch;
private:
	enum_anc_type m_anc_type;
	variant<CC*, RC*, Empty> m_anc;
protected:
	enum_anc_type anc_type() const { return m_anc_type; }
	variant<CC*, RC*, Empty> anc() const { return m_anc; }
	void restore_anc(enum_anc_type anc_type_old, variant<CC*, RC*, Empty> anc_old) { m_anc_type = anc_type_old; m_anc = anc_old; }
public:
	Nd_with_Anc();
	Nd_with_Anc(N* first_old_node, N* second_old_node);
	Nd_with_Anc(N* old_node);

	using Base_Nd<N, C, T>::logger;
	using Base_Nd<N, C, T>::birth_time;

	double time_of_parent() const { return apply_visitor(Time_of_Parent_Visitor<CC, RC>(), m_anc); }

	BOOL is_active() const { return m_anc_type == ACTIVE; }
	BOOL has_coal_anc() const { return (m_anc_type == COAL_ANC_FIRST) || (m_anc_type == COAL_ANC_SECOND); }
	BOOL has_rec_anc() const { return m_anc_type == RECOMB_ANC; }
	BOOL is_first_child() const { return m_anc_type == COAL_ANC_FIRST; }
	BOOL is_second_child() const { return m_anc_type == COAL_ANC_SECOND; }
	CN* parent_node() const { return apply_visitor(Parent_Node_Visitor<CN, CC, RC>(), m_anc); }
	RN* first_parent_node() const { return apply_visitor(First_Parent_Node_Visitor<RN, CC, RC>(), m_anc); }
	RN* second_parent_node() const { return apply_visitor(Second_Parent_Node_Visitor<RN, CC, RC>(), m_anc); }
	PC* parent_contnr() const { return apply_visitor(Parent_Contnr_Visitor<CC, RC, PC>(), m_anc); }
	CC* coal_parent_contnr() const { return apply_visitor(Coal_Parent_Contnr_Visitor<CC, RC, PC>(), m_anc); }
	RC* rec_parent_contnr() const { return apply_visitor(Rec_Parent_Contnr_Visitor<CC, RC, PC>(), m_anc); }
	variant<CC*, RC*, Empty>& ref_parent_contnr() { return m_anc; }

	void set_active();
	void set_to_first_coal_child() { m_anc_type = COAL_ANC_FIRST; }
	void set_to_second_coal_child() { m_anc_type = COAL_ANC_SECOND; }
	void set_to_rec_child() { m_anc_type = RECOMB_ANC; }

	void check() const;
};



template<class N, class C, class T> class Nd_with_Seq : public virtual Base_Nd<N, C, T>
{
private:
	VECTOR(int) m_seq;
	BOOL m_is_init;   // whole sequence (for all positions) is initialized
	VECTOR(BOOL) m_pos_is_init;   // a position is initialized
	int m_num_sites;
protected:
	Nd_with_Seq() { ERROR(logger(), "Illegal constructor"); }
	Nd_with_Seq(int num_sites);
	Nd_with_Seq(VECTOR(int) seq);
	Nd_with_Seq(string acid_seq);
	BOOL is_init() const { return m_is_init; }
public:
	using Base_Nd<N, C, T>::logger;

	void init(const VECTOR(int) seq) { m_seq = seq; m_is_init = true; }

	VECTOR(int) seq() const;
	int seq_pos(int i) const;
	string acid_seq() const;
	char acid(int i) const;

	void set_seq_pos(int pos, int value);

	void check() const;
};

#endif
