#include "standard.h"

#include "base_nd.h"
#include "base_tree.h"
#include "crf.h"
#include "constants.h"
#include "empty.h"
#include "error.h"
#include "files.h"
#include "log_acid_prob.h"
#include "parent_visitors.h"
#include "subst_prob.h"
#include "sum_probs.h"

using namespace boost;
using namespace std;

template<class N, class C, class T> long Base_Nd<N, C, T>::m_count_id = 0;
template<class N, class C, class T> LoggerPtr Base_Nd<N, C, T>::s_logger(Logger::getLogger("node.base_nd"));



// *****************************************************************************
// * Base_Nd
// *****************************************************************************

template<class N, class C, class T> double Base_Nd<N, C, T>::birth_time() const { return unspec_node_contnr()->birth_time(); }



template<class N, class C, class T> void Base_Nd<N, C, T>::init(variant<int, Empty> id)
{
	OUT_FNCTN_HEADER(logger());
	m_tree->m_num_nodes++;
	m_tree->m_num_act_nodes++;
	if (int* i = boost::get<int>(&id))
		m_id = *i;
	else
		if (Empty* e = boost::get<Empty>(&id))
			m_id = m_count_id++;
		else
			throw ERROR(logger(), "Something wrong with id");
	OUT_FNCTN_BOTTOM(logger());
}



template<class N, class C, class T>
int Base_Nd<N, C, T>::calc_num_sites_new_coal_nd(N* first_old_node, N* second_old_node)
{
	if (first_old_node->num_sites() != second_old_node->num_sites())
		throw ERROR(logger(), "Trying to construct new coalescent node with its two successor nodes having different number of sites");
	return first_old_node->num_sites();
}



template<class N, class C, class T> Base_Nd<N, C, T>::Base_Nd(T* tree, int num_sites, variant<int, Empty> id)
: m_tree(tree), m_num_sites(num_sites)
{
	OUT_FNCTN_HEADER(logger());
	init(id);
	VECTOR(BOOL) is_act_site(Base_Nd<N, C, T>::num_sites());
	int left_most_act_site, right_most_act_site;
	for (VECTOR(BOOL)::iterator it = is_act_site.begin(); it < is_act_site.end(); it++)
		*it = true;
	left_most_act_site = 0;
	right_most_act_site = Base_Nd<N, C, T>::num_sites() - 1;
	init_act_sites(is_act_site, left_most_act_site, right_most_act_site);
	OUT_FNCTN_BOTTOM(logger());
}



template<class N, class C, class T>
Base_Nd<N, C, T>::Base_Nd(T* tree, N* first_old_node, N* second_old_node, variant<int, Empty> id)
: m_tree(tree), m_num_sites(calc_num_sites_new_coal_nd(first_old_node, second_old_node))
{
	OUT_FNCTN_HEADER(logger());
	LOG4CXX_DEBUG(logger(), "first_old_node: " << first_old_node->id());
	LOG4CXX_DEBUG(logger(), "second_old_node: " << second_old_node->id() << endl);

	init(id);

	VECTOR(BOOL) is_act_site(Base_Nd<N, C, T>::num_sites());
	int left_most_act_site, right_most_act_site;

	is_act_site.resize(Base_Nd<N, C, T>::num_sites());
	left_most_act_site = GSL_MIN(first_old_node->left_most_act_site(), second_old_node->left_most_act_site());
	right_most_act_site = GSL_MAX(first_old_node->right_most_act_site(), second_old_node->right_most_act_site());
	for (int i = 0; i < left_most_act_site; ++i)
		is_act_site[i] = false;
	for (int i = left_most_act_site; i <= right_most_act_site; ++i)
		is_act_site[i] = first_old_node->is_act_site(i) || second_old_node->is_act_site(i);
	for (int i = right_most_act_site + 1; i < Base_Nd<N, C, T>::num_sites(); ++i)
		is_act_site[i] = false;
	init_act_sites(is_act_site, left_most_act_site, right_most_act_site);

	OUT_FNCTN_BOTTOM(logger());
}



template<class N, class C, class T>
void Base_Nd<N, C, T>::init_act_sites(const VECTOR(BOOL)& is_act_site, int left_most_act_site, int right_most_act_site)
{
	m_is_act_site = is_act_site;
	m_left_most_act_site = left_most_act_site;
	m_right_most_act_site = right_most_act_site;
	m_num_elig_sites = m_right_most_act_site - m_left_most_act_site;

	m_is_act_sites_init = true;
}



// recalculate active sites
template<class N, class C, class T> void Base_Nd<N, C, T>::recalc_act_sites(N* first_child, N* second_child)
{
	OUT_FNCTN_HEADER(logger());
	m_left_most_act_site = GSL_MIN(first_child->left_most_act_site(),
								   second_child->left_most_act_site());
	m_right_most_act_site = GSL_MAX(first_child->right_most_act_site(),
									second_child->right_most_act_site());
	for (int i = 0; i < m_left_most_act_site; ++i)
		m_is_act_site[i] = false;
	for (int i = m_left_most_act_site; i <= m_right_most_act_site; ++i)
		m_is_act_site[i] = first_child->is_act_site(i) || second_child->is_act_site(i);
	for (int i = m_right_most_act_site + 1; i < m_num_sites; ++i)
		m_is_act_site[i] = false;

	m_num_elig_sites = m_right_most_act_site - m_left_most_act_site;
	OUT_FNCTN_BOTTOM(logger());
}



template<class N, class C, class T> string Base_Nd<N, C, T>::acid_seq() const
{
	string str_left(left_most_act_site(), '-');
	string str_mid;
	for (int i = left_most_act_site(); i <= right_most_act_site(); i++)
		if (this->is_act_site(i))
			str_mid += "N";
		else
			str_mid += "-";
	string str_right(num_sites() - 1 - right_most_act_site(), '-');
	return str_left + str_mid + str_right;
}



// corrects active sites, etc. after having moved a hook node
template<class N, class C, class T> void Base_Nd<N, C, T>::correct_after_move_hook_node(Base_Nd<N, C, T>* dest, BOOL is_move_down)
{
	OUT_FNCTN_HEADER(logger());
	if (is_move_down)
	{
		for (int i = 0; i < tree()->num_sites(); ++i)
			dest->m_is_act_site[i] = m_is_act_site[i];
		dest->m_left_most_act_site = m_left_most_act_site;
		dest->m_right_most_act_site = m_right_most_act_site;
		dest->m_num_elig_sites = m_num_elig_sites;
	}
	else
	{
		for (int i = 0; i < tree()->num_sites(); ++i)
			m_is_act_site[i] = dest->m_is_act_site[i];
		m_left_most_act_site = dest->m_left_most_act_site;
		m_right_most_act_site = dest->m_right_most_act_site;
		m_num_elig_sites = dest->m_num_elig_sites;
	}
	OUT_FNCTN_BOTTOM(logger());
}



// check integrity
template<class N, class C, class T> void Base_Nd<N, C, T>::check() const
{
	BOOL is_ok = true;
	ostringstream oss;

	if (!tree() || (id() < 0) || (num_sites() < 0) || !m_is_act_sites_init) {
		is_ok = false;
		oss << "Base_Nd in illegal state";
	}

	if ((left_most_act_site() < 0) || (left_most_act_site() >= num_sites()) ||
		(right_most_act_site() < 0) || (right_most_act_site() >= num_sites()) ||
		(num_elig_sites() != right_most_act_site() - left_most_act_site())) {
		is_ok = false;
		oss << "left/right most active site wrong";
	}

	if (!is_ok)
	{
		oss << "Node " << id() << ": ";
		throw ERROR(logger(), oss.str());
	}
}



// *****************************************************************************
// * Prob_Nd
// *****************************************************************************

template<class N, class C, class T> Prob_Nd<N, C, T>::Prob_Nd()
{
	OUT_FNCTN_HEADER(logger());
	m_log_acid_probs_calced = false;
	m_log_subst_prob.resize(Base_Nd<N, C, T>::tree()->num_mutation_rates());
	for (int i = 0; i < Base_Nd<N, C, T>::tree()->num_mutation_rates(); ++i)
		m_log_subst_prob[i] = gsl_matrix_alloc (NUM_NUCL, NUM_NUCL);
	OUT_FNCTN_BOTTOM(logger());
}



template<class N, class C, class T> Prob_Nd<N, C, T>::~Prob_Nd()
{
	OUT_FNCTN_HEADER(logger());
	for (int i = 0; i < Base_Nd<N, C, T>::tree()->num_mutation_rates(); ++i)
		gsl_matrix_free(m_log_subst_prob[i]);
	for (VECTOR( VECTOR(Log_Acid_Prob*) )::iterator it = m_log_acid_prob_ptr.begin();
		 it != m_log_acid_prob_ptr.end(); ++it)
		for (VECTOR(Log_Acid_Prob*)::iterator jt = it->begin(); jt != it->end(); ++jt)
			checked_delete(*jt);
	OUT_FNCTN_BOTTOM(logger());
}



template<class N, class C, class T> void Prob_Nd<N, C, T>::init_act_groups(const set<int> act_groups)
{
	m_act_groups = act_groups;
	m_is_act_groups_init = true;
}



// corrects active groups and active sites, etc. after having moved a hook node
template<class N, class C, class T> void Prob_Nd<N, C, T>::correct_after_move_hook_node(Prob_Nd<N, C, T>* dest, BOOL is_move_down)
{
	if (is_move_down)
		dest->m_act_groups = m_act_groups;
	else
		m_act_groups = dest->m_act_groups;
	Base_Nd<N, C, T>::correct_after_move_hook_node(dest, is_move_down);
}



// recalculates active groups
template<class N, class C, class T> void Prob_Nd<N, C, T>::recalc_act_grps(N* first_child, N* second_child)
{
	m_act_groups.clear();
	VECTOR(int) v(Base_Nd<N, C, T>::tree()->part_alias()->num_groups());
	VECTOR(int)::iterator it =
		set_union(first_child->m_act_groups.begin(), first_child->m_act_groups.end(),
				  second_child->m_act_groups.begin(), second_child->m_act_groups.end(),
			      v.begin());
	for (VECTOR(int)::iterator jt = v.begin(); jt != it; ++jt) {
		m_act_groups.insert(*jt);
	}
}



template<class N, class C, class T>
void Prob_Nd<N, C, T>::restore_lap(VECTOR( VECTOR(Log_Acid_Prob*) )* log_acid_prob_ptr_old, VECTOR(Log_Acid_Prob*)* site_to_prob_old)
{
	for (VECTOR( VECTOR(Log_Acid_Prob*) )::iterator jt = m_log_acid_prob_ptr.begin();
		 jt != m_log_acid_prob_ptr.end(); ++jt)
		for (VECTOR(Log_Acid_Prob*)::iterator kt = jt->begin(); kt != jt->end(); ++kt)
			checked_delete(*kt);
	m_log_acid_prob_ptr = *log_acid_prob_ptr_old;
	m_site_to_prob = *site_to_prob_old;
	for (VECTOR( VECTOR(Log_Acid_Prob*) )::iterator jt = log_acid_prob_ptr_old->begin();
		 jt != log_acid_prob_ptr_old->end(); ++jt)
		jt->clear();
}



template<class N, class C, class T> VECTOR(double) Prob_Nd<N, C, T>::log_acid_probs() const
{
	OUT_FNCTN_HEADER(logger());

	VECTOR(double) ret(Base_Nd<N, C, T>::num_sites(), 0);
	VECTOR(BOOL) is_def(Base_Nd<N, C, T>::num_sites(), false);

	for (set<int>::const_iterator it = act_groups().begin(); it != act_groups().end(); ++it) {
		map<Log_Acid_Prob*, double> mp;
		for (int sgm = 0; sgm < Base_Nd<N, C, T>::tree()->part_alias()->num_segms(*it); ++sgm)
			for (int site = Base_Nd<N, C, T>::tree()->part_alias()->begin_segm(*it, sgm);
				 site <= Base_Nd<N, C, T>::tree()->part_alias()->end_segm(*it, sgm); ++site) {
				if (mp.find(m_site_to_prob[site]) == mp.end()) {
					VECTOR(double) v(Base_Nd<N, C, T>::tree()->num_mutation_rates());
					for (int mr = 0; mr < Base_Nd<N, C, T>::tree()->num_mutation_rates(); ++mr) {
						double r[NUM_NUCL];
						for (int i = 0; i < NUM_NUCL; i++)
							r[i] = log(Constants::instance()->base_prob(i)) + m_site_to_prob[site]->log_p(mr, i);
						double q = 0;
						sum_probs(&q, r);
						v[mr] = q + Base_Nd<N, C, T>::tree()->log_prob_mutation_rate(mr);
					}
					double p = 0;
					sum_probs(&p, v);
					mp[m_site_to_prob[site]] = p;
					LOG4CXX_DEBUG(logger(), "site: " << site);
				}
				ret[site] = mp[m_site_to_prob[site]];
				is_def[site] = true;
			}
	}

	FOREACH (BOOL b, is_def)
		if (!b)
			throw ERROR (logger(), "Trying to log-prob of node with inactive sites");

	OUT_FNCTN_BOTTOM(logger());
	return ret;
}



// sum log-probs of nucleotides
template<class N, class C, class T> double Prob_Nd<N, C, T>::sum_acid_probs() const
{
	OUT_FNCTN_HEADER(logger());

	double ret = 0;

	for (set<int>::const_iterator it = act_groups().begin(); it != act_groups().end(); ++it) {
		map<Log_Acid_Prob*, double> mp;
		for (int sgm = 0; sgm < Base_Nd<N, C, T>::tree()->part_alias()->num_segms(*it); ++sgm)
			for (int site = Base_Nd<N, C, T>::tree()->part_alias()->begin_segm(*it, sgm);
				 site <= Base_Nd<N, C, T>::tree()->part_alias()->end_segm(*it, sgm); ++site) {
				if (mp.find(m_site_to_prob[site]) == mp.end()) {
					VECTOR(double) v(Base_Nd<N, C, T>::tree()->num_mutation_rates());
					for (int mr = 0; mr < Base_Nd<N, C, T>::tree()->num_mutation_rates(); ++mr) {
						double r[NUM_NUCL];
						for (int i = 0; i < NUM_NUCL; i++)
							r[i] = log(Constants::instance()->base_prob(i)) + m_site_to_prob[site]->log_p(mr, i);
						double q = 0;
						sum_probs(&q, r);
						v[mr] = q + Base_Nd<N, C, T>::tree()->log_prob_mutation_rate(mr);
					}
					double p = 0;
					sum_probs(&p, v);
					mp[m_site_to_prob[site]] = p;
					LOG4CXX_DEBUG(logger(), "site: " << site);
				}
				ret += mp[m_site_to_prob[site]];
			}
	}

	OUT_FNCTN_BOTTOM(logger());
	return ret;
}



template<class N, class C, class T> double Prob_Nd<N, C, T>::lap(const int mut_rate, const int site, const int acid) const
{
	if (!Base_Nd<N, C, T>::is_act_site(site)) {
		stringstream ss;
		ss << "Trying to get log acid probs of node" << Base_Nd<N, C, T>::id() << " at non-active site" << site;;
		throw ERROR(logger(), ss.str());
	}
	return m_site_to_prob[site]->log_p(mut_rate, acid);
}



template<class N, class C, class T> double Prob_Nd<N, C, T>::lap_anc(const int mut_rate, const int site, const int acid) const
{
	if (!Base_Nd<N, C, T>::is_act_site(site)) {
		stringstream ss;
		ss << "Trying to get log anc acid probs of node" << Base_Nd<N, C, T>::id() << " at non-active site" << site;;
		throw ERROR(logger(), ss.str());
	}
	return m_site_to_prob[site]->log_p_anc(mut_rate, acid);
}



template<class N, class C, class T> double Prob_Nd<N, C, T>::get_log_subst_prob(const int index, const int first_acid, const int second_acid) const
{
	double x = gsl_matrix_get(m_log_subst_prob[index], first_acid, second_acid);
	return x;
}



// check integrity
template<class N, class C, class T> void Prob_Nd<N, C, T>::check() const
{
	Base_Nd<N, C, T>::check();
	if (!m_is_act_groups_init) {
		ostringstream oss;
		oss << "Node " << Base_Nd<N, C, T>::id() << ": Prob_Nd in illegal state";
		throw ERROR(logger(), oss.str());
	}
}



// *****************************************************************************
// * Nd_with_NC
// *****************************************************************************

template<class N, class C, class OC, class T> N* Nd_with_NC<N, C, OC, T>::next() const
{ return var_next(node_contnr()->next()); }

template<class N, class C, class OC, class T> N* Nd_with_NC<N, C, OC, T>::prev() const
{ return var_next(node_contnr()->prev()); }



// check integrity
template<class N, class C, class OC, class T> void Nd_with_NC<N, C, OC, T>::check() const
{
	Base_Nd<N, C, T>::check();

	BOOL is_ok = true;
	stringstream oss;

	if (!node_contnr()) {
		is_ok = false;
		oss << "node_contnr is NULL";
	}

	if (!is_ok)
		throw ERROR(logger(), oss.str());
}



// *****************************************************************************
// * Nd_alone_in_NC
// *****************************************************************************

template<class N, class C, class OC, class T> string Nd_alone_in_NC<N, C, OC, T>::to_dot_node_name() const
{
	stringstream ss;
	ss << Nd_with_NC<N, C, OC, T>::node_contnr()->to_dot_node_name() << ":N0";
	return ss.str();
}



// *****************************************************************************
// * Coal_Nd
// *****************************************************************************

template<class N, class C, class OC, class T> N*& Coal_Nd<N, C, OC, T>::first_child_node() const
{ return Nd_with_NC<N, C, OC, T>::node_contnr()->first_child(); }

template<class N, class C, class OC, class T> N*& Coal_Nd<N, C, OC, T>::second_child_node() const
{ return Nd_with_NC<N, C, OC, T>::node_contnr()->second_child(); }



// *****************************************************************************
// * Rec_Nd
// *****************************************************************************

template<class N, class C, class OC, class T> N*& Rec_Nd<N, C, OC, T>::child_node() const
{ return Nd_with_NC<N, C, OC, T>::node_contnr()->child(); }



template<class N, class C, class OC, class T> N* Rec_Nd<N, C, OC, T>::var_next(C* c) const
{
	if (is_first_in_NC())
		return Nd_with_NC<N, C, OC, T>::node_contnr()->second_node();
	else
		return c ? c->unspec_nodes().front() : 0;
}



template<class N, class C, class OC, class T> N* Rec_Nd<N, C, OC, T>::var_prev(C* c) const
{
	if (!is_first_in_NC())
		return Nd_with_NC<N, C, OC, T>::node_contnr()->first_node();
	else
		return c ? c->unspec_nodes().back() : 0;
}



template<class N, class C, class OC, class T> string Rec_Nd<N, C, OC, T>::to_dot_node_name() const
{
	stringstream ss;
	ss << Nd_with_NC<N, C, OC, T>::node_contnr()->to_dot_node_name() << ":N";
	ss << (is_first_in_NC() ? 0 : 1);
	return ss.str();
}



// *****************************************************************************
// * Nd_with_Anc
// *****************************************************************************

template<class N, class CN, class RN, class C, class CC, class RC, class PC, class T>
Nd_with_Anc<N, CN, RN, C, CC, RC, PC, T>::Nd_with_Anc()
: m_anc_type(ACTIVE)
{
	OUT_FNCTN_HEADER(logger());

	VECTOR(BOOL) is_act_site(Base_Nd<N, C, T>::num_sites());
	int left_most_act_site, right_most_act_site;

	for (VECTOR(BOOL)::iterator it = is_act_site.begin(); it < is_act_site.end(); it++)
		*it = true;
	left_most_act_site = 0;
	right_most_act_site = Base_Nd<N, C, T>::num_sites() - 1;
	Base_Nd<N, C, T>::init_act_sites(is_act_site, left_most_act_site, right_most_act_site);

	OUT_FNCTN_BOTTOM(logger());
}



template<class N, class CN, class RN, class C, class CC, class RC, class PC, class T>
Nd_with_Anc<N, CN, RN, C, CC, RC, PC, T>::Nd_with_Anc(N* first_old_node, N* second_old_node)
: m_anc_type(ACTIVE)
{
	OUT_FNCTN_HEADER(logger());
	LOG4CXX_DEBUG(logger(), "first_old_node: " << first_old_node->id());
	LOG4CXX_DEBUG(logger(), "second_old_node: " << second_old_node->id() << endl);

	VECTOR(BOOL) is_act_site(Base_Nd<N, C, T>::num_sites());
	int left_most_act_site, right_most_act_site;

	is_act_site.resize(Base_Nd<N, C, T>::num_sites());
	left_most_act_site = GSL_MIN(first_old_node->left_most_act_site(), second_old_node->left_most_act_site());
	right_most_act_site = GSL_MAX(first_old_node->right_most_act_site(), second_old_node->right_most_act_site());
	for (int i = 0; i < left_most_act_site; ++i)
		is_act_site[i] = false;
	for (int i = left_most_act_site; i <= right_most_act_site; ++i)
		is_act_site[i] = first_old_node->is_act_site(i) || second_old_node->is_act_site(i);
	for (int i = right_most_act_site + 1; i < Base_Nd<N, C, T>::num_sites(); ++i)
		is_act_site[i] = false;
	Base_Nd<N, C, T>::init_act_sites(is_act_site, left_most_act_site, right_most_act_site);

	OUT_FNCTN_BOTTOM(logger());
}



template<class N, class CN, class RN, class C, class CC, class RC, class PC, class T>
Nd_with_Anc<N, CN, RN, C, CC, RC, PC, T>::Nd_with_Anc(N* old_node) : m_anc_type(ACTIVE)
{ }



template<class N, class CN, class RN, class C, class CC, class RC, class PC, class T>
void Nd_with_Anc<N, CN, RN, C, CC, RC, PC, T>::set_active()
{
	if (!is_active())
		Base_Nd<N, C, T>::tree()->m_num_act_nodes++;
	m_anc_type = ACTIVE;
}



// check integrity
template<class N, class CN, class RN, class C, class CC, class RC, class PC, class T>
void Nd_with_Anc<N, CN, RN, C, CC, RC, PC, T>::check() const
{
	Base_Nd<N, C, T>::check();

	BOOL is_ok = true;
	stringstream oss;
	oss << "Node " << Base_Nd<N, C, T>::id() << ": ";

	if (has_coal_anc()) {
		if (!parent_node() || !parent_contnr() || !coal_parent_contnr()) {
			is_ok = false;
			oss << "parent is NULL" << endl;
		}
	}
	if (has_rec_anc()) {
		if (!first_parent_node() || !second_parent_node() || !parent_contnr() || !rec_parent_contnr()) {
			is_ok = false;
			oss << "parent is NULL" << endl;
		}
	}

	if (!is_ok) {
		throw ERROR(logger(), oss.str());
	}
}



template<class N, class CN, class RN, class C, class CC, class RC, class PC, class T> struct node_comp
{
	BOOL operator() (Nd_with_Anc<N, CN, RN, C, CC, RC, PC, T>* lhs, Nd_with_Anc<N, CN, RN, C, CC, RC, PC, T>* rhs)
	{
		return lhs->id() < rhs->id();
	}
};



// *****************************************************************************
// * Nd_with_Seq
// *****************************************************************************

template<class N, class C, class T> Nd_with_Seq<N, C, T>::Nd_with_Seq(int num_sites)
: m_is_init(false), m_num_sites(num_sites), m_pos_is_init(VECTOR(BOOL)(num_sites, false))
{
	m_seq.resize(num_sites);
}



template<class N, class C, class T> Nd_with_Seq<N, C, T>::Nd_with_Seq(VECTOR(int) seq)
: m_seq(seq), m_is_init(true), m_num_sites(seq.size()), m_pos_is_init(VECTOR(BOOL)(seq.size(), true))
{ }

template<class N, class C, class T> Nd_with_Seq<N, C, T>::Nd_with_Seq(string acid_seq)
: m_seq(Constants::instance()->alphabet().indices(acid_seq)), m_is_init(true), m_num_sites(acid_seq.size()),
m_pos_is_init(VECTOR(BOOL)(acid_seq.length(), true))
{ }



template<class N, class C, class T> VECTOR(int) Nd_with_Seq<N, C, T>::seq() const
{
	CHECK_INTGRTY;
	if (!m_is_init)
		throw ERROR(logger(), "Trying to get sequence of Nd_with_Seq not having one");
	return m_seq;
}



template<class N, class C, class T> int Nd_with_Seq<N, C, T>::seq_pos(int i) const
{
	CHECK_INTGRTY;
	if (!m_is_init && !m_pos_is_init[i])
		throw ERROR(logger(), "Trying to get sequence of Nd_with_Seq not having one");
	return m_seq[i];
}



template<class N, class C, class T> string Nd_with_Seq<N, C, T>::acid_seq() const
{
	CHECK_INTGRTY;
	if (!m_is_init)
		throw ERROR(logger(), "Trying to get sequence of Nd_with_Seq not having one");
	return Constants::instance()->alphabet().characters(m_seq);
}



template<class N, class C, class T> char Nd_with_Seq<N, C, T>::acid(int i) const
{
	CHECK_INTGRTY;
	if (!m_is_init && !m_pos_is_init[i])
		throw ERROR(logger(), "Trying to get sequence of Nd_with_Seq not having one");
	return Constants::instance()->alphabet().character(m_seq[i]);
}



template<class N, class C, class T> void Nd_with_Seq<N, C, T>::set_seq_pos(int pos, int value)
{
	CHECK_INTGRTY;
	if (!Constants::instance()->alphabet().is_legal_idx(value)) {
		stringstream ss;
		ss << "Trying to assign value " << value << " at position " << pos;
		throw ERROR(logger(), ss.str());
	}
	m_seq[pos] = value;
	m_pos_is_init[pos] = true;
	for (int i = 0; i < m_num_sites; ++i)
		if (!m_pos_is_init[i])
			return;
	m_is_init = true;
}



// check integrity
template<class N, class C, class T> void Nd_with_Seq<N, C, T>::check() const
{
	Base_Nd<N, C, T>::check();

	BOOL is_ok = true;
	ostringstream oss;
	oss << "Node " << Base_Nd<N, C, T>::id() << ": ";

	// check indices at all positions where available
	for (int i = 0; i < m_num_sites; ++i)
		if (m_pos_is_init[i] && !Constants::instance()->alphabet().is_legal_idx(m_seq[i])) {
			is_ok = false;
			oss << "Illegal index at pos " << i << endl;
		}

	// num_sites matches the one of Base_Nd?
	if (m_num_sites != Base_Nd<N, C, T>::num_sites()) {
		is_ok = false;
		oss << "num_sites does not match num_sites of Base_Nd" << endl;
	}

	if (!is_ok)
		throw ERROR(logger(), oss.str());
}
