#include "standard.h"

#include "changes.h"
#include "dot.h"
#include "files.h"
#include "nd.h"
#include "nd_visitor.h"
#include "nc.h"
#include "tree.h"

LoggerPtr Rev_NC::s_logger(Logger::getLogger("node.nc"));

class Rev_Tree;
class Rev_Nd;



// *****************************************************************************
// * Rev_NC
// *****************************************************************************

ostream& operator<<(ostream& s, const Rev_NC& c) { s << Det_Manip<Rev_NC>(0)(&c); return s; }



void Rev_NC::init()
{
	OUT_FNCTN_HEADER(logger());
	m_is_pos_affected = false;
	m_is_time_affected = false;
	m_is_nd_st_affected = false;

	m_old_num_elig_nodes = 0;
	m_old_num_cum_elig_sites = 0;
	m_index = 0;
	OUT_FNCTN_BOTTOM(logger());
}



Rev_NC::Rev_NC() : m_mark(NONE)
{
	OUT_FNCTN_HEADER(logger());
	init();
	connect_prev_and_next();
	OUT_FNCTN_BOTTOM(logger());
}



Rev_NC::Rev_NC(Rev_Nd* first_old_node, Rev_Nd* second_old_node) : m_mark(NEW)
{
	OUT_FNCTN_HEADER(logger());
	LOG4CXX_DEBUG(logger(), "first_old_node: " << first_old_node->id());
	LOG4CXX_DEBUG(logger(), "second_old_node: " << second_old_node->id() << endl);

	init();
	first_old_node->set_anc_affected();
	second_old_node->set_anc_affected();
	connect_prev_and_next();
	OUT_FNCTN_BOTTOM(logger());
}



Rev_NC::Rev_NC(Rev_Nd* old_node) : m_mark(NEW)
{
	OUT_FNCTN_HEADER(logger());
	LOG4CXX_DEBUG(logger(), "old_node: " << old_node->id() << endl);

	init();
	old_node->set_anc_affected();
	connect_prev_and_next();
	OUT_FNCTN_BOTTOM(logger());
}



void Rev_NC::set_pos_affected()
{
	if ((mark() == NONE) && !is_pos_affected())
	{
		m_next_old = next();
		m_prev_old = prev();
		m_is_pos_affected = true;
		tree()->changes()->add_to_pos_aff(this);
		tree()->changes()->set_rearranging_done();
	}
}



void Rev_NC::set_nd_st_affected()
{
	if ((mark() == NONE) && !is_nd_st_affected())
	{
		m_old_num_cum_elig_sites = num_cum_elig_sites();
		m_old_num_elig_nodes = num_elig_nodes();
		m_is_nd_st_affected = true;
		tree()->changes()->add_to_nd_st_aff(this);
	}
}



void Rev_NC::set_num_cum_elig_sites(int i)
{
	if (!is_nd_st_affected())
		set_nd_st_affected();
	Base_NC<Rev_Nd, Rev_NC, Rev_Intern_NC, Rev_Tree>::ref_num_cum_elig_sites() = i;
}



void Rev_NC::set_num_elig_nodes(int i)
{
	if (!is_nd_st_affected())
		set_nd_st_affected();
	Base_NC<Rev_Nd, Rev_NC, Rev_Intern_NC, Rev_Tree>::ref_num_elig_nodes() = i;
}



void Rev_NC::change_time(double new_time)
{
	if (!is_time_affected() && is_unmarked())
	{
		m_birth_time_old = birth_time();
		m_is_time_affected = true;
		tree()->changes()->add_to_time_aff(this);
	}
	ref_birth_time() = new_time;
}



void Rev_NC::restore_pos()
{
	m_is_pos_affected = false;
	next() = m_next_old;
	prev() = m_prev_old;}



void Rev_NC::restore_time()
{
	m_is_time_affected = false;
	ref_birth_time() = m_birth_time_old;
}



void Rev_NC::restore_nd_st()
{
	Base_NC<Rev_Nd, Rev_NC, Rev_Intern_NC, Rev_Tree>::ref_num_elig_nodes() = m_old_num_elig_nodes;
	Base_NC<Rev_Nd, Rev_NC, Rev_Intern_NC, Rev_Tree>::ref_num_cum_elig_sites() = m_old_num_cum_elig_sites;
	m_is_nd_st_affected = false;
}



void Rev_NC::adjust_next_and_prev()
{
	OUT_FNCTN_HEADER(logger());
	Rev_NC* prev_elig = prev();
	while ((prev_elig) && (prev_elig->mark() != NONE))
		prev_elig = prev_elig->prev();
	if (prev_elig)
		prev_elig->set_pos_affected();

	Rev_NC* next_elig = next();
	while ((next_elig) && (next_elig->mark() != NONE))
		next_elig = next_elig->next();
	if (next_elig)
		next_elig->set_pos_affected();


	if (prev_elig)
		prev_elig->next() = next_elig;
	else
		tree()->m_first_node_contnr = next_elig;

	if (next_elig)
		next_elig->prev() = prev_elig;
	else
		tree()->m_last_node_contnr = prev_elig;
	OUT_FNCTN_BOTTOM(logger());
}



// connects previous and next node container (if this node container is deleted)
void Rev_NC::connect_prev_and_next()
{
	OUT_FNCTN_HEADER(logger());
	Rev_NC* c = tree()->first_node_contnr();

	if (!c) {
		tree()->ref_first_node_contnr() = this;
		tree()->ref_last_node_contnr() = this;
		prev() = 0;
		next() = 0;
	}
	else {
	   	while (!c->is_last() && (c->birth_time() <= birth_time()))
	   		c = c->next();
	   	if (c->birth_time() > birth_time()) {
	   		c->prev()->set_pos_affected();
	 		c->set_pos_affected();
	  		next() = c;
	   		prev() = c->prev();
	   		c->prev()->next() = this;
	   		c->prev() = this;
	   	}
	   	else {
	   		tree()->last_node_contnr()->set_pos_affected();
	   		tree()->last_node_contnr()->next() = this;
	   		prev() = tree()->last_node_contnr();
	   		tree()->m_last_node_contnr = this;
	   		next() = 0;
	   	}
	}
	OUT_FNCTN_BOTTOM(logger());
}



void Rev_NC::check() const
{
	Base_NC<Rev_Nd, Rev_NC, Rev_Intern_NC, Rev_Tree>::check();

	BOOL is_ok = true;
	stringstream oss;
	oss << "Node container " << *this << ": ";

	// old prev ok?
	if (is_pos_affected() &&
		(tree()->old_first_node_contnr() && (tree()->old_first_node_contnr() != this) ||
		 !tree()->old_first_node_contnr() && (tree()->first_node_contnr() != this)) && !old_prev()) {
		is_ok = false;
		oss << "old prev is NULL";
	}

	// old next ok?
	if (is_pos_affected() &&
		(tree()->old_last_node_contnr() && (tree()->old_last_node_contnr() != this) ||
		 !tree()->old_last_node_contnr() && (tree()->last_node_contnr() != this)) && !old_next()) {
		is_ok = false;
		oss << "old next is NULL";
	}

	// old birth time ok?
	if (is_time_affected() && (m_birth_time_old < 0)) {
		is_ok = false;
		oss << "old birth time negative";
	}

	if (!is_ok)
		throw ERROR(logger(), oss.str());
}


// get string for output of container with Dot
string Rev_NC::to_dot_string(Dot_Para dot_para) const
{
	const string COLOR_TIP = "cyan";
	const string COLOR_ORDINARY = "grey78";
	const string COLOR_NEW = "green1";
	const string COLOR_OLD = "red1";
	const int WIDTH_RECOMB = 1;

	// determine color of container
	string color;
	if (is_tip())
		color = COLOR_TIP;
	else
		if (is_new())
			color = COLOR_NEW;
		else
			if (is_old())
				color = COLOR_OLD;
			else
				color = COLOR_ORDINARY;

	// write Dot output for container and its nodes
	stringstream ss;
	ss << to_dot_node_name() << " [color=\"" << color << "\" label=<" << endl;
	ss << "<TABLE CELLSPACING=\"0\" BORDER=\"0\">" << endl;
	ss << "<TR>";

	if (const Rev_Tip_NC* c = dynamic_cast<const Rev_Tip_NC*>(this)) {
		To_Dot_Str_Visitor vst(dot_para);
		c->node()->accept(&vst);
		ss << (string)vst;
		//ss << c->node()->to_dot_string(dot_para);
	}
	else
		if (const Rev_Coal_NC* c = dynamic_cast<const Rev_Coal_NC*>(this))
			ss << c->node()->to_dot_string(dot_para);
		else
			if (const Rev_Rec_NC* c = dynamic_cast<const Rev_Rec_NC*>(this))
			{
				ss << c->first_node()->to_dot_string(dot_para);
				ss << "<TD BGCOLOR=\"black\" WIDTH=\"" << WIDTH_RECOMB << "\"></TD>";
				ss << c->second_node()->to_dot_string(dot_para);
			}
	ss << "</TR>" << endl;
	ss << "</TABLE>>];" << endl;

	return ss.str();
}



// *****************************************************************************
// * Rev_Intern_NC
// *****************************************************************************

void Rev_Intern_NC::set_old_impl()
{
	Rev_NC::set_old_impl();
	tree()->changes()->add_to_old(this);
	tree()->changes()->set_adjusting_pos_necessary();
}



void Rev_Intern_NC::set_suc_affected()
{
	if (is_unmarked() && !is_suc_affected())
	{
		m_is_suc_affected = true;
		tree()->changes()->add_to_suc_aff(this);
		tree()->changes()->set_rearranging_done();
	}
}



// *****************************************************************************
// * Rev_Tip_NC
// *****************************************************************************

const set<int>& Rev_Tip_NC::occur_subs() const { return node()->occur_subs(); }



Rev_Tip_NC::Rev_Tip_NC(Rev_Tree* tr, string acid_seq, int crf_sub_id, BOOL is_pure)
: Base_NC<Rev_Nd, Rev_NC, Rev_Intern_NC, Rev_Tree>(tr),
Rev_NC(),
Tip_NC<Rev_Nd, Rev_Tip_Nd, Rev_NC, Rev_Intern_NC, Rev_Tree>()
{
	OUT_FNCTN_HEADER(logger());
	Rev_Tip_Nd* new_node = new Rev_Tip_Nd(tr, this, acid_seq, crf_sub_id, is_pure);
	init_node(new_node);
	OUT_FNCTN_BOTTOM(logger());
}



Rev_Tip_NC::Rev_Tip_NC(Rev_Tree* tr, int num_sites, int crf_sub_id, BOOL is_pure)
: Base_NC<Rev_Nd, Rev_NC, Rev_Intern_NC, Rev_Tree>(tr),
Rev_NC(),
Tip_NC<Rev_Nd, Rev_Tip_Nd, Rev_NC, Rev_Intern_NC, Rev_Tree>()
{
	OUT_FNCTN_HEADER(logger());
	Rev_Tip_Nd* new_node = new Rev_Tip_Nd(tr, this, num_sites, crf_sub_id, is_pure);
	init_node(new_node);
	OUT_FNCTN_BOTTOM(logger());
}



VECTOR(Rev_Nd*) Rev_Tip_NC::unspec_nodes() const
{
	VECTOR(Rev_Nd*) v;
	v.push_back(node());
	return v;
}



void Rev_Tip_NC::save(ofstream* out_strm) const
{
	*out_strm << "T " << node()->id() << " " << node()->is_pure_node() << " "
			  << (node()->is_pure_node() ? node()->sub() : node()->crf_id()) << " "
			  << node()->acid_seq() << endl;
}



// *****************************************************************************
// * Rev_Coal_NC
// *****************************************************************************

const set<int>& Rev_Coal_NC::occur_subs() const { return node()->occur_subs(); }
BOOL Rev_Coal_NC::is_final() const { return is_single() && node()->is_final_node(); }



Rev_Coal_NC::Rev_Coal_NC(Rev_Tree* tr, double birth_time, Rev_Nd* first_old_node, Rev_Nd* second_old_node)
: Base_NC<Rev_Nd, Rev_NC, Rev_Intern_NC, Rev_Tree>(tr, birth_time, first_old_node, second_old_node),
Rev_NC(first_old_node, second_old_node),
Coal_NC<Rev_Nd, Rev_Coal_Nd, Rev_NC, Rev_Coal_NC, Rev_Intern_NC, Rev_Tree>(birth_time, this, first_old_node, second_old_node)
{
	OUT_FNCTN_HEADER(logger());
	Rev_Coal_Nd* new_node = new Rev_Coal_Nd(tr, this, first_old_node, second_old_node);
	init_node(new_node);
	tree()->changes()->add_to_new(this);
	OUT_FNCTN_BOTTOM(logger());
}



VECTOR(Rev_Nd*) Rev_Coal_NC::unspec_nodes() const
{
	VECTOR(Rev_Nd*) v;
	v.push_back(node());
	return v;
}



VECTOR(Rev_Nd*) Rev_Coal_NC::unspec_old_children() const
{
	VECTOR(Rev_Nd*) v;
	v.push_back(first_old_child_node());
	v.push_back(second_old_child_node());
	return v;
}



void Rev_Coal_NC::set_old()
{
	Rev_Intern_NC::set_old_impl();
	tree()->m_num_nodes--;
	if (node()->is_active())
		tree()->m_num_act_nodes--;
}



void Rev_Coal_NC::set_suc_affected()
{
	if (is_unmarked() && !is_suc_affected()) {
		m_old_children = children();
		Rev_Intern_NC::set_suc_affected();
	}
}



void Rev_Coal_NC::restore_suc()
{
	first_child() = m_old_children.first;
	second_child() = m_old_children.second;
	unset_suc_affected();
}



void Rev_Coal_NC::save(ofstream* out_strm) const
{
	*out_strm << "C " << node()->id() << " " << birth_time() << " "
			  << first_child()->id() << " " << second_child()->id() << endl;
}



void Rev_Coal_NC::check() const
{
	Rev_Intern_NC::check();
	Coal_NC<Rev_Nd, Rev_Coal_Nd, Rev_NC, Rev_Coal_NC, Rev_Intern_NC, Rev_Tree>::check();

	BOOL is_ok = true;
	ostringstream oss;
	oss << "Node container " << str_id() << ": ";

	// old parent container of old children ok?
	if (is_suc_affected())
		if (!first_old_child_node() || !second_old_child_node()) {
			is_ok = false;
			oss << "first or second old child is NULL" << endl;
		}
		else {
			if (first_old_child_node()->is_anc_affected()) {
				if (!first_old_child_node()->old_parent_contnr()) {
					is_ok = false;
					oss << "old parent container of first old child (" << first_old_child_node()->id() << ") is NULL" << endl;
				}
				else
				if (*first_old_child_node()->old_parent_contnr() != *this) {
					is_ok = false;
					oss << "old parent container of first old child (" << first_old_child_node()->id() << ") is wrong: "
						<< first_old_child_node()->old_parent_contnr()->str_id() << endl;
				}
			}
			if (second_old_child_node()->is_anc_affected()) {
				if (!second_old_child_node()->parent_contnr()) {
					is_ok = false;
					oss << "old parent container of second old child (" << second_old_child_node()->id() << ") is NULL" << endl;
				}
				else
				if (*second_old_child_node()->old_parent_contnr() != *this) {
					is_ok = false;
					oss << "old parent container of second old child (" << second_old_child_node()->id() << ") is wrong: "
						<< second_old_child_node()->old_parent_contnr()->str_id() << endl;
				}
			}
		}

	// check occuring subtypes (container <-> children)
	set<int> s(first_child()->occur_subs().begin(), first_child()->occur_subs().end());
	s.insert(second_child()->occur_subs().begin(), second_child()->occur_subs().end());
	if ((s.size() != occur_subs().size()) || !equal(s.begin(), s.end(), occur_subs().begin())) {
		is_ok = false;
		oss << "node container " << *this
		   << " has occuring subtypes not corresponding with the ones of its children" << endl;
	}

	// check occuring subtypes (container <-> node)
	if ((occur_subs().size() != node()->occur_subs().size()) ||
		!equal(occur_subs().begin(), occur_subs().end(), node()->occur_subs().begin())) {
		is_ok = false;
		oss << "node container " << *this << " has occuring subtypes different from its node" << endl;
	}

	if (!is_ok)
		throw ERROR(logger(), oss.str());
}



// *****************************************************************************
// * Rev_Rec_NC
// *****************************************************************************

Rev_Rec_NC::Rev_Rec_NC(Rev_Tree* tr, double birth_time, Rev_Nd* old_node, set<int> subtypes_first,
			           set<int> subtypes_second)
: Base_NC<Rev_Nd, Rev_NC, Rev_Intern_NC, Rev_Tree>(tr, birth_time, old_node),
Rev_NC(old_node),
Rec_NC<Rev_Nd, Rev_Rec_Nd, Rev_NC, Rev_Rec_NC, Rev_Intern_NC, Rev_Tree>(birth_time, this, old_node)
{
	OUT_FNCTN_HEADER(logger());
	Rev_Rec_Nd* first_new_node = new Rev_Rec_Nd(tr, this, old_node, subtypes_first, true);
	Rev_Rec_Nd* second_new_node = new Rev_Rec_Nd(tr, this, old_node, subtypes_second, false);
	init_nodes(first_new_node, second_new_node);
	tree()->changes()->add_to_new(this);
	m_occur_subs = old_node->occur_subs();
	OUT_FNCTN_BOTTOM(logger());
}



VECTOR(Rev_Nd*) Rev_Rec_NC::unspec_nodes() const
{
   	VECTOR(Rev_Nd*) v;
   	v.push_back(first_node());
   	v.push_back(second_node());
   	return v;
}



VECTOR(Rev_Nd*) Rev_Rec_NC::unspec_old_children() const
{
	VECTOR(Rev_Nd*) v;
	v.push_back(old_child_node());
	return v;
}



void Rev_Rec_NC::set_old()
{
	Rev_Intern_NC::set_old_impl();
	tree()->m_num_nodes -= 2;
	if (first_node()->is_active())
		tree()->m_num_act_nodes--;
	if (second_node()->is_active())
		tree()->m_num_act_nodes--;
}



void Rev_Rec_NC::set_suc_affected()
{
	if (is_unmarked() && !is_suc_affected())
	{
		Rev_Intern_NC::set_suc_affected();
		m_old_child = child();
	}
}



void Rev_Rec_NC::restore_suc()
{
	child() = m_old_child;
	unset_suc_affected();
}



void Rev_Rec_NC::save(ofstream* out_strm) const
{
	*out_strm << "R " << first_node()->id() << " "  << second_node()->id() << " "
			  << birth_time() << " " << child()->id() << " ";
	for (set<int>::iterator it = first_node()->occur_subs().begin();
	 	 it != first_node()->occur_subs().end(); ++it)
		*out_strm << *it << " ";
	*out_strm << "- ";
	for (set<int>::iterator it = second_node()->occur_subs().begin();
	 	 it != second_node()->occur_subs().end(); ++it)
		*out_strm << *it << " ";
	*out_strm << endl;
}



// check integrity
void Rev_Rec_NC::check() const
{
	Rev_Intern_NC::check();
	Rec_NC<Rev_Nd, Rev_Rec_Nd, Rev_NC, Rev_Rec_NC, Rev_Intern_NC, Rev_Tree>::check();

	BOOL is_ok = true;
	ostringstream oss;
	oss << "Node container " << *this << ": ";

	if (is_suc_affected() && !old_child_node()) {
		is_ok = false;
		oss << "old child node is NULL" << endl;
	}

	// check occuring subtypes (container <-> child)
	if ((occur_subs().size() != child()->occur_subs().size()) ||
		!equal(occur_subs().begin(), occur_subs().end(),
			   child()->occur_subs().begin())) {
		is_ok = false;
		oss << "node container " << *this
		   << " has occuring subtypes not corresponding with the ones of its child" << endl;
	}

	// check occuring subtypes (first_node <-> second_node)
    set<int> s(first_node()->occur_subs().begin(), first_node()->occur_subs().end());
	set<int> t(second_node()->occur_subs().begin(), second_node()->occur_subs().end());
	VECTOR(int) v(s.size());
	VECTOR(int)::iterator it = set_intersection(s.begin(), s.end(), t.begin(), t.end(), v.begin());
	if (it - v.begin() > 0) {
		is_ok = false;
		oss << "nodes of node container " << *this << " have common occuring subtypes" << endl;
	}

	// check occuring subtypes (container <-> nodes)
	v.resize(s.size() + t.size());
	it = set_union(s.begin(), s.end(), t.begin(), t.end(), v.begin());
	if ((it - v.begin() != occur_subs().size()) ||
		!equal(occur_subs().begin(), occur_subs().end(), v.begin())) {
		is_ok = false;
		oss << "node container " << *this
		   << " has occuring subtypes not corresponding with the ones of its nodes" << endl;
	}

	if (!is_ok)
		throw ERROR(logger(), oss.str());
}


