#ifndef NODE_CONTR_H
#define NODE_CONTR_H

#include "standard.h"

#include "base_nc.h"
#include "empty.h"
#include "trans_probs.h"

template<class N, class C, class PC, class T> class Base_NC;
class Sim_Nd;
class Rev_Nd;
class Rev_Tip_Nd;
class Rev_Intern_Nd;
class Rev_Coal_Nd;
class Rev_Rec_Nd;
class Rev_Intern_NC;
class Rev_Tree;
class Changes;
class S_and_P;
struct Dot_Para;
template<class T> class Bound_Det_Manip;



class Rev_NC : public virtual Base_NC<Rev_Nd, Rev_NC, Rev_Intern_NC, Rev_Tree>
{
	friend class Changes;
	friend ostream& operator<<(ostream& s, const Bound_Det_Manip<Rev_NC>& bdm);
private:
	enum enum_mark { NONE, OLD, NEW };

	Rev_NC* m_next_old;
	Rev_NC* m_prev_old;
	BOOL m_is_pos_affected;
	BOOL m_is_time_affected;
	BOOL m_is_nd_st_affected;
	double m_birth_time_old;
	int m_old_num_elig_nodes;
	int m_old_num_cum_elig_sites;
	enum_mark m_mark;

	S_and_P m_s_and_p;
	int m_index;

	static LoggerPtr s_logger;

	void init();
protected:
	Rev_NC();
	Rev_NC(Rev_Nd* first_old_node, Rev_Nd* second_old_node);
	Rev_NC(Rev_Nd* old_node);
	static LoggerPtr& logger() { return s_logger; }
public:
	S_and_P s_and_p() const { return m_s_and_p; }
	S_and_P& s_and_p() { return m_s_and_p; }
	int index() const { return m_index; }
	int& index() { return m_index; }

	Rev_NC* old_next() const { return m_next_old; }
	Rev_NC* old_prev() const { return m_prev_old; }
	int old_num_elig_nodes() const { return m_old_num_elig_nodes; }
	int old_num_cum_elig_sites() const { return m_old_num_cum_elig_sites; }

	BOOL is_pos_affected() const { return m_is_pos_affected; }
	BOOL is_time_affected() const { return m_is_time_affected; }
	BOOL is_nd_st_affected() const { return m_is_nd_st_affected; }
	BOOL is_unmarked() const { return m_mark == NONE; }
	BOOL is_new() const { return m_mark == NEW; }
	BOOL is_old() const { return m_mark == OLD; }
	virtual BOOL is_final() const { return false; }
	enum_mark mark() const { return m_mark; }
	virtual const set<int>& occur_subs() const = 0;

	void set_old_impl() { m_mark = OLD; }
	void set_unmarked() { m_mark = NONE; }
	void set_pos_affected();
	void set_nd_st_affected();
	virtual void set_num_cum_elig_sites(int i);
	virtual void set_num_elig_nodes(int i);

	void restore_pos();
	void restore_time();
	void restore_nd_st();

	void change_time(double new_time);
	void adjust_next_and_prev();
	void connect_prev_and_next();

	VECTOR(Rev_Nd*) rev_nodes() const { return unspec_nodes(); }

	string to_dot_string(Dot_Para dot_para) const;
	virtual void save(ofstream* out_strm) const = 0;

	void check() const;
};

ostream& operator<<(ostream& s, const Rev_NC& c);



class Rev_Intern_NC : public virtual Rev_NC, public virtual Intern_NC<Rev_NC>
{
	friend class Changes;
private:
	BOOL m_is_suc_affected;
protected:
	Rev_Intern_NC() { m_is_suc_affected = false; }
	void unset_suc_affected() { m_is_suc_affected = false; }
	void set_old_impl();
public:
	virtual VECTOR(Rev_Nd*) unspec_old_children() const = 0;
	BOOL is_suc_affected() const { return m_is_suc_affected; }
	virtual void set_suc_affected();
	virtual void set_old() = 0;
	virtual void restore_suc() = 0;
	void check() const { Rev_NC::check(); }
};



class Rev_Tip_NC : public virtual Rev_NC, public virtual Tip_NC<Rev_Nd, Rev_Tip_Nd, Rev_NC, Rev_Intern_NC, Rev_Tree>
{
private:
	VECTOR(Rev_Nd*) unspec_nodes() const;
public:
	Rev_Tip_NC(Rev_Tree* tr, string acid_seq, int crf_sub_id, BOOL is_pure);
	Rev_Tip_NC(Rev_Tree* tr, int num_sites, int crf_sub_id, BOOL is_pure);
	using Rev_NC::logger;
	const set<int>& occur_subs() const;
	void save(ofstream* out_strm) const;
	void check() const { Rev_NC::check(); Tip_NC<Rev_Nd, Rev_Tip_Nd, Rev_NC, Rev_Intern_NC, Rev_Tree>::check(); }
};



class Rev_Coal_NC
: public virtual Rev_Intern_NC, public virtual Coal_NC<Rev_Nd, Rev_Coal_Nd, Rev_NC, Rev_Coal_NC, Rev_Intern_NC, Rev_Tree>
{
private:
	pair<Rev_Nd*, Rev_Nd*> m_old_children;

	VECTOR(Rev_Nd*) unspec_nodes() const;
public:
	Rev_Coal_NC(Rev_Tree* tr, double birth_time, Rev_Nd* first_old_node, Rev_Nd* second_old_node);

	using Rev_NC::logger;

	const set<int>& occur_subs() const;
	BOOL is_final() const;
	Rev_Nd* first_old_child_node() const { return m_old_children.first; }
	Rev_Nd* second_old_child_node() const { return m_old_children.second; }
	VECTOR(Rev_Nd*) unspec_old_children() const;
	void set_old();
	void set_suc_affected();
	void restore_suc();
	void save(ofstream* out_strm) const;
	void check() const;
};



class Rev_Rec_NC
: public virtual Rev_Intern_NC, public virtual Rec_NC<Rev_Nd, Rev_Rec_Nd, Rev_NC, Rev_Rec_NC, Rev_Intern_NC, Rev_Tree>
{
private:
	Rev_Nd* m_old_child;
	set<int> m_occur_subs;

	VECTOR(Rev_Nd*) unspec_nodes() const;
public:
	Rev_Rec_NC(Rev_Tree* tr, double birth_time, Rev_Nd* old_node, set<int> subtypes_first,
			   set<int> subtypes_second);

	using Rev_NC::logger;

	const set<int>& occur_subs() const { return m_occur_subs; }
	Rev_Nd* old_child_node() const  { return m_old_child; }
	VECTOR(Rev_Nd*) unspec_old_children() const;
	void set_old();
	void set_suc_affected();
	void restore_suc();
	void save(ofstream* out_strm) const;
	void check() const;
};

#endif
