#ifndef NODE_H
#define NODE_H

#include "standard.h"

#include "base_nd.h"

class Dot;
struct Dot_Para;
class Nd_Visitor;
class Rev_Intern_Nd;
class Rev_Coal_Nd;
class Rev_Rec_Nd;
class Rev_NC;
class Rev_Intern_NC;
class Rev_Tip_NC;
class Rev_Coal_NC;
class Rev_Rec_NC;
class Rev_Tree;



class Rev_Nd
: public virtual Prob_Nd<Rev_Nd, Rev_NC, Rev_Tree>,
public virtual Nd_with_Anc<Rev_Nd, Rev_Coal_Nd, Rev_Rec_Nd, Rev_NC, Rev_Coal_NC, Rev_Rec_NC, Rev_Intern_NC, Rev_Tree>
{
private:
	enum_node_type m_type;
	BOOL m_is_anc_affected;
	BOOL m_is_ac_pr_affected;
	int m_crf_id;
	int m_subtype;
	set<int> m_occur_subs;
	variant<Rev_Coal_NC*, Rev_Rec_NC*, Empty> m_anc_old;
	enum_anc_type m_anc_type_old;
	VECTOR(BOOL) m_is_act_segment;

	VECTOR( VECTOR( VECTOR(double) ) ) m_old_log_acid_prob;   // log-prob of nucleic acids (mutation rate, site, acid)
	VECTOR( VECTOR( VECTOR( VECTOR(double) ) ) ) m_old_log_brnch_acid_prob;   // log-prob of nucleic acids (first/second, mutation rate, site, acid)
	VECTOR( VECTOR(Log_Acid_Prob*) ) m_log_acid_prob_ptr_old;   // pointers to log-prob of nucleic acids (group, index)
	VECTOR(Log_Acid_Prob*) m_site_to_prob_old;   // pointers to log-prob of nucleic acids

	BOOL m_is_crf_sub_init;
	BOOL m_is_lap_init;
	static LoggerPtr m_logger;

	void init();
	void init_old_lap();

	void connect_prev_and_next(Rev_Nd* n);
	void connect_prev_and_next(Rev_Nd* first, Rev_Nd* second);
protected:
	static LoggerPtr& logger() { return m_logger; }
	const VECTOR(BOOL)& is_act_segment() const { return m_is_act_segment; }
	VECTOR(BOOL)& is_act_segment() { return m_is_act_segment; }
	void init_crf_sub(const enum_node_type type, const int crf_sub_id, const set<int> occur_subs);
	void reset_occur_subs(const set<int>& occur_subs);
public:
	Rev_Nd();
	~Rev_Nd();

	using Base_Nd<Rev_Nd, Rev_NC, Rev_Tree>::tree;
	using Prob_Nd<Rev_Nd, Rev_NC, Rev_Tree>::log_acid_probs_calced;
	using Prob_Nd<Rev_Nd, Rev_NC, Rev_Tree>::log_acid_prob_ptr;
	using Prob_Nd<Rev_Nd, Rev_NC, Rev_Tree>::site_to_prob;

	Rev_NC* rev_node_contnr() const { return unspec_node_contnr(); }

	BOOL is_anc_affected() const { return m_is_anc_affected; }
	BOOL is_ac_pr_affected() const { return m_is_ac_pr_affected; }
	virtual BOOL is_pos_affected() const = 0;
	virtual BOOL is_time_affected() const = 0;
	BOOL is_nd_st_affected() const;

	enum_node_type type() const { return m_type; }
	int crf_id() const;
	int sub() const;
	const set<int>& occur_subs() const { return m_occur_subs; }

	BOOL had_coal_anc() const { return (m_anc_type_old == COAL_ANC_FIRST) || (m_anc_type_old == COAL_ANC_SECOND); }
	BOOL had_rec_anc() const { return m_anc_type_old == RECOMB_ANC; }
	BOOL was_first_child() const { return m_anc_type_old == COAL_ANC_FIRST; }
	BOOL was_second_child() const { return m_anc_type_old == COAL_ANC_SECOND; }
	Rev_Intern_NC* old_parent_contnr() const;
	Rev_Coal_Nd* old_parent_node() const;
	Rev_Rec_Nd* first_old_parent_node() const;
	Rev_Rec_Nd* second_old_parent_node() const;
	BOOL is_old_first() const;
	BOOL is_old_last() const;
	Rev_Nd* old_next() const;
	Rev_Nd* old_prev() const;
	int old_num_elig_nodes() const;
	int old_num_cum_elig_sites() const;

	void determine_new_anc();
	void set_coal_anc(Rev_Coal_NC* parent, BOOL is_first_child);
	void set_recomb_anc(Rev_Rec_NC* parent);

	void set_anc_affected();
	void set_ac_pr_affected();
	void set_nd_st_affected();
	void reset_anc_affected() { m_is_anc_affected = false; }
	void reset_ac_pr_affected() { m_is_ac_pr_affected = false; }

	BOOL is_pure_node() const { return m_type == PURE_NODE; }
	BOOL is_crf_node() const { return m_type == CRF_NODE; }
	BOOL is_rec_node() const { return m_type == RECOMB_NODE; }
	BOOL is_final_node() const { return m_type == FINAL_NODE; }
	BOOL has_crf() const { return ((type() == CRF_NODE) || (type() == RECOMB_NODE)); }
	virtual BOOL is_unmarked() const = 0;
	virtual BOOL is_new() const = 0;
	virtual BOOL is_old() const = 0;

	double old_log_acid_prob(const int index, const int site, const int acid) const { return m_old_log_acid_prob[index][site][acid]; }
	double old_log_left_acid_prob(const int index, const int site, const int acid) const { return m_old_log_brnch_acid_prob[0][index][site][acid]; }
	double old_log_right_acid_prob(const int index, const int site, const int acid) const { return m_old_log_brnch_acid_prob[1][index][site][acid]; }
	double old_lap(const int mut_rate, const int site, const int acid) const;
	double old_lap_anc(const int mut_rate, const int site, const int acid) const;

	void restore_lap();
	void restore_anc();

	virtual void recalc_act_sites_grps() { }
	virtual void recalc_occur_subs() { }
	void calc_num_elig_nodes();

	string to_dot_string(Dot_Para dot_para) const;
	virtual string to_tree_output_string() const = 0;
	string sub_segments_to_string() const;

	void check() const;
	void accept(Nd_Visitor* v);
};

ostream& operator<<(ostream& s, const Rev_Nd& n);



class Rev_Intern_Nd : public virtual Rev_Nd
{
private:
	Rev_Intern_NC* m_rev_intern_node_contnr;
protected:
	void init_end();
	Rev_Intern_Nd() { }
	Rev_Intern_Nd(Rev_Intern_NC* node_contnr) { m_rev_intern_node_contnr = node_contnr; }
public:
	Rev_Intern_NC* rev_intern_node_contnr() const { return m_rev_intern_node_contnr; }
	BOOL is_suc_affected() const;
	BOOL is_pos_affected() const;
	BOOL is_time_affected() const;
	BOOL is_unmarked() const;
	BOOL is_new() const;
	BOOL is_old() const;

	void change_time(double new_time);
	void set_unmarked();
	virtual void set_old() = 0;
	void set_suc_affected();
	void set_pos_affected();

	void check() const;
	void accept(Nd_Visitor* v);
};



class Rev_Tip_Nd :
public virtual Rev_Nd, public virtual Tip_Nd<Rev_Nd, Rev_NC, Rev_Tip_NC, Rev_Tree>,
public virtual Nd_with_Seq<Rev_Nd, Rev_NC, Rev_Tree>
{
private:
	BOOL m_is_prob_init;

	void init(int num_sites, int crf_sub_id, BOOL is_pure);
public:
	void init_prob();

	Rev_Tip_Nd(Rev_Tree* tr, Rev_Tip_NC* node_contnr, string acid_seq, int crf_sub_id, BOOL is_pure);
	Rev_Tip_Nd(Rev_Tree* tr, Rev_Tip_NC* node_contnr, int num_sites, int crf_sub_id, BOOL is_pure);

	using Rev_Nd::logger;

	BOOL is_pos_affected() const { return false; }
	BOOL is_time_affected() const { return false; }
	BOOL is_unmarked() const { return true; }
	BOOL is_new() const { return false; }
	BOOL is_old() const { return false; }
	Rev_NC* unspec_node_contnr() const;
	void calc_log_acid_probs() { set_log_acid_probs_calced(); }
	void calc_log_subst_probs() { }

	string to_tree_output_string() const;
	string acid_seq() const;
	void check() const;
	void accept(Nd_Visitor* v);
};



class Rev_Coal_Nd : public virtual Rev_Intern_Nd, public virtual Coal_Nd<Rev_Nd, Rev_NC, Rev_Coal_NC, Rev_Tree>
{
private:
public:
	Rev_Coal_Nd(Rev_Tree* tr, Rev_Coal_NC* node_contnr, Rev_Nd* first_old_node, Rev_Nd* second_old_node);

	using Rev_Nd::logger;

	Rev_Nd* first_old_child_node() const;
	Rev_Nd* second_old_child_node() const;
	Rev_NC* unspec_node_contnr() const;
	void set_old();
	void calc_log_acid_probs();
	void calc_log_subst_probs();
	void recalc_act_sites_grps();
	void recalc_occur_subs();

	void correct_after_move_hook_node(Rev_Coal_Nd* dest, BOOL is_move_down);
	string to_tree_output_string() const;
	void check() const;
	void accept(Nd_Visitor* v);
};



class Rev_Rec_Nd : public virtual Rev_Intern_Nd, public virtual Rec_Nd<Rev_Nd, Rev_NC, Rev_Rec_NC, Rev_Tree>
{
private:
public:
	Rev_Rec_Nd(Rev_Tree* tr, Rev_Rec_NC* node_contnr, Rev_Nd* old_node, set<int> subtypes, BOOL is_first_node);

	using Rev_Nd::logger;

	Rev_Nd*& child_node() const;
	BOOL is_first_in_NC() const;
	Rev_Nd* old_child_node() const;
	Rev_NC* unspec_node_contnr() const;
	void set_old();
	Rev_Coal_Nd* hook_in(Rev_Nd* n, double time);
	void calc_log_acid_probs();
	void calc_log_subst_probs();

	string to_tree_output_string() const;
	void check() const;
	void accept(Nd_Visitor* v);
};

#endif
