#include "standard.h"

#include "constants.h"
#include "error.h"
#include "nd.h"
#include "trans_probs.h"

LoggerPtr S_and_P::s_logger(Logger::getLogger("node.trans_probs"));



// calculates one of the sums over k in P_{j, j-i}(t) (see "rearrange_coal.pdf)
// l - the Ls occuring the formula (e.g. L_1^k, L_2^k, L_3 for j == 3)
double S_and_P::calc_P(int j, int i, VECTOR(int) l, int z, double t)
{
	OUT_FNCTN_HEADER(logger());

	if (l.size() != i + 1)
		throw Error("L has wrong length in calc_trans_prob");
	if ((j < 1) || (j > 3))
			throw Error("Wrong value for j in calc_trans_prob");

	const double rho = Constants::instance()->recomb_const();
	double ret;
	switch (i)
	{
	case 0:
		ret = exp(-(j * (j - 1) / 2 + j * z + rho * l[0]) * t);
		break;
	case 1:
		if (j == 1)
			throw Error("Wrong value for i or j in calc_trans_prob");
		ret = (exp(-((j - 1) * (j - 2) / 2 + (j - 1) * z + rho * l[0]) * t)
			  - exp(-(j * (j - 1) / 2 + j* z + rho * l[1]) * t)) / (j - 1 + z + rho * (l[1] - l[0]));
		break;
	case 2: {
		if (j != 3)
			throw Error("Wrong value for i or j in calc_trans_prob");

		double alpha = 1 + z + rho * (l[1] - l[0]);
		double gamma = 3 + 2 * z + rho * (l[2] - l[0]);
		double beta = 1 - exp(-gamma * t);
		double delta = exp(-(1 + z + rho * (l[1] - l[0])) * t) - exp(-gamma * t);
		double epsilon = 2 + z + rho * (l[2] - l[1]);
		ret = (beta / gamma - delta / epsilon) / alpha;
		break;
	}
	default:
		throw Error("Wrong value for i in calc_trans_prob");
		break;
	}

	OUT_FNCTN_BOTTOM(logger());
	return ret;
}



// calculates the transition probs between states describing how many and which nodes are left (not coalesced)
// e.g. P("All nodes are left" -> "No node is left")
void S_and_P::calc_trans_prob(int z, double t)
{
	OUT_FNCTN_HEADER(logger());

	if (!m_elig_sites_init)
		throw Error("Calling calc_trans_prob although m_elig_sites has not been initialized");

	VECTOR(int) l;
	m_p.clear();

	l = VECTOR(int)(1, m_elig_sites[NONE_LEFT]);
	m_p[pair<Nds_Left, Nds_Left>(NONE_LEFT, NONE_LEFT)] = calc_P(1, 0, l, z, t);

	l = VECTOR(int)(1, m_elig_sites[FIRST_LEFT]);
	m_p[pair<Nds_Left, Nds_Left>(FIRST_LEFT, FIRST_LEFT)] = calc_P(2, 0, l, z, t);
	l.push_back(NONE_LEFT);
	m_p[pair<Nds_Left, Nds_Left>(FIRST_LEFT, NONE_LEFT)] = calc_P(2, 1, l, z, t);

	l = VECTOR(int)(1, m_elig_sites[SECOND_LEFT]);
	m_p[pair<Nds_Left, Nds_Left>(SECOND_LEFT, SECOND_LEFT)] = calc_P(2, 0, l, z, t);
	l.push_back(NONE_LEFT);
	m_p[pair<Nds_Left, Nds_Left>(SECOND_LEFT, NONE_LEFT)] = calc_P(2, 1, l, z, t);

	l = VECTOR(int)(1, m_elig_sites[THIRD_LEFT]);
	m_p[pair<Nds_Left, Nds_Left>(THIRD_LEFT, THIRD_LEFT)] = calc_P(2, 0, l, z, t);
	l.push_back(NONE_LEFT);
	m_p[pair<Nds_Left, Nds_Left>(THIRD_LEFT, NONE_LEFT)] = calc_P(2, 1, l, z, t);

	l = VECTOR(int)(1, m_elig_sites[ALL_LEFT]);
	m_p[pair<Nds_Left, Nds_Left>(ALL_LEFT, ALL_LEFT)] = calc_P(3, 0, l, z, t);

	l = VECTOR(int)(1, m_elig_sites[ALL_LEFT]);
	l.push_back(m_elig_sites[FIRST_LEFT]);
	m_p[pair<Nds_Left, Nds_Left>(ALL_LEFT, FIRST_LEFT)] = calc_P(3, 1, l, z, t);
	l.push_back(m_elig_sites[NONE_LEFT]);
	m_q[0] = calc_P(3, 2, l, z, t);

	l = VECTOR(int)(1, m_elig_sites[ALL_LEFT]);
	l.push_back(m_elig_sites[SECOND_LEFT]);
	m_p[pair<Nds_Left, Nds_Left>(ALL_LEFT, SECOND_LEFT)] = calc_P(3, 1, l, z, t);
	l.push_back(m_elig_sites[NONE_LEFT]);
	m_q[1] = calc_P(3, 2, l, z, t);

	l = VECTOR(int)(1, m_elig_sites[ALL_LEFT]);
	l.push_back(m_elig_sites[THIRD_LEFT]);
	m_p[pair<Nds_Left, Nds_Left>(ALL_LEFT, THIRD_LEFT)] = calc_P(3, 1, l, z, t);
	l.push_back(m_elig_sites[NONE_LEFT]);
	m_q[2] = calc_P(3, 2, l, z, t);

	m_p[pair<Nds_Left, Nds_Left>(ALL_LEFT, NONE_LEFT)] = (m_q[0] + m_q[1] + m_q[2]) / 3;

	l = VECTOR(int)(1, m_elig_sites[ALL_LEFT]);
	m_p[pair<Nds_Left, Nds_Left>(FIRST_TWO_LEFT, FIRST_TWO_LEFT)] = calc_P(2, 0, l, z, t);
	l.push_back(m_elig_sites[THIRD_LEFT]);
	m_p[pair<Nds_Left, Nds_Left>(FIRST_TWO_LEFT, NONE_LEFT)] = calc_P(2, 1, l, z, t);

	OUT_FNCTN_BOTTOM(logger());
}




void S_and_P::calc_elig_sites(const int curr_elig_sites, VECTOR(Rev_Nd*) children)
{
	OUT_FNCTN_HEADER(logger());

	VECTOR(int) nes, rmas, lmas;
	nes.push_back(children[0]->num_elig_sites());
	nes.push_back(children[1]->num_elig_sites());
	nes.push_back(children[2]->num_elig_sites());
	rmas.push_back(children[0]->right_most_act_site());
	rmas.push_back(children[1]->right_most_act_site());
	rmas.push_back(children[2]->right_most_act_site());
	lmas.push_back(children[0]->left_most_act_site());
	lmas.push_back(children[1]->left_most_act_site());
	lmas.push_back(children[2]->left_most_act_site());

	m_elig_sites.clear();
	m_elig_sites.push_back(curr_elig_sites - nes[0] - nes[1] - nes[2] + GSL_MAX(rmas[0], GSL_MAX(rmas[1], rmas[2]))
						   - GSL_MIN(lmas[0], GSL_MIN(lmas[1], lmas[2])));
	m_elig_sites.push_back(curr_elig_sites - nes[1] - nes[2] + GSL_MAX(rmas[1], rmas[2]) - GSL_MIN(lmas[1], lmas[2]));
	m_elig_sites.push_back(curr_elig_sites - nes[0] - nes[2] + GSL_MAX(rmas[0], rmas[2]) - GSL_MIN(lmas[0], lmas[2]));
	m_elig_sites.push_back(curr_elig_sites - nes[0] - nes[1] + GSL_MAX(rmas[0], rmas[1]) - GSL_MIN(lmas[0], lmas[1]));
	m_elig_sites.push_back(curr_elig_sites);
	m_elig_sites.push_back(curr_elig_sites);

	m_elig_sites_init = true;

	OUT_FNCTN_BOTTOM(logger());
}



double S_and_P::P(const Nds_Left before, const Nds_Left after) const
{
	map<pair<Nds_Left, Nds_Left>, double>::const_iterator it = m_p.find(pair<Nds_Left, Nds_Left>(before, after));
	if (it == m_p.end())
		return 0;
	else
		return it->second;
}



void S_and_P::set_S_to_start(const BOOL start_at_second_child)
{
	m_s[NONE_LEFT] = 0;
	m_s[FIRST_LEFT] = 0;
	m_s[SECOND_LEFT] = 0;
	m_s[THIRD_LEFT] = 0;
	if (start_at_second_child) {
		m_s[ALL_LEFT] = 0;
		m_s[FIRST_TWO_LEFT] = 1;
	}
	else {
		m_s[ALL_LEFT] = 1;
		m_s[FIRST_TWO_LEFT] = 0;
	}
}



void S_and_P::jump()
{
	m_s[THIRD_LEFT] = m_s[NONE_LEFT];
	m_s[ALL_LEFT] = m_s[FIRST_TWO_LEFT];
	m_s[NONE_LEFT] = 0;
	m_s[FIRST_LEFT] = 0;
	m_s[SECOND_LEFT] = 0;
	m_s[FIRST_TWO_LEFT] = 0;
}








