#ifndef BASE_TREE_DEF_H_
#define BASE_TREE_DEF_H_

#include "standard.h"

#include "base_tree.h"
#include "constants.h"
#include "decomposition.h"
#include "error.h"
#include "gamma.h"
#include "files.h"

using namespace std;

class Rev_Nd;
class Rev_NC;
class Sim_Nd;
class Sim_NC;
class CRF;
class Partition_Tree;
class Tip_Seqs;
class Constants;
class Changes;
class Partition_Aliasing;

template<class N, class CN, class RN, class C> LoggerPtr Base_Tree<N, CN, RN, C>::m_logger(Logger::getLogger("tree.base_tree"));



template<class N, class CN, class RN, class C> Base_Tree<N, CN, RN, C>::Base_Tree(const int num_sites_para, const int num_mutation_rates)
{
	OUT_FNCTN_HEADER(logger());

	m_num_nodes = 0;
	m_num_act_nodes = 0;
	m_num_tips = 0;
	m_num_sites = num_sites_para;
	m_first_node_contnr = 0;
	m_last_node_contnr = 0;
	m_decomp = new Decomposition();

	m_num_mutation_rates = num_mutation_rates;
	m_mutation_rates.resize(m_num_mutation_rates);
	m_log_prob_mutation_rates.resize(m_num_mutation_rates);
	for (int i = 0; i < m_num_mutation_rates; ++i)
	{
		m_mutation_rates[i] = Gamma_Distribution::mutation_rate(i);
		m_log_prob_mutation_rates[i] = Gamma_Distribution::log_prob_mutation_rate(i);
	}

	OUT_FNCTN_BOTTOM(logger());
}



template<class N, class CN, class RN, class C> Base_Tree<N, CN, RN, C>::~Base_Tree()
{
	OUT_FNCTN_HEADER(logger());
	delete m_decomp;
	VECTOR(C*) node_contnrs;
	for (C* c = first_node_contnr(); c; c = c->next())
		node_contnrs.push_back(c);
	for (typename VECTOR(C*)::iterator it = node_contnrs.begin(); it != node_contnrs.end(); it++)
		delete *it;
	OUT_FNCTN_BOTTOM(logger());
}



// calculate prior probability (probability of Rev_Tree) from "start" to last node
template<class N, class CN, class RN, class C> void Base_Tree<N, CN, RN, C>::calc_prior_prob(C* start)
{
	OUT_FNCTN_HEADER(logger());
	LOG4CXX_DEBUG(logger(), "start: " << start->str_id());

	if (num_act_nodes() == 1)
	{
		double time = 0;
		C* c = start;
		stringstream ss;
		ss << "tips: " << flush;
		while (c->is_tip()) {
			c->calc_log_prior_prob();
			ss << c->str_id() << "  ";
			c = c->next();
		}
		LOG4CXX_DEBUG(logger(), ss.str());

		double p = 1;
		ss.str("");
		ss << "interior nodes: ";
		while (c) {
			c->calc_log_prior_prob();
			ss << c->str_id() << "  ";
			c = c->next();
		}
		LOG4CXX_DEBUG(logger(), ss.str());	}
	else
		throw ERROR(logger(),"Trying to calculate prior probability of a tree with number of active nodes unequal 1");

	OUT_FNCTN_BOTTOM(logger());
}



// calculates number of eligible nodes and number of cumulative eligible sites for nodes from start to end
template<class N, class CN, class RN, class C> void Base_Tree<N, CN, RN, C>::calc_num_elig_nd_st(C* start, C* end)
{
	OUT_FNCTN_HEADER(logger());
	LOG4CXX_DEBUG(logger(), "start: " << start->str_id());
	stringstream ss;
	ss << "end: ";
	if (end)
		ss << start->str_id() << endl;
	else
		ss << "NULL" << endl;
	LOG4CXX_DEBUG(logger(), ss.str());

	int curr_num_elig_sites;
	int curr_num_elig_nodes;
	if (*start == *first_node_contnr())
	{
		curr_num_elig_sites = 0;
		curr_num_elig_nodes = 0;
	}
	else
	{
		curr_num_elig_sites = start->prev()->num_cum_elig_sites();
		curr_num_elig_nodes = start->prev()->num_elig_nodes();
	}

	for (C* c = start; c != end; c = c->next())
		c->calc_num_elig_nd_st(&curr_num_elig_sites, &curr_num_elig_nodes);

	OUT_FNCTN_BOTTOM(logger());
}



// find node with a given id
template<class N, class CN, class RN, class C> N* Base_Tree<N, CN, RN, C>::find_node(int id) const
{
	for (N* n = first_node(); n; n = n->next())
		if (n->id() == id)
			return n;
	return 0;
}



// check integrity
template<class N, class CN, class RN, class C> void Base_Tree<N, CN, RN, C>::check() const
{
	BOOL is_ok = true;
	stringstream ss;

	check_node_contnr();

	// check first/last node (container)
	if (!first_node_contnr()) {
		is_ok = false;
		ss << "first nc is NULL" << endl;
	}
	if (!last_node_contnr()) {
		is_ok = false;
		ss << "last nc is NULL" << endl;
	}
	if (!first_node()) {
		is_ok = false;
		ss << "first nd is NULL" << endl;
	}
	if (!last_node()) {
		is_ok = false;
		ss << "last nd is NULL" << endl;
	}

	// check number of active nodes
	int num_act_nodes = 0;
	for (N* n = first_node(); n; n = n->next())
		if (n->is_active())
			++num_act_nodes;
	if (num_act_nodes != m_num_act_nodes) {
		is_ok = false;
		ss << "num_act_nodes is wrong: " << num_act_nodes << " (actual) " << m_num_act_nodes << " (current)" << endl;
	}

	// check for next() circular
	if (first_node_contnr())
	{
		set<string> nodes;
		for (C* c = first_node_contnr(); c; c = c->next())
		{
			pair<set<string>::iterator, BOOL> p = nodes.insert(c->str_id());
			if (!p.second)
			{
				is_ok = false;
				ss << "next() is circular: ";
				C* d = first_node_contnr();
				int k = 0;
				if (first_node_contnr()->str_id() == *p.first)
					k++;
				while (k < 2)
				{
					ss << d->str_id() << " -> ";
					d = d->next();
					if (d->str_id() == *p.first)
						k++;
				}
				ss << d->str_id() << endl;;
				break;
			}
		}
	}

	// check for prev() circular
	if (last_node_contnr())
	{
		set<string> nodes;
		for (C* c = last_node_contnr(); c; c = c->prev())
		{
			pair<set<string>::iterator, BOOL> p = nodes.insert(c->str_id());
			if (!p.second)
			{
				is_ok = false;
				ss << "prev() is circular: ";
				C* d = first_node_contnr();
				int k = 0;
				if (first_node_contnr()->str_id() == *p.first)
					k++;
				while (k < 2)
				{
					ss << d->str_id() << " -> ";
					d = d->prev();
					if (d->str_id() == *p.first)
						k++;
				}
				ss << d->str_id() << endl;;
				break;
			}
		}
	}

	// check integrity of position and birth time
	if (first_node())
	{
		set<string> nodes;
		for (C* c = first_node_contnr(); c; c = c->next())
		{
			if (!nodes.insert(c->str_id()).second)
				break;
			if ((c->next()) && (c->birth_time() > c->next()->birth_time()))
			{
				is_ok = false;
				ss << "time of generation inconsistent (pos): " << c->str_id() << " (" << c->birth_time() << ")"
				   << " -> " << c->next()->str_id() << " (" << c->next()->birth_time() << ")" << endl;
			}
		}
	}

	// check time integrity of ancestors and successors
	if (first_node())
	{
		set<int> nodes;
		for (N* n = first_node(); n; n = n->next())
		{
			if (!nodes.insert(n->id()).second)
				break;

			// check ancestors
			if (n->has_coal_anc())
			{
				if ((n->parent_node()) && (n->birth_time() > n->parent_node()->birth_time()))
				{
					is_ok = false;
					ss << "time of generation inconsistent (anc): " << n->id() << " (" << n->birth_time() << ")"
					   << " -> " << n->parent_node()->id() << " (" << n->parent_node()->birth_time() << ")" << endl;
				}
			}
			if (n->has_rec_anc())
			{
				if ((n->first_parent_node()) && (n->second_parent_node()) &&
					((n->birth_time() > n->first_parent_node()->birth_time()) ||
					(n->birth_time() > n->second_parent_node()->birth_time())))
				{
					is_ok = false;
					ss << "time of generation inconsistent (anc): " << n->id() << " (" << n->birth_time() << ")"
					   << " -> " << n->first_parent_node()->id() << " (" << n->first_parent_node()->birth_time() << ") "
					   << n->second_parent_node()->id() << " (" << n->second_parent_node()->birth_time() << ")"  << endl;
				}
			}

			// check successors
			if (RN* m = dynamic_cast<RN*>(n))
			{
				if ((m->child_node()) && (m->birth_time() < m->child_node()->birth_time()))
				{
					is_ok = false;
					ss << "time of generation inconsistent (suc): " << m->id() << " (" << m->birth_time() << ")"
					   << " -> " << m->child_node()->id() << " (" << m->child_node()->birth_time() << ")" << endl;
				}
			}
			else
				if (CN* m = dynamic_cast<CN*>(n))
				{
					if ((m->first_child_node()) && (m->second_child_node()) &&
						((m->birth_time() < m->first_child_node()->birth_time()) ||
						(m->birth_time() < m->second_child_node()->birth_time())))
					{
						is_ok = false;
						ss << "time of generation inconsistent (suc): " << m->id() << " (" << m->birth_time() << ")"
						   << " -> " << m->first_child_node()->id() << " (" << m->first_child_node()->birth_time() << ") "
						   << m->second_child_node()->id() << " (" << m->second_child_node()->birth_time() << ")"  << endl;
					}
				}
		}
	}

	// check integrity of ancestors
	if (first_node())
	{
		set<int> nodes;
		for (N* n = first_node(); n; n = n->next())
		{
			if (!nodes.insert(n->id()).second)
				break;

			// has coalescent parent
			if (n->has_coal_anc())
			{
				if (n->parent_node())
				{
					// is first child
					if (n->is_first_child())
						if (n->parent_node()->first_child_node())
						{
							if (n->parent_node()->first_child_node()->id() != n->id())
							{
								is_ok = false;
								ss << "ancestor inconsistent: " << n->id() << endl;
							}
					    }
						else
						{
							is_ok = false;
							ss << "child of parent is missing: " << n->id() << endl;
						}

					// is second child
					if (n->is_second_child())
						if (n->parent_node()->second_child_node())
						{
							if (n->parent_node()->second_child_node()->id() != n->id())
							{
								is_ok = false;
								ss << "ancestor inconsistent: " << n->id() << endl;
							}
					    }
						else
						{
							is_ok = false;
							ss << "child of parent is missing: " << n->id() << endl;
						}
				}
				else
				{
					is_ok = false;
					ss << "parent is missing: " << n->id() << endl;
				}
			}

			// has recombination parents
			if (n->has_rec_anc())
			{
				// check first parent
				if (n->first_parent_node())
				{
					if (n->first_parent_node()->child_node())
					{
						if (n->first_parent_node()->child_node()->id() != n->id())
						{
							is_ok = false;
							ss << "ancestor inconsistent: " << n->id() << endl;
						}
				    }
					else
					{
						is_ok = false;
						ss << "child of parent is missing: " << n->id() << endl;
					}
				}
				else
				{
					is_ok = false;
					ss << "parent is missing: " << n->id() << endl;
				}

				// check second parent
				if (n->second_parent_node())
				{
					if (n->second_parent_node()->child_node())
					{
						if (n->second_parent_node()->child_node()->id() != n->id())
						{
							is_ok = false;
							ss << "ancestor inconsistent: " << n->id() << endl;
						}
				    }
					else
					{
						is_ok = false;
						ss << "child of parent is missing: " << n->id() << endl;
					}
				}
				else
				{
					is_ok = false;
					ss << "parent is missing: " << n->id() << endl;
				}
			}
		}
	}

	// check integrity of active sites
	if (first_node()) {
		for (N* n = first_node(); n; n = n->next()) {
			if (n->is_tip()) {
				for (int i = 0; i < num_sites(); ++i)
					if (!n->is_act_site(i)) {
						is_ok = false;
						ss << "site " << i << " of tip " << n->id() << " is not active" << endl;
					}
			}
			else
				if (CN* m = dynamic_cast<CN*>(n)) {
					for (int i = 0; i < num_sites(); ++i)
						if (m->is_act_site(i) !=
							(m->first_child_node()->is_act_site(i) || m->second_child_node()->is_act_site(i))) {
							is_ok = false;
							ss << "site " << i << " of node " << m->id() << " has not the correct active status" << endl;
						}
				}
				else
					if (RN* m = dynamic_cast<RN*>(n)) {
						for (int i = 0; i < num_sites(); ++i)
							if ((m->child_node()->is_act_site(i) !=
								 (m->node_contnr()->first_node()->is_act_site(i) ^
								  m->node_contnr()->second_node()->is_act_site(i))) ||
								m->node_contnr()->first_node()->is_act_site(i) &&
								m->node_contnr()->second_node()->is_act_site(i)) {
								is_ok = false;
								ss << "site " << i << " of node " << m->id() << " has not the correct active status" << endl;
							}
					}

		}
	}

	if (!is_ok)
		throw ERROR(logger(), ss.str());
}

#endif
