#include "standard.h"

#include "changes.h"
#include "error.h"
#include "files.h"
#include "nc.h"
#include "nd.h"
#include "tree.h"

LoggerPtr Changes::s_logger(Logger::getLogger("tree.changes"));



Changes::Changes(Rev_Tree* tree)
{
	m_tree = tree;
	m_rearranging_done = false;
	m_adjusting_pos_necessary = false;
}



void Changes::accept()
{
	OUT_FNCTN_HEADER(logger());
	if (m_rearranging_done) {
		if (m_adjusting_pos_necessary)
			throw ERROR(logger(), "Reverse Changes: Illegal Rev_Tree");

		delete_old();
		unmark_new();

		m_tree->set_old_first_last_to_zero();

		reset_anc();
		reset_ac_pr();
		reset_suc();
		reset_pos();
		reset_time();
		reset_nd_st();

		m_rearranging_done = false;
	}
	OUT_FNCTN_BOTTOM(logger());
}



void Changes::reverse()
{
	OUT_FNCTN_HEADER(logger());
	if (m_rearranging_done) {
		if (m_adjusting_pos_necessary)
			throw ERROR(logger(), "Reverse Changes: Illegal Rev_Tree");

		delete_new();
		unmark_old();

		restore_anc();
		restore_ac_pr();
		restore_suc();
		restore_pos();
		restore_time();
		restore_nd_st();

		m_tree->set_first_node_contnr(m_tree->old_first_node_contnr());
		m_tree->set_last_node_contnr(m_tree->old_last_node_contnr());
		m_tree->set_old_first_last_to_zero();

		reset_anc();
		reset_ac_pr();
		reset_suc();
		reset_pos();
		reset_time();
		reset_nd_st();

		m_tree->recalc_act_sites_grps();
		m_tree->recalc_occur_subs();

		m_rearranging_done = false;
	}
	OUT_FNCTN_BOTTOM(logger());
}



// correct the position of ordinary nodes after some nodes have been set to be old
void Changes::adjust_pos()
{
	for(set<Rev_NC*>::iterator it = m_old_node_contnrs.begin(); it != m_old_node_contnrs.end(); it++)
		(*it)->adjust_next_and_prev();
	m_adjusting_pos_necessary = false;
}



void Changes::delete_new()
{
	for (set<Rev_NC*>::iterator it = m_new_node_contnrs.begin(); it != m_new_node_contnrs.end(); it++)
	{
		delete *it;
	}
	m_new_node_contnrs.clear();
}



void Changes::delete_old()
{
	for (set<Rev_NC*>::iterator it = m_old_node_contnrs.begin(); it != m_old_node_contnrs.end(); it++)
	{
		VECTOR(Rev_Nd*) v = (*it)->rev_nodes();
		for (VECTOR(Rev_Nd*)::iterator jt = v.begin(); jt != v.end(); ++jt)
			m_ac_pr_aff_nodes.erase(*jt);
		delete *it;
	}
	m_old_node_contnrs.clear();
}



void Changes::restore_anc()
{
	for (set<Rev_Nd*>::iterator it = m_anc_aff_nodes.begin(); it != m_anc_aff_nodes.end(); it++)
		(*it)->restore_anc();
	m_anc_aff_nodes.clear();
}



void Changes::restore_ac_pr()
{
	ofstream* out_strm = In_Out_Streams::instance()->ostrm(OUTPUT_FILE);
	for (set<Rev_Nd*>::iterator it = m_ac_pr_aff_nodes.begin(); it != m_ac_pr_aff_nodes.end(); it++)
		(*it)->restore_lap();
	m_ac_pr_aff_nodes.clear();
}



void Changes::restore_suc()
{
	for (set<Rev_Intern_NC*>::iterator it = m_suc_aff_node_contnrs.begin(); it != m_suc_aff_node_contnrs.end(); it++)
		(*it)->restore_suc();
	m_suc_aff_node_contnrs.clear();
}



void Changes::restore_pos()
{
	for (set<Rev_NC*>::iterator it = m_pos_aff_node_contnrs.begin(); it != m_pos_aff_node_contnrs.end(); it++)
		(*it)->restore_pos();
	m_pos_aff_node_contnrs.clear();
}



void Changes::restore_time()
{
	for (set<Rev_NC*>::iterator it = m_time_aff_node_contnrs.begin(); it != m_time_aff_node_contnrs.end(); it++)
	{
		(*it)->restore_time();
		(*it)->m_is_time_affected = false;
	}
	m_time_aff_node_contnrs.clear();
}



void Changes::restore_nd_st()
{
	for (set<Rev_NC*>::iterator it = m_nd_st_aff_node_contnrs.begin(); it != m_nd_st_aff_node_contnrs.end(); it++)
		(*it)->restore_nd_st();
	m_nd_st_aff_node_contnrs.clear();
}



void Changes::unmark_new()
{
	for (set<Rev_NC*>::iterator it = m_new_node_contnrs.begin(); it != m_new_node_contnrs.end(); it++)
		(*it)->set_unmarked();
	m_new_node_contnrs.clear();
}



void Changes::unmark_old()
{
	for (set<Rev_NC*>::iterator it = m_old_node_contnrs.begin(); it != m_old_node_contnrs.end(); it++)
		(*it)->set_unmarked();;
	m_old_node_contnrs.clear();
}



void Changes::reset_anc()
{
	for (set<Rev_Nd*>::iterator it = m_anc_aff_nodes.begin(); it != m_anc_aff_nodes.end(); it++)
		(*it)->reset_anc_affected();
	m_anc_aff_nodes.clear();
}



void Changes::reset_ac_pr()
{
	for (set<Rev_Nd*>::iterator it = m_ac_pr_aff_nodes.begin(); it != m_ac_pr_aff_nodes.end(); it++)
		(*it)->reset_ac_pr_affected();
	m_ac_pr_aff_nodes.clear();
}



void Changes::reset_suc()
{
	for (set<Rev_Intern_NC*>::iterator it = m_suc_aff_node_contnrs.begin(); it != m_suc_aff_node_contnrs.end(); it++)
		(*it)->m_is_suc_affected = false;
	m_suc_aff_node_contnrs.clear();
}



void Changes::reset_pos()
{
	for (set<Rev_NC*>::iterator it = m_pos_aff_node_contnrs.begin(); it != m_pos_aff_node_contnrs.end(); it++)
		(*it)->m_is_pos_affected = false;
	m_pos_aff_node_contnrs.clear();
}



void Changes::reset_time()
{
	for (set<Rev_NC*>::iterator it = m_time_aff_node_contnrs.begin(); it != m_time_aff_node_contnrs.end(); it++)
		(*it)->m_is_time_affected = false;
	m_time_aff_node_contnrs.clear();
}



void Changes::reset_nd_st()
{
	for (set<Rev_NC*>::iterator it = m_nd_st_aff_node_contnrs.begin(); it != m_nd_st_aff_node_contnrs.end(); it++)
		(*it)->m_is_nd_st_affected = false;
	m_nd_st_aff_node_contnrs.clear();
}



ostream& operator<<(ostream& s, const Changes& changes)
{
	s << "*** CHANGES ***" << endl;

	s << "new:         ";
	for (set<Rev_NC*>::const_iterator it = changes.m_new_node_contnrs.begin(); it != changes.m_new_node_contnrs.end(); it++)
	{
		s << (*it)->str_id() << " ";
	}
	s << endl;

	s << "old:         ";
	for (set<Rev_NC*>::const_iterator it = changes.m_old_node_contnrs.begin(); it != changes.m_old_node_contnrs.end(); it++)
	{
		s << (*it)->str_id() << " ";
	}
	s << endl;

	s << "anc_aff:";
	for (set<Rev_Nd*>::const_iterator it = changes.m_anc_aff_nodes.begin(); it != changes.m_anc_aff_nodes.end(); it++)
	{
		s.width(3);
		s << (*it)->id();
	}
	s << endl;

	s << "suc_aff:";
	for (set<Rev_Intern_NC*>::const_iterator it = changes.m_suc_aff_node_contnrs.begin(); it != changes.m_suc_aff_node_contnrs.end(); it++)
	{
		s << (*it)->str_id() << " ";
	}
	s << endl;

	s << "pos_aff:";
	for (set<Rev_NC*>::const_iterator it = changes.m_pos_aff_node_contnrs.begin(); it != changes.m_pos_aff_node_contnrs.end(); it++)
	{
		s << (*it)->str_id() << " ";
	}
	s << endl;

	s << "nd_st_aff:";
	for (set<Rev_NC*>::const_iterator it = changes.m_nd_st_aff_node_contnrs.begin(); it != changes.m_nd_st_aff_node_contnrs.end(); it++)
	{
		s << (*it)->str_id() << " ";
	}
	s << endl;

	s << "ac_pr_aff:";
	for (set<Rev_Nd*>::const_iterator it = changes.m_ac_pr_aff_nodes.begin(); it != changes.m_ac_pr_aff_nodes.end(); it++)
	{
		s.width(3);
		s << (*it)->id();
	}
	s << endl;

	s << endl;

	return s;
}
