#include "standard.h"

#include "constants.h"
#include "error.h"
#include "gamma.h"
#include "files.h"
#include "random.h"

LoggerPtr Gamma_Distribution::s_logger(Logger::getLogger("tree.gamma"));



double Gamma_Distribution::log_prob_mutation_rate(const int idx) { return log(Constants::instance()->prob_mutation_rate(idx)); }



// integrand for integration performed in mutation_rate(...)
double integrand_gamma (double x, void * params)
{
	double alpha = *(double *) params;
	double beta = 1 / alpha;
	double f = x * gsl_ran_gamma_pdf(x, alpha, beta);
    return f;
}



// discretization of Gamma distribution according to Felsenstein, p. 261, (16.23)
double Gamma_Distribution::mutation_rate(const int idx)
{
	OUT_FNCTN_HEADER(logger());

	gsl_integration_workspace * w = gsl_integration_workspace_alloc (1000);

	double result, error;
	double alph = Constants::instance()->alpha();
	double bet = 1 / alph;

	gsl_function F;
	F.function = &integrand_gamma;
	F.params = &alph;

	double p = 0;
	for (int i = 0; i < idx; ++i)
		p += Constants::instance()->prob_mutation_rate(i);
	double a = gsl_cdf_gamma_Pinv(p, alph, bet);
	double b = 0;
	p += Constants::instance()->prob_mutation_rate(idx);
	if (idx + 1 < Constants::instance()->num_mutation_rates())
	{
		b = gsl_cdf_gamma_Pinv(p, alph, bet);
		gsl_integration_qags(&F, a, b, 0, 1e-7, 1000, w, &result, &error);
	}
	else // take care of b = inf
	{
		b = numeric_limits<double>::max();
		gsl_integration_qagiu(&F, a, 0, 1e-7, 1000, w, &result, &error);
	}
	result *= Constants::instance()->num_mutation_rates();

	gsl_integration_workspace_free (w);

	LOG4CXX_DEBUG(logger(), "idx: " << idx);
	LOG4CXX_DEBUG(logger(), "result: " << result);
	LOG4CXX_DEBUG(logger(), "a: " << a);
	LOG4CXX_DEBUG(logger(), "b: " << b)

	OUT_FNCTN_BOTTOM(logger());

	return result;
}



// sample a mutation rate using a Gamma distribution
double Gamma_Distribution::sample_mutation_rate()
{
	const double LOWER_LIMIT = 1e-6;
	double beta = 1 / Constants::instance()->alpha();
	return max(LOWER_LIMIT, gsl_ran_gamma (Random::rng(), Constants::instance()->alpha(), beta));
}


