#include "standard.h"

#include "constants.h"
#include "dot.h"
#include "files.h"
#include "nc.h"
#include "nd.h"
#include "sim_nc.h"
#include "sim_nd.h"
#include "sim_tree.h"
#include "tree.h"

LoggerPtr Sim_Tree::m_logger(Logger::getLogger("tree.sim_tree"));



Sim_Tree::Sim_Tree(const int num_sites, const int num_mutation_rates)
: Base_Tree<Sim_Nd, Sim_Coal_Nd, Sim_Rec_Nd, Sim_NC>(num_sites, num_mutation_rates)
{ }

Sim_Tree::~Sim_Tree() { }



// construct Sim_Tree from Rev_Tree
Sim_Tree::Sim_Tree(const Rev_Tree& tr)
: Base_Tree<Sim_Nd, Sim_Coal_Nd, Sim_Rec_Nd, Sim_NC>(tr.num_sites(), tr.num_mutation_rates())
{
	OUT_FNCTN_HEADER(logger());

	for (Rev_NC* c = tr.first_node_contnr(); c; c = c->next())
	{
		LOG4CXX_DEBUG(logger(), "c: " << c->str_id());

		if (c->is_tip())
		{
			LOG4CXX_DEBUG(logger(), "TIP");
			Rev_Tip_NC* d = dynamic_cast<Rev_Tip_NC*>(c);
			new Sim_Tip_NC(this, d->node()->id());
		}
		else
			if (c->has_coal_suc())
			{
				LOG4CXX_DEBUG(logger(), "COAL");

				Rev_Coal_NC* d = dynamic_cast<Rev_Coal_NC*>(c);
				Sim_Nd* m = first_node();
				while ((m) && (m->id() != d->first_child()->id()))
					m = m->next();
				Sim_Nd* first_old_node = m;

				m = first_node();
				while ((m) && (m->id() != d->second_child()->id()))
					m = m->next();
				Sim_Nd* second_old_node = m;

				LOG4CXX_DEBUG(logger(), "time: " << d->birth_time());
				LOG4CXX_DEBUG(logger(), "first_old_node: " << first_old_node->id());
				LOG4CXX_DEBUG(logger(), "second_old_node: " << second_old_node->id());
				LOG4CXX_DEBUG(logger(), "new node id: " << d->node()->id());
				new Sim_Coal_NC(this, d->birth_time(), first_old_node, second_old_node, d->node()->id());
			}
			else
				if (c->has_rec_suc())
				{
					LOG4CXX_DEBUG(logger(), "REC");

					Rev_Rec_NC* d = dynamic_cast<Rev_Rec_NC*>(c);
					Sim_Nd* m = first_node();
					while ((m) && (m->id() != d->child()->id()))
						m = m->next();
					Sim_Nd* old_node = m;

					VECTOR(int) bps;
					BOOL first_node_has_first_segm = true;
					int i = 0;
					while (i < num_sites())
					{
						if (old_node->is_act_site(i))
						{
							if (d->first_node()->is_act_site(i))
								first_node_has_first_segm = true;
							else
								first_node_has_first_segm = false;
							break;
						}
						++i;
					}

					BOOL first_is_active = first_node_has_first_segm;
					for (int i = 0; i < num_sites(); ++i)
						if (old_node->is_act_site(i))
						{
							if (first_is_active && d->second_node()->is_act_site(i))
							{
								bps.push_back(i - 1);
								first_is_active = false;
							}
							if (!first_is_active && d->first_node()->is_act_site(i))
							{
								bps.push_back(i - 1);
								first_is_active = true;
							}
						}

					int first_node_id, second_node_id;
					if (first_node_has_first_segm)
					{
						first_node_id = d->first_node()->id();
						second_node_id = d->second_node()->id();
					}
					else
					{
						first_node_id = d->second_node()->id();
						second_node_id = d->first_node()->id();
					}

					stringstream ss;
					ss << "first_node_has_first_segm: " << boolalpha << first_node_has_first_segm << endl;
					ss << "time: " << d->birth_time() << endl;
					ss << "old_node: " << old_node->id() << endl;
					ss << "bps: ";
					for (VECTOR(int)::iterator it = bps.begin(); it != bps.end(); ++it)
						ss << *it << " ";
					ss << endl;
					ss << "first_node_id: " << first_node_id << endl;
					ss << "second_node_id: " << second_node_id;
					LOG4CXX_DEBUG(logger(), ss.str());
					new Sim_Rec_NC(this, d->birth_time(), old_node, bps, first_node_id, second_node_id);
				}

		LOG4CXX_DEBUG(logger(), "");
	}

	OUT_FNCTN_BOTTOM(logger());
}



// builds a simulation tree
void Sim_Tree::generate()
{
	OUT_FNCTN_HEADER(logger());

	if (num_nodes() > 0)
		ERROR(logger(), "Trying to apply generate_tree to a Sim_Tree containing nodes");

	ifstream* tree_strm = In_Out_Streams::instance()->istrm(TREE_FILE);
	map<int, Sim_Nd*> nodes;
	map<string, Sim_NC*> node_contnrs;

	// generate nodes
	while (!tree_strm->eof())
	{
		// is current line for coalescent or recombination event?
		stringstream ss;
		string str;
		string token;
		BOOL is_coal;

		getline(*tree_strm, str);
		ss << str;
		ss >> token;
		ss >> token;
		if (token == "-")
			is_coal = false;
		else
			is_coal = true;

		ss.seekg (0, ios::beg);
		if (is_coal)
		// generate coalescent node
		{
			int num_left_child, num_right_child, num_parent;
			double birth_time;
			ss >> num_left_child;
			ss >> num_right_child;
			ss >> token;
			ss >> num_parent;
			ss >> birth_time;
			LOG4CXX_DEBUG(logger(), "COAL");
			LOG4CXX_DEBUG(logger(), "num_left_child: " << num_left_child);
			LOG4CXX_DEBUG(logger(), "num_right_child: " << num_right_child);
			LOG4CXX_DEBUG(logger(), "num_parent: " << num_parent);
			LOG4CXX_DEBUG(logger(), "birth_time: " << birth_time);

			map<int, Sim_Nd*>::iterator it;
			Sim_Nd* left_child;
			it = nodes.find(num_left_child);
			if (it == nodes.end())
			{
				Sim_Tip_NC* new_contnr = new Sim_Tip_NC(this, num_left_child);
				node_contnrs[new_contnr->str_id()] = new_contnr;
				left_child = new_contnr->node();
				nodes[num_left_child] = left_child;
				LOG4CXX_DEBUG(logger(), new_contnr->str_id() << " generated");
			}
			else
				left_child = (*it).second;

			Sim_Nd* right_child;
			it = nodes.find(num_right_child);
			if (it == nodes.end())
			{
				Sim_Tip_NC* new_contnr = new Sim_Tip_NC(this, num_right_child);
				node_contnrs[new_contnr->str_id()] = new_contnr;
				right_child = new_contnr->node();
				nodes[num_right_child] = right_child;
				LOG4CXX_DEBUG(logger(), new_contnr->str_id() << " generated");
			}
			else
				right_child = (*it).second;

			Sim_Coal_NC* parent = new Sim_Coal_NC(this, birth_time, left_child, right_child, num_parent);
			node_contnrs[parent->str_id()] = parent;
			nodes[parent->node()->id()] = parent->node();
			LOG4CXX_DEBUG(logger(), parent->str_id() << " generated");
		}
		else
		// generate recombination node
		{
			int num_child, num_left_parent, num_right_parent;
			double birth_time;
			VECTOR(int) bps;
			ss >> num_child;
			ss >> token;
			ss >> num_left_parent;
			ss >> num_right_parent;
			ss >> birth_time;
			while (!ss.eof())
			{
				int x;
				ss >> x;
				bps.push_back(x);
			}

			ss.str("");
			ss << "REC" << endl;
			ss << "num_child: " << num_child << endl;
			ss << "num_left_parent: " << num_left_parent << endl;
			ss << "num_right_parent: " << num_right_parent << endl;
			ss << "birth_time: " << birth_time << endl;
			ss << "bps: ";
			for (VECTOR(int)::iterator it = bps.begin(); it != bps.end(); ++it)
				ss << *it << " ";
			LOG4CXX_DEBUG(logger(), ss.str());

			map<int, Sim_Nd*>::iterator it;
			Sim_Nd* child;
			it = nodes.find(num_child);
			if (it == nodes.end())
			{
				Sim_Tip_NC* new_contnr = new Sim_Tip_NC(this, num_child);
				node_contnrs[new_contnr->str_id()] = new_contnr;
				child = new_contnr->node();
				nodes[num_child] = child;
				LOG4CXX_DEBUG(logger(), new_contnr->str_id() << " generated");
			}
			else
				child = (*it).second;

			Sim_Rec_NC* parent = new Sim_Rec_NC(this, birth_time, child, bps, num_left_parent, num_right_parent);
			node_contnrs[parent->str_id()] = parent;
			nodes[num_left_parent] = parent->first_node();
			nodes[num_right_parent] = parent->second_node();
			LOG4CXX_DEBUG(logger(), parent->str_id() << " generated");
		}
	}

	stringstream ss;
	ss << "node_contnrs: ";
	for (map<string, Sim_NC*>::iterator it = node_contnrs.begin(); it != node_contnrs.end(); ++it)
		ss << (*it).first << "   " << flush;
	ss << endl;
	LOG4CXX_DEBUG(logger(), ss.str());

	BOOL is_first_pos = true;
	Sim_NC* prev_contnr;
	while (node_contnrs.size() > 0)
	{
		map<string, Sim_NC*>::iterator it;
		string argmin = "";
		double min = 0;
		BOOL is_first_node = true;
		for (it = node_contnrs.begin(); it != node_contnrs.end(); it++)
		{
			if (is_first_node || ((*it).second->birth_time() < min))
			{
				argmin = (*it).first;
				min = (*it).second->birth_time();
			}
			is_first_node = false;
		}

		if (is_first_pos)
		{
			set_first_node_contnr(node_contnrs[argmin]);
			node_contnrs[argmin]->prev() = NULL;
		}
		else
		{
			prev_contnr->next() = node_contnrs[argmin];
			node_contnrs[argmin]->prev() = prev_contnr;
		}

		if (node_contnrs.size() == 1)
		{
			set_last_node_contnr(node_contnrs[argmin]);
			node_contnrs[argmin]->next() = NULL;
		}
		prev_contnr = node_contnrs[argmin];
		is_first_pos = false;
		node_contnrs.erase(argmin);
	}

	OUT_FNCTN_BOTTOM(logger());
}



VECTOR(string) Sim_Tree::sim_seq(const BOOL is_mut_rate_const)
{
	OUT_FNCTN_HEADER(logger());

	if (num_act_nodes() != 1)
		ERROR(logger(), "Trying to simulate sequences using a tree with number of active node != 1");

	double base_probs[NUM_NUCL];
	for (int i = 0; i < NUM_NUCL; ++i)
		base_probs[i] = Constants::instance()->base_prob(i);
	for (int i = 0; i < num_sites(); i++) {
		/*gsl_ran_discrete_t* grd = gsl_ran_discrete_preproc (NUM_NUCL, base_probs);
		int j = gsl_ran_discrete (Random::rng(), grd);
		gsl_ran_discrete_free(grd);
		switch (j) {
		case 0:
			last_node()->m_seq[i] = 'A';
			break;
		case 1:
			last_node()->m_seq[i] = 'C';
			break;
		case 2:
			last_node()->m_seq[i] = 'G';
			break;
		case 3:
			last_node()->m_seq[i] = 'T';
			break;
		default:
			last_node()->m_seq[i] = 'A';
			break;
		}*/
		last_node()->set_seq_pos(i, 0);
	}

	// determine site-wise mutation rates
	VECTOR(double) sitewise_mutation_rates(num_sites());
	for (int i = 0; i < num_sites(); i++)
		if (is_mut_rate_const)
			sitewise_mutation_rates[i] = 1;
		else
			sitewise_mutation_rates[i] = Gamma_Distribution::sample_mutation_rate();

	stringstream ss;
	for (int i = 0; i < num_sites(); i++)
		ss << "site: " << setw(4) << i << " mutation rate: " << sitewise_mutation_rates[i] << endl;
	LOG4CXX_DEBUG(logger(), ss.str());

	for (int i = 0; i < num_sites(); i++)
		last_node()->sim_mutation(i, sitewise_mutation_rates[i]);

	VECTOR(string) str;
	Sim_Nd* n = first_node();
	while (n->is_tip())
	{
		str.push_back(n->acid_seq());
		n = n->next();
	}

	OUT_FNCTN_BOTTOM(logger());

	return str;
}



// get the prior probability
double Sim_Tree::log_prior_prob()
{
	calc_num_elig_nd_st(first_node_contnr(), NULL);
	calc_prior_prob(first_node_contnr());
	return last_node_contnr()->P_prior();
}



// check integrity of all node containers
void Sim_Tree::check_node_contnr() const
{
	set<string> node_contnrs;
	for (Sim_NC* c = first_node_contnr(); c; c = c->next()) {
		if (!node_contnrs.insert(c->str_id()).second)
			break;
		c->check();
	}}



// check integrity
void Sim_Tree::check() const
{
	Base_Tree<Sim_Nd, Sim_Coal_Nd, Sim_Rec_Nd, Sim_NC>::check();

	BOOL is_ok = true;
	stringstream ss;
}



// output nodes to text file
ostream& operator<<(ostream& s, const Sim_Tree& tr)
{
	s << "*** TREE ***" << endl;

	s << "first: ";
	if (tr.first_node())
		s << tr.first_node()->id();
	else
		s << "NULL";
	s << endl;

	s << "last: ";
	if (tr.last_node())
		s << tr.last_node()->id();
	else
		s << "NULL";
	s << endl;

	s << "num_nodes: " << tr.num_nodes() << endl;
	s << "num_tips: " << tr.num_tips() << endl;
	s << "num_act_nodes: " << tr.num_act_nodes() << endl;

	s << "mutation rates: ";
	for (int i = 0; i < tr.num_mutation_rates(); ++i)
		s << tr.mutation_rate(i) << " ";
	s << endl << endl;

	for (Sim_Nd* n = tr.first_node(); n; n = n->next())
	{
		s.width(3);
		s << n->id() << " " << flush;
		s << n->acid_seq() << "   " << flush;

		s.width(8);
		s.setf(ios::fixed,ios::floatfield);
		s.precision(3);
		s << n->birth_time() << "   " << flush;

		s.width(5);
		s << n->num_elig_sites() << "   " << flush;

		if (n->prev())
		{
			s.width(4);
			s << n->prev()->id() << " " << flush;
		}
		else
			s << "NULL " << flush;

		if (n->next())
		{
			s.width(4);
			s << n->next()->id() << " " << flush;
		}
		else
			s << "NULL " << flush;
		s << "   ";

		if (n->is_active())
		{
			s << "A              " << flush;
		}
		if (n->has_coal_anc())
		{
			if (n->is_first_child())
			{
				s << "C1" << flush;
				s.width(3);
				s << n->parent_node()->id() << flush;
				s << "(P) " << flush;
				s.width(3);

				s << n->parent_node()->second_child_node()->id() << flush;
				s << "(B)" << flush;
			}
			else
			{
				s << "C2" << flush;
				s.width(3);
				s << n->parent_node()->id() << flush;
				s << "(P) " << flush;
				s.width(3);
				s << n->parent_node()->first_child_node()->id() << flush;
				s << "(B)" << flush;
			}
		}
		if (n->has_rec_anc())
		{
			s << "R " << flush;
			s.width(3);
			s << n->first_parent_node()->id() << flush;
			s << "(P) " << flush;
			s.width(3);
			s << n->second_parent_node()->id() << flush;
			s << "(P)" << flush;
		}
		s << "   ";


		if (n->is_tip())
		{
			s << "T              " << flush;
		}
		if (n->has_coal_suc())
		{
			Sim_Coal_Nd* m = dynamic_cast<Sim_Coal_Nd*>(n);
			s << "C " << flush;
			s.width(3);
			s << m->first_child_node()->id() << flush;
			s << "(C) " << flush;
			s.width(3);
			s << m->second_child_node()->id() << flush;
			s << "(C)" << flush;
		}
		if (n->has_rec_suc())
		{
			Sim_Rec_Nd* m = dynamic_cast<Sim_Rec_Nd*>(n);
			if (m->is_first_in_NC())
			{
				s << "R1" << flush;
				s.width(3);
				s << m->child_node()->id() << flush;
				s << "(C) " << flush;
				s.width(3);
				s << m->child_node()->second_parent_node()->id() << flush;
				s << "(S)" << flush;
			}
			else
			{
				s << "R2" << flush;
				s.width(3);
				s << m->child_node()->id() << flush;
				s << "(C) " << flush;
				s.width(3);
				s << m->child_node()->first_parent_node()->id() << flush;
				s << "(S)" << flush;
			}
		}
		s << endl;
	}
	s << endl;

	return s;
}



double aux_optim_para(double x, void *params)
{
	Constants::instance()->recomb_const() = x;
	Sim_Tree* tr = (Sim_Tree*) params;
	return -tr->log_prior_prob();
}





void optim_para(Sim_Tree* tr)
{
	int status;
	int iter = 0, max_iter = 100;
	const gsl_min_fminimizer_type *T;
	gsl_min_fminimizer *s;
	double m = Constants::instance()->recomb_const();
	double a = 0.0, b = 10.0;
	gsl_function F;

	F.function = &aux_optim_para;
	F.params = tr;

	T = gsl_min_fminimizer_brent;
	s = gsl_min_fminimizer_alloc (T);
	gsl_min_fminimizer_set (s, &F, m, a, b);

	do
	{
		iter++;
	    status = gsl_min_fminimizer_iterate (s);

	    m = gsl_min_fminimizer_x_minimum (s);
	    a = gsl_min_fminimizer_x_lower (s);
	    b = gsl_min_fminimizer_x_upper (s);

	    status = gsl_min_test_interval (a, b, 0.001, 0.0);
	}
	while (status == GSL_CONTINUE && iter < max_iter);

	gsl_min_fminimizer_free (s);

	cout << "recomb_const: " << Constants::instance()->recomb_const() << endl;
}



