#include "standard.h"

#include "changes.h"
#include "constants.h"
#include "dot.h"
#include "files.h"
#include "nc.h"
#include "nd.h"
#include "prt_nd_conn.h"
#include "random.h"
#include "trans_probs.h"
#include "tr_rearr_coal.h"
#include "tree.h"

LoggerPtr tree_rearranger_coal::s_logger(Logger::getLogger("tree.tr_rearr_coal"));

struct two_coal_intgr_fract_params { int z; double t; double gamma; VECTOR(int) diff_l; };
struct intgrl_time_next_event_params { int num_pas_nodes; double tot_time; Part_Node_Conn conn; double target_intgrl; };

class Dot;



// generate time of one coalescent event
// j - number of active nodes
// z - number of eligible passive nodes
// t - available time interval
double generate_one_coal(int j, int z, int diff_l, double t)
{
	double alpha = j - 1 + z + Constants::instance()->recomb_const() * diff_l;
	double gamma = gsl_rng_uniform(Random::rng());
	double ret = - log(1 - gamma * (1 - exp(-alpha * t))) / alpha;
	return ret;
}



// auxiliary function for 'generate_two_coal'
double two_coal_intgr_fract (double r, void *params)
{
	struct two_coal_intgr_fract_params *p = (struct two_coal_intgr_fract_params *) params;

	int z = p->z;
	double t = p->t;
	double gamma = p->gamma;
	VECTOR(int) diff_l = p->diff_l;
	double rho = Constants::instance()->recomb_const();

	double alpha_3_1 = 3 + 2 * z + rho * diff_l[0];
	double alpha_3_2 = 2 + z + rho * diff_l[1];
	double alpha_2_1 = 1 + z + rho * diff_l[2];

	double x = (alpha_3_2 * (1 - exp(-alpha_3_1 * r)) - alpha_3_1 * (exp(-alpha_2_1 * r) - exp(-alpha_3_1 * r))) /
			   (alpha_3_2 * (1 - exp(-alpha_3_1 * t)) - alpha_3_1 * (exp(-alpha_2_1 * t) - exp(-alpha_3_1 * t))) - gamma;
	return x;
}



// generate time of two coalescent event
void generate_two_coal(int z, double t, VECTOR(int) diff_l, double ret[2])
{
	const double MAX_ERROR = 1e-6;

	double x_lo = 0.0, x_hi = t;
	struct two_coal_intgr_fract_params params = { z, t, gsl_rng_uniform(Random::rng()), diff_l };
	while (x_hi - x_lo > MAX_ERROR)
	{
		if (two_coal_intgr_fract((x_lo + x_hi) / 2, &params) > 0)
			x_hi = (x_lo + x_hi) / 2;
		else
			x_lo = (x_lo + x_hi) / 2;
	}

	ret[1] = (x_lo + x_hi) / 2;
	ret[0] = generate_one_coal(3, z, diff_l[1], (x_lo + x_hi) / 2);
}



// calculate S and P (cf. Kuhner 95, Appendix) and number of eligible passive nodes
// jump - third child (cf. Kuhner 95, Fig. 4)
void tree_rearranger_coal::calculate_S_and_P(Rev_Tree* tr, VECTOR(Rev_Nd*) children, Rev_NC* start, Rev_NC* stop,
					  					     Rev_NC* jump, BOOL start_at_second_child)
{
	OUT_FNCTN_HEADER(logger());
	LOG4CXX_DEBUG(logger(), "start: " << start->str_id());
	LOG4CXX_DEBUG(logger(), "stop: " << stop->str_id());
	LOG4CXX_DEBUG(logger(), "jump: " << (jump ? jump->str_id() : "NULL"));

	double curr_time = 0;
	Rev_NC* c = start;
	Rev_NC* pred = start->prev();
	if (pred == NULL)
		throw ERROR(logger(),"Illegal start node in calculation of S and P");

	while (!pred->is_unmarked())
		Rev_NC* pred = pred->prev();
	while(*pred != *stop)
	{
		LOG4CXX_DEBUG(logger(), "c: " << c->str_id() << " " << c->birth_time());

		c->s_and_p().calc_elig_sites(c->num_cum_elig_sites(), children);
		if (!c->is_last())
			c->s_and_p().calc_trans_prob(c->num_elig_nodes(), c->next()->birth_time() - c->birth_time());

		// calculate S
		if (*c == *start)
			c->s_and_p().set_S_to_start(start_at_second_child);
		else {
			for (int i = 0; i < S_and_P::num_S; i++)
			{
				c->s_and_p().S(i) = 0;
				for (int j = i; j < S_and_P::num_S; j++)
					c->s_and_p().S(i) += pred->s_and_p().S(j) *
									     pred->s_and_p().P((S_and_P::Nds_Left)j, (S_and_P::Nds_Left)i);
			}

			// treat case where n is the third child
			if ((jump) && (*c == *jump))
				c->s_and_p().jump();
		}

		// output S and P
		stringstream ss;
		for (int i = 0; i < S_and_P::num_S; ++i)
			ss << "S[" << i << "]: " << c->s_and_p().S(i) << "   ";
		ss << endl;
		if (!c->is_last()) {
			for (int i = 0; i < S_and_P::num_S; ++i)
				for (int j = 0; j < S_and_P::num_S; ++j)
					if (c->s_and_p().P((S_and_P::Nds_Left)i, (S_and_P::Nds_Left)j) > 0)
						ss << "P[" << i << "," << j << "] = "
								  << c->s_and_p().P((S_and_P::Nds_Left)i, (S_and_P::Nds_Left)j) << endl;
			for (int i = 0; i < 3; ++i)
				ss << "q[" << i << "] = " << c->s_and_p().q(i) << endl;
		}
		for (int i = 0; i < 6; ++i)
			ss << "num_elig_sites[" << i << "] = " << c->s_and_p().elig_sites(i) << endl;
		ss << endl;
		LOG4CXX_DEBUG(logger(), ss.str());

		pred = c;
		if (!c->is_last())
			do
				c = c->next();
			while (!c->is_unmarked() && (c->next()));
	}

	OUT_FNCTN_BOTTOM(logger());
}



// get the time of the two coalescent events (occuring when doing rearrange operation "coalescent")
void tree_rearranger_coal::get_coal_times(Rev_Tree* tr, Rev_NC* end, Rev_NC* start, Rev_NC* third_child,
										  VECTOR(double)* time_of_new_nodes, int* node_left)
{
	OUT_FNCTN_HEADER(logger());

	Rev_NC* c = end;
	Rev_NC* pred = c;
	if (!pred->is_first())
		do
			pred = pred->prev();
		while (!pred->is_unmarked() && !pred->is_first());

	BOOL third_child_passed = false;
	double curr_time = 0;
	while(c != start)
	{
		LOG4CXX_DEBUG(logger(), "pred: " << setw(3) << pred->str_id());
		LOG4CXX_DEBUG(logger(), "c: " << setw(3) << c->str_id());

		pred->index() = c->index();
		curr_time = pred->birth_time();
		double time_diff = c->birth_time() - pred->birth_time();
		if (*c == *third_child)
		{   // n is the third child (index() is increased by one if no coalescent event happens)
			if (c->index() == S_and_P::THIRD_LEFT)
				if (pred->s_and_p().S(S_and_P::NONE_LEFT) * pred->s_and_p().P(S_and_P::NONE_LEFT, S_and_P::NONE_LEFT) <
					gsl_rng_uniform(Random::rng()) * c->s_and_p().S(S_and_P::THIRD_LEFT))
				{
					LOG4CXX_DEBUG(logger(), "generate_one_coal");
					double t =
						generate_one_coal(2, pred->num_elig_nodes(),
										  c->s_and_p().elig_sites(S_and_P::ALL_LEFT)
										  - c->s_and_p().elig_sites(S_and_P::THIRD_LEFT), time_diff);
					time_of_new_nodes->insert(time_of_new_nodes->begin(), curr_time + t);
					pred->index() = S_and_P::FIRST_TWO_LEFT;
					*node_left = 2;
				}
				else
					pred->index() = S_and_P::NONE_LEFT;
			else
				pred->index() = S_and_P::FIRST_TWO_LEFT;
			third_child_passed = true;
		}
		else
		{   // n is not the third child (normal case)
			switch (c->index())
			{
				case S_and_P::NONE_LEFT:
				{
					double p[3] = { GSL_MAX(pred->s_and_p().q(0) * pred->s_and_p().S(S_and_P::FIRST_LEFT), 0),
									GSL_MAX(pred->s_and_p().q(1) * pred->s_and_p().S(S_and_P::SECOND_LEFT), 0),
									GSL_MAX(pred->s_and_p().q(2) * pred->s_and_p().S(S_and_P::THIRD_LEFT), 0)};
					LOG4CXX_DEBUG(logger(), "p: " << p[0] << " " << p[1] << " " << p[2]);
					gsl_ran_discrete_t* grd = gsl_ran_discrete_preproc (3, p);
					int i = gsl_ran_discrete (Random::rng(), grd);
					gsl_ran_discrete_free(grd);

					VECTOR(int) diff_l(3);
					diff_l[0] = c->s_and_p().elig_sites(S_and_P::ALL_LEFT) - c->s_and_p().elig_sites(S_and_P::NONE_LEFT);
					switch (i)	{
					case 0:
						diff_l[1] = c->s_and_p().elig_sites(S_and_P::ALL_LEFT) - c->s_and_p().elig_sites(S_and_P::S_and_P::FIRST_LEFT);
						diff_l[2] = c->s_and_p().elig_sites(S_and_P::S_and_P::FIRST_LEFT) - c->s_and_p().elig_sites(S_and_P::NONE_LEFT);
						*node_left = 0;
						break;
					case 1:
						diff_l[1] = c->s_and_p().elig_sites(S_and_P::ALL_LEFT) - c->s_and_p().elig_sites(S_and_P::SECOND_LEFT);
						diff_l[2] = c->s_and_p().elig_sites(S_and_P::SECOND_LEFT) - c->s_and_p().elig_sites(S_and_P::NONE_LEFT);
						*node_left = 1;
						break;
					case 2:
						diff_l[1] = c->s_and_p().elig_sites(S_and_P::ALL_LEFT) - c->s_and_p().elig_sites(S_and_P::THIRD_LEFT);
						diff_l[2] = c->s_and_p().elig_sites(S_and_P::THIRD_LEFT) - c->s_and_p().elig_sites(S_and_P::NONE_LEFT);
						*node_left = 2;
						break;
					}

					double x = gsl_rng_uniform(Random::rng());
					if (pred->s_and_p().S(S_and_P::NONE_LEFT) * pred->s_and_p().P(S_and_P::NONE_LEFT, S_and_P::NONE_LEFT) < x * c->s_and_p().S(S_and_P::NONE_LEFT))
						if ((pred->s_and_p().S(S_and_P::NONE_LEFT) * pred->s_and_p().P(S_and_P::NONE_LEFT, S_and_P::NONE_LEFT) +
							 pred->s_and_p().S(S_and_P::FIRST_LEFT) * pred->s_and_p().P(S_and_P::FIRST_LEFT, S_and_P::NONE_LEFT) +
							 pred->s_and_p().S(S_and_P::SECOND_LEFT) * pred->s_and_p().P(S_and_P::SECOND_LEFT, S_and_P::NONE_LEFT) +
							 pred->s_and_p().S(S_and_P::THIRD_LEFT) * pred->s_and_p().P(S_and_P::THIRD_LEFT, S_and_P::NONE_LEFT) < x * c->s_and_p().S(S_and_P::NONE_LEFT)
							 && !third_child_passed))
						{
							LOG4CXX_DEBUG(logger(), "generate_two_coal");
							double t[2];
							generate_two_coal(pred->num_elig_nodes(), time_diff, diff_l, t);
							time_of_new_nodes->insert(time_of_new_nodes->begin(), curr_time + t[0]);
							time_of_new_nodes->insert(time_of_new_nodes->begin(), curr_time + t[1]);
							pred->index() = S_and_P::ALL_LEFT;
						}
						else
						{
							LOG4CXX_DEBUG(logger(), "generate_one_coal");
							double t = generate_one_coal(2, pred->num_elig_nodes(), diff_l[1], time_diff);
							time_of_new_nodes->insert(time_of_new_nodes->begin(), curr_time + t);
							if (third_child_passed)
								pred->index() = S_and_P::FIRST_TWO_LEFT;
							else
								switch (i)	{
								case 0:
									pred->index() = S_and_P::FIRST_LEFT;
									break;
								case 1:
									pred->index() = S_and_P::SECOND_LEFT;
									break;
								case 2:
									pred->index() = S_and_P::THIRD_LEFT;
									break;
								}
						}
					break;
				}
				case S_and_P::FIRST_LEFT:
					if (!third_child_passed)
						if (pred->s_and_p().S(S_and_P::FIRST_LEFT) * pred->s_and_p().P(S_and_P::FIRST_LEFT, S_and_P::FIRST_LEFT) < gsl_rng_uniform(Random::rng()) * c->s_and_p().S(S_and_P::FIRST_LEFT))
						{
							LOG4CXX_DEBUG(logger(), "generate_one_coal");
							double t =
								generate_one_coal(3, pred->num_elig_nodes(),
												  c->s_and_p().elig_sites(S_and_P::ALL_LEFT) - c->s_and_p().elig_sites(S_and_P::FIRST_LEFT), time_diff);
							time_of_new_nodes->insert(time_of_new_nodes->begin(), curr_time + t);
							pred->index() = S_and_P::ALL_LEFT;
						}
					break;
				case S_and_P::SECOND_LEFT:
					if (!third_child_passed)
						if (pred->s_and_p().S(S_and_P::SECOND_LEFT) * pred->s_and_p().P(S_and_P::SECOND_LEFT, S_and_P::SECOND_LEFT) < gsl_rng_uniform(Random::rng()) * c->s_and_p().S(S_and_P::SECOND_LEFT))
						{
							LOG4CXX_DEBUG(logger(), "generate_one_coal");
							double t =
								generate_one_coal(3, pred->num_elig_nodes(),
												  c->s_and_p().elig_sites(S_and_P::ALL_LEFT) - c->s_and_p().elig_sites(S_and_P::SECOND_LEFT), time_diff);
							time_of_new_nodes->insert(time_of_new_nodes->begin(), curr_time + t);
							pred->index() = S_and_P::ALL_LEFT;
						}
					break;
				case S_and_P::THIRD_LEFT:
					if (!third_child_passed)
						if (pred->s_and_p().S(S_and_P::THIRD_LEFT) * pred->s_and_p().P(S_and_P::THIRD_LEFT, S_and_P::THIRD_LEFT) < gsl_rng_uniform(Random::rng()) * c->s_and_p().S(S_and_P::THIRD_LEFT))
						{
							LOG4CXX_DEBUG(logger(), "generate_one_coal");
							double t =
								generate_one_coal(3, pred->num_elig_nodes(),
												  c->s_and_p().elig_sites(S_and_P::ALL_LEFT) - c->s_and_p().elig_sites(S_and_P::THIRD_LEFT), time_diff);
							time_of_new_nodes->insert(time_of_new_nodes->begin(), curr_time + t);
							pred->index() = S_and_P::ALL_LEFT;
						}
					break;
				default:
					break;
				}
		}

		LOG4CXX_DEBUG(logger(), "pred index():" << pred->index() << endl);

		c = pred;
		if (!pred->is_first())
			do
				pred = pred->prev();
			while (!pred->is_unmarked() && !pred->is_first());
	}

	// output time of the new nodes
	stringstream ss;
	ss << "time_of_new_nodes: ";
	for (VECTOR(double)::iterator jt = time_of_new_nodes->begin(); jt != time_of_new_nodes->end(); jt++)
		ss << (*jt) << " ";
	ss << endl;
	LOG4CXX_DEBUG(logger(), ss.str());

	if (time_of_new_nodes->size() > 2)
		throw ERROR(logger(), "Generate more than two times");

	OUT_FNCTN_BOTTOM(logger());
}



BOOL compare_birth_time (Rev_Nd* first, Rev_Nd* second)
{
	if (first->birth_time() < second->birth_time())
		return true;
	if (first->birth_time() == second->birth_time())
	{
		Rev_Nd* n = first;
		while (n)
		{
			if (*n == *second)
				return true;
			n = n->next();
		}
	}
	return false;
}



// perform rearrange opeartion "coalescent" (cf. Kuhner 95)
// topology_changeable - may the topology of the nodes be changed
void tree_rearranger_coal::coal(Rev_Tree* tr, Rev_Coal_Nd* target, BOOL topology_changeable)
{
	OUT_FNCTN_HEADER(logger());
	LOG4CXX_DEBUG(logger(), "target:       " << target->id());

	const BOOL OUTPUT_DOT = 0;

	if (tr->changes()->rearranging_done())
		throw ERROR(logger(),"Rearranging coalescent: Illegal Rev_Tree");

	tr->changes()->set_rearranging_done();
	tr->save_first_last();

	// get the children of the target and its parent
	list<Rev_Nd*> lst;
	lst.push_back(target->first_child_node());
	lst.push_back(target->second_child_node());
	if (target->is_first_child())
		lst.push_back(target->parent_node()->second_child_node());
	else
		lst.push_back(target->parent_node()->first_child_node());

	lst.sort(compare_birth_time);
	VECTOR(Rev_Nd*) children (lst.begin(), lst.end());
	VECTOR(Rev_Nd*)::iterator jt = children.begin();
	Rev_Nd* first_child = *jt++;
	Rev_Nd* second_child = *jt++;
	Rev_Nd* third_child = *jt;
	first_child->set_anc_affected();
	second_child->set_anc_affected();
	third_child->set_anc_affected();

	// get parent and grandparent of target
	Rev_Intern_Nd* parent = target->parent_node();
	Rev_Intern_NC* grandparent = 0;
	BOOL has_grandparent;
	BOOL parent_is_recomb = (parent->has_rec_anc());
	if (!parent->is_last()) {
		has_grandparent = true;
		grandparent = parent->parent_contnr();
	}
	else
		has_grandparent = false;

	// output target, parent, grandparent and children
	LOG4CXX_DEBUG(logger(), "parent:       " << parent->id());
	if (has_grandparent) {
		LOG4CXX_DEBUG(logger(), "grandparent:  " << grandparent->str_id());
	}
	else
		LOG4CXX_DEBUG(logger(), "grandparent: none");
	LOG4CXX_DEBUG(logger(), "first_child:  " << first_child->id());
	LOG4CXX_DEBUG(logger(), "second_child: " << second_child->id());
	LOG4CXX_DEBUG(logger(), "third_child:  " << third_child->id());

	// set target and parent to be old
	target->set_old();
	parent->set_old();
	tr->adjust_pos();
	tr->calc_num_elig_nd_st(target->prev()->rev_node_contnr(), tr->last_node_contnr());

	BOOL third_coals_second;   // is the third child not involved in the first coalescent event?
	if ((first_child->parent_node()->id() == target->id()) && (second_child->parent_node()->id() == target->id()))
		third_coals_second = true;
	else
		third_coals_second = false;

	// cf. 'rearrange_coal.pdf'
	BOOL no_coal_occured = true;
	Rev_Nd* last_child;
	if (topology_changeable || has_grandparent || third_coals_second)
	{
		if (has_grandparent)
			grandparent->set_suc_affected();

		// calculate S and P
		if (has_grandparent && (topology_changeable || third_coals_second))
			calculate_S_and_P(tr, children, second_child->rev_node_contnr(), grandparent, third_child->rev_node_contnr(), true);
		else
			if (has_grandparent)
				calculate_S_and_P(tr, children, third_child->rev_node_contnr(), grandparent, third_child->rev_node_contnr(), false);
			else
				calculate_S_and_P(tr, children, second_child->rev_node_contnr(), third_child->rev_node_contnr(),
						          third_child->rev_node_contnr(), true);

		// set index() and end node
		Rev_NC* start;
		Rev_NC* end;
		if (has_grandparent)
		{
			end = grandparent;
			end->index() = S_and_P::NONE_LEFT;
		}
		else
		{
			end = third_child->rev_node_contnr();

			double p[4] = { end->s_and_p().S(S_and_P::FIRST_LEFT), end->s_and_p().S(S_and_P::SECOND_LEFT),
							end->s_and_p().S(S_and_P::THIRD_LEFT), end->s_and_p().S(S_and_P::ALL_LEFT) };
			gsl_ran_discrete_t* grd = gsl_ran_discrete_preproc (3, p);
			int i = gsl_ran_discrete (Random::rng(), grd);
			gsl_ran_discrete_free(grd);
			switch (i) {
			case 0:
				end->index() = S_and_P::FIRST_LEFT;
				break;
			case 1:
				end->index() = S_and_P::SECOND_LEFT;
				break;
			case 2:
				end->index() = S_and_P::THIRD_LEFT;
				break;
			case 3:
				end->index() = S_and_P::ALL_LEFT;
				break;
			}
		}

		// start node
		if (topology_changeable || third_coals_second)
			start = second_child->rev_node_contnr();
		else
			start = third_child->rev_node_contnr();

		// get time of the two coalescent events
		VECTOR(double) time_of_new_nodes;
		int node_left;
		if (end->birth_time() > start->birth_time())
			get_coal_times(tr, end, start, third_child->rev_node_contnr(), &time_of_new_nodes, &node_left);

		stringstream ss;
		ss << "time_of_new_nodes: ";
		for (VECTOR(double)::iterator it = time_of_new_nodes.begin(); it != time_of_new_nodes.end(); ++it)
			ss << *it << " ";
		LOG4CXX_DEBUG(logger(), ss.str());

		// output Dot file of S and P
		if (OUTPUT_DOT)
		{
			if (has_grandparent && (topology_changeable || third_coals_second))
				Dot::to_dot(second_child->rev_node_contnr(), grandparent, third_child->rev_node_contnr());
			else
				if (has_grandparent)
					Dot::to_dot(third_child->rev_node_contnr(), grandparent, third_child->rev_node_contnr());
				else
					Dot::to_dot(second_child->rev_node_contnr(), third_child->rev_node_contnr(),
								third_child->rev_node_contnr());
		}

		// generate two new coalescent nodes
		BOOL is_first_coal = true;
		Rev_Coal_Nd* first_new_node;
		Rev_Coal_Nd* second_new_node;
		for (VECTOR(double)::iterator jt = time_of_new_nodes.begin(); jt != time_of_new_nodes.end(); jt++)
		{
			if (is_first_coal)
			{
				if (!topology_changeable)
				{
					Rev_Coal_NC* new_node_contnr =
						new Rev_Coal_NC(tr, *jt, target->first_child_node(), target->second_child_node());
					first_new_node = new_node_contnr->node();

					if (target->is_first_child())
						last_child = target->parent_node()->second_child_node();
					else
						last_child = target->parent_node()->first_child_node();
				}
				else
					if (((*jt) < third_child->birth_time()))
					{   // third child cannot take part in first coalescent event
						Rev_Coal_NC* new_node_contnr = new Rev_Coal_NC(tr, *jt, first_child, second_child);
						first_new_node = new_node_contnr->node();
						last_child = third_child;
					}
					else
					{
						Rev_Nd* first;
						Rev_Nd* second;
						switch (node_left) {
						case 0:
							first = second_child;
							second = third_child;
							last_child = first_child;
							break;
						case 1:
							first = first_child;
							second = third_child;
							last_child = second_child;
							break;
						case 2:
							first = first_child;
							second = second_child;
							last_child = third_child;
							break;
						}

						Rev_Coal_NC* new_node_contnr = new Rev_Coal_NC(tr, *jt, first, second);
						first_new_node = new_node_contnr->node();
					}
				is_first_coal = false;
			}
			else
			{
				Rev_Coal_NC* new_node_contnr = new Rev_Coal_NC(tr, *jt, last_child, first_new_node);
				second_new_node = new_node_contnr->node();
				if (has_grandparent)
					if (parent_is_recomb)
						second_new_node->set_recomb_anc(parent->rec_parent_contnr());
					else
						second_new_node->set_coal_anc(parent->coal_parent_contnr(), parent->is_first_child());
			}
		}
		no_coal_occured = is_first_coal;
	} // if (topology_changeable || has_grandparent || third_coals_second)

	// generate coalescent events if target has no grandparent
	if (!has_grandparent)
	{
		if (no_coal_occured)
		{
			first_child->set_active();
			second_child->set_active();
			third_child->set_active();
		}
		else
			last_child->set_active();
		tr->generate_final();
	}

	OUT_FNCTN_BOTTOM(logger());
}
