#include "standard.h"

#include "changes.h"
#include "constants.h"
#include "dot.h"
#include "files.h"
#include "nc.h"
#include "nd.h"
#include "random.h"
#include "tr_rearr_sngl.h"
#include "tree.h"

LoggerPtr tree_rearranger_sngl::s_logger(Logger::getLogger("tree.tr_rearr_sngl"));



double tree_rearranger_sngl::gen_single_event(double lambda, double T)
{
	if (lambda > 0)
	{
		double F = gsl_cdf_exponential_P(T, 1 / lambda);
		return gsl_cdf_exponential_Pinv(gsl_rng_uniform(Random::rng()) * F, 1 / lambda);
	}
	else
	{
		double F = gsl_cdf_exponential_P(T, -1 / lambda);
		double t = T - gsl_cdf_exponential_Pinv(gsl_rng_uniform(Random::rng()) * F, -1 / lambda);
	}
}



double tree_rearranger_sngl::get_ratio_mcmc(Rev_NC* target, const BOOL is_down_op, double* t_new)
{
	OUT_FNCTN_HEADER(logger());

	int num_1, num_down_2, num_up_2, num_up_3;
	double T_down, T_up;
	double t_down, t_up;
	BOOL is_target_rec = target->is_double();
	BOOL is_mid_rec;
	double r = Constants::instance()->recomb_const();
	double lambda_down, lambda_up;
	double lambda_2, mu_2;
	double s, delta_1, delta_2;
	Rev_NC* n_down;
	Rev_NC* n_up;

	if (is_down_op)
	{
		n_down = target->prev()->prev();
		n_up = target->prev();
	}
	else
	{
		n_down = target->prev();
		n_up = target->next();
	}

	is_mid_rec = n_up->is_double();
	num_1 = n_down->num_elig_nodes() + 1;

	if (is_target_rec)
		num_down_2 = num_1 + 1;
	else
		num_down_2 = num_1 - 1;

	if (is_mid_rec)
		num_up_2 = num_1 + 1;
	else
		num_up_2 = num_1 - 1;
	if (is_target_rec)
		num_up_3 = num_up_2 + 1;
	else
		num_up_3 = num_up_2 - 1;

	if (is_down_op)
	{
		T_down = target->prev()->birth_time() - target->prev()->prev()->birth_time();
		T_up = target->next()->birth_time() - target->prev()->birth_time();
		t_up = target->birth_time() - target->prev()->birth_time();
	}
	else
	{
		T_down = target->next()->birth_time() - target->prev()->birth_time();
		T_up = target->next()->next()->birth_time() - target->next()->birth_time();
		t_down = target->birth_time() - target->prev()->birth_time();
	}

	s = n_down->num_elig_sites();
	if (is_target_rec) {
		Rev_Rec_NC* c = dynamic_cast<Rev_Rec_NC*>(target);
		delta_1 = c->num_elig_sites() - c->child()->num_elig_sites();
	}
	else {
		Rev_Coal_NC* c = dynamic_cast<Rev_Coal_NC*>(target);
		delta_1 = c->num_elig_sites() - c->first_child()->num_elig_sites() - c->second_child()->num_elig_sites();
	}
	if (is_mid_rec) {
		Rev_Rec_NC* c = dynamic_cast<Rev_Rec_NC*>(n_up);
		delta_2 = c->num_elig_sites() - c->child()->num_elig_sites();
	}
	else {
		Rev_Coal_NC* c = dynamic_cast<Rev_Coal_NC*>(n_up);
		delta_2 = c->num_elig_sites() - c->first_child()->num_elig_sites() - c->second_child()->num_elig_sites();
	}

	lambda_2 = (num_down_2) * (num_down_2 - 1) / 2 + r * (s + delta_1);
	mu_2 = num_up_2 * (num_up_2 - 1) / 2 + r * (s + delta_2);
	lambda_down = num_1 * (num_1 - 1) / 2 - (num_down_2) * (num_down_2 - 1) / 2 - r * delta_1;
	lambda_up = num_up_2 * (num_up_2 - 1) / 2 - (num_up_3) * (num_up_3 - 1) / 2 - r * delta_2;

	if (is_down_op)
	{
		t_down = tree_rearranger_sngl::gen_single_event(lambda_down, T_down);
		*t_new = n_down->birth_time() + t_down;
	}
	else
	{
		t_up = tree_rearranger_sngl::gen_single_event(lambda_up, T_up);
		*t_new = n_up->birth_time() + t_up;
	}

	double L_ratio = mu_2 / lambda_2 * exp(lambda_down * T_down)
					 * exp(-lambda_down * t_down) / exp(-lambda_up * t_up);
	double Q_ratio = mu_2 / lambda_2 * exp(lambda_down * T_down)
					 * abs(lambda_up / lambda_down)
					 * abs((1 - exp(-lambda_down * T_down)) / (1 - exp(-lambda_up * T_up)));

	LOG4CXX_DEBUG(logger(), "n_down: " << n_down->str_id());
	LOG4CXX_DEBUG(logger(), "n_up: " << n_up->str_id());
	LOG4CXX_DEBUG(logger(), "num_1: " << num_1);
	LOG4CXX_DEBUG(logger(), "num_down_2: " << num_down_2);
	LOG4CXX_DEBUG(logger(), "num_up_2: " << num_up_2);
	LOG4CXX_DEBUG(logger(), "num_up_3: " << num_up_3);
	LOG4CXX_DEBUG(logger(), "T_down: " << T_down);
	LOG4CXX_DEBUG(logger(), "T_up: " << T_up);
	LOG4CXX_DEBUG(logger(), "t_down: " << t_down);
	LOG4CXX_DEBUG(logger(), "t_up: " << t_up);
	LOG4CXX_DEBUG(logger(), "lambda_down: " << lambda_down);
	LOG4CXX_DEBUG(logger(), "lambda_up: " << lambda_up);
	LOG4CXX_DEBUG(logger(), "L_ratio: " << L_ratio);
	LOG4CXX_DEBUG(logger(), "Q_ratio: " << Q_ratio);

	double x = L_ratio * Q_ratio;

	OUT_FNCTN_BOTTOM(logger());

	if (is_down_op)
		return x;
	else
		return 1 / x;
}



// adjust prev and next after up or down move
void tree_rearranger_sngl::adjust_prev_and_next(Rev_Tree* tr, Rev_NC* target, BOOL is_move_down)
{
	OUT_FNCTN_HEADER(logger());

	if (is_move_down)
	{
		target->set_pos_affected();

		Rev_NC* old_next = target->next();
		Rev_NC* old_prev = target->prev();
		Rev_NC* new_next = target->prev();
		Rev_NC* new_prev = target->prev()->prev();

		old_next->set_pos_affected();
		old_prev->set_pos_affected();
		new_next->set_pos_affected();
		new_prev->set_pos_affected();

		old_next->prev() = old_prev;
		old_prev->next() = old_next;

		new_next->prev() = target;
		new_prev->next() = target;
		target->next() = new_next;
		target->prev() = new_prev;
	}
	else
	{
		target->set_pos_affected();

		Rev_NC* old_next = target->next();
		Rev_NC* old_prev = target->prev();
		Rev_NC* new_next = target->next()->next();
		Rev_NC* new_prev = target->next();

		old_next->set_pos_affected();
		old_prev->set_pos_affected();
		new_next->set_pos_affected();
		new_prev->set_pos_affected();

		old_next->prev() = old_prev;
		old_prev->next() = old_next;

		new_next->prev() = target;
		new_prev->next() = target;
		target->next() = new_next;
		target->prev() = new_prev;
	}

	OUT_FNCTN_BOTTOM(logger());
}



void tree_rearranger_sngl::move_node_contnr(Rev_Tree* tr, Rev_Intern_NC* target, enum_direction dir, stringstream& ss_dot)
{
	OUT_FNCTN_HEADER(logger());

	switch(dir)
	{
		case DIR_STAY:
		{
			LOG4CXX_DEBUG(logger(), "DIR_STAY");
			LOG4CXX_DEBUG(logger(), (target->has_coal_suc() ? "Coal" : "Recomb"));
			LOG4CXX_DEBUG(logger(), "Rev_Nd id: " << target->str_id());
			LOG4CXX_DEBUG(logger(), "old time of generation: " << target->birth_time());
			ss_dot << "STAY ";
			ss_dot << target->str_id();

			if (target->has_coal_suc())
			{
				Rev_Coal_NC* c = dynamic_cast<Rev_Coal_NC*>(target);
				int m = target->num_elig_nodes();
				double delta = target->num_elig_nodes() - c->first_child()->num_elig_sites()
							   - c->second_child()->num_elig_sites();
				double lambda = m + Constants::instance()->recomb_const() * delta;
				double F = gsl_cdf_exponential_P(target->next()->birth_time() - target->prev()->birth_time(), 1 / lambda);
				double t = gsl_cdf_exponential_Pinv(gsl_rng_uniform(Random::rng()) * F, 1 / lambda);
				target->change_time(target->prev()->birth_time() + t);
			}
			else
			{
				Rev_Rec_NC* c = dynamic_cast<Rev_Rec_NC*>(target);
				int m = target->num_elig_nodes();
				double delta = target->num_elig_nodes() - c->child()->num_elig_sites();
				double lambda = (m - 1) + Constants::instance()->recomb_const() * delta;
				double F = gsl_cdf_exponential_P(target->next()->birth_time() - target->prev()->birth_time(), 1 / lambda);
				double t = gsl_cdf_exponential_Pinv(gsl_rng_uniform(Random::rng()) * F, 1 / lambda);
				target->change_time(target->next()->birth_time() - t);
			}
			LOG4CXX_DEBUG(logger(), "new time of generation: " << target->birth_time());
			break;
		}
		case DIR_UP:
		{
			LOG4CXX_DEBUG(logger(), "UP_OP");
			ss_dot << "UP ";
			LOG4CXX_DEBUG(logger(), (target->has_coal_suc() ? "Coal" : "Recomb"));
			LOG4CXX_DEBUG(logger(), "Rev_Nd id: " << target->str_id());
			LOG4CXX_DEBUG(logger(), "old time of generation: " << target->birth_time() << endl);
			ss_dot << target->str_id() << " ";

			double* t_new = new double;
			double r = get_ratio_mcmc(target, false, t_new);

			LOG4CXX_DEBUG(logger(), "t_new: " << *t_new);
			LOG4CXX_DEBUG(logger(), "r: " << r);

			VECTOR(Rev_Intern_NC*) elig_targets_up;
			VECTOR(Rev_Intern_NC*) elig_targets_down;
			get_elig_targets_single(tr, DIR_UP, &elig_targets_up);
			get_elig_targets_single(tr, DIR_DOWN, &elig_targets_down);
			int num_down = elig_targets_down.size();
			if (find(elig_targets_down.begin(), elig_targets_down.end(), target) == elig_targets_down.end())
				++num_down;
			if (*target->next()->next() == *target->youngest_parent_contnr())
				--num_down;
			int num_up = elig_targets_up.size();
			double q = num_down / num_up;

			if (r * q > gsl_rng_uniform(Random::rng()))
			{
				LOG4CXX_DEBUG(logger(), "ACC");
				ss_dot << "ACC";

				target->set_pos_affected();

				Rev_NC* old_next = target->next();
				Rev_NC* old_prev = target->prev();
				Rev_NC* new_next = target->next()->next();
				Rev_NC* new_prev = target->next();

				old_next->set_pos_affected();
				old_prev->set_pos_affected();
				new_next->set_pos_affected();
				new_prev->set_pos_affected();

				old_next->prev() = old_prev;
				old_prev->next() = old_next;

				new_next->prev() = target;
				new_prev->next() = target;
				target->next() = new_next;
				target->prev() = new_prev;

				target->change_time(*t_new);
			}
			else
			{
				LOG4CXX_DEBUG(logger(), "REJ");
				ss_dot << "REJ";
			}
			break;
		}
		case DIR_DOWN:
		{
			LOG4CXX_DEBUG(logger(), "DOWN_OP");
			ss_dot << "DOWN ";
			LOG4CXX_DEBUG(logger(), (target->has_coal_suc() ? "Coal" : "Recomb"));
			LOG4CXX_DEBUG(logger(), "Rev_Nd id: " << target->str_id());
			LOG4CXX_DEBUG(logger(), "old time of generation: " << target->birth_time() << endl);
			ss_dot << target->str_id() << " ";

			double* t_new = new double;
			double r = get_ratio_mcmc(target, true, t_new);

			LOG4CXX_DEBUG(logger(), "t_new: " << *t_new);
			LOG4CXX_DEBUG(logger(), "r: " << r);

			VECTOR(Rev_Intern_NC*) elig_targets_up;
			VECTOR(Rev_Intern_NC*) elig_targets_down;
			get_elig_targets_single(tr, DIR_UP, &elig_targets_up);
			get_elig_targets_single(tr, DIR_DOWN, &elig_targets_down);
			int num_down = elig_targets_down.size();
			int num_up = elig_targets_up.size();
			if (find(elig_targets_up.begin(), elig_targets_up.end(), target) == elig_targets_up.end())
				++num_up;
			if (*target->prev()->prev() == *target->oldest_child_contnr())
				--num_up;
			double q = num_up / num_down;

			if (r * q > gsl_rng_uniform(Random::rng()))
			{
				LOG4CXX_DEBUG(logger(), "ACC");
				ss_dot << "ACC";

				target->set_pos_affected();

				Rev_NC* old_next = target->next();
				Rev_NC* old_prev = target->prev();
				Rev_NC* new_next = target->prev();
				Rev_NC* new_prev = target->prev()->prev();

				old_next->set_pos_affected();
				old_prev->set_pos_affected();
				new_next->set_pos_affected();
				new_prev->set_pos_affected();

				old_next->prev() = old_prev;
				old_prev->next() = old_next;

				new_next->prev() = target;
				new_prev->next() = target;
				target->next() = new_next;
				target->prev() = new_prev;

				target->change_time(*t_new);
			}
			else
			{
				LOG4CXX_DEBUG(logger(), "REJ");
				ss_dot << "REJ";
			}
			break;
		}
	}

	OUT_FNCTN_BOTTOM(logger());
}



void tree_rearranger_sngl::move_node_contnr(Rev_Tree* tr, Rev_Intern_NC* target, enum_direction dir)
{
	stringstream dummy;
	move_node_contnr(tr, target, dir, dummy);
}



// get eligible targets for rearrange operation "single"
void tree_rearranger_sngl::get_elig_targets_single(Rev_Tree* tr, enum_direction dir, VECTOR(Rev_Intern_NC*)* elig_targets)
{
	OUT_FNCTN_HEADER(logger());

	switch(dir)
	{
		case DIR_STAY:
		{
			Rev_NC* c = tr->first_node_contnr();
			while (c)
			{
				if ((!c->is_old()) && !c->is_tip() && !c->is_last())
					elig_targets->push_back(dynamic_cast<Rev_Intern_NC*>(c));
				c = c->next();
			}
			break;
		}
		case DIR_UP:
		{
			Rev_NC* c = tr->first_node_contnr();
			while (c)
			{
				Rev_Intern_NC* d = dynamic_cast<Rev_Intern_NC*>(c);
				if (d && (!d->is_old()) && !d->is_last() &&
					!d->next()->is_last() && (*d->next() != *d->youngest_parent_contnr())) {
					VECTOR(int) v(d->occur_subs().size());
					VECTOR(int)::iterator it =
						set_intersection(d->occur_subs().begin(), d->occur_subs().end(),
										 d->next()->occur_subs().begin(), d->next()->occur_subs().end(), v.begin());
					if (!d->next()->is_final() || d->is_final() || (it - v.begin() == 0))
						elig_targets->push_back(d);
				}
				c = c->next();
			}
			break;
		}
		case DIR_DOWN:
		{
			Rev_NC* c = tr->first_node_contnr();
			while (c)
			{
				Rev_Intern_NC* d = dynamic_cast<Rev_Intern_NC*>(c);
				if (d && (!d->is_old()) && !d->is_last() && !d->prev()->is_tip() &&
					(*d->prev() != *d->oldest_child_contnr())) {
					VECTOR(int) v(d->occur_subs().size());
					VECTOR(int)::iterator it =
						set_intersection(d->occur_subs().begin(), d->occur_subs().end(),
										 d->prev()->occur_subs().begin(), d->prev()->occur_subs().end(), v.begin());
					if (d->prev()->is_final() || !d->is_final() || (it - v.begin() == 0))
						elig_targets->push_back(d);
				}
				c = c->next();
			}
			break;
		}
	}
	OUT_FNCTN_BOTTOM(logger());
}



// perform rearrange operation "single"
void tree_rearranger_sngl::single(Rev_Tree* tr)
{
	OUT_FNCTN_HEADER(logger());

	if (tr->changes()->rearranging_done())
		throw ERROR(logger(), "Rearranging coalescent: Illegal Rev_Tree");

	tr->changes()->set_rearranging_done();
	tr->save_first_last();

	const int NUM_OPS = 3;
	double FRACTION_OPS[NUM_OPS] = { 1, 1, 1 };
	gsl_ran_discrete_t* grd = gsl_ran_discrete_preproc (NUM_OPS, FRACTION_OPS);
	enum_direction chosen_dir = (enum_direction)gsl_ran_discrete (Random::rng(), grd);
	gsl_ran_discrete_free(grd);

	VECTOR(Rev_Intern_NC*) elig_targets;
	get_elig_targets_single(tr, chosen_dir, &elig_targets);

	// output possible targets
	LOG4CXX_DEBUG(logger(), "chosen_dir: " << chosen_dir);
	stringstream ss;
	ss << "elig_targets (single): ";
	for (VECTOR(Rev_Intern_NC*)::iterator it = elig_targets.begin(); it != elig_targets.end(); it++)
		ss << (*it)->str_id() << " ";
	LOG4CXX_DEBUG(logger(), ss.str());

	if (elig_targets.size() > 0)
	{
		int r = gsl_rng_uniform_int(Random::rng(), elig_targets.size());
		Rev_Intern_NC* target = elig_targets[r];
		LOG4CXX_DEBUG(logger(), "target: " << target->str_id());
		move_node_contnr(tr, target, chosen_dir);
	}

	OUT_FNCTN_BOTTOM(logger());
}
