#include "standard.h"

#include "changes.h"
#include "crf.h"
#include "constants.h"
#include "dot.h"
#include "error.h"
#include "files.h"
#include "nc.h"
#include "nd.h"
#include "part_tree.h"
#include "random.h"
#include "trans_probs.h"
#include "tree.h"

LoggerPtr Rev_Tree::s_logger(Logger::getLogger("tree.tree"));



Rev_Nd* Rev_Tree::old_first_node() const
{ return m_old_first_node_contnr ? m_old_first_node_contnr->unspec_nodes().front() : 0; }

Rev_Nd* Rev_Tree::old_last_node() const
{ return m_old_last_node_contnr ? m_old_last_node_contnr->unspec_nodes().back() : 0; }

void Rev_Tree::accept_changes() { changes()->accept(); }
void Rev_Tree::reverse_changes() { changes()->reverse(); }
void Rev_Tree::adjust_pos() { changes()->adjust_pos(); }



// initialize tree
void Rev_Tree::init()
{
	OUT_FNCTN_HEADER(logger());
	m_old_first_node_contnr = 0;
	m_old_last_node_contnr = 0;
	m_changes = new Changes(this);
	m_num_subtypes = 0;
	m_all_tips_added = false;
	m_part_alias_calced = false;
	m_acid_prob_calced = false;
	OUT_FNCTN_BOTTOM(logger());
}



// construct empty tree
Rev_Tree::Rev_Tree(const int num_sites, const int num_mutation_rates)
: Base_Tree<Rev_Nd, Rev_Coal_Nd, Rev_Rec_Nd, Rev_NC>(num_sites, num_mutation_rates)
{
	OUT_FNCTN_HEADER(logger());
	init();
	OUT_FNCTN_BOTTOM(logger());
}



// copy constructor
Rev_Tree::Rev_Tree(const Rev_Tree& tr)
: Base_Tree<Rev_Nd, Rev_Coal_Nd, Rev_Rec_Nd, Rev_NC>(tr.num_sites(), tr.num_mutation_rates())
{
	OUT_FNCTN_HEADER(logger());	init();

	// create CRFs
	LOG4CXX_DEBUG(logger(), "crfs:");
	for (int i = 0; i < tr.num_crfs(); ++i) {
		LOG4CXX_DEBUG(logger(), tr.crf(i)->id());
		int num_bps = tr.crf(i)->num_bps();
		VECTOR(int) bps(num_bps);
		for (int j = 0; j < num_bps; ++j)
			bps[j] = tr.crf(i)->bp(j);
		VECTOR(int) segm_subs(tr.crf(i)->num_segms());
		for (int j = 0; j < tr.crf(i)->num_segms(); ++j)
			segm_subs[j] = tr.crf(i)->segm_sub(j);
		new CRF(this, num_bps, bps, segm_subs);
	}

	// create node containers
	LOG4CXX_DEBUG(logger(), "nodes:");
	map<int, Rev_Nd*> old_id_to_new_node;
	for (Rev_NC* c = tr.first_node_contnr(); c; c = c->next()) {
		LOG4CXX_DEBUG(logger(), c->str_id());
		if (Rev_Tip_NC* tc = dynamic_cast<Rev_Tip_NC*>(c)) {
			Rev_Tip_NC* td;
			if (tc->node()->is_pure_node())
				td = new Rev_Tip_NC(this, tc->node()->acid_seq(), tc->node()->sub(), true);
			else
				td = new Rev_Tip_NC(this, tc->node()->acid_seq(), tc->node()->crf_id(), false);
			old_id_to_new_node[tc->node()->id()] = td->node();
		}
		else {
			set_to_all_tips_added();
			if (c->has_coal_suc()) {
				Rev_Coal_NC* d = dynamic_cast<Rev_Coal_NC*>(c);
				Rev_Coal_NC* e =
					new Rev_Coal_NC(this, d->birth_time(),
									old_id_to_new_node[d->first_child()->id()],
									old_id_to_new_node[d->second_child()->id()]);
				old_id_to_new_node[d->node()->id()] = e->node();
			}
			else {
				Rev_Rec_NC* d = dynamic_cast<Rev_Rec_NC*>(c);
				Rev_Rec_NC* e =
					new Rev_Rec_NC(this, d->birth_time(),
							  	   old_id_to_new_node[d->child()->id()], d->first_node()->occur_subs(),
							  	   d->second_node()->occur_subs());
				old_id_to_new_node[d->first_node()->id()] = e->first_node();
				old_id_to_new_node[d->second_node()->id()] = e->second_node();
			}
		}
	}

	accept_changes();
	OUT_FNCTN_BOTTOM(logger());
}



Rev_Tree::~Rev_Tree()
{
	OUT_FNCTN_HEADER(logger());
	delete m_changes;
	OUT_FNCTN_BOTTOM(logger());
}



// get final node of subtype s
Rev_Nd* Rev_Tree::final_node(const int s)
{
	for (Rev_Nd* n = last_node(); n; n = n->prev())
		if (n->is_pure_node() && (n->sub() == s))
			return n;
	throw Error("Trying to get final node of non-existing subtype");
}



// generates a tree with starting time of the last node
void Rev_Tree::generate_final()
{
	OUT_FNCTN_HEADER(logger());

	if (!all_tips_added())
		throw ERROR(logger(), "not all tips have been added");

	// calculate start time of generation process
	double curr_time = 0;
	Rev_Nd* n = first_node();
	while (n)
	{
		curr_time = GSL_MAX(curr_time, n->birth_time());
		n = n->next();
	}

	while (num_act_nodes() > 1)
	{
		LOG4CXX_DEBUG(logger(), "num_act_nodes: " << num_act_nodes() << endl);

		VECTOR( VECTOR(Rev_Nd*) ) pure_nodes(num_subtypes());
		VECTOR( VECTOR(Rev_Nd*) ) crf_nodes(num_crfs());
		VECTOR(Rev_Nd*) final_nodes;
		VECTOR(Rev_Nd*) recomb_nodes;
		double time_of_next_node;

		// classify nodes
		Rev_Nd* n = first_node();
		while (n)
		{
			if (!n->is_old() && n->is_active())
			{
				switch (n->type())
				{
					case PURE_NODE:
						pure_nodes[n->sub()].push_back(n);
						break;
					case CRF_NODE:
						crf_nodes[n->crf_id()].push_back(n);
						break;
					case FINAL_NODE:
						final_nodes.push_back(n);
						break;
					case RECOMB_NODE:
						recomb_nodes.push_back(n);
						break;
				}
			}
			n = n->next();
		}

		// move single pure nodes to final_nodes
		for (int i = 0; i < num_subtypes(); i++)
		{
			if (pure_nodes[i].size() == 1)
			{
				BOOL is_in_crf = false;
				for (int j = 0; j < num_crfs(); j++)
					if ((crf_nodes[j].size() > 0) &&
						(crfs_containing_subtype(i)->find(j) != crfs_containing_subtype(i)->end()))
						is_in_crf = true;

				BOOL is_in_recomb = false;
				for (VECTOR(Rev_Nd*)::iterator it = recomb_nodes.begin(); it != recomb_nodes.end(); it++)
					if ((*it)->occur_subs().find(i) != (*it)->occur_subs().end())
						is_in_recomb = true;

				if (!is_in_crf && !is_in_recomb)
				{
					final_nodes.push_back(pure_nodes[i].back());
					pure_nodes[i].pop_back();
				}
			}
		}

		// move single crfs to recomb_nodes
		for (VECTOR( VECTOR(Rev_Nd*) )::iterator it = crf_nodes.begin(); it != crf_nodes.end(); it++)
			if (it->size() == 1)
			{
				recomb_nodes.push_back(it->back());
				it->pop_back();
			}

		stringstream ss;
		for (int i = 0; i < num_subtypes(); i++)
		{
			ss << "pure_nodes Subtype " << i << ": ";
			for (VECTOR(Rev_Nd*)::iterator it = pure_nodes[i].begin(); it != pure_nodes[i].end(); it++)
				ss << (*it)->id() << " ";
			ss << endl;
		}

		for (int i = 0; i < num_crfs(); i++)
		{
			ss << "crf_nodes CRF " << i << ": ";
			for (VECTOR(Rev_Nd*)::iterator it = crf_nodes[i].begin(); it != crf_nodes[i].end(); it++)
				ss << (*it)->id() << " ";
			ss << endl;
		}

		ss << "final_nodes: ";
		for (VECTOR(Rev_Nd*)::iterator it = final_nodes.begin(); it != final_nodes.end(); it++)
			ss << (*it)->id() << " ";
		ss << endl;

		ss << "recomb_nodes: ";
		for (VECTOR(Rev_Nd*)::iterator it = recomb_nodes.begin(); it != recomb_nodes.end(); it++)
			ss << (*it)->id() << " (" << (*it)->occur_subs() << ")   ";
		ss << endl << endl;

		LOG4CXX_DEBUG(logger(), ss.str());

		// calculate total intensity
		double total_intensity = 0;
		n = first_node();
		while (n)
		{
			if (!n->is_old() && n->is_active())
				total_intensity += Constants::instance()->recomb_const() * n->num_elig_sites();
			n = n->next();
		}
		total_intensity += num_act_nodes() * (num_act_nodes() - 1) / 2;

		// calculate coalescent intensities
		VECTOR(double) pure_coal_intensity(num_subtypes());
		double sum_pure_coal_intensity = 0;
		for (int i = 0; i < num_subtypes(); i++)
		{
			pure_coal_intensity[i] = pure_nodes[i].size() * (pure_nodes[i].size() - 1) / 2;
			sum_pure_coal_intensity += pure_coal_intensity[i];
		}

		VECTOR(double) crf_coal_intensity(num_crfs());
		double sum_crf_coal_intensity = 0;
		for (int i = 0; i < num_crfs(); i++)
		{
			crf_coal_intensity[i] = crf_nodes[i].size() * (crf_nodes[i].size() - 1) / 2;
			sum_crf_coal_intensity += crf_coal_intensity[i];
		}

		double final_coal_intensity = final_nodes.size() * (final_nodes.size() - 1) / 2;

		// calculate recombination intensity
		double recomb_intensity = 0;
		for (VECTOR(Rev_Nd*)::iterator it = recomb_nodes.begin(); it != recomb_nodes.end(); it++)
			recomb_intensity += Constants::instance()->recomb_const() * (*it)->num_elig_sites();

		LOG4CXX_DEBUG(logger(), "total_intensity: " << total_intensity);
		for (int i = 0; i < num_subtypes(); i++)
			LOG4CXX_DEBUG(logger(), "pure_coal_intensity Subtype " << i << ": " << pure_coal_intensity[i]);
		for (int i = 0; i < num_crfs(); i++)
			LOG4CXX_DEBUG(logger(), "crf_coal_intensity CRF " << i << ": " << crf_coal_intensity[i]);
		LOG4CXX_DEBUG(logger(), "final_coal_intensity: " << final_coal_intensity);
		LOG4CXX_DEBUG(logger(), "recomb_intensity: " << recomb_intensity << endl);

		// carry out coalescent or recombination operation
		const int NUM_OPS = 4;
		enum enum_ops { PURE_COAL_OP, CRF_COAL_OP, FINAL_COAL_OP, RECOMB_OP };
		double p[NUM_OPS] = { sum_pure_coal_intensity, sum_crf_coal_intensity, final_coal_intensity, recomb_intensity};
		gsl_ran_discrete_t* grd = gsl_ran_discrete_preproc (NUM_OPS, p);
		enum_ops chosen_op = (enum_ops)gsl_ran_discrete (Random::rng(), grd);
		gsl_ran_discrete_free(grd);
		double time_to_next_node = gsl_ran_exponential(Random::rng(), 1 / total_intensity);
		curr_time += time_to_next_node;
		switch (chosen_op)
		{
			case PURE_COAL_OP:
			{
				// chose Random subtype
				double q[num_subtypes()];
				for (int i = 0; i < num_subtypes(); i++)
					q[i] = pure_coal_intensity[i];
				gsl_ran_discrete_t* grd = gsl_ran_discrete_preproc (num_subtypes(), q);
				int s = gsl_ran_discrete (Random::rng(), grd);
				gsl_ran_discrete_free(grd);

				// chose Random pair of nodes
				int i, j;
				Random::get_rnd_pair(pure_nodes[s].size(), &i, &j);
				Rev_Nd* first_old_node = pure_nodes[s][i];
				Rev_Nd* second_old_node = pure_nodes[s][j];

				Rev_Coal_NC* new_node_contnr = new Rev_Coal_NC(this, curr_time, first_old_node, second_old_node);

				LOG4CXX_DEBUG(logger(), "PURE_COAL_OP");
				LOG4CXX_DEBUG(logger(), "Subtype " << s);
				LOG4CXX_DEBUG(logger(), first_old_node->id() << " " << second_old_node->id()
									    << " -> " << new_node_contnr->str_id());
				break;
			}
			case CRF_COAL_OP:
			{
				// chose Random crf
				double q[num_crfs()];
				for (int i = 0; i < num_crfs(); i++)
					q[i] = crf_coal_intensity[i];
				gsl_ran_discrete_t* grd = gsl_ran_discrete_preproc (num_crfs(), q);
				int c = gsl_ran_discrete (Random::rng(), grd);
				gsl_ran_discrete_free(grd);

				// chose Random pair of nodes
				int i, j;
				Random::get_rnd_pair(crf_nodes[c].size(), &i, &j);
				Rev_Nd* first_old_node = crf_nodes[c][i];
				Rev_Nd* second_old_node = crf_nodes[c][j];

				Rev_Coal_NC* new_node_contnr = new Rev_Coal_NC(this, curr_time, first_old_node, second_old_node);

				LOG4CXX_DEBUG(logger(), "CRF_COAL_OP");
				LOG4CXX_DEBUG(logger(), "CRF " << c);
				LOG4CXX_DEBUG(logger(), first_old_node->id() << " " << second_old_node->id()
										<< " -> " << new_node_contnr->str_id());
			break;
			}
			case FINAL_COAL_OP:
			{
				// chose Random pair of nodes
				int i, j;
				Random::get_rnd_pair(final_nodes.size(), &i, &j);
				Rev_Nd* first_old_node = final_nodes[i];
				Rev_Nd* second_old_node = final_nodes[j];

				Rev_Coal_NC* new_node_contnr = new Rev_Coal_NC(this, curr_time, first_old_node, second_old_node);

				LOG4CXX_DEBUG(logger(), "FINAL_COAL_OP");
				LOG4CXX_DEBUG(logger(), first_old_node->id() << " " << second_old_node->id()
											<< " -> " << new_node_contnr->str_id());
				break;
			}
			case RECOMB_OP:
			{
				int i = gsl_rng_uniform_int(Random::rng(), recomb_nodes.size());
				Rev_Nd* old_node = recomb_nodes[i];

				// distribute subtypes to parent nodes of target
				set<int> subtypes_first;
				set<int> subtypes_second;
				if (old_node->type() == CRF_NODE)
					for (set<int>::const_iterator it = crf(old_node->crf_id())->occur_subs().begin();
						 it != crf(old_node->crf_id())->occur_subs().end(); it++)
						if (gsl_rng_uniform(Random::rng()) < 0.5)
							subtypes_first.insert(*it);
						else
							subtypes_second.insert(*it);
				else
					for (set<int>::const_iterator it = old_node->occur_subs().begin();
						 it != old_node->occur_subs().end(); it++)
					{
						if (gsl_rng_uniform(Random::rng()) < 0.5)
							subtypes_first.insert(*it);
						else
							subtypes_second.insert(*it);
					}

				// assure that each parent containes at least one subtypes
				if (subtypes_first.size() == 0)
				{
					int i = gsl_rng_uniform_int(Random::rng(), subtypes_second.size());
					int j = 0;
					set<int>::iterator it = subtypes_second.begin();
					while (i < j)
					{
						it++;
						j++;
					}

					subtypes_first.insert(*it);
					subtypes_second.erase(it);
				}
				else
					if (subtypes_second.size() == 0)
					{
						int i = gsl_rng_uniform_int(Random::rng(), subtypes_first.size());
						int j = 0;
						set<int>::iterator it = subtypes_first.begin();
						while (i < j)
						{
							it++;
							j++;
						}
						subtypes_second.insert(*it);
						subtypes_first.erase(it);
					}

				Rev_Rec_NC* new_node_contnr = new Rev_Rec_NC(this, curr_time, old_node, subtypes_first, subtypes_second);

				LOG4CXX_DEBUG(logger(), "RECOMB_OP");
				LOG4CXX_DEBUG(logger(), old_node->id() << " -> " << flush
										<< new_node_contnr->str_id());
				break;
			}
		}
	}

	OUT_FNCTN_BOTTOM(logger());
}



// generate the rest of a tree after applying rearrange operation "recomb" with subtypes only occuring in one CRF
void Rev_Tree::generate_unknown()
{
	OUT_FNCTN_HEADER(logger());

	if (!all_tips_added())
		throw ERROR(logger(), "not all tips have been added");

	calc_num_elig_nd_st(first_node_contnr(), last_node_contnr());

	// determine last pure nodes
	Rev_Nd* n = last_node();
	set<int> s;
	set<int> ids_last_pure_nodes;
	while(n) {
		if (n->is_pure_node() && (s.find(n->sub()) == s.end())) {
			s.insert(n->sub());
			ids_last_pure_nodes.insert(n->id());
		}
		n = n->prev();
	}

	// calculate start time of generation process and current final nodes
	set<Rev_Nd*> curr_pas_final_nodes;
	n = first_node();
	while (!n->is_last() && !n->is_active()) {
		if (n->is_final_node() || ids_last_pure_nodes.count(n->id()))
			curr_pas_final_nodes.insert(n);
		if (n->has_coal_suc()) {
			VECTOR(Rev_Nd*) v = n->unspec_children();
			for (VECTOR(Rev_Nd*)::iterator it = v.begin(); it != v.end(); ++it)
				curr_pas_final_nodes.erase(*it);
		}
		n = n->next();
	}
	double curr_time = n->birth_time();

	set<Rev_Nd*> curr_act_final_nodes;
	int curr_num_act_nodes = 0;
	BOOL node_just_generated = false;
	while ((num_act_nodes() > 1) && !n->is_last())
	{
		// update final nodes
		if (!node_just_generated) {
			if (n->is_final_node() || ids_last_pure_nodes.count(n->id()))
				if (n->is_active())
					curr_act_final_nodes.insert(n);
				else
					curr_pas_final_nodes.insert(n);
			if (n->has_coal_suc()) {
				VECTOR(Rev_Nd*) v = n->unspec_children();
				for (VECTOR(Rev_Nd*)::iterator it = v.begin(); it != v.end(); ++it)
					curr_pas_final_nodes.erase(*it);
			}
			curr_num_act_nodes = curr_act_final_nodes.size();
		}

		LOG4CXX_DEBUG(logger(), "n: " << n->id() << " (" << n->birth_time() << ")");
		LOG4CXX_DEBUG(logger(), "num_act_nodes: " << num_act_nodes());
		LOG4CXX_DEBUG(logger(), "curr_num_act_nodes: " << curr_num_act_nodes);
		LOG4CXX_DEBUG(logger(), "curr_num_pas_final_nodes: " << curr_pas_final_nodes.size());
		LOG4CXX_DEBUG(logger(), "curr_num_act_final_nodes: " << curr_act_final_nodes.size() << endl);

		// calculate virtual intensities and time to next event
		int curr_num_elig_nodes = n->unspec_node_contnr()->num_elig_nodes();
		int curr_num_elig_sites = n->unspec_node_contnr()->num_elig_sites();
		double virt_coal_intensity = curr_num_act_nodes * (curr_num_act_nodes - 1) / 2 +
									 curr_num_elig_nodes * curr_num_act_nodes;
		double virt_rec_intensity = Constants::instance()->recomb_const() * curr_num_elig_sites;
		double total_intensity = virt_coal_intensity + virt_rec_intensity;
		double time_to_next_event = gsl_ran_exponential(Random::rng(), 1 / total_intensity);
		double time_to_next_node = n->next()->birth_time() - n->birth_time();
		double coal_intensity_act_act = curr_act_final_nodes.size() * (curr_act_final_nodes.size() - 1) / 2;
		double coal_intensity_act_pas = curr_act_final_nodes.size() * curr_pas_final_nodes.size();
		LOG4CXX_DEBUG(logger(), "curr_num_elig_nodes: " << curr_num_elig_nodes);
		LOG4CXX_DEBUG(logger(), "virt_coal_intensity: " << virt_coal_intensity);
		LOG4CXX_DEBUG(logger(), "virt_rec_intensity: " << virt_rec_intensity);
		LOG4CXX_DEBUG(logger(), "coal_intensity_act_act: " << coal_intensity_act_act);
		LOG4CXX_DEBUG(logger(), "coal_intensity_act_pas: " << coal_intensity_act_pas << endl);

		if ((time_to_next_event < time_to_next_node) &&
			(coal_intensity_act_act + coal_intensity_act_pas > 0)) {   // event takes place?
			// determine event

			LOG4CXX_DEBUG(logger(), "time of event: " << setprecision(4) << curr_time + time_to_next_event);

			BOOL is_act_act;
			const int NUM = 2;
			double p[NUM] = { coal_intensity_act_act, coal_intensity_act_pas };
			gsl_ran_discrete_t* grd = gsl_ran_discrete_preproc (NUM, p);
			int r = gsl_ran_discrete (Random::rng(), grd);
			gsl_ran_discrete_free(grd);
			if (r == 0)
				is_act_act = true;
			else
				is_act_act = false;

			if (is_act_act) {   // two active nodes coalesce?
				// coalesce two active nodes
				LOG4CXX_DEBUG(logger(), "act_act ");

				// choose target nodes
				VECTOR(Rev_Nd*) v;
				copy(curr_act_final_nodes.begin(), curr_act_final_nodes.end(), back_inserter(v));
				int first, second;
				Random::get_rnd_pair(v.size(), &first, &second);
				Rev_Nd* first_old_node = v[first];
				Rev_Nd* second_old_node = v[second];

				Rev_Coal_NC* new_node_contnr =
					new Rev_Coal_NC(this, curr_time + time_to_next_event, first_old_node, second_old_node);
				curr_act_final_nodes.erase(first_old_node);
				curr_act_final_nodes.erase(second_old_node);
				curr_act_final_nodes.insert(new_node_contnr->node());

				LOG4CXX_DEBUG(logger(), "first_old_node: " << first_old_node->id());
				LOG4CXX_DEBUG(logger(), "second_old_node: " << second_old_node->id());
				LOG4CXX_DEBUG(logger(), "new node: " << new_node_contnr->node()->id() << endl);
			}
			else {
				// coalesce an active and a passive node
				LOG4CXX_DEBUG(logger(), "act_pas ");

				// choose target nodes
				VECTOR(Rev_Nd*) v;
				copy(curr_act_final_nodes.begin(), curr_act_final_nodes.end(), back_inserter(v));
				int act_idx = gsl_rng_uniform_int(Random::rng(), v.size());
				Rev_Nd* act_old_node = v[act_idx];

				VECTOR(Rev_Nd*) w;
				copy(curr_pas_final_nodes.begin(), curr_pas_final_nodes.end(), back_inserter(w));
				int pas_idx = gsl_rng_uniform_int(Random::rng(), w.size());
				Rev_Nd* pas_old_node = w[pas_idx];

				Rev_Coal_NC* new_node_contnr;
				Rev_Coal_NC* parent = pas_old_node->parent_node()->node_contnr();
				if (pas_old_node->is_first_child()) {
					new_node_contnr =
						new Rev_Coal_NC(this, curr_time + time_to_next_event, pas_old_node, act_old_node);
				}
				else {
					new_node_contnr =
						new Rev_Coal_NC(this, curr_time + time_to_next_event, act_old_node, pas_old_node);
				}
				new_node_contnr->node()->set_coal_anc(parent, pas_old_node->is_first_child());
				curr_act_final_nodes.erase(act_old_node);
				curr_pas_final_nodes.erase(pas_old_node);
				curr_pas_final_nodes.insert(new_node_contnr->node());

				LOG4CXX_DEBUG(logger(), "act_old_node: " << act_old_node->id());
				LOG4CXX_DEBUG(logger(), "pas_old_node: " << pas_old_node->id());
				LOG4CXX_DEBUG(logger(), "new node: " << new_node_contnr->node()->id() << endl);
			}

			node_just_generated = true;
		}
		else
			node_just_generated = false;

		LOG4CXX_DEBUG(logger(), "");

		n = n->next();
		curr_time = n->birth_time();
	}

	OUT_FNCTN_BOTTOM(logger());
}



// correct ancestors and successors of ordinary nodes after some nodes have been set to be old
void Rev_Tree::adjust_anc_suc(const set<Rev_NC*>& old_node_contnrs)
{
	OUT_FNCTN_HEADER(logger());

	for (set<Rev_NC*>::iterator it = old_node_contnrs.begin(); it != old_node_contnrs.end(); it++)
	{
		LOG4CXX_DEBUG(logger(), "id: " << (*it)->str_id());
		if ((*it)->has_coal_suc()) {
			VECTOR(Rev_Nd*) v = (*it)->unspec_children();
			for (VECTOR(Rev_Nd*)::iterator jt = v.begin(); jt != v.end(); ++jt)
				if ((*jt)->is_unmarked())
					(*jt)->determine_new_anc();
		}
	}

	LOG4CXX_DEBUG(logger(), "");
	OUT_FNCTN_BOTTOM(logger());
}



// sets the affected nodes above a recombination Rev_Nd to old in order to prepare rearrange operation "recombination"
void Rev_Tree::set_nodes_old_recomb(Rev_Nd* n)
{
	n->parent_contnr()->set_old();

	if (n->has_rec_anc())
	{
		set_nodes_old_recomb(n->first_parent_node());
		set_nodes_old_recomb(n->second_parent_node());
	}
}



// get first final Rev_Nd which contains the given subtype
Rev_Nd* Rev_Tree::get_first_final(const int i) const
{
	Rev_Nd* n = first_node();
	while ((n) && (!n->is_unmarked() || !n->is_final_node() ||
		   (n->occur_subs().find(i) == n->occur_subs().end())))
		n = n->next();
	return n;
}



// get nodes above which a recomb node can be hooked in the rest of the tree (see hook_in.txt)
void Rev_Tree::get_hook_in_targets(VECTOR(Rev_Nd*)* hook_nodes, const double time, const int sub) const
{
	OUT_FNCTN_HEADER(logger());
	LOG4CXX_DEBUG(logger(), "sub: " << sub);
	LOG4CXX_DEBUG(logger(), "time: " << time);
	LOG4CXX_DEBUG(logger(), "sub_only_in_crfs: " << sub_only_in_crfs(sub) << endl);

	Rev_Nd* n = first_node();
	while (!n->is_last()) {
		if (sub_only_in_crfs(sub)) {
			if (!n->is_old() && !n->is_active() && (n->birth_time() <= time) && (n->time_of_parent() >= time) &&
				n->has_coal_anc())
				hook_nodes->push_back(n);
		}
		else {
			if (!n->is_old() && !n->is_active() && (n->birth_time() <= time) && (n->time_of_parent() >= time) &&
				n->has_coal_anc() && n->is_pure_node() && (n->sub() == sub))
				hook_nodes->push_back(n);
		}
		n = n->next();
	}

	OUT_FNCTN_BOTTOM(logger());
}



// calculate the number of subtypes and of crfs and determine for each subtype which crf contains it
void Rev_Tree::calc_num_subtypes_crfs()
{
	OUT_FNCTN_HEADER(logger());

	Rev_Nd* n = first_node();
	set<int> occuring_subtypes;
	set<int> subs_in_pure_tips;
	while (n)
	{
		if (n->is_tip())
			if (n->is_pure_node()) {
				occuring_subtypes.insert(n->sub());
				subs_in_pure_tips.insert(n->sub());
			}
			else
				occuring_subtypes.insert(n->occur_subs().begin(), n->occur_subs().end());
		n = n->next();
	}
	m_num_subtypes = occuring_subtypes.size();
	VECTOR(int) v(occuring_subtypes.size());
	VECTOR(int)::iterator it = set_difference(occuring_subtypes.begin(), occuring_subtypes.end(),
				   							  subs_in_pure_tips.begin(), subs_in_pure_tips.end(), v.begin());
	m_subs_only_in_crfs = set<int>(v.begin(), it);

	stringstream ss;
	ss << "num_subtypes: " << num_subtypes() << endl;
	ss << "subtypes: ";
	for (set<int>::iterator it = occuring_subtypes.begin(); it != occuring_subtypes.end(); it++)
		ss << *it << " ";
	ss << endl;
	ss << "subtypes only occuring in CRFs: ";
	for (set<int>::iterator it = m_subs_only_in_crfs.begin(); it != m_subs_only_in_crfs.end(); it++)
		ss << *it << " ";
	LOG4CXX_DEBUG(logger(), ss.str());

	n = first_node();
	set<int> occuring_crfs;
	while (n)
	{
		if (n->is_tip() && n->is_crf_node())
			occuring_crfs.insert(n->crf_id());
		n = n->next();
	}
	m_num_crfs = occuring_crfs.size();

	ss.str("");
	ss << "num_crfs: " << num_crfs() << endl;
	ss << "crfs: ";
	for (set<int>::iterator it = occuring_crfs.begin(); it != occuring_crfs.end(); it++)
		ss << *it << " ";
	ss << endl;
	LOG4CXX_DEBUG(logger(), ss.str());

	m_crfs_containing_subtype.clear();
	m_crfs_containing_subtype.resize(num_subtypes());
	for (int i = 0; i < num_crfs(); i++)
	{
		for (int j = 0; j < crf(i)->num_segms(); ++j)
			m_crfs_containing_subtype.at(crf(i)->segm_sub(j)).insert(i);
	}

	ss.str("");
	for (int i = 0; i < num_subtypes(); i++) {
		ss << "Subtype " << i << " is contained by the CRFs: ";
		for (set<int>::iterator it = crfs_containing_subtype(i)->begin(); it != crfs_containing_subtype(i)->end(); it++)
			ss << *it << " ";
	}
	LOG4CXX_DEBUG(logger(), ss.str());

	OUT_FNCTN_BOTTOM(logger());
}



// calculate probability of data given the tree
double Rev_Tree::log_prob_acids()
{
	log_prob_acids_impl();
	return last_node()->sum_acid_probs();
}



//
VECTOR(double) Rev_Tree::log_prob_acids_pos()
{
	VECTOR(double) v;

	log_prob_acids_impl();
	v = last_node()->log_acid_probs();

	return v;
}



// get the prior probability
double Rev_Tree::log_prior_prob()
{
	OUT_FNCTN_HEADER(logger());
	changes()->reset_nd_st();
	calc_num_elig_nd_st(first_node_contnr(), NULL);
	check_num_elig_sites_nodes();
	calc_prior_prob(first_node_contnr());
	OUT_FNCTN_BOTTOM(logger());
	return log(last_node_contnr()->P_prior());
}



void Rev_Tree::check_num_elig_sites_nodes()
{
	OUT_FNCTN_HEADER(logger());

	BOOL is_ok = true;
	stringstream ss;

	// check integrity of num_elig_nodes
	if (first_node_contnr())
	{
		set<string> contnr;
		int k = 0;
		for (Rev_NC* c = first_node_contnr(); c; c = c->next())
		{
			if (!contnr.insert(c->str_id()).second)
				break;

			if (c->is_tip())
				k++;
			if (c->has_coal_suc())
				k--;
			if (c->has_rec_suc())
				k++;

			if (c->num_elig_nodes() != k)
			{
				is_ok = false;
				ss << "number of eligible nodes wrong: " << c->str_id() << "   "
				   << c->num_elig_nodes() << "(current)  " << k << "(actual)" << endl;
			}
		}
	}

	if (!is_ok)
		throw ERROR(logger(), ss.str());

	OUT_FNCTN_BOTTOM(logger());
}



// check integrity of all node containers
void Rev_Tree::check_node_contnr() const
{
	set<string> node_contnrs;
	for (Rev_NC* c = first_node_contnr(); c; c = c->next()) {
		if (!node_contnrs.insert(c->str_id()).second)
			break;
		c->check();
	}
}



// calculate the log-prob of the nucleotides of the last node
void Rev_Tree::log_prob_acids_impl()
{
	OUT_FNCTN_HEADER(logger());

	if (!all_tip_laps_calced())
		throw ERROR(logger(), "not all log acid probs of the tips have been calculated");

	if (num_act_nodes() == 1) {
		// calculate acid probabilities for all nodes
		last_node()->calc_log_subst_probs();
		Rev_Nd* n = first_node();
		while (n)
		{
			LOG4CXX_DEBUG(logger(), "n: " << n->id());
			n->calc_log_acid_probs();
			n = n->next();
		}
		LOG4CXX_DEBUG(logger(), "");

		m_acid_prob_calced = true;

		OUT_FNCTN_BOTTOM(logger());
	}
	else
		throw ERROR(logger(), "Trying to calculate data probability given a Rev_Tree with number of active nodes unequal 1");
}



// check integrity of the order of pure and final nodes
BOOL Rev_Tree::check_pure_final()
{
	BOOL ret = true;
	for (Rev_Nd* n = first_node(); n; n = n->next())
		if (n->is_final_node())
			for (Rev_Nd* m = n->next(); m; m = m->next())
				if (m->is_pure_node() && (n->occur_subs().find(m->sub()) != n->occur_subs().end()))
					ret = false;

	return ret;
}



// prepare tree after all tips have been added
void Rev_Tree::set_to_all_tips_added()
{
	calc_num_subtypes_crfs();
	for (int i = 0; i < num_crfs(); ++i) {
		set<int> s;
		for (set<int>::iterator it = crf(i)->occur_subs().begin(); it != crf(i)->occur_subs().end(); ++it)
			if (sub_only_in_crfs(*it))
				s.insert(*it);
		crf(i)->ref_part_tree() = new Partition_Tree(crf(i)->occur_subs(), s);
	}
	m_all_tips_added = true;
}



// calculate log acid probs of all tips
void Rev_Tree::calc_tip_laps()
{
	for (Rev_Nd* n = first_node(); n; n = n->next())
		if (Rev_Tip_Nd* m = dynamic_cast<Rev_Tip_Nd*>(n))
			m->init_prob();
	m_all_tip_laps_calced = true;
}



// calculate the partition due to 'aliasing'
void Rev_Tree::calc_part_alias()
{
	const BOOL OUTPUT_DEBUG = 0;
	ofstream* out_strm = In_Out_Streams::instance()->ostrm(OUTPUT_FILE);
	m_part_alias = new Partition_Aliasing(m_crfs, num_sites());
	if (OUTPUT_DEBUG)
		*out_strm << *m_part_alias;
	m_part_alias_calced = true;
}



// recalculates act_sites/groups of all nodes having coal suc
void Rev_Tree::recalc_act_sites_grps()
{
	for (Rev_Nd* n = first_node(); n; n = n->next())
		n->recalc_act_sites_grps();
}



// recalculates occuring subtypes of all nodes having coal suc
void Rev_Tree::recalc_occur_subs()
{
	for (Rev_Nd* n = first_node(); n; n = n->next())
		n->recalc_occur_subs();
}



// set old first and last node container to zero
void Rev_Tree::set_old_first_last_to_zero()
{
	m_old_first_node_contnr = 0;
	m_old_last_node_contnr = 0;
}



// saves first and last node container in old first and last node container
void Rev_Tree::save_first_last()
{
	m_old_first_node_contnr = first_node_contnr();
	m_old_last_node_contnr = last_node_contnr();
}



// save tree to file
void Rev_Tree::save()
{
	ofstream* tree_strm = In_Out_Streams::instance()->ostrm(SAVED_TREE_FILE, true);
	
	*tree_strm << ">tree   # number of sites" << endl;
	*tree_strm << num_sites() << endl;

	*tree_strm << ">crfs   # id, number of breakpoints, '-', breakpoints, '-', subtypes of segments "
			   << endl;
	for (int i = 0; i < num_crfs(); ++i)
	{
		*tree_strm << crf(i)->id() << " " << crf(i)->num_bps() << " - ";
		for (int j = 0; j < crf(i)->num_bps(); ++j)
			*tree_strm << crf(i)->bp(j) << " ";
		*tree_strm << "- ";
		for (int j = 0; j < crf(i)->num_segms(); ++j)
			*tree_strm << crf(i)->segm_sub(j) << " ";
		*tree_strm << endl;
	}

	*tree_strm << ">nodes   # type (tip, coal, rec),"
			   << "tip: id of node, is pure node?, id of subtype / crf, sequence, "
			   << "coal: id of node, time, id of first and second child, "
			   << "rec: ids of nodes, time, id of child, subtypes of first node, '-', subtypes of second node"
			   << endl;
	for (Rev_NC* c = first_node_contnr(); c; c = c->next())
		c->save(tree_strm);

	*tree_strm << ">end" << endl;
}



// load tree from file
Rev_Tree* load_tree()
{
	OUT_FNCTN_HEADER(Rev_Tree::logger());

	ifstream* tree_strm = In_Out_Streams::instance()->istrm(SAVED_TREE_FILE);
	string str;
	getline(*tree_strm, str);
	LOG4CXX_DEBUG(Rev_Tree::logger(), str);
	if (str.find(">tree") != 0) {
		stringstream ss;
		ss << "File format error in load_tree (tree): " << In_Out_Streams::instance()->filename(SAVED_TREE_FILE);
		throw Error(ss.str());
	}

	int num_sites;
	getline(*tree_strm, str);
	LOG4CXX_DEBUG(Rev_Tree::logger(), str);
	stringstream ss(str);
	ss >> num_sites;
	Rev_Tree* tr = new Rev_Tree(num_sites, Constants::instance()->num_mutation_rates());

	getline(*tree_strm, str);
	LOG4CXX_DEBUG(Rev_Tree::logger(), str);
	if (str.find(">crfs") != 0) {
		stringstream ss;
		ss << "File format error in load_tree (crfs): " << In_Out_Streams::instance()->filename(SAVED_TREE_FILE);
		throw Error(ss.str());
	}

	do {
		getline(*tree_strm, str);
		LOG4CXX_DEBUG(Rev_Tree::logger(), str);
		if (str[0] != '>') {
			stringstream ss(str);
			int id;
			ss >> id;
			int num_bps;
			ss >> num_bps;
			string dummy;
			ss >> dummy;
			VECTOR(int) bps(num_bps);
			for (int i = 0; i < num_bps; ++i)
				ss >> bps[i];
			ss >> dummy;
			VECTOR(int) segm_subs(num_bps + 1);
			for (int i = 0; i < num_bps + 1; ++i)
				ss >> segm_subs[i];

			new CRF(tr, num_bps, bps, segm_subs);
		}
	}
	while (!tree_strm->eof() && (str[0] != '>'));
	if (str.find(">nodes") != 0) {
		stringstream ss;
		ss << "File format error in load_tree (nodes): " << In_Out_Streams::instance()->filename(SAVED_TREE_FILE);
		throw Error(ss.str());
	}

	tr->calc_part_alias();

	map<int, Rev_Nd*> saved_id_to_new_node;
	do {
		getline(*tree_strm, str);
		LOG4CXX_DEBUG(Rev_Tree::logger(), str);
		if (str[0] != '>') {
			stringstream ss(str);
			char type;
			ss >> type;
			switch (type) {
			case 'T': {
				int id;
				ss >> id;
				BOOL is_pure;
				ss >> is_pure;
				int sub_crf_id;
				ss >> sub_crf_id;
				string seq;
				ss >> seq;
				Rev_Tip_NC* c = new Rev_Tip_NC(tr, seq, sub_crf_id, is_pure);
				saved_id_to_new_node[id] = c->node();
				break;
			}
			case 'C': {
				tr->set_to_all_tips_added();
				int id;
				ss >> id;
				double time;
				ss >> time;
				int id_first_child;
				ss >> id_first_child;
				int id_second_child;
				ss >> id_second_child;
				Rev_Coal_NC* c = new Rev_Coal_NC(tr, time, saved_id_to_new_node[id_first_child],
								                 saved_id_to_new_node[id_second_child]);
				saved_id_to_new_node[id] = c->node();
				break;
			}
			case 'R': {
				tr->set_to_all_tips_added();
				int first_id;
				ss >> first_id;
				int second_id;
				ss >> second_id;
				double time;
				ss >> time;
				int id_child;
				ss >> id_child;

				string str_sub;
				set<int> subtypes_first;
				do {
					ss >> str_sub;
					if (str_sub.compare("-") != 0) {
						stringstream tt(str_sub);
						int sub;
						tt >> sub;
						subtypes_first.insert(sub);
					}
				}
				while (!ss.eof() && (str_sub.compare("-") != 0));

				set<int> subtypes_second;
				do {
					int sub;
					ss >> sub;
					subtypes_second.insert(sub);
				}
				while (!ss.eof());

				Rev_Rec_NC* c = new Rev_Rec_NC(tr, time, saved_id_to_new_node[id_child],
								   			   subtypes_first, subtypes_second);
				saved_id_to_new_node[first_id] = c->first_node();
				saved_id_to_new_node[second_id] = c->second_node();
				break;
			}
			default: {
				stringstream ss;
				ss << "File format error in load_tree (node type): " << In_Out_Streams::instance()->filename(SAVED_TREE_FILE);
				throw Error(ss.str());
				break;
			}
			}
		}
	}
	while (!tree_strm->eof() && (str[0] != '>'));
	if (str.find(">end") != 0) {
		stringstream ss;
		ss << "File format error in load_tree (end): " << In_Out_Streams::instance()->filename(SAVED_TREE_FILE);
		throw Error(ss.str());
	}

	tr->accept_changes();
	tr->calc_tip_laps();
	tr->log_prob_acids();

	if (Constants::instance()->messages_on())
		cout << "tree loaded" << endl;

	OUT_FNCTN_BOTTOM(Rev_Tree::logger());

	return tr;
}



// check integrity
void Rev_Tree::check() const
{
	Base_Tree<Rev_Nd, Rev_Coal_Nd, Rev_Rec_Nd, Rev_NC>::check();

	BOOL is_ok = true;
	stringstream ss;

	// check for old_next() circular
	if (old_first_node_contnr())
	{
		set<string> nodes;
		Rev_NC* c = old_first_node_contnr();
		while (c)
		{
			pair<set<string>::iterator, BOOL> p = nodes.insert(c->str_id());
			if (!p.second)
			{
				is_ok = false;
				ss << "old_next() is circular: ";
				Rev_NC* d = old_first_node_contnr();
				int k = 0;
				if (old_first_node_contnr()->str_id() == *p.first)
					k++;
				while (k < 2)
				{
					ss << d->str_id() << " -> ";
					if (d->is_pos_affected())
						d = d->old_next();
					else
						d = d->next();
					if (d->str_id() == *p.first)
						k++;
				}
				ss << d->str_id() << endl;;
				break;
			}

			if (c->is_pos_affected())
				c = c->old_next();
			else
				c = c->next();
		}
	}

	// check for old_prev() circular
	if (old_last_node_contnr())
	{
		set<string> nodes;
		Rev_NC* c = old_last_node_contnr();
		while (c)
		{
			pair<set<string>::iterator, BOOL> p = nodes.insert(c->str_id());
			if (!p.second)
			{
				is_ok = false;
				ss << "old_prev() is circular: ";
				Rev_NC* d = old_last_node_contnr();
				int k = 0;
				if (old_last_node_contnr()->str_id() == *p.first)
					k++;
				while (k < 2)
				{
					ss << d->str_id() << " -> ";
					if (d->is_pos_affected())
						d = d->old_prev();
					else
						d = d->prev();
					if (d->str_id() == *p.first)
						k++;
				}
				ss << d->str_id() << endl;;
				break;
			}

			if (c->is_pos_affected())
				c = c->old_prev();
			else
				c = c->prev();
		}
	}

	if (!is_ok)
		throw ERROR(logger(), ss.str());
}



// output nodes to text file
ostream& operator<<(ostream& s, const Rev_Tree& tr)
{
	s << "*** TREE ***" << endl;

	s << "first: ";
	if (tr.first_node())
		s << tr.first_node()->id();
	else
		s << "NULL";
	s << endl;

	s << "last: ";
	if (tr.last_node())
		s << tr.last_node()->id();
	else
		s << "NULL";
	s << endl;

	s << "first_old: ";
	if (tr.old_first_node())
		s << tr.old_first_node()->id();
	else
		s << "NULL";
	s << endl;

	s << "last_old: ";
	if (tr.old_last_node())
		s << tr.old_last_node()->id();
	else
		s << "NULL";
	s << endl << endl;

	s << "num_nodes: " << tr.num_nodes() << endl;
	s << "num_tips: " << tr.num_tips() << endl;
	s << "num_act_nodes: " << tr.num_act_nodes() << endl;

	s << "mutation rates: ";
	for (int i = 0; i < tr.num_mutation_rates(); ++i)
		s << tr.mutation_rate(i) << " ";
	s << endl << endl;

	for (Rev_Nd* n = tr.first_node(); n; n = n->next())
	{
		s.width(3);
		s << n->id() << " " << flush;
		s << n->acid_seq() << "   " << flush;

		if (n->is_unmarked())
			s << "U" << flush;
		if (n->is_new())
			s << "N" << flush;
		if (n->is_old())
			s << "O" << flush;
		s << " ";
		s.width(5);
		s << boolalpha << n->is_anc_affected() << "(A) " << flush;
		s.width(5);
		if (Rev_Intern_Nd* m = dynamic_cast<Rev_Intern_Nd*>(n))
			s << boolalpha << m->rev_intern_node_contnr()->is_suc_affected() << "(S) " << flush;
		s.width(5);
		s << boolalpha << n->is_pos_affected() << "(P) " << flush;
		s.width(5);
		s << boolalpha << n->is_time_affected() << "(T)   " << flush;

		s.width(8);
		s.setf(ios::fixed,ios::floatfield);
		s.precision(3);
		s << n->birth_time() << "   " << flush;

		s.width(5);
		s << n->num_elig_sites() << "   " << flush;

		s.width(5);
		s << GSL_MIN(10000, n->unspec_node_contnr()->index()) << "   " << flush;

		if (n->prev())
		{
			s.width(4);
			s << n->prev()->id() << " " << flush;
		}
		else
			s << "NULL " << flush;

		if (n->next())
		{
			s.width(4);
			s << n->next()->id() << " " << flush;
		}
		else
			s << "NULL " << flush;
		s << "   ";

		if (n->is_active())
		{
			s << "A              " << flush;
		}
		if (n->has_coal_anc())
		{
			if (n->is_first_child())
			{
				s << "C1" << flush;
				s.width(3);
				s << n->parent_node()->id() << flush;
				s << "(P) " << flush;
				s.width(3);

				s << n->parent_node()->second_child_node()->id() << flush;
				s << "(B)" << flush;
			}
			else
			{
				s << "C2" << flush;
				s.width(3);
				s << n->parent_node()->id() << flush;
				s << "(P) " << flush;
				s.width(3);
				s << n->parent_node()->first_child_node()->id() << flush;
				s << "(B)" << flush;
			}
		}
		if (n->has_rec_anc())
		{
			s << "R " << flush;
			s.width(3);
			s << n->first_parent_node()->id() << flush;
			s << "(P) " << flush;
			s.width(3);
			s << n->second_parent_node()->id() << flush;
			s << "(P)" << flush;
		}
		s << "   ";

		s << endl;

		s << "                                                                              ";
		if (n->is_pos_affected())
		{
			if (n->old_prev())
			{
				s.width(4);
				s << n->old_prev()->id() << " " << flush;
			}
			else
				s << "NULL " << flush;

			if (n->old_next())
			{
				s.width(4);
				s << n->old_next()->id() << " " << flush;
			}
			else
				s << "NULL " << flush;
			s << "   ";
		}
		else
			s << "             ";

		if (n->is_anc_affected())
		{
			if (n->had_coal_anc())
			{
				if (n->was_first_child())
				{
					s << "C1" << flush;
					s.width(3);
					s << n->old_parent_node()->id() << flush;
					s << "(P) " << flush;
				}
				else
				{
					s << "C2" << flush;
					s.width(3);
					s << n->old_parent_node()->id() << flush;
					s << "(P) " << flush;
				}
			}
			if (n->had_rec_anc())
			{
				s << "R " << flush;
				s.width(3);
				s << n->first_old_parent_node()->id() << flush;
				s << "(P) " << flush;
				s.width(3);
				s << n->second_old_parent_node()->id() << flush;
				s << "(P)" << flush;
			}
			s << "     ";
		}
		else
			s << "                  ";

		if (Rev_Intern_Nd* m = dynamic_cast<Rev_Intern_Nd*>(n))
			if (m->rev_intern_node_contnr()->is_suc_affected())
			{
				if (Rev_Coal_Nd* l = dynamic_cast<Rev_Coal_Nd*>(m))
				{
					s << "C " << flush;
					s.width(3);
					s << l->first_old_child_node()->id() << flush;
					s << "(C) " << flush;
					s.width(3);
					s << l->second_old_child_node()->id() << flush;
					s << "(C)" << flush;
				}
				else
					if (Rev_Rec_Nd* l = dynamic_cast<Rev_Rec_Nd*>(m))
					{
						if (l->is_first_in_NC())
						{
							s << "R1" << flush;
							s.width(3);
							s << l->old_child_node()->id() << flush;
							s << "(C) " << flush;
						}
						else
						{
							s << "R2" << flush;
							s.width(3);
							s << l->old_child_node()->id() << flush;
							s << "(C) " << flush;
						}
					}
			}
		s << endl << endl;
	}
	s << endl;

	return s;
}



ostream& operator<<(ostream& s, const enum_direction& dir)
{
	switch (dir)
	{
		case DIR_STAY:
			s << "DIR_STAY";
			break;
		case DIR_UP:
			s << "DIR_UP";
			break;
		case DIR_DOWN:
			s << "DIR_DOWN";
			break;
	}
	return s;
}







