/*********************************************************************************
 filename	: alignment.cc
 licence        : GNU Licence, see file LICENCE
 description    : -> alignment.hh
 author		: Anne-Kathrin Schultz, anne@gobics.de
 date		: Jan. 2004

content: classes Alignment, Profile

  |   date    |    author            |  changes
 -|-----------|----------------------|-------------------------------
  | April 06  | Anne-Kathrin Schultz | reading in LINUX AND windows alignments 

**********************************************************************************/
#include "const_def.hh"
#include "alignment.hh"
#include <iostream>
#include <fstream>
#include <math.h>
#include <string>
#include <vector>
#include <list>
#include <cstdlib>

#define MSA_OUTPUT "msa_output.txt"
#define PROFIL     "profile.txt"


using namespace std;


/*******************************************************************\
                                                                  
 *                       class Alignment                           *
                                                                 
\*******************************************************************/

//-----------------------------------------------------------------------------
//   Alignment(ifstream&)
//-----------------------------------------------------------------------------

Alignment::Alignment(ifstream& input) {      

  anz_Subtypes  = 1;              
  anz_Sequenzen = 1;                              
  anz_Spalten   = 0;
  anz_Residues  = 0;              // in alignment!
  prof          = NULL;
  eval_Alignment(input);
}

//-----------------------------------------------------------------------------
//   Alignment(ifstream&, int, int, int) 
//-----------------------------------------------------------------------------

Alignment::Alignment(ifstream& input,
		     int subtypes,
		     int sequenzen, 
		     int spalten){    

  anz_Subtypes    = subtypes;
  anz_Sequenzen   = sequenzen;
  anz_Spalten     = spalten;
  anz_Residues    = 0;
  prof            = NULL;
  eval_Alignment(input);
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// ~Alignment()
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------

Alignment::~Alignment(){                     

  for (int i=0; i<anz_Sequenzen; i++)
    delete [] alignment[i];
  delete [] alignment;
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// eval_Alignment(ifstream&):
//
//    evaluation of the whole alignment: read in, fill alignment[][], determineConsensusColumns()
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
void Alignment::eval_Alignment(ifstream& input) {

  int max_anz_Spalten = 0;

  if ((anz_Spalten == 0) || (anz_Sequenzen) == 1)
    max_anz_Spalten = anzahlZeilenUndSpalten(input);
  else
    max_anz_Spalten = anz_Spalten;
  speicherBereitstellen();
  dateiEinlesen(input, max_anz_Spalten);
  determineConsensusColumns();
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// eval_Subalignment(int):
//
//    evaluation of the subalignments
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
Profile* Alignment::eval_Subalignment(int noOfSubtype) {
 
  prof = NULL;         // set NULL => new profile of the corresponding sub-alignment will be computed
  char **sub_Alignment;
  int start = 0;
  for (int i=0; i<noOfSubtype; i++)
    start += numSeqsOfSubtypes[i];
  int numSeqs = numSeqsOfSubtypes[noOfSubtype];
  sub_Alignment    = new char*[numSeqs];
  for (int j=0; j<numSeqs; j++)          
    sub_Alignment[j]= new char[anz_Spalten +1];
  for (int j=0; j<numSeqs; j++)
    for (int k=0; k<anz_Spalten; k++)
      sub_Alignment[j][k] = alignment[start+j][k];
  
  Profile* p = getProfil(sub_Alignment, numSeqs, anz_Spalten);
  return p;
}


//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// anzahlZeilenUndSpalten(ifstream&):
//
//    determination of the number of rows and columns of the alignment (necessary to determine size of alignment[][])
//    and of the number of sequences(=rows) of each subtype
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
int Alignment::anzahlZeilenUndSpalten(ifstream& eingabe){
  int  numSeqsOfSubtype = 0;
  int  max_anz_Spalten  = 0;                     // length of the longest row
  int  akt_anz_Spalten  = 0;                     // length of the actual (eingelesenen) ROW
  int  name             = 0;                     // =0: name of a sequence, =1: name of a subtype (?)
  int  neuerSubtype     = 0;
  char Z;                                        // character which is read in/memorized
  int  wind             = 0;                     // query of line feed (Abfrage fuer Zeilenvorschub-Zeichen '\r' (windows) (=1 if found))
  
  while (eingabe.get(Z)) {
    if (Z != '\r') {   // falls kein Zeilenvorschub ('\r') (in Windows eingefuegt, wuerde sonst als Extra-Zeichen gezaehlt werden -> Fehler in der Laenge der eingelesenen Sequenzen)
      if (Z == '>') { // '>' : sequence name     '>>' : subtype name
	if (name == 1){       // name of a subtype
	  neuerSubtype = 1;
	}
	else {    // name of a sequence
	  name = 1;
	  neuerSubtype = 0;
	}
      }
      if (name == 1 && neuerSubtype == 1 && anz_Sequenzen != 1) {
	// now new subtype => count number of precedent sequences (except of first line)
	anz_Subtypes++;
	numSeqsOfSubtypes.push_back(numSeqsOfSubtype);
	numSeqsOfSubtype = 0;
	neuerSubtype = 0;
      }
      if (Z == '\n') {  // || Z == '\r'){ // || Z == '\x0D' || Z == '\x0A') {   // end of line (unter Beruecksichtigung der Zeilenenden unter windows z.B.) => Vergleich: Zeilenlaengen
	if (max_anz_Spalten < akt_anz_Spalten)    
	  max_anz_Spalten = akt_anz_Spalten;
	if (name != 1) {
	  anz_Sequenzen    += 1;
	  numSeqsOfSubtype += 1;
	  if (akt_anz_Spalten != anz_Spalten) {   // (1st sequence > subsequent sequence) => error 
	    cerr << "Alignment incorrect, sequences don't have the same length. Please check this!" << endl
		 << "last line must be a blank line! " 
		 << endl;	    
	    exit(1);
	  }
	}
	name            = 0;
	neuerSubtype    = 0;
	akt_anz_Spalten = 0;             
      }
      else {
	akt_anz_Spalten += 1;
	if (name != 1) {
	  if (anz_Sequenzen == 1) {
	    anz_Spalten += 1;                  // => length of the first sequence
	  }
	  else {
	    if (akt_anz_Spalten > anz_Spalten) {  // (1st sequence < following/last(!) sequence) => error
	      cerr << "Alignment incorrect, sequences don't have the same length. Please check this!" << endl
		   << "Number of columns read in in the actually considered sequence: " << akt_anz_Spalten << " " << "length of 1st seq: " << anz_Spalten << " " << "Number of sequences read in (incl. act. seq.): " << anz_Sequenzen << endl;
	      exit(1);
	    }
	  }
	}
      }
    }
    else {
      wind = 1; // fuer das Zaehlen der max. Spaltenanzahl, ('\r' ist extra Zeichen) fuer Fkt. dateiEinlesen() benoetigt.
    }
  }
  numSeqsOfSubtypes.push_back(numSeqsOfSubtype);
  anz_Sequenzen--; // since the last line break is also counted, i.e. at the end the counter 'anz_Sequenzen' is increased again (it counts the blank line at the end); thus -1.
  if (wind == 1)
    max_anz_Spalten++;
  return max_anz_Spalten;
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// speicherBereitstellen():
//
//    memory allocation for alignment[][]
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
void Alignment::speicherBereitstellen(){
 
 alignment    = new char*[anz_Sequenzen];
  for (int j=0; j<anz_Sequenzen; j++) {          
    alignment[j]= new char[anz_Spalten +1];
  }
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// dateiEinlesen(ifstream&, int):
//
//    reads in the alignment from the file
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
void Alignment::dateiEinlesen(ifstream& eingabe, int max_anz_Spalten){ 

  int   zeile            = 0;
  char *eingeleseneZeile = new char[max_anz_Spalten+1];

  eingabe.clear();
  eingabe.seekg(0);
  
  if (numSeqsOfSubtypes.size() == 1) { // was ist bei allg. Profil? Wenn nicht jede Sequenz als einzelner Subtyp betrachtet wird?     AENDERN!!
    // d.h. keine Subtypen, sondern alle Sequenzen werden einzeln betrachtet => d.h. jeder Subtyp (= Zeile) enthaelt eine Sequenz
    numSeqsOfSubtypes.clear();
    for (int i=0; i<anz_Sequenzen; i++){
      numSeqsOfSubtypes.push_back(1);
    }  
  }
  /*else {
    for (unsigned int i=0; i<numSeqsOfSubtypes.size(); i++) {
    }
  }*/
  
  while (eingabe.getline( eingeleseneZeile, max_anz_Spalten+1)) {
    if ((eingeleseneZeile[0] != '\0') && (eingeleseneZeile[0] != '>')){        // => sequence 
      for (int spalte=0; spalte< anz_Spalten; spalte++) { 
	alignment[zeile][spalte] = eingeleseneZeile[spalte];
	if (eingeleseneZeile[spalte] != '*' && eingeleseneZeile[spalte] != '-')
	  anz_Residues++;
      }
      zeile++;
    }
    if (eingeleseneZeile[0] == '>' && eingeleseneZeile[1] == '>') { // name of subtype
      string s = eingeleseneZeile;
      unsigned int k=0;
      while (k < s.size()) {
	if (s[k] == '\x0D' || s[k] == '\x0A' || s[k] == '>') {
	  s.erase(k,1);
	}
	else 
	  k++;
      }
      names.push_back(s);
    }
  }
  anz_Sequenzen = zeile;                    // Veraenderung dieser Anzahl falls vorher Leerzeilen eingelesen wurden
                                            // oder Ueberpruefen?

  delete [] eingeleseneZeile;
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// determineConsensusColumn():
//
//    determines the first and the last consensus column of each subtype
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
void Alignment::determineConsensusColumns() {
  
  int spalte    = 0;
  int num_gaps  = 0;
  int first_row = 0;
  int consensus = -1;       
  
  first_ConsensusColumn = 0;
  last_ConsensusColumn  = anz_Spalten;
  
  for (unsigned int noOfSubtype = 0; noOfSubtype < numSeqsOfSubtypes.size(); noOfSubtype++) {
    // Searching for first consensus column (starting at the beginning of the alignment):
    spalte = 0;
    consensus = -1;
    while (consensus == -1) {
      num_gaps = 0;
      for (int zeile = first_row; zeile < first_row + numSeqsOfSubtypes[noOfSubtype]; zeile++) {
	if ((alignment[zeile][spalte] == '-') || (alignment[zeile][spalte] == '*')) 
	  num_gaps++;
      }
      if (((numSeqsOfSubtypes[noOfSubtype]-num_gaps) >= consensus_Percentage*numSeqsOfSubtypes[noOfSubtype]) || ((numSeqsOfSubtypes[noOfSubtype]-num_gaps) >= consensus_Threshold)){
	consensus = spalte;
	if (first_ConsensusColumn < consensus) 
	  first_ConsensusColumn = consensus;
      }
      spalte++;
    }
    spalte = anz_Spalten-1;
    consensus = -1;
    // Searching for last consensus column (starting at the end of the alignment)
    while (consensus == -1) {
      num_gaps = 0;
      for (int zeile = first_row; zeile < first_row + numSeqsOfSubtypes[noOfSubtype]; zeile++) {
	if ((alignment[zeile][spalte] == '-') || (alignment[zeile][spalte] == '*')) 
	  num_gaps++;
      }
      if (((numSeqsOfSubtypes[noOfSubtype]-num_gaps) >= consensus_Percentage*numSeqsOfSubtypes[noOfSubtype]) || ((numSeqsOfSubtypes[noOfSubtype]-num_gaps) >= consensus_Threshold)){
	consensus = spalte;
	if (last_ConsensusColumn > consensus) 
	  last_ConsensusColumn = consensus;
      }
      spalte--;
    }
    first_row = first_row + numSeqsOfSubtypes[noOfSubtype];
  }
}


//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// getProfil(char**&, int, int):
//
//    generates the profile of the alignment with the aid of class Profile
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
Profile* Alignment::getProfil(char** &alignment, int numSequenzen, int numSpalten) {

  prof = new Profile(alignment, numSequenzen, numSpalten, first_ConsensusColumn, last_ConsensusColumn);
  return prof;
}

/*******************************************************\

 *                    class Profile                    *

\*******************************************************/

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Profile(char**&, int, int, int, int)
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
Profile::Profile(char** &alignment, int numSequences, int numColumns, int first_ConsensusColumn, int last_ConsensusColumn) {

  first_Consensus = first_ConsensusColumn;
  last_Consensus  = last_ConsensusColumn;
  numSpalten      = numColumns;
  numSequenzen    = numSequences;
  profilErstellen(alignment);
  eval_Profil();
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// ~Profile()
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
Profile::~Profile() {

  for (int i=0; i<numSpalten; i++) 
    delete [] profil[i];
  delete [] profil;
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// profilErstellen(char**&):
//
//    generates the profile of the alignment/subtype
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
void Profile::profilErstellen(char** &alignment) {

  // Profile: columns = frequencies of the nucleotides/amino acids (A, C, D, E, ...... , Y)  in the rows of the alignment
  //          rows    = columns in the alignment

  int NUM;
  if (alphabet == protein)
    NUM = NUM_AA;
  else 
    NUM = NUM_NUCL;
  
  profil = new int*[numSpalten];  
  for (int i=0; i<numSpalten; i++) {
    profil[i] = new int[NUM+3];                   // +1 for stop-character oder unknown!
                                                  // +2 for gap-character
  }

  for (int j=0; j<numSpalten; j++)
    for (int k=0; k<NUM+3; k++)
      profil[j][k] = 0;
  for (int j=0; j<numSpalten; j++){
    for (int k=0; k<numSequenzen; k++){
      char c = alignment[k][j];
      int  d = Const_Def::indexOfAlphabet[(int) c];
      if (d != -1)
	profil[j][d]++;
      else {
	// Increment the special/unknown letter at position NUM in the corresponding set of letters
	if (alphabet == protein) 
	  profil[j][Const_Def::indexOfAlphabet[(int) amino[NUM]]]++;
	else 
	  profil[j][Const_Def::indexOfAlphabet[(int) nucleic[NUM]]]++;
      }
    }
  }
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// eval_Profil():
//
//    evaluation of the alignment (profile): 
//    consensus or Insert column?
//    if anz_gaps > 50%  => Insert, else consensus (match)
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
void Profile::eval_Profil() {
  
  int NUM;
  if (alphabet == protein)
    NUM = NUM_AA;
  else 
    NUM = NUM_NUCL;
  
  int num_gaps = 0;
  
  konsensusColumns = new int[numSpalten];
    
  for (int i=0; i<numSpalten; i++) {
    int sum = 0;
    // gaps werden nicht addiert
    for (int j=0; j<NUM; j++) 
      sum += profil[i][j];
    num_gaps = numSequenzen-sum;
    // consider number of gaps; if in a column the number of gaps is > 'Threshold*anz_Zeilen' => consensus columns (here (1-Threshold, because we compare for Gaps)
    if ((((numSequenzen-num_gaps) >= consensus_Percentage*numSequenzen) || ((numSequenzen-num_gaps) >= consensus_Threshold))) {
      if (((consensus == commonConsensus) && (i >= first_Consensus) && (i <= last_Consensus)) || (consensus == noCommonConsensus))
	konsensusColumns[i] = 1;          // Match => consensus column
      else
	konsensusColumns[i] = 0;
    }
    else
      konsensusColumns[i] = 0;          // Insert => no consensus column
  }
}

