/**************************************************************
 filename	: HMM.CC 
 licence        : GNU Licence, see file LICENCE
 description    : -> hmm.hh
 author		: Anne-Kathrin Schultz, anne@gobics.de
 date		: 03/04 - 08/05

content: classes HMM, ActiveStates


  |   date    |    author            |  changes
 -|-----------|----------------------|-------------------------------
  | April 06  | Anne-Kathrin Schultz | sequences containing gaps

***************************************************************/
#include "const_def.hh"
#include "profileHMM.hh"
#include "hmm.hh"
#include "alignment.hh"
#include <math.h>
#include <float.h>
#include <vector>
#include <list>
#include <iostream>
#include <sstream>
#include <iomanip>
#include <string>
#include <time.h>
#include <cstdlib>

#define VITERBI_PATH_GFF                "viterbi_Path.gff"
#define RECOMBINATION                   "recombination.txt"
#define RECOMBINATION_GAPPED            "recombination_incl_gaps.txt"
#define RECOMBINATION_WITHOUT_POSITIONS "recombination_withoutPositions.txt"

#define MAX_ROW_LEN 20000            // max. length of query sequence, arbitrarily chosen, can be changed


using namespace std;


/******************************************************\

 *                   class HMM                        *

\******************************************************/

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// HMM(Alignment, int, ifstream&, ifstream&)
//
//    calls function createJumpingProfileHMM(...): 
//
//-------------------------------------------------------------------------------------------------------------------------------------------------------------------- 

HMM::HMM(Alignment *al,
	 int        numSubtypes,
	 ifstream&  emissionPriors,
	 ifstream&  querySequence) {

  createJumpingProfileHMM(al, numSubtypes, emissionPriors, querySequence);
 
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// ~HMM()
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
HMM::~HMM() {

}

// -------------------------------------------------------------------------------------------------------------------------------------------------------------------
// provideMemory(State****&, int, int):  
//
//    memory allocation for allStates[][][];
//
// -------------------------------------------------------------------------------------------------------------------------------------------------------------------
void HMM::provideMemory(State**** &allStates, 
			int        numSubtypes,
			int        numSpalten) {

  allStates = new State***[numSubtypes];
  for (int i=0; i<numSubtypes; i++) {
    allStates[i] = new State**[numSpalten];
    for (int j=0; j<numSpalten; j++) {
      allStates[i][j] = new State*[NUM_ACT_STATES];
      for (int k=0; k<NUM_ACT_STATES; k++) 
	allStates[i][j][k] = NULL;
    }
  }
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// fill_HMM_States(State****&):  
//
//    fills allStates[][][] with states:
//
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
void HMM::fill_HMM_States(State**** &allStates) {
  
  for (unsigned int state=0; state<HMM_States.size(); state++)
    allStates[HMM_States[state]->row][HMM_States[state]->column][(int)HMM_States[state]->stateType] = HMM_States[state];
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// addTransitions(State*, State****&, int, int, int&, int&): 
//
//   determines some additional predecessor and successor states of a state:
//
//   in profileHMM:addTransitions() the predecessors and successors of a state in the same subtype are determined
//   now the following transitions are added: from/to BEGIN, N_DELETE, N_INSERT, C_INSERT, C_DELETE, END 
//                                            and from/to certain states of the other subtypes
//
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
void HMM::addTransitions(State     *akt_Zustand, 
			 State**** &allStates,
			 int        numSubtypes,
			 int        numSpalten,
			 int       &num_JumpsToProfile,
			 int       &num_JumpsToSubtypes) {

  // Uebergaenge von N_Delete zu Matches und von Matches zu C_Delete...

  Transition *Uebergang; 
  int         zeile            = akt_Zustand->row;
  int         spalte           = akt_Zustand->column;
  int         gap_Laenge       = 0;
  int         sp               = spalte;
  int         vorgaenger       = 0;
  int         nachfolger       = 0;
  int         anz_SubProfile   = 0;
  
  num_JumpsToProfile  = 0;
  num_JumpsToSubtypes = 0;

  if ((akt_Zustand->stateType == Match) || (akt_Zustand->stateType == Delete)) {
    for (unsigned vorg=0; vorg<akt_Zustand->pred.size(); vorg++) {
      if ((akt_Zustand->pred[vorg]->whereFrom->row == akt_Zustand->row ) && (akt_Zustand->pred[vorg]->whereFrom->column < akt_Zustand->column)) {
	vorgaenger = 1;
	break;
      }
    }
    if (vorgaenger == 0) {  // BEGIN->akt_Zustand
      Uebergang = new Transition(HMM_States.front(), akt_Zustand);
      HMM_States.front()->succ.push_back( Uebergang);
      akt_Zustand->pred.push_back(Uebergang);

      if (akt_Zustand->stateType == Match) {  // N_INSERT->MATCH
	Uebergang = new Transition(HMM_States[2], akt_Zustand);
	(HMM_States[2])->succ.push_back( Uebergang);
	akt_Zustand->pred.push_back(Uebergang);
      }
    }
  }    
  if (akt_Zustand->stateType == Match) {  // N_DELETE->MATCH
    Uebergang = new Transition(HMM_States[1], akt_Zustand); 
    akt_Zustand->pred.push_back(Uebergang);
    HMM_States[1]->succ.push_back(Uebergang);
    
    // Match->C_Delete
    Uebergang = new Transition(akt_Zustand, HMM_States[HMM_States.size()-2]);
    HMM_States[HMM_States.size()-2]->pred.push_back(Uebergang);
    akt_Zustand->succ.push_back(Uebergang);
  }
  
  //  transitions to other subtypes (allStates[z][sp+1][(int)Match] is the successional (naechst(!)moeglicher) successor in the subtype) (at most the column before the last column is considered)
  if (spalte != numSpalten-1) {
       // Determination of the column of the successional successor of the current state akt_Zustand (of the same type of state) = spalte+gap_Laenge
    while (((spalte+gap_Laenge) < numSpalten-2) && (allStates[zeile][spalte+gap_Laenge+1][(int)Match] == NULL)) 
      gap_Laenge++;

    if (profile_mode == nocommon)
      anz_SubProfile = numSubtypes;
    else
      anz_SubProfile = numSubtypes + 1;

    // Loop over all subtypes
    for (int z=0; z<anz_SubProfile; z++) {
      sp = spalte;
      // search for the successional successor in another subtype (according to the above condition)
      while ((sp < spalte+gap_Laenge) && (allStates[z][sp+1][(int)Match] == NULL) ) 
	sp++;
      if ((sp<numSpalten) && (allStates[z][sp+1][(int)Match] != NULL))  { 
	if (z != zeile) {
	  Transition *Uebergang = new Transition(akt_Zustand, allStates[z][sp+1][(int)Match]);
	  if (z != numSubtypes) 
	    num_JumpsToSubtypes++;
	  if ((zeile != numSubtypes) && (z == numSubtypes)) 
	    num_JumpsToProfile = 1;
	  
	  akt_Zustand->succ.push_back( Uebergang);
	  allStates[z][sp+1][(int)Match]->pred.push_back( Uebergang);
	}
	if (akt_Zustand->stateType == Match) {
	  if (allStates[z][sp+1][(int)Delete] != NULL) {
	    if (z != zeile) {
	      Uebergang = new Transition(akt_Zustand, allStates[z][sp+1][(int)Delete]);
	      akt_Zustand->succ.push_back( Uebergang);
	      allStates[z][sp+1][(int)Delete]->pred.push_back( Uebergang);
	    }	 
	  }	
	}
      }
    }
  }
  if ((akt_Zustand->stateType == Match) || (akt_Zustand->stateType == Delete)) {
    for (unsigned nachf=0; nachf<akt_Zustand->succ.size(); nachf++) {
      if ((akt_Zustand->succ[nachf]->whereTo->row == akt_Zustand->row ) && (akt_Zustand->succ[nachf]->whereTo->column > akt_Zustand->column)) {
	nachfolger = 1;
	break;
      }
    }
    if (nachfolger == 0) { // akt_Zustand->END
      Uebergang = new Transition(akt_Zustand, HMM_States.back());
      HMM_States.back()->pred.push_back( Uebergang);
      akt_Zustand->succ.push_back(Uebergang);
      if (akt_Zustand->stateType == Match) { // MATCH->C_INSERT
	Uebergang = new Transition(akt_Zustand, HMM_States[HMM_States.size()-3]);
	HMM_States[HMM_States.size()-3]->pred.push_back( Uebergang);
	akt_Zustand->succ.push_back(Uebergang);
      }
    }
  }
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// createJumpingProfileHMM(Alignment*, int, ifstream&, ifstream&):
//
//    generates the jpHMM:
//
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
void HMM::createJumpingProfileHMM(Alignment *al, 
				  int        numSubtypes,
				  ifstream&  emissionPriors,
				  ifstream&  querySequence){
  
  State      *akt_Zustand;
  State****   allStates;
  Transition *Uebergang;
  Transition *t                   = new Transition();
  int         numSpalten          = al->getAnzSpalten();
  int         numResidues         = al->getAnzResidues();
  Profile    *p                   = al->getProfil(al->alignment, al->getAnzSequenzen(), al->getAnzSpalten());
  int**       profil              = p->getProfil();
  int         num_JumpsToSubtypes = 0;
  int         num_JumpsToProfile  = 0;
  int         spalte              = 0;
  
  int anz_SubProfile = 0;
  if (profile_mode == nocommon)
    anz_SubProfile = numSubtypes;
  else
    anz_SubProfile = numSubtypes+1;
  int **gaps = new int*[anz_SubProfile];
  for (int i=0; i<anz_SubProfile; i++)
    gaps[i] = new int[al->getAnzSpalten()];
  
  State*      begin_HMM           = new State(Begin, -3, -3);                            // Begin state of the des HMM (5'-end)
  State*      n_Delete_HMM        = new State(N_Delete, -2, -2);                         // N-terminal delete state (5'-end)
  State*      n_Insert_HMM        = new State(N_Insert, -1, -1);                         // N-terminal unaligned sequence state (5'-end)
  State*      c_Insert_HMM        = new State(C_Insert, anz_SubProfile+1, numSpalten);   // C-terminal unaligned sequence state (3'-end)
  State*      c_Delete_HMM        = new State(C_Delete, anz_SubProfile+2, numSpalten+1); // C-terminal delete state (3'-end)
  State*      end_HMM             = new State(End, anz_SubProfile+3, numSpalten+2);      // End state of the HMM (3'-end)

  ProfileHMM    *pHMM                = new ProfileHMM();
  vector<State*> profileHMM_States;
  int NUM; 
  if (alphabet == protein)
    NUM = NUM_AA;
  else 
    NUM = NUM_NUCL;
  
  // frequencies of all(!) letters that the alphabet doesn't contain are added (in the whole alignment) and saved in the variable
  int num_X = 1;       // for DNA and protein sequences      =1, => pseudo count, because probability of N(unknown) must be given
  for (int i=0; i<al->getAnzSpalten(); i++) 
    num_X     += profil[i][NUM];

  mytype prob_X = (mytype)num_X/((mytype)numResidues);
  
  vector<mytype> insert_emProbs;
  mytype        *n_insert_aa_Frequencies = new mytype[NUM];
  mytype        *c_insert_aa_Frequencies = new mytype[NUM];
  for (int i=0; i<NUM; i++){
    n_insert_aa_Frequencies[i] = 0;
    c_insert_aa_Frequencies[i] = 0;
  }

  // INSERT emission probabilities for N_INSERT and C_INSERT: 
  if (profile_mode == common) { // profile of the whole alignment
    gaps[numSubtypes] = p->getKonsensusColumns(); 
  }
  for (int noOfSubtype=0; noOfSubtype<numSubtypes; noOfSubtype++){
    p       = al->eval_Subalignment(noOfSubtype);
    profil  = p->getProfil();
    gaps[noOfSubtype] = p->getKonsensusColumns();           // profile of the corresponding subtype
    
    spalte=-1;
    // N_INSERT emission frequencies:
    spalte = 0;
    while ((spalte < al->getAnzSpalten()) && (gaps[noOfSubtype][spalte] != 1)) {  
      for (int a=0; a<NUM; a++)
	n_insert_aa_Frequencies[a] += profil[spalte][a];
      spalte++;
    }
    
    // C_INSERT emission frequencies:
    spalte = al->getAnzSpalten()-1;
    while ((spalte > -1) && (gaps[noOfSubtype][spalte] != 1)) {  
      for (int a=0; a<NUM; a++)
	c_insert_aa_Frequencies[a] += profil[spalte][a];
      spalte--;
    }
  }
  
  // N_INSERT emission probabilities:  
  insert_emProbs     = State::dirichlet_EmissionPriors(Insert, n_insert_aa_Frequencies, emissionPriors);  
  for (int j=0; j<NUM; j++)
    insert_emProbs[j] = ((mytype)1 - prob_X)*insert_emProbs[j];
  insert_emProbs.push_back(prob_X);
  n_Insert_HMM->log_emProb = State::log_Probs(insert_emProbs);
  
  // C_INSERT emission probabilities:  
  insert_emProbs     = State::dirichlet_EmissionPriors(Insert, c_insert_aa_Frequencies, emissionPriors);  
  for (int j=0; j<NUM; j++) 
    insert_emProbs[j] = ((mytype)1 - prob_X)*insert_emProbs[j];
  insert_emProbs.push_back(prob_X);
  c_Insert_HMM->log_emProb = State::log_Probs(insert_emProbs);

  provideMemory(allStates, anz_SubProfile, numSpalten);
 
  for (int typ=0; typ<anz_SubProfile; typ++) {
    profileHMM_States = pHMM->createProfileHMM(al, numSubtypes, typ, anz_SubProfile, emissionPriors);
    for (unsigned int length=0; length<profileHMM_States.size(); length++)
      HMM_States.push_back(profileHMM_States[length]);
    profileHMM_States.clear();
  }
  

  // transfer all states from HMM_States in allStates[][][] in fill_HMM_States(...);
  // clear HMM_States (delete all states); 
  // reassignment/rearrangement of the states in HMM_States by reading them in column by column from allStates[][][] (beginning in column 0);
  // HMM_States = (BEGIN, N_DELETE, N_INSERT, ..., C_INSERT, C_DELETE, END);
  fill_HMM_States(allStates);  
  HMM_States.clear();       
  HMM_States.push_back(begin_HMM);
  HMM_States.push_back(n_Delete_HMM);
  HMM_States.push_back(n_Insert_HMM);
  HMM_States.push_back(c_Insert_HMM);
  HMM_States.push_back(c_Delete_HMM);
  HMM_States.push_back(end_HMM); 
  
  // special transitions:
  Uebergang = new Transition(begin_HMM, n_Delete_HMM);
  begin_HMM->succ.push_back(Uebergang);
  n_Delete_HMM->pred.push_back(Uebergang);
  
  Uebergang = new Transition(begin_HMM, n_Insert_HMM);
  begin_HMM->succ.push_back(Uebergang);
  n_Insert_HMM->pred.push_back(Uebergang);
  
  Uebergang = new Transition(n_Insert_HMM, n_Insert_HMM);
  n_Insert_HMM->succ.push_back(Uebergang);
  n_Insert_HMM->pred.push_back(Uebergang);
  
  Uebergang = new Transition(c_Insert_HMM, end_HMM);
  Uebergang->transProb = epsilon;
  c_Insert_HMM->succ.push_back(Uebergang);
  end_HMM->pred.push_back(Uebergang);
  
  Uebergang = new Transition(c_Delete_HMM, end_HMM);
  Uebergang->transProb = 1;
  c_Delete_HMM->succ.push_back(Uebergang);
  end_HMM->pred.push_back(Uebergang);
  
  Uebergang = new Transition(c_Insert_HMM, c_Insert_HMM);
  Uebergang->transProb = (mytype)1-epsilon; 
  c_Insert_HMM->succ.push_back(Uebergang);
  c_Insert_HMM->pred.push_back(Uebergang);

  // transitions and transition probabilities:
  for (int spalte=0; spalte<numSpalten; spalte++) {
    for (int zeile=0; zeile<anz_SubProfile; zeile++) {
      if (allStates[zeile][spalte][(int) Match] != NULL) {
	for (int typ=(int)Match; typ<=(int)Delete; typ++) {
	  if (allStates[zeile][spalte][typ] != NULL) {
	    akt_Zustand = allStates[zeile][spalte][typ];
	    HMM_States.insert(--(--(--HMM_States.end())), akt_Zustand);
	    addTransitions(akt_Zustand, allStates, numSubtypes, numSpalten, num_JumpsToProfile, num_JumpsToSubtypes);
	    t->ermittleTransProbs(akt_Zustand, allStates, numSubtypes, numSpalten, gaps, num_JumpsToProfile, num_JumpsToSubtypes);
	  }
	}
      }
    }
  }
  // only now transition probabilities for N_INSERT and N_DELETE, because only now all transitions are determined
  t->ermittleTransProbs(n_Insert_HMM, allStates, numSubtypes, numSpalten, gaps, num_JumpsToProfile, num_JumpsToSubtypes);
  t->ermittleTransProbs(n_Delete_HMM, allStates, numSubtypes, numSpalten, gaps, num_JumpsToProfile, num_JumpsToSubtypes); 
 
  // transition probabilities for BEGIN   
  for (unsigned int pos=0; pos<HMM_States[0]->succ.size(); pos++) {
    if (HMM_States[0]->succ[pos]->whereTo->stateType == N_Insert) 
      HMM_States[0]->succ[pos]->transProb = (mytype)1-epsilon-d_init; 
    else if (HMM_States[0]->succ[pos]->whereTo->stateType == N_Delete) 
      HMM_States[0]->succ[pos]->transProb = d_init;
    else {
      if (profile_mode == nocommon){
	if (HMM_States[0]->succ[pos]->whereTo->stateType == Match || HMM_States[0]->succ[pos]->whereTo->stateType == Delete) 
	  HMM_States[0]->succ[pos]->transProb = (epsilon/2)*((mytype)1/(numSubtypes));
      }
      else {
	if (HMM_States[0]->succ[pos]->whereTo->row == numSubtypes) { 
	  if (HMM_States[0]->succ[pos]->whereTo->stateType == Match || HMM_States[0]->succ[pos]->whereTo->stateType == Delete) 
	    HMM_States[0]->succ[pos]->transProb = 0.5*epsilon;
	}
	else {
	  if (HMM_States[0]->succ[pos]->whereTo->stateType == Match || HMM_States[0]->succ[pos]->whereTo->stateType == Delete) 
 	    HMM_States[0]->succ[pos]->transProb = 0.5*(epsilon/2)*((mytype) 1)/(numSubtypes);
	} 
      }
    }
  }

  mytype prob = 0;
  mytype sum  = 0;

  // standardization of the transition probabilities:
  for (unsigned int j=0; j<HMM_States.size()-1; j++) {
    sum  = 0;
    prob = 0;
    akt_Zustand = HMM_States[j];
    for (unsigned int i = 0; i<akt_Zustand->succ.size(); i++) 
      sum += akt_Zustand->succ[i]->transProb;
    for (unsigned int k = 0; k<akt_Zustand->succ.size(); k++) {
      prob = akt_Zustand->succ[k]->transProb/sum;
      akt_Zustand->succ[k]->transProb = prob;
    }
  }
  
  // logarithm of the transition probabilities (successor):  (without END)
  for (unsigned int pos=0; pos<HMM_States.size()-1; pos++) {
    akt_Zustand = HMM_States[pos];
    for (unsigned int i=0; i<akt_Zustand->succ.size(); i++) {
      akt_Zustand->succ[i]->log_transProb = log(akt_Zustand->succ[i]->transProb);
    }
  }

  // logarithm of the transition probabilities (predecessor):  (without BEGIN) (really necessary?)
  for (unsigned int pos=1; pos<HMM_States.size(); pos++) {
    akt_Zustand = HMM_States[pos];
    for (unsigned int k=0; k<akt_Zustand->pred.size(); k++)
      akt_Zustand->pred[k]->log_transProb = log(akt_Zustand->pred[k]->transProb);
  }
  
  // to each state its position in the list of all states (HMM_States[]) is assigned to
  for (unsigned int pos=0; pos<HMM_States.size(); pos++) 
    HMM_States[pos]->pos_In_HMM_States = pos;

  evaluationOfHMM(al, numSubtypes, querySequence);

  for (int i=0; i<anz_SubProfile; i++)
    delete [] gaps[i];
  delete [] gaps;
  
  for (int i=0; i<anz_SubProfile; i++) { 
    for (int j=0; j<numSpalten; j++) 
      delete [] allStates[i][j];
    delete [] allStates[i];
  }
  delete [] allStates;
  
  for (unsigned int i=0; i<HMM_States.size(); i++) {
    HMM_States[i]->succ.clear();
    HMM_States[i]->pred.clear();
  }
  HMM_States.clear();
}


//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// evaluationOfHMM(Alignment*, ifstream&): 
//
//    computes the viterbi and the forward table
//
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
void HMM::evaluationOfHMM(Alignment* al,
			  int        numSubtypes,
			  ifstream&  querySequence) {
  
  ofstream VITERBI_GFF(VITERBI_PATH_GFF, ios::out);
  stringstream ss;
  ss << "recombination_" << appl_id << ".txt";
  ofstream RECOMB(ss.str().c_str(), ios::out);
  ofstream RECOMB_GAPPED(RECOMBINATION_GAPPED, ios::out);
  ofstream RECOMB_WITHOUT_POS(RECOMBINATION_WITHOUT_POSITIONS, ios::out);

  vector<int> viterbi_Path_aktSubtype;
  vector<int> viterbi_Path_startPosition;
  vector<int> viterbi_Path_endPosition;
  
  mytype       beam_width       = (mytype)BEAM_WIDTH;
  unsigned int posInQuery       = 0;
  int          numGaps          = 0;              // counts the number of gaps that are located in front a certain position in the query sequence (-> required for the correct recombination output)
  vector<int>  list_of_gaps;                      // retains for each position in the query the number of preceding gaps (=numGaps)
  int          numGaps_Gesamt   = 0;              // number of gaps in a sequence

  vector<char> name;
  char        *eingeleseneZeile = new char[MAX_ROW_LEN];
  int          seqsCounter      = 0;
  int          querySpalte      = 0;
  float        viterbi_Max      = -INF;
  list<string>           predictedSubtypes;
  list<string>::iterator subtypeIterator;
  list< list<ActiveStates> >           listOfLists;
  list< list<ActiveStates> >::iterator mainListIterator;
  list<ActiveStates>           allActiveStates;
  list<ActiveStates>::iterator listIterator;
  list<int>           possibleActiveStates;      // Indices of possible active states for a certain position in the query sequence
  list<int>::iterator posActiveStates;
  State* akt_Zustand;
  State* akt_Nachfolger;
  mytype viterbi_Score = 0;
  mytype forward_Score = 0;
  vector<State*> viterbi_Path;
  State* akt_bestPred;
  State* akt_Subtype;
  int    anz_SubProfile = 0;
  if (profile_mode == nocommon)
    anz_SubProfile = numSubtypes;
  else
    anz_SubProfile = numSubtypes+1;

  /****************************************************************/
 
  // For the computation of the viterbi and the forward table, the following LOCAL arrays are needed.
  // They are only required for the transition from position i -> i+1 in the query to obtain the corresponding active states.
  // Initialization with 0; Only positions of states (=pos in HMM_States) which are possibly active for this query position obtain the value 1.
  // The final list of active states of a certain query position is saved in a smaller list and is later used for the viterbi backtracking.
  // These local arrays are overwritten/deleted for a new query position
  
  int **activatedStates = new int*[2];
  activatedStates[0] = new int[HMM_States.size()];       // list of potentially active states at pos i in query
  activatedStates[1] = new int[HMM_States.size()];       // list of potentially active states at pos i+1 in query
  int *dummy_Int = new int[HMM_States.size()];
  
  mytype **viterbi = new mytype*[2];
  viterbi[0] = new mytype[HMM_States.size()];    // Viterbi values of the potentially active states at pos i in query
  viterbi[1] = new mytype[HMM_States.size()];    // Viterbi values of the potentially active states at pos i+1 in query
  
  mytype **forward = new mytype*[2];
  forward[0] = new mytype[HMM_States.size()];    // forward values of the potentially active states at pos i in query
  forward[1] = new mytype[HMM_States.size()];    // forward values of the potentially active states at pos i+1 in query
  mytype *dummy_Double = new mytype[HMM_States.size()];
  
  int **posOfBestPred = new int*[2];
  posOfBestPred[0] = new int[HMM_States.size()];       // list of the best predecessors of the corresponding states in the list of the potentially active states at pos i in query
  posOfBestPred[1] = new int[HMM_States.size()];       // list of the best predecessors of the corresponding states in the list of the potentially active states at pos i+1 in query


  RECOMB << "# Output of the jpHMM. " << endl
         << "#" << endl
	 << "#   chosen parameters: " << endl
	 << "#" << "\t" << "jump probability: " << "\t" << jumpRowToRow << endl
         << "#" << endl
	 << "#   format: " << endl
	 << "#   sequence_name " << endl
	 << "#   start_position\t end_position\t predicted_subtype " << endl
	 << "#" << endl
	 << "# predicted recombination with breakpoints based on raw sequence positions " << endl
	 << "# of the following sequence(s): " << endl
	 << endl;
  RECOMB_GAPPED << "# Output of the jpHMM. " << endl
		<< "#" << endl
		<< "#   chosen parameters: " << endl
		<< "#" << "\t" << "jump probability: " << "\t" << jumpRowToRow << endl
		<< "#" << endl
		<< "#   format: " << endl
		<< "#   sequence_name " << endl
		<< "#   start_position\t end_position\t predicted_subtype " << endl
		<< "#" << endl
		<< "# predicted recombination with breakpoints based on sequence positions with gaps " << endl
		<< "# of the following sequence(s): " << endl
		<< endl;
  RECOMB_WITHOUT_POS << "# Output of the jpHMM. " << endl
		     << "#" << endl
		     << "#   format: " << endl
		     << "#   sequence_name\t chosen_parameters\t predicted_subtypes" << endl 
		     << "#" << endl		
		     << "#   (bw=beam-width, jp=jump probability)" << endl
		     << "#" << endl
		     << "# predicted recombination of the following sequence(s): " << endl
		     << endl;

     

  /****************************************************************/
  
  while (querySequence.getline(eingeleseneZeile, MAX_ROW_LEN)) {
    list_of_gaps.clear();
    numGaps_Gesamt = 0;
    predictedSubtypes.clear();
    if (eingeleseneZeile[0] == '>') {
      name.clear();
      querySpalte = 0; 
      while (eingeleseneZeile[querySpalte] && (eingeleseneZeile[querySpalte] != '\x0D') && (eingeleseneZeile[querySpalte] != '\x0A')) {
	if (eingeleseneZeile[querySpalte] != '>') {  // nur der Name, nicht im FASTA-Format
	  name.push_back(eingeleseneZeile[querySpalte]);
	}
	querySpalte++;
      }
    }
    else {
      // counting the query sequences read in
      seqsCounter++;
      if (name.size()==0) {
	RECOMB << "Sequence_" << seqsCounter << endl;
	RECOMB_WITHOUT_POS << "Sequence_" << seqsCounter << endl;
      }
      else {
	for (unsigned int i=0; i<name.size(); i++) {
	  if (name[i] != '>') {
	    RECOMB << name[i];
	    RECOMB_WITHOUT_POS << name[i];
	  }
	}
	RECOMB << endl;
      }
      
      querySpalte = 0;
      while (eingeleseneZeile[querySpalte] && (eingeleseneZeile[querySpalte] != '\x0D') && (eingeleseneZeile[querySpalte] != '\x0A')) {
	if (eingeleseneZeile[querySpalte] == '-' || eingeleseneZeile[querySpalte] == '*') { // Luecke in query
	  numGaps++;
	}
	else { // gaps are not copied in querySeq but saved separately -> for the recombination output
	  querySeq.push_back(eingeleseneZeile[querySpalte]);
	  list_of_gaps.push_back(numGaps);   // Beginn bei 0. d.h. falls am Anfang der query keine Luecke, dann ist list_of_gaps[0] = 0 (=Anzahl Luecken am Anfang); beachte dies bei Ausgabe
                                             // = Anzahl der Position 0 in der query vorhergehender Luecken
	  //numGaps = 0;                     erstmal so, d.h. zaehlen Luecken insgesamt, mal sehen ob das passt
	}
	querySpalte++;
      }
      
      // adapt the beam-width to the length of the query sequence for the forward and Viterbi algorithm 
      // (a beam-width of 1e-20 is too large to allow the alignment of a short sequence to the end of the given alignment (remember that the logarithm of the beam-width is used)
      if (querySeq.size() <= 1000)
	beam_width = 1e-45;
      else if (1000 < querySeq.size() && querySeq.size() <= 1500)
	beam_width = 1e-40;
      else if (1500 < querySeq.size() && querySeq.size() <= 2500)
	beam_width = 1e-35;
      else if (2500 < querySeq.size() && querySeq.size() <= 4000)
	beam_width = 1e-30;
      else if (4000 < querySeq.size() && querySeq.size() <= 7000)
	beam_width = 1e-25;
      else
	beam_width = BEAM_WIDTH;

      RECOMB_WITHOUT_POS << "\t" << "bw=" << beam_width << ", jp=" << jumpRowToRow << "\t";
      
      // Computation of the Viterbi and the forward table:
      // Initialization
      for (int i=0; i<2; i++) {
	for (unsigned int j=0; j<HMM_States.size(); j++) {
	  activatedStates[i][j] = 0;      // only =1, if the corresponding state is active
	  viterbi[i][j]         = -INF;  // since the Viterbi values < 0, and need to find the maximum for the best predecessor
	  forward[i][j]         = -INF;     
	  posOfBestPred[i][j]   = -1;    // since 0 is the position for the BEGIN state
	}
      }
      
      // Determination of the active list for each position in the query sequence and of the Viterbi & forward values
      for (posInQuery=0; posInQuery<querySeq.size()+1; posInQuery++) { 
	viterbi_Max = -INF;
	if (posInQuery == 0) { // BEGIN
	  activatedStates[1][HMM_States[0]->pos_In_HMM_States] = 1;
	  viterbi[1][HMM_States[0]->pos_In_HMM_States]         = 0;
	  forward[1][HMM_States[0]->pos_In_HMM_States]         = 0;
	  possibleActiveStates.push_back(HMM_States[0]->pos_In_HMM_States);
	  
	  // activation of DELETEs directly following to BEGIN and computation of the viterbi values (=> deletion at the beginning of the sequence possible)
	  for (unsigned k=0; k<HMM_States[0]->succ.size(); k++) {
	    if (HMM_States[0]->succ[k]->whereTo->stateType == Delete || HMM_States[0]->succ[k]->whereTo->stateType == N_Delete) {
	      activatedStates[1][HMM_States[0]->succ[k]->whereTo->pos_In_HMM_States] = 1;
	      viterbiAndForward(HMM_States[0]->succ[k]->whereTo->pos_In_HMM_States, 0, activatedStates, posOfBestPred, viterbi, forward); 
	      possibleActiveStates.push_back(HMM_States[0]->succ[k]->whereTo->pos_In_HMM_States);   // adding the index of the active state to the list of potentially active states
	    }
	  }
	  viterbi_Max = viterbi[1][HMM_States[0]->pos_In_HMM_States];  // = viterbi value of BEGIN
	}
	else {
	  // at position posInQuery=0: liste for BEGIN => only DELETEs and N_Delete activ
	  // new active list:  potential(!) successors of the anterior active list
	  
	  // Reihenfolge ist fuer Rueckwaertsalgo wichtig, hier nicht, also vielleicht wieder loeschen
	  for (listIterator = (*mainListIterator).begin(); listIterator != (*mainListIterator).end(); ++listIterator) {
	    akt_Zustand = HMM_States[(*listIterator).posState];
	    for (unsigned int i=0; i<akt_Zustand->succ.size(); i++) {
	      akt_Nachfolger = akt_Zustand->succ[i]->whereTo;
	      // first: add all INSERTs to the list, then MATCHes, wegen Berechnung im Rueckwaertsalgo
	      if (akt_Nachfolger->stateType == Insert || akt_Nachfolger->stateType == N_Insert || akt_Nachfolger->stateType == C_Insert) {
		// DELETEs: in anterior list, DELETEs of this list are added after adding MATCHes and INSERTs
		// END can only be reached after the emission of the whole query sequence
		if (activatedStates[1][akt_Nachfolger->pos_In_HMM_States] == 0) {
		  activatedStates[1][akt_Nachfolger->pos_In_HMM_States] = 1;
		  viterbiAndForward(akt_Nachfolger->pos_In_HMM_States, posInQuery, activatedStates, posOfBestPred, viterbi, forward);
		  possibleActiveStates.push_back(akt_Nachfolger->pos_In_HMM_States);
		  if (viterbi[1][akt_Nachfolger->pos_In_HMM_States] > viterbi_Max)
		    viterbi_Max = viterbi[1][akt_Nachfolger->pos_In_HMM_States];
		}
	      }
	    }
	  }
	  for (listIterator = (*mainListIterator).begin(); listIterator != (*mainListIterator).end(); ++listIterator) {
	    akt_Zustand = HMM_States[(*listIterator).posState];
	    for (unsigned int i=0; i<akt_Zustand->succ.size(); i++) {
	      akt_Nachfolger = akt_Zustand->succ[i]->whereTo;
	      // jetzt alle Matches in die Liste, wegen Berechnung im Rueckwaertsalgo
	      if (akt_Nachfolger->stateType == Match) {
		// DELETEs: in anterior list, DELETEs of this list are added after adding MATCHes and INSERTs
		// END can only be reached after the emission of the whole query sequence
		if (activatedStates[1][akt_Nachfolger->pos_In_HMM_States] == 0) {
		  activatedStates[1][akt_Nachfolger->pos_In_HMM_States] = 1;
		  viterbiAndForward(akt_Nachfolger->pos_In_HMM_States, posInQuery, activatedStates, posOfBestPred, viterbi, forward);
		  possibleActiveStates.push_back(akt_Nachfolger->pos_In_HMM_States);
		  if (viterbi[1][akt_Nachfolger->pos_In_HMM_States] > viterbi_Max)
		    viterbi_Max = viterbi[1][akt_Nachfolger->pos_In_HMM_States];
		}
	      }
	    }
	  }
	}	
	// N_INSERT and all INSERTs mustn't be active at the end since no path to the END state without emitting further symbols exists
	// N_Insert und C_Insert muessen immer aktiv bleiben N_Delete fuer posInQuery=0, aber eh nur dort moeglich, da (Delete-)Nachfolger von Begin
	if (posInQuery != querySeq.size()) {
	  // BEAM algorithm on the same list (for all query positions except the last one. At the last position all states and their successional DELETEs are active to guarantee at least one path from BEGIN to END
	  for (posActiveStates = possibleActiveStates.begin(); posActiveStates != possibleActiveStates.end(); ++posActiveStates) {
	    if ( HMM_States[(*posActiveStates)]->stateType != N_Insert &&  HMM_States[(*posActiveStates)]->stateType != C_Insert && HMM_States[(*posActiveStates)]->stateType != N_Delete) {
	      if (viterbi[1][(*posActiveStates)] < (log(beam_width)+viterbi_Max)){
		activatedStates[1][(*posActiveStates)] = 0;
		viterbi[1][(*posActiveStates)]         = -INF;
		forward[1][(*posActiveStates)]         = -INF;
		posOfBestPred[1][(*posActiveStates)]   = -1;     
		possibleActiveStates.erase(posActiveStates);   // clears element at position pos and returns position of the next element (=> ++pos)
		--posActiveStates;
	      }
	    }
	  }
	}
	else {  // at the end all successors of the anterior active lsit are active, without using the BEAM algorithm
	  for (posActiveStates = possibleActiveStates.begin(); posActiveStates != possibleActiveStates.end(); ++posActiveStates) {
	    if ((HMM_States[(*posActiveStates)]->stateType == N_Insert) || (HMM_States[(*posActiveStates)]->stateType == Insert)) {
	      activatedStates[1][(*posActiveStates)] = 0;
	      viterbi[1][(*posActiveStates)]         = -INF;
	      forward[1][(*posActiveStates)]         = -INF;
	      posOfBestPred[1][(*posActiveStates)]   = -1;     
	      possibleActiveStates.erase(posActiveStates);   // clears element at position pos and returns position of the next element (=> ++pos)
	      --posActiveStates;
	    }
	  }
	}
	
	// successional DELETEs:
	// by adding more DELETEs viterbiMax remains unmodified (since the probability for these additonal transitions is <0)
	for (posActiveStates = possibleActiveStates.begin(); posActiveStates != possibleActiveStates.end(); ++posActiveStates) {
	  akt_Zustand = HMM_States[(*posActiveStates)];
	  if (akt_Zustand->stateType != Begin) {
	    // already considered for BEGIN
	    deleteActivationRecursion(akt_Zustand, posInQuery, viterbi_Max, activatedStates, viterbi, forward, posOfBestPred, possibleActiveStates, beam_width);
	  }
	}
	
	if (posInQuery == querySeq.size()) { // C_DELETE is added at the end of the list; this state can only be active at the end of the query sequence
	  activatedStates[1][HMM_States.size()-2] = 1;
	  viterbiAndForward(HMM_States.size()-2, posInQuery, activatedStates, posOfBestPred, viterbi, forward);
	  possibleActiveStates.push_back(HMM_States.size()-2);
	}

	// transfer of the real active states to the list allActiveStates  // doppelt??? gleich in (*mainListIterator?)
	for(posActiveStates = possibleActiveStates.begin(); posActiveStates != possibleActiveStates.end(); ++posActiveStates) {
	  if (activatedStates[1][(*posActiveStates)] != 0) {
	    // check again in order to avoid that the list contains some states twicenochmal
	    allActiveStates.push_back(ActiveStates((*posActiveStates), posOfBestPred[1][(*posActiveStates)], forward[1][(*posActiveStates)]));
	  }
	}
	
	if (posInQuery != 0) {
	  // fuer Begin gab es noch keine vorherige Position
	  // Clear array of the active states at the previous position in the query sequence
	  for (listIterator = (*mainListIterator).begin(); listIterator != (*mainListIterator).end(); ++listIterator) {
	    activatedStates[0][(*listIterator).posState] = 0;
	    viterbi[0][(*listIterator).posState]         = -INF;
	    forward[0][(*listIterator).posState]         = -INF;
	    posOfBestPred[0][(*listIterator).posState]   = -1;
	  }
	}

	dummy_Int = activatedStates[0];  activatedStates[0] = activatedStates[1];  activatedStates[1] = dummy_Int;
	dummy_Int = posOfBestPred[0];    posOfBestPred[0] = posOfBestPred[1];      posOfBestPred[1] = dummy_Int;
	dummy_Double = viterbi[0];       viterbi[0] = viterbi[1];                  viterbi[1] = dummy_Double;
	dummy_Double = forward[0];       forward[0] = forward[1];                  forward[1] = dummy_Double;
		
	// Clear the 'local' list of the potentially active states (=possibleActiveStates)
	possibleActiveStates.clear();
	// transfer of the actual active list to listOfLists
	listOfLists.push_back(allActiveStates);
	mainListIterator = --listOfLists.end(); // set iterator at the end
	allActiveStates.clear();
      }
      // END:
      posInQuery                              = querySeq.size()+1;
      activatedStates[1][HMM_States.size()-1] = 1;
      viterbiAndForward(HMM_States.size()-1, posInQuery, activatedStates, posOfBestPred, viterbi, forward);
      allActiveStates.push_back( ActiveStates(HMM_States.size()-1, posOfBestPred[1][HMM_States.size()-1], forward[1][HMM_States.size()-1]));
      listOfLists.push_back(allActiveStates);
      allActiveStates.clear();
      
      // Scores: 
      viterbi_Score = viterbi[1][HMM_States.size()-1];
      forward_Score = forward[1][HMM_States.size()-1]; 
      

      /***************************************************************\
                       Viterbi algorithm: backtracking:
      \***************************************************************/
      
      posInQuery   = querySeq.size()+1;
      akt_Zustand  = HMM_States.back();
      akt_bestPred = akt_Zustand;
      akt_Subtype  = akt_bestPred;
      
      viterbi_Path.push_back(HMM_States.back());
      viterbi_Path_endPosition.push_back(posInQuery); 
      viterbi_Path_aktSubtype.push_back(akt_Zustand->row);
      
      mainListIterator = --listOfLists.end();
      
      while ((mainListIterator != --listOfLists.begin()) && (akt_Zustand->stateType != Begin) ) {
	if (akt_Zustand->stateType == Delete || akt_Zustand->stateType == N_Delete || akt_Zustand->stateType == C_Delete) {
	  for (listIterator = (*mainListIterator).begin(); listIterator != (*mainListIterator).end(); ++listIterator) {
	    if ((*listIterator).posState == akt_Zustand->pos_In_HMM_States) {
	      akt_bestPred = HMM_States[(*listIterator).posBestPred];
	      break;
	    }
	  }
	}
	else {
	  for (listIterator = (*mainListIterator).begin(); listIterator != (*mainListIterator).end(); ++listIterator) {
	    if ((*listIterator).posState == akt_Zustand->pos_In_HMM_States) {
	      if ((*listIterator).posBestPred != -1) {
		// to guarantee that the program stops, even if for a certain sequence the beam-width is chosen too small and quasi END doesn't possess an active predecessor
		akt_bestPred = HMM_States[(*listIterator).posBestPred];
	      }
	      break;
	    }
	  }
	  --mainListIterator;
	  --posInQuery;
	}
	akt_Zustand = akt_bestPred;
	if (akt_Zustand->row != akt_Subtype->row) {
	  akt_Subtype = akt_Zustand;
	  viterbi_Path_startPosition.push_back(posInQuery+1);
	  viterbi_Path_endPosition.push_back(posInQuery); 
	  viterbi_Path_aktSubtype.push_back(akt_Subtype->row);
	  
	  if (akt_Subtype->stateType != Begin) {
	    if (akt_Subtype->stateType == N_Insert) { 
	    }
	    else if (akt_Subtype->stateType == C_Insert) {
	    }
	    else if (akt_Subtype->stateType == N_Delete) {
	    } 
	    else if (akt_Subtype->stateType == C_Delete) {
	    } 
	    else { // subtype
	      // Adding all predicted subtypes (!= insertion) to predictedSubtypes
	      predictedSubtypes.push_back(al->names[akt_Subtype->row]);
	    }
	  }
	}
	viterbi_Path.insert(viterbi_Path.begin(), akt_bestPred);
      }
      viterbi_Path_startPosition.push_back(0);
      
      predictedSubtypes.sort();
      predictedSubtypes.unique();  //=> each subtype may only appear once in this list
      
      // Output of the predicted subtypes in RECOMBINATION_WITHOUT_POSITIONS
      // Output of the viterbi path in GFF format in VITERBI_GFF
      // Output of the recombination based on raw sequence positions in RECOMB
      // Output of the recombination based on sequence positions with gaps in RECOMB_GAPPED


      for (subtypeIterator = predictedSubtypes.begin(); subtypeIterator != predictedSubtypes.end(); subtypeIterator++) 
	RECOMB_WITHOUT_POS << (*subtypeIterator);
      RECOMB_WITHOUT_POS << endl;

      VITERBI_GFF << "# Output of the jpHMM. " << endl
                  << "#" << endl
                  << "# predicted recombination of sequence " << seqsCounter;

      if (name.size() != 0) {
        VITERBI_GFF << " ( name = ";
        for (unsigned int i=0; i<name.size(); i++){
          if (name[i] != '>')
            VITERBI_GFF << name[i];
        }
        VITERBI_GFF << ", length = " << querySeq.size() << " )";
      }
      VITERBI_GFF << endl
                  << "#" << endl
                  << "# predicted subtypes for sequence '";
      if (name.size() == 0) 
        VITERBI_GFF << seqsCounter << "': " << endl;
      else {
        for (unsigned int i=0; i<name.size(); i++){
          if (name[i] != '>')
            VITERBI_GFF << name[i];
        }
        VITERBI_GFF << "': " << endl;
      }

      // total number of gaps in the query sequence
      for (unsigned int g=0; g < list_of_gaps.size(); g++)
	numGaps_Gesamt += list_of_gaps[g];
      // falls Luecken in der Sequenz wird zusaetzlich die Rekombination dieser Sequenz basierend auf den Seq-Positionen mit Gaps in die Datei RECOMB_GAPPED geschrieben
      // hier zunaechst der Name der Sequenz mit Luecken
      if (numGaps_Gesamt != 0) {
	if (name.size()==0) 
	  RECOMB_GAPPED << "Sequence_" << seqsCounter << endl;
	else {
	  for (unsigned int i=0; i<name.size(); i++) {
	    if (name[i] != '>')
	      RECOMB_GAPPED << name[i];
	  }
	  RECOMB_GAPPED << endl;
	}
      }
      for (unsigned int v = viterbi_Path_aktSubtype.size()-1; v>0; v--) {
	// ending at 1, since 0 is the END state
	if ((viterbi_Path_aktSubtype[v] != HMM_States[0]->row) && (viterbi_Path_aktSubtype[v] != HMM_States[1]->row) && (viterbi_Path_aktSubtype[v] != HMM_States[HMM_States.size()-2]->row) && (viterbi_Path_aktSubtype[v] != HMM_States[HMM_States.size()-1]->row)) { 
	  // != BEGIN, N_DELETE, C_DELETE, END
	  if (name.size() == 0)
	    VITERBI_GFF << "Sequence_" << seqsCounter << "\t";
	  else {
	    for (unsigned int i=0; i<name.size(); i++){
	      if (name[i] != '>')
		VITERBI_GFF << name[i];
	    }
	    VITERBI_GFF << "\t";
	  }
	  VITERBI_GFF << "jpHMM\t";
	  
	  // Output of the breakpoints based on raw sequence positions
	  RECOMB << viterbi_Path_startPosition[v] << "\t" << viterbi_Path_endPosition[v] << "\t";
	
	  // if query sequence contains gaps: Output of the breakpoints based on sequence positions WITH gaps
	  if (numGaps_Gesamt != 0) {
	    // Anzahl vorhergehender Luecken werden dazu gerechnet
	    // -1, da in viterbi_Path_startPosition die 'realen' Positionen gespeichert sind, d.h. beginnend bei 1 statt bei 0;
	    RECOMB_GAPPED << viterbi_Path_startPosition[v]+ list_of_gaps[viterbi_Path_startPosition[v]-1] << "\t" << viterbi_Path_endPosition[v]+ list_of_gaps[viterbi_Path_endPosition[v]-1] << "\t";
	  }
	  if (viterbi_Path_aktSubtype[v] == -1) { // N_INSERT 
	    VITERBI_GFF << "5'-Insertion\t";
	    RECOMB << "5'-Insertion" << endl;
	    if (numGaps_Gesamt != 0) 
	      RECOMB_GAPPED << "5'-Insertion" << endl;
	  }
	  else if (viterbi_Path_aktSubtype[v] == anz_SubProfile+1) { // C_INSERT
	    VITERBI_GFF << "3'-Insertion\t";
	    RECOMB << "3'-Insertion" << endl;
	    if (numGaps_Gesamt != 0) 
	      RECOMB_GAPPED << "3'-Insertion" << endl;
	  }
	  else {
	    for (unsigned int i=0; i<al->names[viterbi_Path_aktSubtype[v]].size(); i++) {
	      if (al->names[viterbi_Path_aktSubtype[v]][i] != '>'){
		VITERBI_GFF << al->names[viterbi_Path_aktSubtype[v]][i];
		RECOMB      << al->names[viterbi_Path_aktSubtype[v]][i];
		if (numGaps_Gesamt != 0) {
		  RECOMB_GAPPED << al->names[viterbi_Path_aktSubtype[v]][i];
		}
	      }
	    }
	    RECOMB      << endl;
	    VITERBI_GFF << "\t";
	    if (numGaps_Gesamt != 0) 
	      RECOMB_GAPPED << endl;
	  }
	  // Output of the breakpoints in gff format based on raw sequence positions 
	  VITERBI_GFF << viterbi_Path_startPosition[v] << "\t" << viterbi_Path_endPosition[v] << "\t.\t+\t." << endl;
	}
      } 
      VITERBI_GFF << endl;
      RECOMB << endl;
      if (numGaps_Gesamt != 0) 
	RECOMB_GAPPED << endl;
      
      viterbi_Path_aktSubtype.clear();
      viterbi_Path_startPosition.clear();
      viterbi_Path_endPosition.clear();

      querySeq.clear();
      viterbi_Path.clear();
      listOfLists.clear();
    }
  }
  
  for (int i=0; i<2; i++) {
    delete [] activatedStates[i];
    delete [] viterbi[i];
    //delete [] forward[i];
    delete [] posOfBestPred[i];
  }
  delete [] activatedStates;
  delete [] viterbi;
  //delete [] forward;
  delete [] posOfBestPred;
  delete [] eingeleseneZeile;
  //delete [] dummy_Int;
  delete [] dummy_Double;

}
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// deleteActivationRecursion(State*, unsigned int, mytype, int**&, mytype**&, mytype**&, int**&, list<int>&, mytype):
//
//    considers successional DELETEs of an actual state:
//    these DELETEs are, if they are active, in the same list as their predecessors, since the position in the query sequence doesn't change
//    DELETE is only added to the list if it satisfies the beam condition!
//    
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------

void HMM::deleteActivationRecursion(State*       akt_Zustand, 
				    unsigned int queryPos, 
				    mytype       viterbi_Max, 
				    int**       &activatedStates, 
				    mytype**    &viterbi, 
				    mytype**    &forward,
				    int**       &posOfBestPred, 
				    list<int>   &possibleActiveStates,
				    mytype       beam_width) {
  // Aendern, damit immer angewendet werden kann
  // nicht dieser spezielle Fall, sondern mit bestPred, also Viterbi
  // nur fuer "normale" Deletes, C_Delete oben extra betrachtet 
  
  mytype      viterbi_Value = 0;
  Transition* akt_Uebergang;
  
  for (unsigned int k=0; k<akt_Zustand->succ.size(); k++) {
    akt_Uebergang = akt_Zustand->succ[k];
    // only transitions to DELETE are added. MATCHes will be add to the subsequent active list, this is done in evaluationOfHMM
    if (akt_Uebergang->whereTo->stateType == Delete) {
      viterbiAndForward(akt_Uebergang->whereTo->pos_In_HMM_States, queryPos, activatedStates, posOfBestPred, viterbi, forward);  
      viterbi_Value = viterbi[1][akt_Uebergang->whereTo->pos_In_HMM_States];
      if ((viterbi_Value >= (log(beam_width)+viterbi_Max)) || (queryPos == querySeq.size())) {
	// for the last position in the query all DELETEs are activated/added to the list, i.e. no beam-width is taken into account
      	if (activatedStates[1][akt_Uebergang->whereTo->pos_In_HMM_States ] == 0) {
	  // Adding the index of the active state to the list of potentially active states
	  activatedStates[1][akt_Uebergang->whereTo->pos_In_HMM_States ] = 1;
	  possibleActiveStates.push_back(akt_Zustand->succ[k]->whereTo->pos_In_HMM_States);   
	  deleteActivationRecursion(akt_Zustand->succ[k]->whereTo, queryPos, viterbi_Max, activatedStates, viterbi, forward, posOfBestPred, possibleActiveStates, beam_width);
	}
      }
      else {
	// erase from the list (wirklich noetig?)
	activatedStates[1][akt_Uebergang->whereTo->pos_In_HMM_States] = 0;
	viterbi[1][akt_Uebergang->whereTo->pos_In_HMM_States]         = -INF;
	forward[1][akt_Uebergang->whereTo->pos_In_HMM_States]         = -INF;
	posOfBestPred[1][akt_Uebergang->whereTo->pos_In_HMM_States]   = -1;
      }
    }
  }
}
 
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
// viterbiAndForward(int, unsigned int, int**&, int**&, mytype**&, mytype**&): 
//
//    computes the viterbi and forward variables for a state (at position posOfStates in the list HMM_States of all states) successively
//     (Attention: is a certain transition possible, i.e. is the predecessor really active for the anterior position?)
//
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------
void HMM::viterbiAndForward(int       posOfState, 
			    unsigned int       queryPos, 
			    int**    &activatedStates, 
			    int**    &posOfBestPred,
			    mytype** &viterbi, 
			    mytype** &forward) {
  
  State      *akt_Zustand    = HMM_States[posOfState];
  Transition *akt_Uebergang;
  State      *akt_Vorgaenger;
  mytype      max            = -INF;    // log(0);
  mytype      log_sum        = -INF;   
  mytype      log_p          = -INF;   
  mytype      log_f          = -INF;

  for (unsigned int i=0; i<akt_Zustand->pred.size(); i++) {
    akt_Uebergang  = akt_Zustand->pred[i];
    akt_Vorgaenger = akt_Uebergang->whereFrom;
    if (akt_Zustand->stateType == Delete || akt_Zustand->stateType == N_Delete || akt_Zustand->stateType == C_Delete) {
      // we have to consider the list for THIS query position, thus list[1]
      if (activatedStates[1][akt_Vorgaenger->pos_In_HMM_States] == 1) {
	log_p = viterbi[1][akt_Vorgaenger->pos_In_HMM_States] + akt_Uebergang->log_transProb;
	log_f = forward[1][akt_Vorgaenger->pos_In_HMM_States] + akt_Uebergang->log_transProb;
	// The log transformation:
	if (log_sum > log_f) 
	  log_sum = log_sum + log(1 + exp( log_f-log_sum));
	else
	  log_sum = log_f + log(1 + exp( log_sum-log_f));
      }   
    }
    else if (akt_Zustand->stateType == End) {
      if (activatedStates[0][akt_Vorgaenger->pos_In_HMM_States] == 1) {
	log_p = (viterbi[0][akt_Vorgaenger->pos_In_HMM_States])+(akt_Uebergang->log_transProb);
	log_f = (forward[0][akt_Vorgaenger->pos_In_HMM_States])+(akt_Uebergang->log_transProb);
	if (log_sum > log_f) 
	  log_sum = log_sum + log(1 + exp( log_f-log_sum));
	else
	  log_sum = log_f + log(1 + exp( log_sum-log_f));
      }
    }
    else { //if (akt_Zustand->stateType == Match || akt_Zustand->stateType == Insert)
      if (activatedStates[0][akt_Vorgaenger->pos_In_HMM_States] == 1) {
	log_p = (viterbi[0][akt_Vorgaenger->pos_In_HMM_States])+(akt_Uebergang->log_transProb)+(akt_Zustand->log_emProb[Const_Def::indexOfAlphabet[ (int) querySeq[queryPos-1]]]);
	log_f = (forward[0][akt_Vorgaenger->pos_In_HMM_States])+(akt_Uebergang->log_transProb)+(akt_Zustand->log_emProb[Const_Def::indexOfAlphabet[ (int) querySeq[queryPos-1]]]);
	if (log_sum > log_f) 
	  log_sum = log_sum + log(1 + exp( log_f-log_sum));
	else
	  log_sum = log_f + log(1 + exp( log_sum-log_f));
      }
    }

    if (log_p > max) {
      max                          = log_p;
      posOfBestPred[1][posOfState] = akt_Vorgaenger->pos_In_HMM_States;
    }
  }
  viterbi[1][akt_Zustand->pos_In_HMM_States] = max;
  forward[1][akt_Zustand->pos_In_HMM_States] = log_sum;
}


/******************************************************\

 *                 class ActiveStates                 *

\******************************************************/

ActiveStates::ActiveStates(int pos, int bestPred, mytype forward_Value){
  posState     = pos;
  posBestPred  = bestPred;
  forwardValue = forward_Value;
}
