/**************************************************************
 filename	: jpHMM.cc
 licence        : GNU Licence, see file LICENCE
 description    : this is the main function for the jphmm program
                  it reads in the command line, assigns the given parameter to the corresponding variables
                  and calls the classes Alignment (reads in the alignment and determines the profile and the consensus columns of the subtypes)
                  and HMM (generates the jphmm and computes the viterbi path which defines the recombination for a given query sequence (read in in class HMM))
 author		: Anne-Kathrin Schultz, anne@gobics.de
 date		: 03/04 - 08/05

Content: main function!

  |   date    |    author            |  changes
 -|-----------|----------------------|-------------------------------------------
  | April 06  | Anne-Kathrin Schultz | add. output for sequences containing gaps
  | 06.10.06  | Anne-Kathrin Schultz | now also accepts alignments with only 1 subtype

***************************************************************/

#include "alignment.hh"
#include "profileHMM.hh"
#include "hmm.hh"
#include "profileHMM.hh"
#include "const_def.hh"

#include <iostream>
#include <fstream>
#include <string>
#include <cstring>
#include <cstdlib>
#include <cstdio>

#include <getopt.h>
#include <dirent.h>


// auxiliary function: usage
//
// short output of an instruction for use
void usage (char *prog, 
	    string errstr) {
    if (errstr.length())
	cerr << errstr << endl << endl;
    cerr << "usage: " << prog 
	 << "\n\t[ -v  type of the given sequences (HIV or HCV) (in capital letters) ] "
	 << "\n\t[ -i  file containing the given multiple sequence alignment ] "
	 << "\n\t[ -s  file containing the query sequence(s) ] "
      // << "\n\t[ -a  file containing the priors for the estimation of the emission (and transition) probabilities ] "
      // << "\n\t[ -N  (opt.) if -N: no profile HMM of the whole alignment, else: profile HMM of the alignment (= default) ] "
      // << "\n\t[ -e  (opt.) 1-epsilon = transition probability from N_INSERT to N_INSERT and from C_INSERT to C_INSERT    (default: 0.01) ] "
      // << "\n\t[ -x  probability of a jump from the 'common profile mode' to the 'subtype mode' (only required if -N is not chosen) ] "
      // << "\n\t[ -y  probability of a jump from the 'subtype mode' to the 'common profile mode' (only required if -N is not chosen) ] "
	 << "\n\t[ -j  (opt.) probability of a jump from one subtype to any other subtype (default: 1e-9) ]\n";
      // << "\n\t[ -c  (opt.) probability of a jump from MATCH to C_DELETE taking into account the corresponding consensus column of the MATCH    (default: 0.01) ] "
      // << "\n\t[ -d  (opt.) transition probability from N_DELETE to N_DELETE and from C_DELETE to C_DELETE    (default: 0.99) ] "
      // << "\n\t[ -p  (opt.) consensus column criteria: required percentage of non-gaps in a column (0.0 - 1.0)   OR! -t    (default: 0.5) ] "
      // << "\n\t[ -t  (opt.) consensus column criteria: required number of non-gaps in a column                   OR! -p    (default: 5)   ] "
      // << "\n\t[ -T  (opt.) if -T: first/last consensus column of all subtypes are equal, else: individual (=default) ] "
      // << "\n\t[ -B  (opt.) beam-width for the beam algorithm    (default: 1e-20) ] "
      // << "\n\t[ -D  (opt.) alphabetic mode = dna:     nucleic sequences considered   OR! -P ] "
      // << "\n\t[ -P  (opt.) alphabetic mode = protein: protein sequences considered   OR! -D    (default: protein) ]\n";
    exit(1);
}

long appl_id;
float jumpProfileToRow;         // -x
float jumpRowToProfile;         // -y 
float jumpRowToRow;             // -j
mytype epsilon;                 // -e
mytype d_init;                  // -c
mytype d_ext;                   // -d
float consensus_Percentage;     // -p
int   consensus_Threshold;      // -t 
float BEAM_WIDTH;               // -B
Alphabetic_Mode alphabet;       // -D OR -P
Profile_Mode    profile_mode;   // -N
Consensus_Mode  consensus;      // -T
//Virus_Type virus_type;        // -v    


int main(int argc, char *argv[]) {

  if (argc == 1)
    usage(argv[0], "no arguments given");

  // Flags fuer die einzelnen Optionen
  //bool has_x = false, has_y = false;
  bool has_j = false;// has_e = false;
  //bool has_c = false, has_d = false;
  bool has_i = false; 
  bool has_s = false;
  //bool has_a = false, 
  bool has_v = false; 
  //bool has_p = false, has_t = false;
  //bool has_T = false;
  //bool has_B = false, has_N = false;
  //bool has_D = false, has_P = false;

  alphabet             = dna;                // Default   : dna sequences considered
  profile_mode         = nocommon;           // Default   : no profile HMM of the whole alignment considered
  consensus_Percentage = 0.5;                // Default   : required percentage of non-gaps in a column 
  consensus            = commonConsensus;    // Default   : common first/last consensus column for each subtype
  consensus_Threshold  = 5;                  // Default   : required number of non-gaps in a column  (if f.ex. =1000 => no effect)
  BEAM_WIDTH           = 1e-20;              // Default   : beam_width
  epsilon              = 0.01;               // Default   : 1-epsilon = transProb(N_Insert->N_Insert) = TransProb(C_Insert->C_Insert)
  d_init               = 0.01;               // Default   : d_init = jumpProb(MATCH->C_DELETE) taking into account the corresponding consensus column of the MATCH
  d_ext                = 0.99;               // Default   : d_ext = transProb(N_DELETE->N_DELETE) = TransProb(C_DELETE->C_DELETE)    
  jumpRowToRow         = 1e-9;               // Default   : probability of a jump from one subtype to any other subtype

  
  // Variables for the option parameters
  string infile, querySeq_file, emissionPriors_file, virus_type; // recomb_out;
  int argument;

  // die einzelnen Argumente werden auf 0 (false) gesetzt. Falls sie dann vorhanden sind, auf 1 
  // Falls bei weiterem Durchlaufen der Parameter dieser Parameter nochmal auftaucht: Fehlermeldung, denn ist schon auf 1 gesetzt
  // Falls zum ersten Mal: Ausfuehren der entsprechenden Aktion/Zuweisung der Parameter 
  //  while (-1 != (argument = getopt(argc, argv, "i:s:o:a:x:y:z:c:d:e:p:t:B:NLGDPT"))) {
  while (-1 != (argument = getopt(argc, argv, "v:i:s:j:X:"))) {
    switch (argument) {
      /*     case 'N':
	     if (has_N)
	     cerr << "multiple option -N given,"
	     << " ignore further declaration\n";
	     else {
	     profile_mode = nocommon;
	     }
	     has_N = true;
	     break;*/
      /*    case 'x':
	    if (has_N) {
	    cerr << "this transition probability is not necessary"
	    << " since no profil HMM of the whole alignment is considered\n";
	    }
	    else {
	    if (has_x) 
	    cerr << "multiple option -x given,"
	    << " ignore further declaration\n";
	    else
	    jumpProfileToRow = atof(optarg); 
	    }
	    has_x = true;
	    break;*/
      /*    case 'y':
	    if (has_N) {
	    cerr << "this transition probability is not necessary"
	    << " since no profil HMM of the whole alignment is considered\n";
	    }
	    else {
	    if (has_y) {
	    cerr << "multiple option -y given,"
	    << " ignore further declaration\n";
	    }    
	    else
	    jumpRowToProfile = atof(optarg); 
	    }      
	    has_y = true;
	    break;*/
    case 'X':
      appl_id = atoi(optarg); 
      break;
    case 'j':
      if (has_j)
	cerr << "multiple option -j given,"
	     << " ignore further declaration\n";
      else
	jumpRowToRow = atof(optarg); 
      has_j = true;
      break;
      /*    case 'c':
	    if (has_c)
	    cerr << "multiple option -c given,"
	    << " ignore further declaration\n";
	    else
	    d_init = atof(optarg);
	    has_c = true;
	    break;*/
      /*    case 'd':
	    if (has_d)
	    cerr << "multiple option -d given,"
	    << " ignore further declaration\n";
	    else
	    d_ext = atof(optarg); 
	    has_d = true;
	    break;*/
      /*    case 'e':
	    if (has_e)
	    cerr << "multiple option -e given,"
	    << " ignore further declaration\n";
	    else
	    epsilon = (mytype)atof(optarg);
	    has_e = true;
	    break;*/
    case 'i':
      if (has_i)
	cerr << "multiple option -i given,"
	     << " ignore further declaration\n";
      else 
	infile.assign(optarg);
      // Eingabe-Alignment
      has_i = true;
      break;
    case 'v':
      if (has_v)
	cerr << "multiple option -v given,"
	     << " ignore further declaration\n";
      else 
	virus_type.assign(optarg);
      // Virustyp
      has_v = true;
      break;
    /*case 'o':
      if (has_o)
	cerr << "multiple option -o given,"
	     << " ignore further declaration\n";
      else 
	recomb_out.assign(optarg);
      has_o = true;
      break;*/
    case 's':
      if (has_s)
	cerr << "multiple option -s given,"
	     << " ignore further declaration\n";
      else 
	querySeq_file.assign(optarg);
      // Anfragesequenz
      has_s = true;
      break; 
      /*    case 'a':
	    if (has_a)
	    cerr << "multiple option -a given,"
	    << " ignore further\n";
	    else 
	    emissionPriors_file.assign(optarg);
	    // emissionPriors
	    has_a = true;
	    break;*/
      /*    case 'p':
	    if (has_p)
	    cerr << "multiple option -p given,"
	    << " ignore further declaration\n";
	    else 
	    consensus_Percentage = atof(optarg);
	    has_p = true;
	    break;*/
      /*    case 't':
	    if (has_t)
	    cerr << "multiple option -t given,"
	    << " ignore further declaration\n";
	    else 
	    consensus_Threshold = atoi(optarg);
	    has_t = true;
	    break;*/
      /*    case 'T':
	    if (has_T)
	    cerr << "multiple option -T given,"
	    << " ignore further declaration\n";
	    else 
	    consensus = commonConsensus;
	    has_T = true;
	    break;*/
      /*    case 'B':
	    if (has_B)
	    cerr << "multiple option -B given,"
	    << " ignore further declaration\n";
	    else 
	    BEAM_WIDTH = atof(optarg);
	    has_B = true;
	    break; */
      /*    case 'D':
	    if (has_D)
	    cerr << "multiple option -D given,"
	    << " ignore further declaration\n";
	    if (has_P)
	    cerr << "mode -P chosen: protein sequences,"
	    << " ignore this (-D) declaration\n";
	    else 
	    alphabet = dna;
	    has_D = true;
	    break;*/
      /*    case 'P':
	    if (has_P)
	    cerr << "multiple option -P given,"
	    << " ignore further declaration\n";
	    if (has_D)
	    cerr << "mode -D chosen: dna sequences,"
	    << " ignore this (-P) declaration\n";
	    else 
	    alphabet = protein;
	    has_P = true;
	    break;*/
    default:
      usage(argv[0], (string)"Invalid Option");
      
    } // switch
  } // while (-1 ...
  if (optind < argc) {
    usage(argv[0], (string)"Invalid Option: " + argv[optind]);
  }
  /*if (!has_i)
    usage(argv[0], "no alignment input file given (-i)\n");*/
  if (!has_s)
    usage(argv[0], "no query sequence input file given (-s)\n");
  if (!has_v)
    usage(argv[0], "no type of the given sequences given (HIV or HCV) (-v)\n");
  /*if (!has_N) {
    if (!has_x)
    usage(argv[0], "no jump probability from 'common profile mode' to 'subtype mode' given (-x)\n");
    if(!has_y)
    usage(argv[0], "no jump probability from 'subtype mode' to 'common profile mode' given (-y)\n");
    }*/
  /*if (!has_z)
    usage(argv[0], "no probability for a jump from one subtype to any other subtype given (-z)\n");
    if (!has_a)
    usage(argv[0], "no prior input file given (-a)\n");*/

  if (virus_type == "HIV") {
    emissionPriors_file.assign("nucleic_Alphas_estimated_HIV.txt");
    //infile.assign("HIV-alignment_pureSubtypes.fas");
  }
  else if (virus_type == "HCV") {
    emissionPriors_file.assign("nucleic_Alphas_estimated_HCV.txt");
    //infile.assign("HCV-alignment.fasta");
  }
  else {
    cerr << "\nfailure: given type of sequences (virus) is different from HIV or HCV (in capital letters).\n" << endl;
    return 1;
  }

  ifstream ifstrm(infile.c_str(), ios::in); 
  ifstream queryfile(querySeq_file.c_str(), ios::in);
  ifstream emissionPriors(emissionPriors_file.c_str(), ios::in);

  if (!ifstrm){ 
    cerr << "\nfailure of the alignment input file.\n" << endl;
    return 1;
  }
  if(!queryfile) {
    cerr << "\nfailure of the query sequence input file.\n" << endl;
    return 1;
  }
  if (!emissionPriors) {
    cerr << "\nthe prior input file could not be opened.\n" << endl; //<< "( for dna sequences: nucleic_Alphas.txt, for protein sequences: amino_Alphas.txt )\n" << endl;
    exit(-1);
  }
  
  // start of the main program -> reading in the alignment, generating the jphmm and computing the viterbi path for a given query sequence

  Const_Def::Const_Def();
  Alignment *al = new Alignment(ifstrm);
  int numSubtypes = 0;
  /*  if (al->getAnzSubtypes() == 1) 
      numSubtypes = al->getAnzSequenzen();
      else */
  numSubtypes = al->getAnzSubtypes();

  HMM::HMM(al, numSubtypes, emissionPriors, queryfile);

  delete al;
  return 0;
}
