/**************************************************************
 filename	: profileHMM.HH
 licence        : GNU Licence, see file LICENCE

 author		: Anne-Kathrin Schultz
 date		: 03/04 - 08/05

Content:
 classes: ProfileHMM,
          Transition,
          State

  |   date    |    author            |  changes
 -|-----------|----------------------|-------------------------------------------
  | 06.10.06  | Anne-Kathrin Schultz | now also accepts only 1 subtype in the alignment

***************************************************************/

/************************************************************************************************************************************************************************************************
description:

ProfileHMM:
-----------
The class ProfileHMM generates a profile HMM for each subtype:

funktions:
 provideMemory(...)              memory allocation for the local array allStates which contains all states of the considered profile HMM;
 fillStates(...)                 assignes to each consensus column of the subtype three states: MATCH, INSERT, DELETE;
                                  assignment of the emission probabilities to MATCHes and INSERTs;
 addStates(...)                  determination of the transitions in the profile HMM: 
                                  assigns to each state its potential predecessors and successors in the profile HMM of this subtype;
 createProfileHMM(...)           generates the profile HMM of this subtype;
                                  the transition probabilities of the transitions determined in addStates() are computed;

variables:
 anz_Spalten                     number of columns of this subtype (=number of columns of the given alignment)
************************************************************************************************************************************************************************************************/

/************************************************************************************************************************************************************************************************
Transition:
-----------
The class Transition assigns to each transition the predecessor, the successor and the probability of this transition:

funktions:
 Transition(...)                       (= construktor) assigns to each transition the predecessor and the successor state, i.e. the initial (whereFrom)(=predecessor) and the terminal state (whereTo)(=successor) of this transition
 ProfileHMM_TransitionFrequencies(..)  to each state of the profile HMM the transition frequencies to the successional states of the profile HMM are assigned to; 
 inner_Transition_Probs(...)           from the transition frequencies of a state the probabilities of the transitions to the successors (of the same profile HMM!) 
                                       are computed taking into account the transition priors;
 ermittleTransProbs(...)               computation of all(!-> in the whole jphmm, not only in one subtype) transition probabilities,
 
variables:
 to each transition belongs a predecessor, a successor and a transition probability:
 
 whereFrom                             initial state of the transition
 whereTo                               terminal state of the transition
 transProb                             transition probability of the transition whereFrom->whereTo
 log_transProb                         logarithm of the transition probability

************************************************************************************************************************************************************************************************/

/************************************************************************************************************************************************************************************************
State:
------
The class State assigns to each state four values (row(=subtype) in the jphmm, column in the alignment, type of the state and its position in the list HMM_States of all states of the jphmm), 
                                      a vector containing the emission probabilities and two lists containing the predecessors and successors of this state in the jphmm:

in general: To each consensus column of a subtype three states are assigned to: MATCH, INSERT, DELETE; each of these states is associated with its subtype and the corresponding column in the alignment;

functions:
 State(HMM_StateType, int, int)  (= construktor) assigns to each state the type of the state, the subtype the state belongs to and the corresponding column in the alignment                           
 dirichlet_EmissionPriors(...)   computes the emission probabilities for MATCH and INSERT 
 log_Probs(...);                 computes the logarithm of a vector containing the transition probabilities (of a state)

variables:
 pos_In_HMM_States               position of the state in the list HMM_States(see class HMM) of all states in the jphmm;
 stateType                       type of the state (f.ex. MATCH, END,...);
 row                             corresponding subtype of the state, = profile HMM the state belongs to (= row on the jphmm-array);
 column                          corresponding column of the state in the alignment;
 pred                            vector of all(!) transitions TO this state;
 succ                            vector of all(!) transitions starting in this state;
 emProb                          vector of all emission probabilities (for DNA: 4 +'X' entries, for proteins: 20 + 'X' entries);
 log_emProb                      vector of the logarithms of all emission probabilities;

**************************************************************************************************************************************************************************************************/

#ifndef _PROFILEHMM_HH
#define _PROFILEHMM_HH

#include "const_def.hh"
#include "alignment.hh"
#include <vector>
#include <list>
#include <iostream>

using namespace std;
 

class State;

extern float  jumpProfileToRow;       // jump probability: profile HMM of whole alignment -> subtypes
extern float  jumpRowToProfile;       // jump probability: subtype -> profile HMM of whole alignment
extern float  jumpRowToRow;           // jump probability: subtype -> other subtype
extern mytype epsilon;                // ->maintest.cc
extern mytype d_init;
extern mytype d_ext;

/******************************************************\
 *                class ProfileHMM                    *
\******************************************************/

class ProfileHMM {
public:
  ProfileHMM();
  ~ProfileHMM();
  void           provideMemory(State*** &allStates, int numSpalten);                   
  void           fillStates(Alignment* al, State*** &allStates, int** &profil, int* &gaps, int numResidues, int numSubtype, int anz_Subtypes, ifstream& emissionPriors);     
  void           addStates(State* akt_Zustand, State*** &allStates, int numSpalten, vector<mytype> innerTransProbs);    
  vector<State*> createProfileHMM(Alignment* al, int numSubtypes, int noOfSubtype, int anz_Subtypes, ifstream& emissionPriors);   

private:
  int anz_Spalten;
};

/******************************************************\
 *                 class Transition                   *
\******************************************************/

class Transition{
public:
  Transition();
  Transition(State* woher, State* wohin);                                                // transition between 2 states  
  ~Transition();  
  static vector<int>    profileHMM_TransitionFrequencies(Alignment* al, int akt_Zeile, int akt_spalte, int numSubtypes, int* &konsensus_Spalte, int** &profil);
                                                                                         // Zaehlen der transition frequencies in the profile HMM for all sequences
  static vector<mytype> inner_Transition_Probs(vector<int> transition_frequencies);      // Berechnung der Uebergangswkten innerhalb von ZeilenHMMs und ProfilHMM
  void                  ermittleTransProbs(State* akt_Zustand, State**** &allStates, int numSubtypes, int numSpalten, int** &gaps, int num_JumpsToProfile, int num_JumpsToSubtypes);  
                                                                                         // Berechnung der transition probabilities for all successors    

  State* whereFrom;                                                                      // transition between 2 states: begin of the transition 
  State* whereTo;                                                                        //                              end of the transition
  mytype transProb;                                                                      // transition probability between the 2 states 'whereFrom' and 'whereTo'
  mytype log_transProb;                                                                  // logarithm of the transition probabilities
   
private:
};

/******************************************************\
 *                   class State                      *
\******************************************************/

class State{
 public:
   State();
   State(HMM_StateType type, int zeile, int spalte);                                    // fuer jeden Zustand: Zustandstyp, Zeile im Alignment, Spalte im Alignment
   ~State();

   static vector<mytype> dirichlet_EmissionPriors(HMM_StateType type, mytype* frequencies, ifstream& emissionPriors);    // Dirichlet priors for the emission probabilities
   static vector<mytype> log_Probs(vector<mytype> probs);                               // logarithm of the emission and transition probabilities
                                                                                        // vielleicht ueberfluessig, aber ? ob vielleicht auch "normale" Wkten benoetigt 
   vector<Transition*> pred;                                                            // list of all possible predecessors of a state with corresponding transition probabilities
   vector<Transition*> succ;                                                            // list of all possible successors of a state with corresponding transition probabilities
   int                 row;                                                             // row of the state in the alignment
   int                 column;                                                          // column of the state in the alignment
   HMM_StateType       stateType;                                                       // type of state
   vector<mytype>      emProb;                                                          // emission probabilities
   vector<mytype>      log_emProb;                                                      // logarithm of the emission probabilities
   int                 pos_In_HMM_States;                                               // position of the state in the list of all states
 private:
};

#endif        // _PROFILEHMM_HH
